/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The archive format pipe device.
 * This writes all of its input files into one output "ar" file.
 */

#include <ar.h>
#include <unistd.h>
#include <time.h>

#include "upPrivate.h"


/*
 * Phases of writing special data.
 * This is the magic begin line, the header for a file, and padding
 * at the end of a file to make the next header be word aligned.
 */
typedef enum
{
	PHASE_MAGIC,
	PHASE_HEADER,
	PHASE_PADDING
} PHASE;


typedef	struct
{
	char *		outputFileName;	/* output file name */
	PHASE		phase;		/* phase of writing special data. */
	off_t		left;		/* data left to write in file */
	off_t		pos;		/* current position in output file */
	const char  *	dataAddress;	/* current address of data to write */
	int		dataLength;	/* length left of data to write */
	struct ar_hdr	header;		/* file header to write */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_ARCHIVE,
	"Archive",
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Local procedures.
 */
static	UP_BOOL	writeFileName(UP_DEV *);
static	UP_BOOL	readFileName(UP_DEV *);
static	UP_BOOL	transferData(UP_DEV *);
static	UP_BOOL	readEndOfFile(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	UP_BOOL	putNumber(char * cp, int len, long value, UP_BOOL octal);
static	void	formatHeader(struct ar_hdr *, const char *, const UP_STAT *);


UP_ID
upAddArchive(UP * up, const char * name)
{
	UP_DEV	*	dev;
	DATA	*	data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	data = DEVTODATA(dev);

	/*
	 * Default the output name if required.
	 */
	if (name == NULL)
		name = "output.ar";

	/*
	 * Initialize our private data.
	 */
	data = DEVTODATA(dev);
	data->dataAddress = NULL;
	data->dataLength = 0;
	data->phase = PHASE_MAGIC;
	data->left = 0;
	data->pos = 0;

	/*
	 * Copy the output file name.
	 */
	data->outputFileName = upDeviceCopyString(dev, name);

	if (data->outputFileName == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Set our initial state to the writing of a file name.
	 */
	upSetState(dev, UP_STATE_WRITE_NAME);

	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the state which we can act on.
	 */
	state = upCheckState(dev);

	/*
	 * Switch on the state to execute it.
	 */
	switch (state)
	{
	case UP_STATE_NONE:
		/*
		 * We can do nothing right now.
		 */
		return UP_TRUE;

	case UP_STATE_READ_NAME:
		/*
		 * We have a new filename available.
		 */
		return readFileName(dev);

	case UP_STATE_READ_END_OF_BUFFER:
		/*
		 * There are no more files to put into the archive.
		 * Acknowledge the end of buffer.
		 */
		upBufferReadEndOfBuffer(readBuffer);

		/*
		 * Write an end of file and end of buffer to the
		 * output buffer.
		 */
		upBufferWriteEof(writeBuffer);
		upBufferWriteEndOfBuffer(writeBuffer);

		/*
		 * Say we are finished.
		 */
		return upSetFinished(dev);

	case UP_STATE_TRANSFER_DATA:
		/*
		 * We can transfer data from the input to output buffer.
		 */
		return transferData(dev);

	case UP_STATE_TRANSFER_EOF:
		/*
		 * We have an end of file during the transfer.
		 * This means that the input file was too short.
		 */
		upDeviceError(dev, UP_ERR_SHORTFILE);

		return UP_FALSE;

	case UP_STATE_READ_DATA:
		/*
		 * We have some extra data from the input file.
		 * This means that the input file was too long.
		 */
		upDeviceError(dev, UP_ERR_LONGFILE);

		return UP_FALSE;

	case UP_STATE_READ_EOF:
		/*
		 * We have an end of file from the input file.
		 */
		return readEndOfFile(dev);

	case UP_STATE_WRITE_NAME:
		/*
		 * We can write the output file name.
		 */
		return writeFileName(dev);

	case UP_STATE_WRITE_DATA:
		/*
		 * We can write some special data to the output buffer.
		 */
		return writeData(dev);

	default:
		/*
		 * We have an unknown state.
		 */
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read the file name and status from the input buffer, format the
 * archive header for that file name, and set up for that header
 * to be written to the output buffer.
 */
static UP_BOOL
readFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	fileName;
	const UP_STAT *	statBuf;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the input file's name and status.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	statBuf = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Make sure there is a file name.
	 */
	if (fileName == NULL)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If the file name is too long then do something about it.
	 */
	if (strlen(fileName) > sizeof(data->header.ar_name))
	{
		fileName = "name-too-long";
	}

	/*
	 * Format the header from the file information.
	 */
	formatHeader(&data->header, fileName, statBuf);

	/*
	 * Set up to write the header block to the output buffer.
	 * Also save the amount to write for the file.
	 */
	data->phase = PHASE_HEADER;
	data->dataAddress = (char *) &data->header;
	data->dataLength = sizeof(data->header);
	data->left = statBuf->size;

	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read an end of file from the input buffer.
 * This is called after we have transferred all of the file's data
 * to the output file, and we just wanted to confirm that there
 * was no more input data.  We might need to write a padding byte
 * to finish the output file.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Read the end of file from the input buffer.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * If the file position is on an even word boundary,
	 * then the file is completely written.
	 */
	if ((data->pos & 0x01) == 0)
	{
		/*
		 * Set the state to read another file name.
		 */
		upSetState(dev, UP_STATE_READ_NAME);

		/*
		 * Return saying we have done I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * We must write a newline padding character to align the
	 * output file on the next word boundary.
	 */
	data->phase = PHASE_PADDING;
	data->dataAddress = "\n";
	data->dataLength = 1;

	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * We have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer data from the input buffer to the output buffer.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *	data;
	int	cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Try to transfer some data.
	 */
	cc = upTransferData(dev, UP_LIMIT(data->left));

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Do the accounting.
	 */
	data->pos += cc;
	data->left -= cc;

	/*
	 * If there is no more data to transfer for the file then
	 * set our state to read any additional data input data.
	 * We expect the read to get an immediate end of file.
	 * If more data exists, then the file size was incorrect
	 * and we will generate an error.
	 */
	if (data->left == 0)
		upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Return saying we did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write the initial file name to the output buffer.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *	data;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Write the file name with default status to the output buffer.
	 */
	upBufferWriteFileName(dev->writeBuffer, data->outputFileName, NULL);

	/*
	 * Now set our state to write the magic archive line to
	 * the output buffer.
	 */
	data->dataAddress = ARMAG;
	data->dataLength = SARMAG;
	data->phase = PHASE_MAGIC;

	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine to write special data to the output buffer.
 * This is either the magic line at the beginning of the file,
 * a file header line, or the padding newline after a file.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	cc = upWriteData(dev, data->dataAddress, data->dataLength);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Do accounting.
	 */
	data->dataAddress += cc;
	data->pos += cc;
	data->dataLength -= cc;

	/*
	 * If there is still more to be written then just return
	 * success for now.
	 */
	if (data->dataLength > 0)
		return upSetDidReadWrite(dev);

	/*
	 * We have finished writing the special data.
	 * Switch on the writing phase to see what we do next.
	 */
	switch (data->phase)
	{
	case PHASE_MAGIC:
	case PHASE_PADDING:
		/*
		 * We have written the magic beginning line or
		 * the padding byte after the end of a file.
		 * Set our state to read a file name now.
		 */
		upSetState(dev, UP_STATE_READ_NAME);

		break;

	case PHASE_HEADER:
		/*
		 * We have written the header for a file.
		 * Set our state to transfer the data now.
		 */
		upSetState(dev, UP_STATE_TRANSFER_DATA);

		break;

	default:
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Return saying we did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Format the specified archive header given the specified file name and
 * status information.  The file name is assumed to fit.
 */
static void
formatHeader(struct ar_hdr * header,
	const char * fileName,
	const UP_STAT * statBuf)
{
	int	len;
	time_t	mtime;
	uid_t	uid;
	gid_t	gid;
	mode_t	mode;

	/*
	 * Set the structure to blanks.
	 */
	memset((char *) header, ' ', sizeof(*header));

	/*
	 * Store the file name with a trailing space, if there is room.
	 */
	len = strlen(fileName);

	memcpy(header->ar_name, fileName, len);

	if (len < sizeof(header->ar_name))
		header->ar_name[len] = '/';

	/*
	 * Pick out the other details for the header,
	 * generating them if they weren't specified.
	 */
	mtime = statBuf->mtime;
	uid = statBuf->uid;
	gid = statBuf->gid;
	mode = statBuf->mode;

	if ((statBuf->flags & UP_STAT_MTIME) == 0)
		mtime = time(NULL);

	if ((statBuf->flags & UP_STAT_UID) == 0)
		uid = getuid();

	if ((statBuf->flags & UP_STAT_GID) == 0)
		gid = getgid();

	if ((statBuf->flags & UP_STAT_MODE) == 0)
		mode = 0666 | S_IFREG;

	/*
	 * Store the other fields.
	 */
	putNumber(header->ar_date, sizeof(header->ar_date), mtime, UP_FALSE);
	putNumber(header->ar_uid, sizeof(header->ar_uid), uid, UP_FALSE);
	putNumber(header->ar_gid, sizeof(header->ar_gid), gid, UP_FALSE);
	putNumber(header->ar_mode, sizeof(header->ar_mode), mode, UP_TRUE);

	putNumber(header->ar_size, sizeof(header->ar_size),
		statBuf->size, UP_FALSE);

	/*
	 * Add the magic ending characters.
	 */
	memcpy(header->ar_fmag, ARFMAG, sizeof(header->ar_fmag));
}


/*
 * Format a decimal or octal number into the specified field.
 * The number is space padded on the right to fill the field.
 * Returns TRUE if successful.
 */
static UP_BOOL
putNumber(char * cp, int len, long value, UP_BOOL octal)
{
	int	bufferLength;
	char	buffer[32];

	/*
	 * Format the decimal or octal number.
	 */
	sprintf(buffer, (octal ? "%o" : "%d"), value);

	/*
	 * If the formatted number is too large then fail.
	 */
	bufferLength = strlen(buffer);

	if (bufferLength > len)
		return UP_FALSE;

	/*
	 * Copy the string to the field.
	 */
	memcpy(cp, buffer, bufferLength);

	/*
	 * Fill the rest of the field with spaces.
	 */
	memset(cp + bufferLength, ' ', len - bufferLength);

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the output file name.
	 */
	free(data->outputFileName);

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	data->dataAddress = NULL;
	data->dataLength = 0;
	data->phase = PHASE_MAGIC;
	data->left = 0;
	data->pos = 0;

	/*
	 * Set our initial state to the writing of a file name.
	 */
	upSetState(dev, UP_STATE_WRITE_NAME);

	return UP_TRUE;
}
