/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The cat device.
 * This just concatenates all input files into one output file.
 */

#include "upPrivate.h"


/*
 * The default file name to generate.
 */
#define	FILE_PATTERN	"cat.%"


static	UP_BOOL	doReadWrite(UP_DEV * dev);
static	UP_BOOL	doClose(UP_DEV * dev);


static UP_FUNC	funcs =
{
	UP_TYPE_CAT,
	"Cat",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef	struct
{
	char *	fileName;
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


UP_ID
upAddCat(UP * up, const char * fileName, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialize our private data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Copy the fileName if it was given, or else generate a name.
	 */
	if (fileName)
		data->fileName = upDeviceCopyString(dev, fileName);
	else
		data->fileName = upGenerateFileName(dev, FILE_PATTERN, -1);

	if (data->fileName == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Set the initial state for the device which is to write
	 * our output file name.
	 */
	upSetState(dev, UP_STATE_WRITE_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	/*
	 * Point to our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state which we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
	case UP_STATE_NONE:
		/*
		 * We can do nothing right now.
		 */
		return UP_TRUE;

	case UP_STATE_WRITE_NAME:
		/*
		 * We are able to write the output file name.
		 */
		if (!upBufferWriteFileName(writeBuffer, data->fileName, NULL))
			return UP_FALSE;

		/*
		 * Set our state to the reading of a file name now.
		 */
		upSetState(dev, UP_STATE_READ_NAME);

		break;

	case UP_STATE_READ_NAME:
		/*
		 * We can read the next file name or end of file.
		 * Do that and then ignore it.
		 */
		upBufferReadFileName(readBuffer);

		/*
		 * Set our state to the transferring of data.
		 */
		upSetState(dev, UP_STATE_TRANSFER_DATA);

		break;

	case UP_STATE_READ_END_OF_BUFFER:
		/*
		 * There are no more files to be read.
		 * Mark the end of file and the end of the buffer
		 * on the output buffer and set ourself as done.
		 */
		upBufferWriteEof(writeBuffer);
		upBufferWriteEndOfBuffer(writeBuffer);

		upSetFinished(dev);

		break;

	case UP_STATE_TRANSFER_DATA:
		/*
		 * We can transfer data from the input to the output buffer.
		 * Do that and remain in the same state for more.
		 */
		if (upTransferData(dev, 0) < 0)
			return UP_FALSE;

		break;

	case UP_STATE_TRANSFER_EOF:
		/*
		 * We have an end of file from the input buffer.
		 * Acknowledge that, but don't pass it to the output buffer.
		 */
		upBufferReadEof(readBuffer);

		/*
		 * Change our state back to the reading of the next
		 * file name from the input buffer.
		 */
		upSetState(dev, UP_STATE_READ_NAME);

		break;

	case UP_STATE_TRANSFER_FILE_ERROR:
	case UP_STATE_FATAL_ERROR:
		/*
		 * There was a fatal error of some kind.
		 */
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;

	default:
		/*
		 * We have an unknown state.
		 */
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Say we have done some I/O, and return success.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the allocated file name.
	 */
	free(data->fileName);
	data->fileName = NULL;

	return UP_TRUE;
}
