/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The filtering pipe device.
 * This allows files to be filtered according to their name, file type,
 * size, or other criteria.  The file name and status values can also
 * be modified.
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <regex.h>

#include "upPrivate.h"


/*
 * The reallocation size for the regular expressions.
 */
#define	REGEX_ALLOC_SIZE	20


/*
 * Local routines.
 */
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL readFileName(UP_DEV * dev);
static	int	compilePattern(UP_DEV * dev, const char * pattern, UP_FLAG);
static	UP_BOOL	testPattern(UP_DEV *, const char *, const UP_FILTER_ITEM *);
static	void	freeTables(UP_DEV * dev);

static	UP_BOOL checkItem
	(UP_DEV * dev, const char *, const UP_STAT *, const UP_FILTER_ITEM *);

static	const char *	extractPart(UP_DEV *, const char *, int);


static UP_FUNC	funcs =
{
	UP_TYPE_FILTER,
	"Filter",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef struct
{
	char *			buffer;
	int			bufferSize;
	UP_FILTER_ITEM *	itemTable;
	int			itemCount;
	regex_t **		regexTable;
	int			regexCount;
	int			regexAlloc;
	regex_t *		quickRegexTable[REGEX_ALLOC_SIZE];
} DATA;

#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


/*
 * The parts of a path that we want to extract.
 */
enum
{
	PART_PATH,
	PART_DIR,
	PART_FILE,
	PART_BASE,
	PART_EXTENSION
};


UP_ID
upAddFilter(UP * up, const UP_FILTER_ITEM * itemTable, int itemCount)
{
	UP_DEV *	dev;
	DATA *		data;
	UP_FILTER_ITEM *item;
	int		index;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);

	data->itemCount = 0;
	data->itemTable = NULL;

	data->regexTable = data->quickRegexTable;
	data->regexAlloc = REGEX_ALLOC_SIZE;
	data->regexCount = 0;

	/*
	 * Set our initial state to read the first file name so that
	 * it can be checked whether or not the file is wanted.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * Validate the arguments.
	 */
	if (itemTable == NULL)
	{
		upError(up, UP_ERR_NULLPOINTER);

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	if (itemCount <= 0)
	{
		upError(up, UP_ERR_BADCOUNT);

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Allocate our own item table and copy the specified one into it.
	 */
	data->itemTable = (UP_FILTER_ITEM *)
		upDeviceMalloc(dev, sizeof(UP_FILTER_ITEM) * itemCount);

	if (data->itemTable == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	memcpy(data->itemTable, itemTable, sizeof(UP_FILTER_ITEM) * itemCount);

	/*
	 * Clear all of the string pointers in our memory in case of an error.
	 */
	for (index = 0; index < itemCount; index++)
		data->itemTable[index].string = NULL;

	/*
	 * Check each of the items in the table.
	 */
	for (index = 0; index < itemCount; index++)
	{
		item = &data->itemTable[index];

		/*
		 * The TRUE index value must be one of the special ones
		 * or else an index within the table and further on.
		 */
		if ((item->trueIndex != UP_FILTER_INDEX_ACCEPT) &&
			(item->trueIndex != UP_FILTER_INDEX_REJECT) &&
			((item->trueIndex <= index) ||
			(item->trueIndex >= itemCount)))
		{
			upDeviceError(dev, UP_ERR_BADINDEX);

			freeTables(dev);

			upFreeDevice(dev);

			return UP_NULL_ID;
		}

		/*
		 * The FALSE index value must be one of the special ones
		 * or else an index within the table and further on.
		 */
		if ((item->falseIndex != UP_FILTER_INDEX_ACCEPT) &&
			(item->falseIndex != UP_FILTER_INDEX_REJECT) &&
			((item->falseIndex <= index) ||
			(item->falseIndex >= itemCount)))
		{
			upDeviceError(dev, UP_ERR_BADINDEX);

			freeTables(dev);

			upFreeDevice(dev);

			return UP_NULL_ID;
		}

		/*
		 * Check the type.
		 */
		switch (item->type)
		{
			case UP_FILTER_TYPE_NONE:
			case UP_FILTER_TYPE_FILE_NUMBER:
			case UP_FILTER_TYPE_IS_ABSOLUTE:
			case UP_FILTER_TYPE_IS_ENTRY:
			case UP_FILTER_TYPE_ENTRY_COUNT:
			case UP_FILTER_TYPE_IS_REGULAR:
			case UP_FILTER_TYPE_IS_DIRECTORY:
			case UP_FILTER_TYPE_IS_SPECIAL:
			case UP_FILTER_TYPE_FILE_SIZE:
			case UP_FILTER_TYPE_FILE_UID:
			case UP_FILTER_TYPE_FILE_GID:
			case UP_FILTER_TYPE_FILE_ATIME:
			case UP_FILTER_TYPE_FILE_MTIME:
			case UP_FILTER_TYPE_FILE_CTIME:
			case UP_FILTER_TYPE_FILE_DEV:
			case UP_FILTER_TYPE_FILE_INODE:
			case UP_FILTER_TYPE_FILE_NLINKS:
			case UP_FILTER_TYPE_HAS_ANY_MODE_BITS:
			case UP_FILTER_TYPE_HAS_ALL_MODE_BITS:
			case UP_FILTER_TYPE_HAS_DATA:
				/*
				 * These items just use numbers.
				 * We don't have to do anything for them.
				 */
				break;

			case UP_FILTER_TYPE_HAS_PATH:
			case UP_FILTER_TYPE_HAS_ENTRY:
			case UP_FILTER_TYPE_HAS_DIR:
			case UP_FILTER_TYPE_HAS_FILE:
			case UP_FILTER_TYPE_HAS_BASE:
			case UP_FILTER_TYPE_HAS_EXTENSION:
			case UP_FILTER_TYPE_HAS_PATH_PREFIX:
			case UP_FILTER_TYPE_HAS_ENTRY_PREFIX:
			case UP_FILTER_TYPE_HAS_DIR_PREFIX:
			case UP_FILTER_TYPE_HAS_FILE_PREFIX:
			case UP_FILTER_TYPE_HAS_BASE_PREFIX:
			case UP_FILTER_TYPE_HAS_EXTENSION_PREFIX:
			case UP_FILTER_TYPE_HAS_PATH_SUFFIX:
			case UP_FILTER_TYPE_HAS_ENTRY_SUFFIX:
			case UP_FILTER_TYPE_HAS_DIR_SUFFIX:
			case UP_FILTER_TYPE_HAS_FILE_SUFFIX:
			case UP_FILTER_TYPE_HAS_BASE_SUFFIX:
			case UP_FILTER_TYPE_HAS_EXTENSION_SUFFIX:
				/*
				 * These items use a literal string.
				 * Copy the string from the user's table
				 * into our own table.
				 */
				item->string = upDeviceCopyString(dev,
					itemTable[index].string);

				if (item->string == NULL)
				{
					freeTables(dev);

					upFreeDevice(dev);

					return UP_NULL_ID;
				}

				/*
				 * Remember the length of the string
				 * for convenience in using it.
				 */
				item->integer = strlen(item->string);

				break;

			case UP_FILTER_TYPE_MATCHES_PATH:
			case UP_FILTER_TYPE_MATCHES_ENTRY:
			case UP_FILTER_TYPE_MATCHES_DIR:
			case UP_FILTER_TYPE_MATCHES_FILE:
			case UP_FILTER_TYPE_MATCHES_BASE:
			case UP_FILTER_TYPE_MATCHES_EXTENSION:
				/*
				 * These items use a regex pattern string.
				 * Allocate a new compiled pattern buffer
				 * and store its index as the integer value.
				 */
				item->integer = compilePattern(dev,
					itemTable[index].string,
					item->flags);

				if (item->integer < 0)
				{
					freeTables(dev);

					upFreeDevice(dev);

					return UP_NULL_ID;
				}

				break;

			default:
				/*
				 * This is an unknown type.
				 */
				upDeviceError(dev, UP_ERR_BADITEM);

				freeTables(dev);

				upFreeDevice(dev);

				return UP_NULL_ID;
		}

		/*
		 * If none of the comparison flags are set for the item
		 * then set the equality flag for our convenience.
		 */
		if ((item->flags &
			(UP_FILTER_ITEM_FLAG_LT |
			UP_FILTER_ITEM_FLAG_EQ |
			UP_FILTER_ITEM_FLAG_GT)) != 0)
		{
			item->flags |= UP_FILTER_ITEM_FLAG_EQ;
		}

		/*
		 * Increment the item counter.
		 */
		data->itemCount++;
	}

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_READ_NAME:
			/*
			 * There is a file name to be read from the input
			 * buffer.  Read it and see whether we want to
			 * transfer or consume this file.  The routine
			 * will change our state as required.
			 */
			return readFileName(dev);

		case UP_STATE_WRITE_NAME:
			/*
			 * The output buffer has room for a file name
			 * that we want to transfer.  Copy it from the
			 * input buffer.
			 */
			upBufferWriteFileName(writeBuffer,
				upBufferGetFileName(readBuffer),
				upBufferGetReadFileStatus(readBuffer));

			/*
			 * Set our state to transfer the file data now.
			 */
			upSetState(dev, UP_STATE_TRANSFER_DATA);

			break;

		case UP_STATE_READ_DATA:
			/*
			 * There is data to be read (and ignored) from
			 * the input buffer.  Consume it and stay in the
			 * same state.
			 */
			upBufferReadClear(readBuffer);

			break;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Transfer as much data as we can from the input to
			 * the output buffer.  Remain in the same state.
			 */
			if (upTransferData(dev, 0) < 0)
				return UP_FALSE;

			break;

		case UP_STATE_READ_EOF:
			/*
			 * We have finished consuming the input file.
			 * Acknowledge the end of file.
			 */
			upBufferReadEof(readBuffer);

			/*
			 * Set our state to read the next file name.
			 */
			upSetState(dev, UP_STATE_READ_NAME);

			break;

		case UP_STATE_TRANSFER_EOF:
			/*
			 * There is an end of file to be transferred
			 * from the input to the output buffer.
			 * Acknowledge the end of file and pass it on.
			 */
			upBufferReadEof(readBuffer);
			upBufferWriteEof(writeBuffer);

			/*
			 * Set our state to read the next file name.
			 */
			upSetState(dev, UP_STATE_READ_NAME);

			break;

		case UP_STATE_READ_END_OF_BUFFER:
			/*
			 * There is an end of buffer to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say that we are finished.
			 */
			return upSetFinished(dev);

		case UP_STATE_READ_FILE_ERROR:
		case UP_STATE_TRANSFER_FILE_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read the file name from the input buffer, and decide whether or not
 * we want to transfer the file or consume the file.  According to this
 * decision, we change our state to either read the input file, or to
 * write the output file name.
 */
static UP_BOOL
readFileName(UP_DEV * dev)
{
	DATA *			data;
	UP_BUF *		readBuffer;
	const char *		fileName;
	const UP_STAT *		status;
	const UP_FILTER_ITEM *	item;
	char *			newBuffer;
	int			newBufferSize;
	int			length;
	int			index;

	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;

	/*
	 * Get the file name and status.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	status = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Make sure that our temporary buffer is as long as our filename.
	 */
	length = strlen(fileName) + 1;

	if (length > data->bufferSize)
	{
		newBufferSize = length + 128;

		newBuffer = upDeviceRealloc(dev, data->buffer, newBufferSize);

		if (newBuffer == NULL)
			return UP_FALSE;

		data->buffer = newBuffer;
		data->bufferSize = newBufferSize;
	}

	/*
	 * Loop starting with the first item, checking them as directed
	 * by their indexes until we find an item which decides the result.
	 */
	index = 0;

	do
	{
		item = &data->itemTable[index];

		index = checkItem(dev, fileName, status, item);
	}
	while ((index > 0) && (index < data->itemCount));

	/*
	 * Now decide whether or not we want this file, and set our
	 * state according to how we want to handle the file.
	 */
	switch (index)
	{
		case UP_FILTER_INDEX_ACCEPT:
			/*
			 * Set our state to write the file name to the
			 * output buffer.  (After that, the file data
			 * will be transferred.)
			 */
			upSetState(dev, UP_STATE_WRITE_NAME);

			break;

		case UP_FILTER_INDEX_REJECT:
			/*
			 * Set our state to read the file data from the
			 * input buffer.  The data will be ignored.
			 */
			upSetState(dev, UP_STATE_READ_DATA);

			break;

		default:
			/*
			 * There was an error during the checking.
			 */
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Check the specified item to see if its condition is satisfied.
 * Returns the index of the next item to be checked (which may be
 * the special accept, reject, or error index values).
 */
static int
checkItem(UP_DEV * dev, const char * pathName, const UP_STAT * status,
	const UP_FILTER_ITEM * item)
{
	DATA *		data;
	UP_BOOL		haveMode;
	UP_BOOL		result;
	UP_BOOL		useComparison;
	int		comparison;
	int		index;
	int		length;
	int		intValue;
	const char *	stringValue;
	const char *	cp;

	/*
	 * Set up a few common values.
	 */
	data = DEVTODATA(dev);
	haveMode = ((status->flags & UP_STAT_MODE) != 0);
	stringValue = item->string;
	intValue = (int) item->integer;
	length = 0;
	comparison = 0;
	useComparison = UP_FALSE;
	result = UP_FALSE;

	/*
	 * Switch on the type of item to calculate either the a result
	 * value or else a comparison value.
	 */
	switch (item->type)
	{
		case UP_FILTER_TYPE_HAS_DATA:
			result = ((status->flags & UP_STAT_DATA) != 0);
			break;

		case UP_FILTER_TYPE_IS_ABSOLUTE:
			result = (*pathName == '/');
			break;

		case UP_FILTER_TYPE_IS_ENTRY:
			result = (strchr(pathName, '/') == NULL);
			break;

		case UP_FILTER_TYPE_IS_REGULAR:
			result = (!haveMode || S_ISREG(status->mode));

			break;

		case UP_FILTER_TYPE_IS_DIRECTORY:
			result = (haveMode && S_ISDIR(status->mode));

			break;

		case UP_FILTER_TYPE_IS_SPECIAL:
			result = (haveMode &&
				!S_ISREG(status->mode) &&
				!S_ISDIR(status->mode));

			break;

		case UP_FILTER_TYPE_FILE_SIZE:
			/*
			 * If the size is not known then ignore it.
			 */
			if ((status->flags & UP_STAT_SIZE) == 0)
			{
				result = UP_FALSE;
				break;
			}

			/*
			 * Calculate a comparison value for the size.
			 */
			if (status->size < item->offset)
				comparison = -1;

			if (status->size > item->offset)
				comparison = 1;

			useComparison = UP_TRUE;

			break;

		case UP_FILTER_TYPE_HAS_PATH_SUFFIX:
			/*
			 * Check whether the path ends with the specified
			 * suffix string.
			 */
			length = strlen(pathName);

			if (length < intValue)
				return UP_FALSE;

			cp = pathName + length - intValue;

			result = (memcmp(cp, stringValue, intValue) == 0);

			break;

		case UP_FILTER_TYPE_HAS_FILE_PREFIX:
			/*
			 * Check whether the file name starts with
			 * the specified prefix string.
			 */
			cp = extractPart(dev, pathName, PART_FILE);

			result = (memcmp(cp, stringValue, intValue) == 0);

			break;

		case UP_FILTER_TYPE_HAS_PATH_PREFIX:
			/*
			 * Check whether the path name starts with the
			 * specified prefix string.
			 */
			length = strlen(pathName);

			if (length < intValue)
				return UP_FALSE;

			result = (memcmp(pathName, stringValue, intValue)
				== 0);

			break;

		case UP_FILTER_TYPE_HAS_PATH:
			/*
			 * Check to see if the path name is an exact match.
			 */
			result = (strcmp(pathName, stringValue) == 0);

			break;

		case UP_FILTER_TYPE_HAS_FILE:
			/*
			 * Check if the file name is an exact match.
			 */
			cp = extractPart(dev, pathName, PART_FILE);

			result = (strcmp(cp, stringValue) == 0);

			break;

		case UP_FILTER_TYPE_HAS_DIR:
			/*
			 * Check if the path to the last pathName entry
			 * is as specified.
			 */
			cp = extractPart(dev, pathName, PART_DIR);

			result = (memcmp(cp, stringValue, length) == 0);

			break;

		case UP_FILTER_TYPE_HAS_ENTRY:
			/*
			 * Check if any of the entry components match
			 * the one specified.
			 */
			result = UP_FALSE;

			break;

		case UP_FILTER_TYPE_MATCHES_ENTRY:
			/*
			 * Check whether any entry matches the pattern.
			 */
			result = UP_FALSE;

			break;

		case UP_FILTER_TYPE_MATCHES_PATH:
		case UP_FILTER_TYPE_MATCHES_DIR:
		case UP_FILTER_TYPE_MATCHES_FILE:
		case UP_FILTER_TYPE_MATCHES_BASE:
		case UP_FILTER_TYPE_MATCHES_EXTENSION:
			/*
			 * These are pattern matches.
			 * Call a special routine for them.
			 */
			result = testPattern(dev, pathName, item);

			break;

		default:
			/*
			 * The item type is not known.
			 * Return an error value.
			 */
			return UP_FILTER_INDEX_ERROR;
	}

	/*
	 * If a comparison value was calculated then set the result
	 * on whether the comparison value matches the item's flags.
	 */
	if (useComparison)
	{
		if ((comparison < 0) &&
			((item->flags & UP_FILTER_ITEM_FLAG_LT) != 0))
		{
			result = UP_TRUE;
		}

		if ((comparison == 0) &&
			((item->flags & UP_FILTER_ITEM_FLAG_EQ) != 0))
		{
			result = UP_TRUE;
		}

		if ((comparison > 0) &&
			((item->flags & UP_FILTER_ITEM_FLAG_GT) != 0))
		{
			result = UP_TRUE;
		}

	}

	/*
	 * Select the appropriate next index according to whether
	 * the condition was satisfied.
	 */
	if (result)
		index = item->trueIndex;
	else
		index = item->falseIndex;

	/*
	 * Return the next index value.
	 */
	return index;
}


/*
 * Test whether the indicated pattern matches the specified path name.
 * This should only be called when the item type is actually a pattern.
 */
static UP_BOOL
testPattern(UP_DEV * dev, const char * pathName, const UP_FILTER_ITEM * item)
{
	DATA *		data;
	const char *	str;
	regex_t *	regex;
	int		part;
	int		index;
	UP_BOOL		result;
	regmatch_t	matchInfo[1];

	data = DEVTODATA(dev);

	/*
	 * Switch on the part of the path name to be checked.
	 */
	switch (item->type)
	{
		case UP_FILTER_TYPE_MATCHES_PATH:
			part = PART_PATH;
			break;

		case UP_FILTER_TYPE_MATCHES_DIR:
			part = PART_DIR;
			break;

		case UP_FILTER_TYPE_MATCHES_FILE:
			part = PART_FILE;
			break;

		case UP_FILTER_TYPE_MATCHES_BASE:
			part = PART_BASE;
			break;

		case UP_FILTER_TYPE_MATCHES_EXTENSION:
			part = PART_EXTENSION;
			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Extract the indicated part of the path name.
	 */
	str = extractPart(dev, pathName, part);

	/*
	 * Get the regular expression to be checked.
	 */
	index = (int) item->integer;

	if ((index < 0) || (index >= data->regexCount))
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	regex = data->regexTable[index];

	/*
	 * See if there is a match and return its location.
	 */
	result = (regexec(regex, str, 1, matchInfo, 0) == 0);

	/*
	 * If there was no match then return false.
	 */
	if (!result)
		return UP_FALSE;

	/*
	 * There was a match.  If it wasn't the entire string then
	 * return false.
	 */
	if (matchInfo[0].rm_so != 0)
		return UP_FALSE;

	if (matchInfo[0].rm_eo != strlen(str))
		return UP_FALSE;

	/*
	 * The match was the entire string.
	 */
	return UP_TRUE;
}


/*
 * Utility routine to find the various bits of a file name.
 * This returns a string stored in our data's buffer, or else a part
 * of the original supplied path name.  If the part is not present then
 * an empty string is returned.  Warning: The buffer must be allocated
 * large enough to hold the complete path name if necessary.
 */
static const char *
extractPart(UP_DEV * dev, const char * pathName, int part)
{
	DATA *		data;
	char *		buffer;
	const char *	fileName;
	const char *	cp;
	int		length;

	data = DEVTODATA(dev);
	buffer = data->buffer;

	switch (part)
	{
		case PART_PATH:
			return pathName;

		case PART_DIR:
			cp = strrchr(pathName, '/');

			if (cp == NULL)
				break;

			length = cp - pathName;

			if (length == 0)
				return "/";

			memcpy(buffer, pathName, length);
			buffer[length] = '\0';

			return buffer;

		case PART_FILE:
			fileName = strrchr(pathName, '/');

			if (fileName == NULL)
				return pathName;

			return fileName + 1;

		case PART_BASE:
			fileName = strrchr(pathName, '/');

			if (fileName)
				fileName++;
			else
				fileName = pathName;

			cp = strrchr(fileName, '.');

			if (cp == NULL)
				return fileName;

			length = cp - fileName;

			memcpy(buffer, fileName, length);
			buffer[length] = '\0';

			return buffer;

		case PART_EXTENSION:
			fileName = strrchr(pathName, '/');

			if (fileName)
				fileName++;
			else
				fileName = pathName;

			cp = strrchr(fileName, '.');

			if (cp == NULL)
				break;

			return cp + 1;

		default:
			break;
	}

	/*
	 * The specified part is empty.
	 */
	return "";
}


/*
 * Allocate a new regex structure for our table and compile the specified
 * pattern into it.  Returns the index into the table or -1 on an error.
 */
static int
compilePattern(UP_DEV * dev, const char * pattern, UP_FLAG flags)
{
	DATA *		data;
	regex_t *	regex;
	regex_t **	newRegexTable;
	int		newRegexAlloc;
	int		regexFlags;

	data = DEVTODATA(dev);

	/*
	 * Grow the pattern table if necessary.
	 */
	if (data->regexCount >= data->regexAlloc)
	{
		newRegexAlloc = data->regexAlloc + REGEX_ALLOC_SIZE;

		newRegexTable = (regex_t **)
			upGrowTable(dev, data->regexTable,
			data->quickRegexTable, data->regexCount,
			newRegexAlloc, sizeof(regex_t *));

		if (newRegexTable == NULL)
			return -1;

		data->regexTable = newRegexTable;
		data->regexAlloc = newRegexAlloc;
	}

	/*
	 * Allocate the new pattern buffer.
	 */
	regex = (regex_t *) upDeviceMalloc(dev, sizeof(regex_t));

	if (regex == NULL)
		return -1;

	/*
	 * Build the regex flags.
	 */
	regexFlags = REG_EXTENDED;

	if ((flags & UP_FILTER_ITEM_FLAG_IGNORE_CASE) != 0)
		regexFlags |= REG_ICASE;

	/*
	 * Compile the pattern.
	 */
	if (regcomp(regex, pattern, regexFlags) != 0)
	{
		upDeviceError(dev, UP_ERR_BADPATTERN);

		free(regex);

		return -1;
	}

	/*
	 * Add the pattern into the table and return its index.
	 */
	data->regexTable[data->regexCount] = regex;

	return data->regexCount++;
}


/*
 * Free all information related to the item table.
 * This is the strings contained within it, and any compiled patterns.
 */
static void
freeTables(UP_DEV * dev)
{
	DATA *			data;
	UP_FILTER_ITEM *	item;
	regex_t *		regex;
	int			index;

	data = DEVTODATA(dev);

	/*
	 * Free all of the strings in the item table.
	 */
	for (index = 0; index < data->itemCount; index++)
	{
		item = &data->itemTable[index];

		free(item->string);
	}

	/*
	 * Free the item table itself.
	 */
	free(data->itemTable);

	/*
	 * Free each pattern in the pattern table.
	 */
	for (index = 0; index < data->regexCount; index++)
	{
		regex = data->regexTable[index];

		regfree(regex);
	}

	/*
	 * Free the pattern table itself if it was allocated.
	 */
	if (data->regexTable != data->quickRegexTable)
		free(data->regexTable);

	/*
	 * Free the buffer we may have allocated.
	 */
	free(data->buffer);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	freeTables(dev);

	return UP_TRUE;
}
