/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The pipe-like device to find the file's data.
 * This is used to read the file data from the input file names.
 * The data for the files may or may not already be supplied.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>

#include "upPrivate.h"


/*
 * Local data for this device.
 */
typedef	struct
{
	int	fd;			/* currently opened file */
	UP_BOOL	isRegular;		/* file is a regular file */
	UP_BOOL	ignoreErrors;		/* ignore any file errors */
	UP_BOOL	alwaysRead;		/* read files even if there is data */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	transferFileName(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV * dev);


static	UP_FUNC	funcs =
{
	UP_TYPE_FIND_FILE_DATA,
	"FindFileData",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Find the data for the input files (if necessary) by opening the
 * specified file names.
 */
UP_ID
upAddFindFileData(UP * up, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;

	/*
	 * Allocate the device.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Fill in our private data.
	 */
	data = DEVTODATA(dev);

	data->fd = -1;

	data->ignoreErrors =
		((flags & UP_FLAG_FINDFILEDATA_IGNORE_ERRORS) != 0);

	data->alwaysRead = 
		((flags & UP_FLAG_FINDFILEDATA_ALWAYS_READ) != 0);

	/*
	 * Set our state to begin by transferring a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *		data;

	data = DEVTODATA(dev);

	/*
	 * Assume that we have nothing to poll for and that
	 * we have an indefinite timeout.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the state that we can act on.
	 */
	switch (upCheckState(dev))
	{
		case UP_STATE_NONE:
			/*
			 * Nothing can be done right now.
			 */
			return UP_TRUE;

		case UP_STATE_WRITE_DATA:
			/*
			 * Data can be stored in the output buffer.
			 * Break out of the switch to handle this.
			 */
			break;

		default:
			/*
			 * For all other states we want to run now.
			 */
			dev->pollTimeout = 0;

			return UP_TRUE;
	}

	/*
	 * Here when file data is wanted by the output buffer.
	 * Make sure that we have a file opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If the file descriptor is a regular file then no polling is
	 * required and there is a timeout value of zero since I/O can
	 * be done immediately.
	 */
	if (data->isRegular)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for reading of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLIN;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred from
			 * the input to the output buffer.
			 */
			return transferFileName(dev);

			break;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Transfer data from the input to the output buffer.
			 */
			if (upTransferData(dev, 0) < 0)
				return UP_FALSE;

			break;

		case UP_STATE_TRANSFER_EOF:
			/*
			 * There is an end of file to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEof(readBuffer);
			upBufferWriteEof(writeBuffer);

			/*
			 * Set our state to transfer another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * There is an end of buffer to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say that we are finished.
			 */
			upSetFinished(dev);

			break;

		case UP_STATE_WRITE_DATA:
			/*
			 * Data can be written to the output buffer.
			 */
			return writeData(dev);

		case UP_STATE_READ_DATA:
			/*
			 * Data can be read from the input buffer.
			 * We just ignore all such data.
			 */
			upBufferReadClear(readBuffer);

			break;

		case UP_STATE_READ_EOF:
			/*
			 * An end of file can be read from the input buffer.
			 */
			upBufferReadEof(readBuffer);

			/*
			 * Change our state to transfer another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		case UP_STATE_READ_FILE_ERROR:
		case UP_STATE_TRANSFER_FILE_ERROR:
			/*
			 * There was a file error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer a file name from the input buffer to the output buffer.
 * If there is no file data coming for the file (or if we are always
 * told to re-read it), then open the file to get its data.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	fileName;
	const UP_STAT *	fileStatus;
	struct stat	statBuf;
	UP_STAT		newStatBuf;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Read the file name and status.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	fileStatus = upBufferGetReadFileStatus(readBuffer);

	/*
	 * If there is file data coming and we can use it then just
	 * set up to transfer the data.
	 */
	if (!data->alwaysRead && (fileStatus->flags & UP_STAT_DATA) != 0)
	{
		/*
		 * Write the file name and status to the output buffer.
		 */
		upBufferWriteFileName(writeBuffer, fileName, fileStatus);

		/*
		 * Change our state to transfer data.
		 */
		upSetState(dev, UP_STATE_TRANSFER_DATA);

		/*
		 * Return saying we have done I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * We must read the file data.
	 * Try to open the file for reading without blocking.
	 */
	data->fd = open(fileName, O_RDONLY | O_NONBLOCK);

	/*
	 * If the open failed then handle it.
	 */
	if (data->fd < 0)
	{
		/*
		 * If we don't ignore open failures then fail.
		 */
		if (!data->ignoreErrors)
		{
			upDeviceFileSystemError(dev, "open", fileName);

			return UP_FALSE;
		}

		/*
		 * We want to ignore this file.
		 * Just set our state to reading so that we read and
		 * ignore the input file data.
		 */
		upSetState(dev, UP_STATE_READ_DATA);

		/*
		 * Return saying we have done I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * Get the status of the file that we have opened.
	 */
	if (fstat(data->fd, &statBuf) < 0)
	{
		upDeviceFileSystemError(dev, "fstat", fileName);

		(void) close(data->fd);

		data->fd = -1;

		return UP_FALSE;
	}

	/*
	 * Save the stat information so that it can be returned later.
	 */
	upCopyStatus(&newStatBuf, &statBuf);

	/*
	 * Remember whether the file is a regular file.
	 * We need this fact to know whether the file handle
	 * requires polling when we read it.
	 */
	data->isRegular = S_ISREG(statBuf.st_mode);

	/*
	 * If this is a directory say that is has no data.
	 */
	if (S_ISDIR(statBuf.st_mode))
		newStatBuf.flags &= ~UP_STAT_DATA;

	/*
	 * Tell the output buffer what the new file name and status is.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer, fileName, &newStatBuf))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If the file is a directory then write an end of file to the buffer
	 * and close the file right now so that we don't try to read it.
	 * (Reading a directory might generate read error anyway.)
	 */
	if (S_ISDIR(statBuf.st_mode))
	{
		upBufferWriteEof(dev->writeBuffer);

		(void) close(data->fd);

		data->fd = -1;

		/*
		 * Set our state to consume any file data.
		 */
		upSetState(dev, UP_STATE_READ_DATA);

		/*
		 * Return saying we have done I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * Set our state to say that we are writing the file's data.
	 * The input buffer will be ignored until the output file is
	 * done, at which time we will begin reading and ignoring the
	 * input file's data (if any).
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Try to write some more file data to the buffer.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	/*
	 * Point at our private data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Make sure that a file is opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If we are polling and there is no data ready, then do nothing.
	 */
	if ((dev->pollCount > 0) &&
		((dev->pollEntry.revents & (POLLIN | POLLHUP)) == 0))
	{
		return UP_TRUE;
	}

	/*
	 * Reserve as much space as we can for file data.
	 */
	writeLength = upBufferRequestSpace(writeBuffer, UP_BUFFER_MAX_SIZE);

	if (writeLength < 0)
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * There should have been some room.
	 */
	if (writeLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Get the address and length of the requested space.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Try to read some data from the file into the buffer.
	 */
	cc = read(data->fd, writeAddress, writeLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "read");

		return UP_FALSE;
	}

	/*
	 * If we read some data then inform the buffer about it.
	 */
	if (cc > 0)
	{
		upBufferWriteAdjust(writeBuffer, cc);

		/*
		 * Return saying we have done some I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * End of file has occurred.
	 * Tell the buffer about the end of file.
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Close the file.
	 */
	if (close(data->fd) < 0)
	{
		upDeviceSystemError(dev, "close");

		data->fd = -1;

		return UP_FALSE;
	}

	data->fd = -1;

	/*
	 * Set our state to read the input buffer's file data now
	 * which will be ignored.
	 */
	upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if ((data->fd >= 0) && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if (data->fd >= 0)
	{
		(void) close(data->fd);

		data->fd = -1;
	}

	return UP_TRUE;
}
