/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The find size device.
 * This buffers input data if necessary to calculate the file size,
 * so that the correct file size can be supplied to the next device.
 * To accomplish this, the file is not passed on to the output buffer
 * until all of the data has been read.  If a file size was given on
 * input, then the buffering is not performed, and the file data is
 * passed on immediately.  However, in this case this also verifies
 * that the actual file size matches the input file size.  If there is
 * a mismatch then the file can optionally be truncated or padded.
 * Data is buffered in memory until the specified amount is reached,
 * at which point it is then put into a temporary file.
 */

#include <stdio.h>
#include <unistd.h>

#include "upPrivate.h"


#define	MIN_BUFFER_SIZE	8192


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);

static	UP_BOOL transferFileName(UP_DEV * dev);
static	UP_BOOL transferData(UP_DEV * dev);
static	UP_BOOL	transferEndOfFile(UP_DEV * dev);
static	UP_BOOL readData(UP_DEV * dev);
static	UP_BOOL	readEndOfFile(UP_DEV * dev);
static	UP_BOOL	writeFileName(UP_DEV * dev);
static	UP_BOOL writeData(UP_DEV * dev);
static	UP_BOOL padData(UP_DEV * dev);
static	UP_BOOL	copyBufferToTemporaryFile(UP_DEV * dev);
static	UP_BOOL	copyTemporaryFileToBuffer(UP_DEV * dev);


static UP_FUNC	funcs =
{
	UP_TYPE_FIND_SIZE,
	"FindSize",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_TRUE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef struct
{
	int	fd;
	char *	buffer;
	int	bufferSize;
	int	bufferWriteOffset;
	int	bufferReadOffset;
	off_t	fileReadOffset;
	off_t	fileWriteOffset;
	off_t	targetFileSize;
	UP_BOOL	truncateFlag;
	UP_BOOL	padFlag;
	UP_BOOL	bufferFlag;
	UP_BOOL	ignoreDataFlag;
	UP_BOOL	paddingDataFlag;
} DATA;

#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


/*
 * Add a device to calculate the sizes of all the files which pass
 * through it.  The specified amount of memory is used for holding
 * file data.  If a file is larger than that, then a temporary file is
 * created to hold the file data.  The flags specify special handling.
 */
UP_ID
upAddFindSize(UP * up, int memorySize, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;
	char *		buffer;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Allocate the memory buffer.
	 * Make sure that it is a minimum useful size.
	 */
	if (memorySize < MIN_BUFFER_SIZE)
		memorySize = MIN_BUFFER_SIZE;

	buffer = malloc(memorySize);

	if (buffer == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);

	data->fd = -1;
	data->buffer = buffer;
	data->bufferSize = memorySize;
	data->bufferWriteOffset = 0;
	data->bufferReadOffset = 0;
	data->fileWriteOffset = 0;
	data->fileReadOffset = 0;
	data->targetFileSize = 0;
	data->ignoreDataFlag = UP_FALSE;
	data->paddingDataFlag = UP_FALSE;

	/*
	 * Set the flags.
	 */
	data->padFlag = ((flags & UP_FLAG_FINDSIZE_PAD) != 0);
	data->truncateFlag = ((flags & UP_FLAG_FINDSIZE_TRUNCATE) != 0);
	data->bufferFlag = ((flags & UP_FLAG_FINDSIZE_BUFFER) != 0);

	/*
	 * Set our initial state to transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *	data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * A file name can be read from the input buffer
			 * and also passed to the output buffer if we want.
			 * We have to read it and decide whether to pass
			 * it straight through, or wait until we have
			 * read all of the data.
			 */
			return transferFileName(dev);

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * There is no more data in the pipeline.
			 * Acknowledge it and tell the output buffer
			 * that we are done.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say that we are finished.
			 */
			return upSetFinished(dev);

		case UP_STATE_WRITE_NAME:
			/*
			 * We can write the file name to the output buffer.
			 * This is used after we have buffered all of the
			 * input file, and know its actual size.
			 */
			return writeFileName(dev);

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Data can be transferred directly from the
			 * input to the output buffer.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_EOF:
			/*
			 * An end of file can be transferred from the
			 * input buffer to the output buffer.  Check
			 * if the amount of data read was correct.
			 */
			return transferEndOfFile(dev);

		case UP_STATE_TRANSFER_FILE_ERROR:
			/*
			 * A file error can be transferred.
			 */
			return UP_FALSE;

		case UP_STATE_READ_DATA:
			/*
			 * Data can be read from the input buffer.
			 * It can either be ignored, or else it can be
			 * buffered until we have read the whole file.
			 */
			return readData(dev);

		case UP_STATE_READ_EOF:
			/*
			 * An end of file can be read from the input buffer.
			 */
			return readEndOfFile(dev);

		case UP_STATE_READ_FILE_ERROR:
			/*
			 * A file error can be read from the input buffer.
			 */
			return UP_FALSE;

		case UP_STATE_WRITE_DATA:
			/*
			 * Data can be written to the output buffer.
			 * This is either data we had buffered when it was
			 * read, or else padding data for a short file.
			 */
			if (data->paddingDataFlag)
				return padData(dev);
			else
				return writeData(dev);

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We successfully transferred something.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Function to read the file name from the input buffer and
 * decide what we should do to handle the file's data.
 * We either pass the file name through and set up to pass the data,
 * or else we set up to buffer all of the input file in order to
 * calculate its size.  We set up to consume all directory data.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	fileName;
	const UP_STAT *	status;

	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Reset some information in preparation for the new file.
	 */
	data->bufferReadOffset = 0;
	data->bufferWriteOffset = 0;
	data->fileReadOffset = 0;
	data->fileWriteOffset = 0;
	data->targetFileSize = 0;
	data->ignoreDataFlag = UP_FALSE;

	/*
	 * Read the file name and status from the input buffer.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	status = upBufferGetReadFileStatus(readBuffer);

	/*
	 * If the input file is a directory then just pass on the
	 * directory entry without attempting to transfer its data.
	 */
	if ((status->flags & UP_STAT_MODE) && S_ISDIR(status->mode))
	{
		/*
		 * Tell the output buffer about the directory.
		 */
		if (!upBufferWriteFileName(writeBuffer, fileName, status))
			return UP_FALSE;

		/*
		 * Weite an immediate end of file.
		 */
		upBufferWriteEof(writeBuffer);

		/*
		 * Set up to read input data but also to ignore it.
		 */
		upSetState(dev, UP_STATE_READ_DATA);
		data->ignoreDataFlag = UP_TRUE;

		/*
		 * Return saying we have done I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * If the file has no data then complain.
	 */
	if ((status->flags & UP_STAT_DATA) == 0)
	{
		upDeviceError(dev, UP_ERR_NODATA);

		return UP_FALSE;
	}

	/*
	 * If the always buffer flag is set, the size value is not set,
	 * or else the file type is unknown or not a regular file,
	 * then we must buffer the file data in order to calculate the
	 * actual file size.
	 */
	if (data->bufferFlag ||
		((status->flags & UP_STAT_SIZE) == 0) ||
		((status->flags & UP_STAT_MODE) == 0) ||
		!S_ISREG(status->mode))
	{
		/*
		 * We don't know the file size up front, so we have to
		 * buffer up all of the file's data before telling the
		 * output buffer the file name.  Set our state to do this.
		 */
		upSetState(dev, UP_STATE_READ_DATA);

		/*
		 * If we have a temporary file in use then seek back
		 * to the beginning of the file so that we can store
		 * new data into it if necessary.
		 */
		if ((data->fd >= 0) && (lseek(data->fd, 0L, SEEK_SET) < 0))
		{
			upDeviceSystemError(dev, "lseek");

			return UP_FALSE;
		}

		/*
		 * Return saying we have done I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * We can use the given file size.  Tell the output buffer
	 * what the file name and status is.
	 */
	if (!upBufferWriteFileName(writeBuffer, fileName, status))
		return UP_FALSE;

	/*
	 * Save the expected file size, and set our state
	 * to copy the file data to the output buffer without
	 * any buffering.
	 */
	data->targetFileSize = status->size;

	upSetState(dev, UP_STATE_TRANSFER_DATA);

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer file data directly between the input and output buffers,
 * while checking whether too much data was read from the input.
 * This is only used when the file size was known at the beginning.
 * Returns true if successful.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	long		leftLength;
	int		cc;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the amount of data left to transfer.
	 */
	leftLength = data->targetFileSize - data->fileWriteOffset;

	/*
	 * If the file size has been exceeded then the indicated
	 * file size was too long.  We have to either generate an
	 * error or else truncate the file.
	 */
	if (leftLength <= 0)
	{
		/*
		 * If we aren't set to truncate the file then
		 * return an error.
		 */
		if (!data->truncateFlag)
		{
			upDeviceError(dev, UP_ERR_LONGFILE);

			return UP_FALSE;
		}

		/*
		 * Generate a warning about the file truncation.
		 */
		upWarning(dev->owner,
			"%s: file too long (expected %ld bytes)",
			upBufferGetFileName(readBuffer),
			data->targetFileSize);

		/*
		 * Change our state to consume all of the remaining
		 * data from the input file.
		 */
		upSetState(dev, UP_STATE_READ_DATA);

		data->ignoreDataFlag = UP_TRUE;

		return upSetDidReadWrite(dev);
	}

	/*
	 * Reduce the amount left to a convenient size which will fit
	 * into an int.
	 */
	cc = 1024 * 16;

	if (cc > leftLength)
		cc = (int) leftLength;

	/*
	 * Transfer as much data as we can between the buffers.
	 */
	cc = upTransferData(dev, cc);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Increment the file offset by the amount transferred.
	 */
	data->fileWriteOffset += cc;

	/*
	 * Remember that we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Function called on an end of file while we were transferring data
 * straight from the input buffer to the output buffer.  We make sure
 * that the required amount of data was read, and if not, set up to
 * pad the file.
 */
static UP_BOOL
transferEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;

	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * If the input file size was large enough then we just need
	 * to pass the end of file to the output buffer and reset our
	 * state to transfer another file name.
	 */
	if (data->fileWriteOffset >= data->targetFileSize)
	{
		upBufferWriteEof(writeBuffer);

		upSetState(dev, UP_STATE_TRANSFER_NAME);

		return upSetDidReadWrite(dev);
	}

	/*
	 * The file size was shorter than specified.
	 * If padding is not enabled then give an error.
	 */
	if (!data->padFlag)
	{
		upDeviceError(dev, UP_ERR_SHORTFILE);

		return UP_FALSE;
	}

	/*
	 * The caller specified to do padding.
	 * Generate a warning message for this file.
	 */
	upWarning(dev->owner, "%s: short read (%ld of %ld bytes)",
		upBufferGetFileName(readBuffer), data->fileWriteOffset,
		data->targetFileSize);

	/*
	 * Set our state to write zeroes to the output file.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);
	data->paddingDataFlag = UP_TRUE;

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Function called when reading the end of file from the input buffer.
 * We either set up to read the next file, or set up to write the file
 * name and data to the output buffer.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * If we were ignoring the file data, then we have already
	 * written the output file name.  So just set up to transfer
	 * another file name.
	 */
	if (data->ignoreDataFlag)
	{
		upSetState(dev, UP_STATE_TRANSFER_NAME);

		return upSetDidReadWrite(dev);
	}

	/*
	 * If we were saving data into a temporary file then
	 * we must set things up to read it back again later.
	 */
	if (data->fd >= 0)
	{
		/*
		 * Append the last of the buffer data to the
		 * temporary file.
		 */
		if (!copyBufferToTemporaryFile(dev))
			return UP_FALSE;

		/*
		 * Seek back to the beginning of the file.
		 */
		if (lseek(data->fd, 0L, SEEK_SET) < 0)
		{
			upDeviceSystemError(dev, "lseek");

			return UP_FALSE;
		}
	}

	/*
	 * Set our state to write the file name and status to
	 * the output buffer.
	 */
	upSetState(dev, UP_STATE_WRITE_NAME);
	data->paddingDataFlag = UP_FALSE;

	return upSetDidReadWrite(dev);
}


/*
 * Called when there is data available to be read from the input buffer.
 * We either ignore it, or buffer it up into memory or a temporary file.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	int		cc;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * If we are to ignore the data, then do that.
	 */
	if (data->ignoreDataFlag)
	{
		/*
		 * Consume any input data.
		 */
		upBufferReadClear(readBuffer);

		/*
		 * Return success.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * We want to read the data and save it.
	 * Calculate the remaining amount of room in the memory buffer.
	 */
	cc = data->bufferSize - data->bufferWriteOffset;

	/*
	 * If the buffer is full then we must empty it.
	 */
	if (cc <= 0)
	{
		/*
		 * Append the buffer data into a temporary file.
		 */
		if (!copyBufferToTemporaryFile(dev))
			return UP_FALSE;

		/*
		 * Now we have the full buffer space available.
		 */
		cc = data->bufferSize;
	}

	/*
	 * Copy data from the input buffer into our buffer.
	 */
	cc = upReadData(dev, data->buffer + data->bufferWriteOffset, cc);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Increment our offsets by the amount copied.
	 */
	data->bufferWriteOffset += cc;
	data->fileWriteOffset += cc;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write the file name and updated status to the output buffer.
 * This is called after all of the input file has been read and saved,
 * at which point we know the true file size.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	readFileName;
	const UP_STAT *	readStatus;
	UP_STAT		newStatus;

	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * We can write the file name and status.
	 * Copy the input status and force the file size to be the
	 * amount of data we read.
	 */
	readFileName = upBufferGetFileName(readBuffer);
	readStatus = upBufferGetReadFileStatus(readBuffer);

	if (readStatus == NULL)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	memcpy(&newStatus, readStatus, sizeof(UP_STAT));

	/*
	 * Set the actual file size.
	 */
	newStatus.flags |= UP_STAT_SIZE;
	newStatus.size = data->fileWriteOffset;

	/*
	 * Tell the output buffer the file name and new status.
	 */
	if (!upBufferWriteFileName(writeBuffer, readFileName, &newStatus))
		return UP_FALSE;

	/*
	 * Set our state to write the file data now.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	data->paddingDataFlag = UP_FALSE;
	data->bufferReadOffset = 0;
	data->fileReadOffset = 0;

	/*
	 * We have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write data to the output buffer.
 * This is called when the file data was buffered in order to get its
 * actual size, and then the buffer data needs to be written to the
 * output buffer.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		cc;

	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * If there is no more data to be written, then write
	 * the end of file and get ready for a new file name.
	 */
	if (data->fileReadOffset >= data->fileWriteOffset)
	{
		upBufferWriteEof(writeBuffer);

		upSetState(dev, UP_STATE_TRANSFER_NAME);

		return upSetDidReadWrite(dev);
	}

	/*
	 * Get the amount of data we can write from the buffer.
	 */
	cc = data->bufferWriteOffset - data->bufferReadOffset;

	/*
	 * If there is no more data in the memory buffer then load
	 * some more from the temporary file.
	 */
	if (cc <= 0)
	{
		/*
		 * Read data back from the temporary file.
		 */
		if (!copyTemporaryFileToBuffer(dev))
			return UP_FALSE;

		/*
		 * Now we have some more data to read.
		 */
		cc = data->bufferWriteOffset;
	}

	/*
	 * Try to write the available data to the output buffer.
	 */
	cc = upWriteData(dev, data->buffer + data->bufferReadOffset, cc);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Increment the amount done.
	 */
	data->bufferReadOffset += cc;
	data->fileReadOffset += cc;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Function called when we had a short file to provide padding of
 * the output file so that it is the correct size.  We just provide
 * zero data until the amount written is correct.
 */
static UP_BOOL
padData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeCount;
	long		padCount;

	data = DEVTODATA(dev);

	writeBuffer = dev->writeBuffer;

	/*
	 * There is room to write some data.
	 * Get the minimum of the amount left to be padded
	 * and the amount that can be written now.
	 */
	padCount = data->targetFileSize - data->fileWriteOffset;

	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeCount = upBufferGetWriteLength(writeBuffer);

	if (writeCount > padCount)
		writeCount = (int) padCount;

	/*
	 * Set the data to zeroes and account for it.
	 */
	memset(writeAddress, 0, writeCount);

	upBufferWriteAdjust(writeBuffer, writeCount);

	data->fileWriteOffset += writeCount;
	padCount -= writeCount;

	/*
	 * If there is no more padding to do then set
	 * the end of file for the output buffer and
	 * make ourself ready for another file name.
	 */
	if (padCount <= 0)
	{
		upBufferWriteEof(writeBuffer);

		upSetState(dev, UP_STATE_TRANSFER_NAME);
	}

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Save the current data in our buffer into a temporary file.
 * This is called when there is no more room in the buffer for data.
 * The temporary file is created on the first call.
 */
static UP_BOOL
copyBufferToTemporaryFile(UP_DEV * dev)
{
	DATA *		data;
	char *		buffer;
	int		length;
	int		cc;

	data = DEVTODATA(dev);

	/*
	 * If the temporary file isn't opened yet, then do so.
	 */
	if (data->fd < 0)
	{
		data->fd = upCreateTemporaryFile(dev);

		if (data->fd < 0)
			return UP_FALSE;
	}

	/*
	 * Keep looping writing the buffer into the file until it
	 * is all stored.
	 */
	buffer = data->buffer;
	length = data->bufferWriteOffset;

	while (length > 0)
	{
		cc = write(data->fd, buffer, length);

		if (cc < 0)
		{
			upDeviceSystemError(dev, "write");

			return UP_FALSE;
		}

		buffer += cc;
		length -= cc;
	}

	/*
	 * Reset the buffer offset so that it can store more data.
	 */
	data->bufferWriteOffset = 0;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


/*
 * Load the data into our buffer from the temporary file.
 * This is called when the buffer is empty.
 */
static UP_BOOL
copyTemporaryFileToBuffer(UP_DEV * dev)
{
	DATA *		data;
	char *		buffer;
	off_t		fileLeftCount;
	int		length;
	int		cc;

	data = DEVTODATA(dev);

	/*
	 * If the temporary file isn't opened then something is wrong.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Get the amount of data left in the file.
	 */
	fileLeftCount = data->fileWriteOffset - data->fileReadOffset;

	/*
	 * Get the length of the data to be read, which is the
	 * minimum of the amount left in the file and the size
	 * of the buffer.
	 */
	length = data->bufferSize;

	if (length > fileLeftCount)
		length = (int) fileLeftCount;

	/*
	 * The length should be positive.
	 */
	if (length <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Keep looping reading the file data into the buffer until all
	 * of the indicated amount of data is read.
	 */
	buffer = data->buffer;

	while (length > 0)
	{
		/*
		 * Read some more.
		 */
		cc = read(data->fd, buffer, length);

		/*
		 * Check for a read error.
		 */
		if (cc < 0)
		{
			upDeviceSystemError(dev, "read");

			return UP_FALSE;
		}

		/*
		 * We should not get an end of file.
		 * This is because we are simply reading back the same
		 * data which we had previously written to the file.
		 */
		if (cc == 0)
		{
			upDeviceError(dev, UP_ERR_SHORTFILE);

			return UP_FALSE;
		}

		/*
		 * Adjust the buffer address and length by the amount read.
		 */
		buffer += cc;
		length -= cc;
	}

	/*
	 * Set the buffer write offset to the amount of data stored
	 * and clear the buffer read offset so the data can be read.
	 */
	data->bufferWriteOffset = buffer - data->buffer;
	data->bufferReadOffset = 0;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the data buffer.
	 */
	free(data->buffer);

	/*
	 * Close the temporary file.
	 */
	if ((data->fd >= 0) && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	return UP_TRUE;
}
