/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The grep device.
 * This looks for lines which contain the specified pattern.
 * Optionally, this selects the whole file if the pattern is found
 * anywhere within it.
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <regex.h>
#include <ctype.h>

#include "upPrivate.h"


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


/*
 * Extra space to allocate for line when it is reallocated.
 */
#define	ALLOC_LINE_SLOP	128


/*
 * The phases of writing data.
 */
typedef	enum
{
	PHASE_FILE_NAME,
	PHASE_COLON,
	PHASE_LINE_NUMBER,
	PHASE_LINE_DATA
} PHASE;


/*
 * Local data for this device.
 */
typedef	struct
{
	PHASE		phase;
	char *		pattern;
	int		patternLength;
	char *		line;
	int		lineLength;
	int		lineAlloc;
	long		lineNumber;
	const char *	writeAddress;
	int		writeCount;
	UP_BOOL		showFileNames;
	UP_BOOL		showLineNumbers;
	UP_BOOL		showNonMatches;
	UP_BOOL		ignoreCase;
	UP_BOOL		stringMatch;
	UP_BOOL		matchWholeWord;
	UP_BOOL		matchWholeLine;
	char		lineNumberBuffer[32];
	regex_t		regexBuffer;
} DATA;


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	readLine(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	UP_BOOL	checkMatch(UP_DEV *);
static	UP_BOOL	setWritePhase(UP_DEV *, PHASE);


static UP_FUNC	funcs =
{
	UP_TYPE_GREP,
	"Grep",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


UP_ID
upAddGrep(UP * up, const char * pattern, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;
	int		regexFlags;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialize our private data.
	 */
	data = DEVTODATA(dev);

	data->stringMatch = ((flags & UP_FLAG_GREP_STRING_MATCH) != 0);
	data->showFileNames = ((flags & UP_FLAG_GREP_SHOW_FILE_NAMES) != 0);
	data->showLineNumbers = ((flags & UP_FLAG_GREP_SHOW_LINE_NUMBERS) != 0);
	data->showNonMatches = ((flags & UP_FLAG_GREP_SHOW_NON_MATCHES) != 0);
	data->ignoreCase = ((flags & UP_FLAG_GREP_IGNORE_CASE) != 0);
	data->matchWholeWord = ((flags & UP_FLAG_GREP_MATCH_WHOLE_WORD) != 0);
	data->matchWholeLine = ((flags & UP_FLAG_GREP_MATCH_WHOLE_LINE) != 0);

	/*
	 * Copy the pattern string.
	 */
	data->pattern = upDeviceCopyString(dev, pattern);

	if (data->pattern == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	data->patternLength = strlen(pattern);

	/*
	 * Set the regular expression flags.
	 */
	regexFlags = REG_EXTENDED;

	if (data->ignoreCase)
		regexFlags |= REG_ICASE;

	/*
	 * Compile the pattern string if this isn't a string match.
	 */
	if (!data->stringMatch &&
		(regcomp(&data->regexBuffer, pattern, regexFlags) != 0))
	{
		upDeviceError(dev, UP_ERR_BADPATTERN);

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Set our initial state to transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name that can be transferred.
			 */
			upBufferReadFileName(readBuffer);

			upBufferWriteFileName(writeBuffer,
				upBufferGetFileName(readBuffer),
				upBufferGetReadFileStatus(readBuffer));

			/*
			 * Set our state to read a new line now.
			 */
			data->lineNumber = 0;
			data->lineLength = 0;

			upSetState(dev, UP_STATE_READ_LINE);

			break;

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * Read the end of buffer, set it on the output
			 * buffer, and say we are finished.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			return upSetFinished(dev);

		case UP_STATE_READ_LINE:
			/*
			 * There is a line that can be read.
			 */
			return readLine(dev);

		case UP_STATE_READ_EOF:
			/*
			 * There is an end of file to be read.
			 */
			upBufferReadEof(readBuffer);
			upBufferWriteEof(writeBuffer);

			/*
			 * Set our state to read another file.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		case UP_STATE_WRITE_DATA:
			/*
			 * We can write data to the output buffer.
			 */
			return writeData(dev);
	
		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read the next line (or partial line) from the input buffer and see
 * what we want to do with it.
 */
static UP_BOOL
readLine(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	line;
	char *		newLine;
	int		lineLength;
	int		totalLineLength;
	UP_BOOL		haveLine;
	UP_BOOL		hasNewLine;
	UP_BOOL		matchFlag;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the length of a line and its address, assuming
	 * that it is complete.
	 */
	line = upBufferGetReadAddress(readBuffer);
	lineLength = upBufferGetReadLineLength(readBuffer);
	haveLine = UP_TRUE;

	/*
	 * If the line is too long then get the amount of data
	 * which is available, and remember we don't have a full
	 * line yet.
	 */
	if (lineLength < 0)
	{
		lineLength = upBufferGetReadLength(readBuffer);
		haveLine = UP_FALSE;
	}

	/*
	 * There should be data present.
	 */
	if (lineLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If this data plus the amount already stored doesn't fit into
	 * our string buffer with a trailing null then reallocate it.
	 */
	totalLineLength = lineLength + data->lineLength;

	if (totalLineLength >= data->lineAlloc)
	{
		newLine = upDeviceRealloc(dev, data->line,
			totalLineLength + ALLOC_LINE_SLOP);

		if (newLine == NULL)
			return UP_FALSE;

		data->line = newLine;
		data->lineAlloc = totalLineLength + ALLOC_LINE_SLOP;
	}

	/*
	 * Append the line to the data already in the line (if any)
	 * and null-terminate it.
	 */
	memcpy(data->line + data->lineLength, line, lineLength);

	data->lineLength = totalLineLength;

	data->line[totalLineLength] = '\0';

	/*
	 * Account for the input data read.
	 */
	upBufferReadAdjust(readBuffer, lineLength);

	/*
	 * If this isn't all of the line then return success for now
	 * and wait for more.
	 */
	if (!haveLine)
		return upSetDidReadWrite(dev);

	/*
	 * The line is complete!  Count it.
	 */
	data->lineNumber++;

	/*
	 * See if it ends in a newline.
	 */
	hasNewLine = (data->line[totalLineLength - 1] == '\n');

	/*
	 * Trim off the newline in preparation for the match.
	 */
	if (hasNewLine)
	{
		data->line[totalLineLength - 1] = '\0';
		data->lineLength--;
	}

	/*
	 * See if the line is matched by the pattern string.
	 */
	matchFlag = checkMatch(dev);

	/*
	 * If we are showing non-matches then invert the result.
	 */
	if (data->showNonMatches)
		matchFlag = !matchFlag;

	/*
	 * If we don't want this line then reset the line to be empty
	 * and return to wait for the next line.
	 */
	if (!matchFlag)
	{
		data->lineLength = 0;

		return upSetDidReadWrite(dev);
	}

	/*
	 * The line is wanted.
	 * Restore the newline if required.
	 */
	if (hasNewLine)
		data->line[data->lineLength++] = '\n';

	/*
	 * Set the correct phase for writing the line to the output buffer.
	 */
	setWritePhase(dev, PHASE_FILE_NAME);

	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write data to the output buffer.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	int		cc;

	data = DEVTODATA(dev);

	/*
	 * Try to write some data.
	 */
	cc = upBufferWrite(dev->writeBuffer,
		data->writeAddress, data->writeCount);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Do the accounting.
	 */
	data->writeAddress += cc;
	data->writeCount -= cc;

	/*
	 * If there is still some left to write then return for now.
	 */
	if (data->writeCount > 0)
		return upSetDidReadWrite(dev);

	/*
	 * Switch on the phase that we were writing to change it
	 * to the next phase.
	 */
	switch (data->phase)
	{
		case PHASE_FILE_NAME:
			if (!setWritePhase(dev, PHASE_COLON))
				return UP_FALSE;

			break;

		case PHASE_COLON:
			if (!setWritePhase(dev, PHASE_LINE_NUMBER))
				return UP_FALSE;

			break;

		case PHASE_LINE_NUMBER:
			if (!setWritePhase(dev, PHASE_LINE_DATA))
				return UP_FALSE;

			break;

		case PHASE_LINE_DATA:
			/*
			 * Change state back to reading another line.
			 */
			data->lineLength = 0;

			upSetState(dev, UP_STATE_READ_LINE);

			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Setup to write the appropriate data for the specified phase of writing.
 * If the specified phase is not desired, then the next one which is
 * desired is picked instead.
 */
static UP_BOOL
setWritePhase(UP_DEV * dev, PHASE phase)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if ((phase == PHASE_FILE_NAME) && !data->showFileNames)
		phase = PHASE_LINE_NUMBER;

	if (phase == PHASE_LINE_NUMBER && !data->showLineNumbers)
		phase = PHASE_LINE_DATA;

	switch (phase)
	{
		case PHASE_FILE_NAME:
			data->writeAddress =
				upBufferGetFileName(dev->readBuffer);

			data->writeCount = strlen(data->writeAddress);

			break;

		case PHASE_COLON:
			data->writeAddress = ": ";
			data->writeCount = 2;

			break;

		case PHASE_LINE_NUMBER:
			/*
			 * Generate the line number and point at it.
			 */
			sprintf(data->lineNumberBuffer, "%ld: ",
				data->lineNumber);

			data->writeAddress = data->lineNumberBuffer;
			data->writeCount = strlen(data->lineNumberBuffer);

			break;

		case PHASE_LINE_DATA:
			/*
			 * Point at the line data.
			 */
			data->writeAddress = data->line;
			data->writeCount = data->lineLength;

			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	data->phase = phase;

	return UP_TRUE;
}


/*
 * Check whether the current line is matched by the pattern string.
 * Returns UP_TRUE if so.
 */
static UP_BOOL
checkMatch(UP_DEV * dev)
{
	DATA *		data;
	int		result;
	int		ch;
	const char *	line;
	regmatch_t	matchInfo[1];

	data = DEVTODATA(dev);

	line = data->line;

	/*
	 * If we aren't doing a string match then see if the
	 * regular expression matches.
	 */
	if (!data->stringMatch)
	{
		result = regexec(&data->regexBuffer, line, 1, matchInfo, 0);

		/*
		 * If there was not a match at all then return that.
		 */
		if (result != 0)
			return UP_FALSE;

		/*
		 * If we must match the whole line then see if that is
		 * the case.
		 */
		if (data->matchWholeLine)
		{
			if (matchInfo[0].rm_so != 0)
				return UP_FALSE;

			if (matchInfo[0].rm_eo != strlen(line))
				return UP_FALSE;
		}

		/*
		 * If we must match a whole word then see if that is
		 * the case.  Here a word is an alphabetic, digit,
		 * or underscore.
		 */
		if (data->matchWholeWord)
		{
			if (matchInfo[0].rm_so > 0)
			{
				ch = line[matchInfo[0].rm_so - 1];

				if (isalnum(ch) || (ch == '_'))
					return UP_FALSE;
			}

			if (matchInfo[0].rm_eo < strlen(line))
			{
				ch = line[matchInfo[0].rm_eo];

				if (isalnum(ch) || (ch == '_'))
					return UP_FALSE;
			}
		}

		/*
		 * The pattern matches.
		 */
 		return UP_TRUE;
	}

	/*
	 * This is a string match.
	 * If this is case sensitive then it is easy.
	 */
	if (!data->ignoreCase)
		return (strstr(line, data->pattern) != 0);

	/*
	 * We must do a case insensitive match.
	 * TODO:
	 */
	UP_DEVICE_INTERNAL_ERROR(dev);

	return UP_FALSE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the regular expression buffer if we weren't doing
	 * a string match.
	 */
	if (!data->stringMatch)
		regfree(&data->regexBuffer);

	/*
	 * Free the pattern string.
	 */
	free(data->pattern);

	/*
	 * Free the line buffer.
	 */
	free(data->line);

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	return UP_TRUE;
}
