/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The gunzip pseudo-devices.
 * This uncompresses the input data using the gzip algorithm.
 * This calls the zlib library by Jean-loup Gailly and Mark Adler.
 */

#include "upPrivate.h"


#ifndef	USE_ZLIB


/*
 * Add the "gunzip" pseudo-device.
 */
UP_ID
upAddGnuUnzip(UP * up)
{
	return upAddProcessString(up, "gzip -d -q", UP_FLAG_NONE, NULL);
}


#else


#include <zlib.h>


/*
 * Phases of input when we are reading special data from the input buffer.
 * We have to read the various gzip header fields and the gzip trailer.
 * The gzip header is a pain to parse because it has several optional
 * variable-sized fields.  I sure wish zlib had a in-memory function to
 * do all of the gzip format for us!
 */
typedef enum
{
	PHASE_MAGIC,
	PHASE_FLAGS,
	PHASE_OPTIONS,
	PHASE_EXTRA_FIELD_COUNT,
	PHASE_STRING,
	PHASE_TRAILER
} PHASE;


/*
 * The first two bytes of a file which identify it as a gzip archive.
 */
#define	MAGIC_1	0x1f
#define	MAGIC_2	0x8b


/*
 * Flag bits in the gzip header which enable extra header information.
 * Based on these bits being present, extra information has to be read
 * and discarded from the gzip header.
 */
#define	FLAG_ASCII		0x01		/* file is probably ASCII */
#define	FLAG_HEADER_CRC		0x02		/* has header CRC */
#define	FLAG_EXTRA_FIELD	0x04		/* has "extra field" */
#define	FLAG_ORIGINAL_NAME	0x08		/* has original file name */
#define	FLAG_COMMENT		0x10		/* has comment field */
#define	FLAG_RESERVED		0xe0		/* reserved flag bits */


/*
 * Local data for this device.
 */
typedef	struct
{
	PHASE		phase;		/* current uncompresssion phase */
	char *		dataAddress;	/* address of data left to read */
	int		dataLength;	/* length of data left to read */
	int		headerBytes;	/* bytes of header written */
	int		flags;		/* flags byte from header */
	char		data[12];	/* data that has been read */
	z_stream	stream;		/* zlib uncompression stream */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL transferFileName(UP_DEV * dev);
static	UP_BOOL transferData(UP_DEV * dev);
static	UP_BOOL	readData(UP_DEV * dev);
static	UP_BOOL readEndOfFile(UP_DEV * dev);


static	UP_FUNC	funcs =
{
	UP_TYPE_GUNZIP,
	"Gunzip",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Translation table for file name extensions when uncompressing.
 * The input and output extensions form pairs, and the table is terminated
 * by a NULL string.
 */
static const char * const translateTable[] =
{
	".gz",	"",
	".tgz",	".tar",
	NULL,	NULL
};


UP_ID
upAddGnuUnzip(UP * up)
{
	UP_DEV *	dev;
	DATA *		data;
	z_streamp	stream;
	int		status;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);
	data->phase = PHASE_MAGIC;
	data->dataAddress = NULL;
	data->dataLength = 0;

	/*
	 * Set our state to start by transferring a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Initialise the zlib structure.
	 */
	stream = &data->stream;

	stream->zalloc = Z_NULL;
	stream->zfree = Z_NULL;
	stream->opaque = Z_NULL;
	stream->next_in = "";
	stream->avail_in = 0;
	stream->next_out = "";
	stream->avail_out = 0;

	/*
	 * Call the zlib library to prepare for uncompression.
	 */
	status = inflateInit2(stream, -MAX_WBITS);

	if (status != Z_OK)
	{
		upError(up, UP_ERR_ZLIBFAILED);

		return UP_NULL_ID;
	}

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	int		state;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred.
			 */
			return transferFileName(dev);

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * The end of the input buffer has been seen.
			 * Acknowledge it and pass it on to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say we are finished now.
			 */
			return upSetFinished(dev);

		case UP_STATE_READ_DATA:
			/*
			 * We can read special data from the input buffer.
			 */
			return readData(dev);

		case UP_STATE_READ_EOF:
			/*
			 * We had an end of file while reading special data.
			 */
			return readEndOfFile(dev);

		case UP_STATE_TRANSFER_DATA:
			/*
			 * We can transfer some data from the input to
			 * the output buffer.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_EOF:
			/*
			 * We got an end of file while uncompressing.
			 */
			upDeviceError(dev, UP_ERR_SHORTFILE);

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Say we have done I/O.
	 */
	upSetDidReadWrite(dev);
}


/*
 * Routine called when reading the file name from the input buffer.
 * This translates the input name according to gzip convenstions, and sets
 * that new file name for the output buffer.  Returns UP_TRUE on success.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	inputFileName;
	const char *	outputFileName;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Read the input file name.
	 */
	upBufferReadFileName(readBuffer);

	inputFileName = upBufferGetFileName(readBuffer);

	if (inputFileName == NULL)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Translate the input file name according to our table.
	 */
	outputFileName = upTranslateFileName(dev, inputFileName,
		translateTable);

	if (outputFileName == NULL)
		return UP_FALSE;

	/*
	 * Tell the output buffer the output file name.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer, outputFileName, NULL))
		return UP_FALSE;

	/*
	 * OK, set the phase to read the magic header bytes.
	 */
	data->phase = PHASE_MAGIC;
	data->dataAddress = data->data;
	data->dataLength = 2;

	upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to transfer data from the input to the output buffer.
 * The input is the compressed file data, and the output is the
 * uncompressed data.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	z_streamp	stream;
	const char *	readAddress;
	char *		writeAddress;
	int		readLength;
	int		writeLength;
	int		status;
	int		readCount;
	int		writeCount;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;
	stream = &data->stream;

	/*
	 * Get the addresses and lengths of the input and output buffers.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Make sure that we have some input data and some room for
	 * output data.
	 */
	if ((readLength <= 0) || (writeLength <= 0))
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Set up the read and write buffers for the zlib call.
	 */
	stream->next_in = (char *) readAddress;
	stream->avail_in = readLength;
	stream->next_out = writeAddress;
	stream->avail_out = writeLength;

	/*
	 * Call zlib to try to do some decompression.
	 */
	status = inflate(stream, 0);

	/*
	 * If the decompression was not ok and was not the end of the stream
	 * then complain about an error.
	 */
	if ((status != Z_OK) && (status != Z_STREAM_END))
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	/*
	 * Get the amount of data read and written.
	 */
	readCount = readLength - stream->avail_in;
	writeCount = writeLength - stream->avail_out;

	/*
	 * Update the input and output buffers by the amount of data that
	 * the zlib call processed.
	 */
	upBufferReadAdjust(readBuffer, readCount);
	upBufferWriteAdjust(writeBuffer, writeCount);

	/*
	 * Check for all the output data being complete from zlib.
	 * If so, then change the phase to read the trailer CRC bytes next.
	 */
	if (status == Z_STREAM_END)
	{
		data->phase = PHASE_TRAILER;
		data->dataAddress = NULL;
		data->dataLength = 2;

		upSetState(dev, UP_STATE_READ_DATA);
	}

	/*
	 * We did some successful I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Here when an end of file can be read from the input buffer.
 * We make sure that it occurred at the correct time, mark the end of
 * file for the output buffer, and set up to read the next file name.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Make sure we got the end of file at the proper phase.
	 */
	if (data->phase != PHASE_TRAILER)
	{
		upDeviceError(dev, UP_ERR_SHORTFILE);

		return UP_FALSE;
	}

	/*
	 * Acknowledge that we have read the end of file.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * Mark the output buffer as having an end of file.
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Reset our state to be ready for another file name.
	 */
	if (!doReset(dev))
		return UP_FALSE;

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when reading data which is not the compressed data stream.
 * This parses the header and trailer information from the gzip file.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	readAddress;
	int		readLength;
	int		readCount;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the address and length of the available data.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * Keep examining the available data until there is no more to
	 * process or until we decide to quit looking for special data.
	 */
	while (readLength > 0)
	{
		/*
		 * Calculate the amount of data which we can exammine.
		 * This is the minimum of the amount desired and the
		 * amount available.
		 */
		readCount = readLength;

		if (readCount > data->dataLength)
			readCount = data->dataLength;

		/*
		 * If the incoming data should be stored then do that
		 * and advance the pointer address past it.  Otherwise
		 * the incoming data will just be ignored.
		 */
		if (data->dataAddress != NULL)
		{
			memcpy(data->dataAddress, readAddress, readCount);

			data->dataAddress += readCount;
		}

		/*
		 * Do accounting for the reading that we did.
		 */
		upBufferReadAdjust(readBuffer, readCount);

		upSetDidReadWrite(dev);

		readAddress += readCount;
		readLength -= readCount;
		data->dataLength -= readCount;

		/*
		 * If there is still some more data to be read for this
		 * phase then we can do nothing more now.
		 */
		if (data->dataLength > 0)
			break;

		/*
		 * All the data has been collected for this phase so that
		 * it is complete.  Process the data that was collected
		 * and then advance to the next phase.
		 */
		switch (data->phase)
		{
		case PHASE_MAGIC:
			/*
			 * This phase collected the first two bytes of the
			 * file.  Check that the magic values are correct.
			 */
			if (((data->data[0] & 0xff) != MAGIC_1) ||
				((data->data[1] & 0xff) != MAGIC_2))
			{
				upDeviceError(dev, UP_ERR_NOTGZIP);

				return UP_FALSE;
			}

			/*
			 * Now set the phase to collect the compression
			 * type and flag bytes.
			 */
			data->phase = PHASE_FLAGS;
			data->dataAddress = data->data;
			data->dataLength = 2;

			break;


		case PHASE_FLAGS:
			/*
			 * This phase collected the compression type and flag
			 * bytes.  Save the flags byte since it specifies
			 * which of the optional header fields are present.
			 */
			data->flags = data->data[1];

			/*
			 * Make sure that we understand the compression type
			 * and that no unknown flags are set.
			 */
			if (((data->data[0] & 0xff) != Z_DEFLATED) ||
				((data->flags & FLAG_RESERVED) != 0))
			{
				upDeviceError(dev, UP_ERR_GZIPRESERVED);

				return UP_FALSE;
			}

			/*
			 * OK, set up to read (and ignore) the last six
			 * bytes of the fixed header, and then enter the
			 * checking of the options phase.
			 */
			data->phase = PHASE_OPTIONS;
			data->dataAddress = NULL;
			data->dataLength = 6;

			break;


		case PHASE_OPTIONS:
			/*
			 * This phase is possibly entered multiple times to
			 * examine the flags byte for optional header stuff.
			 * Check the flags byte to see which of the optional
			 * header information is present.  When we find
			 * something then turn its flag off so we won't do
			 * it again.  Warning: The order of the following
			 * checks for the optional header information is
			 * important.
			 */
			if ((data->flags & FLAG_EXTRA_FIELD) != 0)
			{
				/*
				 * Set up to read the byte count for the
				 * extra header field, and to enter that
				 * special phase when it is done.
				 */
				data->flags &= ~FLAG_EXTRA_FIELD;
				data->phase = PHASE_EXTRA_FIELD_COUNT;
				data->dataAddress = data->data;
				data->dataLength = 2;

				break;
			}

			if ((data->flags & FLAG_ORIGINAL_NAME) != 0)
			{
				/*
				 * Set up to read and ignore a null-terminated
				 * string which is the original file name.
				 */
				data->flags &= ~FLAG_ORIGINAL_NAME;
				data->phase = PHASE_STRING;
				data->dataLength = 1;

				break;
			}

			if ((data->flags & FLAG_COMMENT) != 0)
			{
				/*
				 * Set up to read and ignore a null-terminated
				 * comment string.
				 */
				data->flags &= ~FLAG_COMMENT;
				data->phase = PHASE_STRING;
				data->dataLength = 1;

				break;
			}

			if ((data->flags & FLAG_HEADER_CRC) != 0)
			{
				/*
				 * Set up to read and ignore a header CRC.
				 */
				data->flags &= FLAG_HEADER_CRC;
				data->phase = PHASE_OPTIONS;
				data->dataLength = 2;

				break;
			}

			/*
			 * There are no more optional header fields left.
			 * Now we can begin reading of the actual
			 * compressed data in the main code!
			 */
			upSetState(dev, UP_STATE_TRANSFER_DATA);

			return upSetDidReadWrite(dev);


		case PHASE_EXTRA_FIELD_COUNT:
			/*
			 * Set up to read and ignore the specified number of
			 * bytes, and then return back to the options phase.
			 */
			data->phase = PHASE_OPTIONS;
			data->dataAddress = NULL;

			data->dataLength = ((data->data[0] & 0xff) |
				((data->data[1] & 0xff) << 8));

			break;


		case PHASE_STRING:
			/*
			 * This phase is used to read (and ignore) a null
			 * terminated string byte by byte.  If we haven't
			 * seen the terminating null character, then
			 * remain in this phase for the next byte.
			 */
			if (data->data[0] != '\0')
			{
				data->dataAddress = data->data;
				data->dataLength = 1;

				break;
			}

			/*
			 * The terminating null has been seen.
			 * Change the phase back to examine the options again.
			 */
			data->phase = PHASE_OPTIONS;
			data->dataAddress = NULL;
			data->dataLength = 0;
	
			break;


		case PHASE_TRAILER:
			/*
			 * This phase is used when the decompression is done.
			 * Ignore all data that is remaining until the
			 * end of file is read.
			 */
			upBufferReadAdjust(readBuffer, readLength);
			readLength = 0;

			break;


		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
		}
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Tell the zlib library to reset itself in preparation for
	 * another decompression.
	 */
	if (inflateReset(&data->stream) != Z_OK)
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	/*
	 * Reset the other information about the compression.
	 */
	data->phase = PHASE_MAGIC;
	data->dataAddress = NULL;
	data->dataLength = 0;

	upSetState(dev, UP_STATE_TRANSFER_NAME);

	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;
	int	status;
	
	data = DEVTODATA(dev);

	/*
	 * Close the zlib stream.
	 */
	status = inflateEnd(&data->stream);

	if (status != Z_OK)
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	return UP_TRUE;
}

#endif
