/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The gzip device.
 * This compresses the input data using the gzip algorithm.
 * This calls the zlib library by Jean-loup Gailly and Mark Adler.
 */

#include "upPrivate.h"


#ifndef	USE_ZLIB


/*
 * Add the "gzip" pseudo-device.
 * It accepts a level of compression.
 * The level zero means use the default level.
 */
UP_ID
upAddGnuZip(UP * up, int level)
{
	char	cmd[80];

	if ((level < 0) || (level > 9)) {
		upError(up, UP_ERR_BADGZIPLEVEL);

		return UP_NULL_ID;
	}

	if (level > 0)
		sprintf(cmd, "gzip -q -%d", level);
	else
		strcpy(cmd, "gzip -q");

	return upAddProcessString(up, cmd, UP_FLAG_NONE, NULL);
}

#else


#include <zlib.h>


/*
 * The first few bytes of a gzip file which is a header.
 * These are:
 *	0-1:	magic number
 *	2:	compression type
 *	3:	flags
 *	4-7:	time
 *	8:	xflags
 *	9:	OS code
 */
#define	HEADER_SIZE	10
#define	TRAILER_SIZE	8

static	char	gzipHeader[HEADER_SIZE] =
{
	0x1f, 0x8b, Z_DEFLATED, 0, 0, 0, 0, 0, 0, 0x03
};


/*
 * Phases of writing when we are not transferring data.
 * We have to output the gzip header, the zlib produced compressed data,
 * and then the gzip trailer.  I sure wish zlib had a in-memory function
 * to do all of the gzip format for us!
 */
typedef enum
{
	PHASE_HEADER,
	PHASE_REMAINDER,
	PHASE_TRAILER
} PHASE;


/*
 * Local data for this device.
 */
typedef	struct
{
	PHASE		phase;		/* current compresssion phase */
	int		level;		/* compression level */
	const char *	dataAddress;	/* address of data left to write */
	int		dataLength;	/* length of data left to write */
	int		headerBytes;	/* bytes of header written */
	unsigned long	crc;		/* crc value */
	char		trailer[TRAILER_SIZE];	/* trailer bytes */
	z_stream	stream;		/* zlib compression stream */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	transferFileName(UP_DEV *);
static	UP_BOOL	transferData(UP_DEV *);
static	UP_BOOL	transferEndOfFile(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_GZIP,
	"Gzip",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Translation table for file name extensions when compressing.
 * The input and output extensions form pairs, and the table is terminated
 * by a NULL string.
 */
static const char * const translateTable[] =
{
	".tar",	".tgz",
	"",	".gz",
	NULL,	NULL
};


UP_ID
upAddGnuZip(UP * up, int level)
{
	UP_DEV *	dev;
	DATA *		data;
	z_streamp	stream;
	int		status;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);
	data->level = level;
	data->phase = PHASE_HEADER;
	data->dataAddress = NULL;
	data->dataLength = 0;
	data->crc = crc32(0L, Z_NULL, 0);

	/*
	 * Set our state to transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Initialise the zlib structure.
	 */
	stream = &data->stream;

	stream->zalloc = Z_NULL;
	stream->zfree = Z_NULL;
	stream->opaque = Z_NULL;
	stream->next_in = "";
	stream->avail_in = 0;
	stream->next_out = "";
	stream->avail_out = 0;

	/*
	 * Call the zlib library to prepare for compression.
	 * We have to call the more complicated init routine in order
	 * to use the *UNDOCUMENTED* feature to suppress the zlib header,
	 * and then apply our own gzip header and trailers.  This is
	 * really ugly.  Zlib should provide a gzip interface for
	 * read/write similar to the low level zlib interface.
	 */
	status = deflateInit2(stream, level, Z_DEFLATED, -MAX_WBITS, 8, 0);

	if (status != Z_OK)
	{
		upError(up, UP_ERR_ZLIBFAILED);

		return UP_NULL_ID;
	}

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	int		state;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred.
			 */
			return transferFileName(dev);

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * The end of the input buffer has been seen.
			 * Acknowledge it and pass it on to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say we are finished now.
			 */
			return upSetFinished(dev);

		case UP_STATE_TRANSFER_DATA:
			/*
			 * We can transfer some data from the input to
			 * the output buffer.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_EOF:
			/*
			 * An end of file has been read from the input file.
			 */
			return transferEndOfFile(dev);

		case UP_STATE_WRITE_DATA:
			/*
			 * There is room to write header or trailer data,
			 * or the final bit of compressed data.
			 */
			return writeData(dev);

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Say we have done I/O.
	 */
	upSetDidReadWrite(dev);
}


/*
 * Routine called when reading the file name from the input buffer.
 * This translates the input name according to gzip convenstions, and sets
 * that new file name for the output buffer.  Returns UP_TRUE on success.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	inputFileName;
	const char *	outputFileName;
	UP_STAT		statBuf;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the input file name and a copy of the status.
	 */
	upBufferReadFileName(readBuffer);

	inputFileName = upBufferGetFileName(readBuffer);
	statBuf = *upBufferGetReadFileStatus(readBuffer);

	/*
	 * Translate the input file name according to our table.
	 */
	outputFileName = upTranslateFileName(dev, inputFileName,
		translateTable);

	if (outputFileName == NULL)
		return UP_FALSE;

	/*
	 * Clear the the file size since we don't know what it will be.
	 * Leave the execute bits set in the mode even though that is
	 * senseless, because that is what the gzip program itself does.
	 * Force the file type to be a regular file.
	 */
	statBuf.flags &= ~UP_STAT_SIZE;
	statBuf.flags |= UP_STAT_MODE;
	statBuf.type = '-';
	statBuf.mode = (statBuf.mode & 0777) | S_IFREG;
	statBuf.size = 0;

	/*
	 * Tell the output buffer the output file name.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer, outputFileName, &statBuf))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * Tell the zlib library to reset itself
	 * in preparation for another compression.
	 */
	if (deflateReset(&data->stream) != Z_OK)
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	/*
	 * Initialise our data for the next file.
	 */
	data->phase = PHASE_HEADER;
	data->dataAddress = gzipHeader;
	data->dataLength = HEADER_SIZE;
	data->crc = crc32(0L, Z_NULL, 0);

	/*
	 * OK, set the phase to send the gzip file header next.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Run one iteration of transferring data through zlib.
 * This handles both the case of input data being available
 * and the case when the input file has reached end of file.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	z_streamp	stream;
	const char *	readAddress;
	char *		writeAddress;
	int		flushFlag;
	int		status;
	int		readLength;
	int		readCount;
	int		writeLength;
	int		writeCount;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;
	stream = &data->stream;

	/*
	 * Get the address and length of the input buffer,
	 * or else a zero length if the phase indicates that
	 * we have had end of file on the input.
	 */
	readAddress = "";
	readLength = 0;

	if (data->phase != PHASE_REMAINDER)
	{
		readAddress = upBufferGetReadAddress(readBuffer);
		readLength = upBufferGetReadLength(readBuffer);

		/*
		 * Make sure there is some input data.
		 */
		if (readLength <= 0)
		{
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
		}
	}

	/*
	 * Get the address and length of the output buffer.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Make sure there is some room in the output buffer.
	 */
	if (writeLength <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * We have (possibly) some input data and room for output.
	 * Set up the read and write buffers for the zlib call.
	 */
	stream->next_in = (char *) readAddress;
	stream->avail_in = readLength;
	stream->next_out = writeAddress;
	stream->avail_out = writeLength;

	/*
	 * If there is no data for input then set the finish flag for zlib.
	 */
	flushFlag = 0;

	if (readLength == 0)
		flushFlag = Z_FINISH;

	/*
	 * Call zlib to try to do some compression.
	 */
	status = deflate(stream, flushFlag);

	/*
	 * If the compression was not ok and was not the end of the stream
	 * then complain about an error.
	 */
	if ((status != Z_OK) && (status != Z_STREAM_END))
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	/*
	 * Get the amount of data read and written.
	 */
	readCount = readLength - stream->avail_in;
	writeCount = writeLength - stream->avail_out;

	/*
	 * Update the CRC value for the data that was read.
	 */
	if (readCount > 0)
	{
		data->crc = crc32(data->crc,
			(const unsigned char *) readAddress, readCount);
	}

	/*
	 * Update the input and output buffers by the amount of data that
	 * the zlib call processed.  If there is no input data, then do
	 * nothing to the input buffer since we are called after the end
	 * of file in that case.
	 */
	if (readLength != 0)
		upBufferReadAdjust(readBuffer, readCount);

	upBufferWriteAdjust(writeBuffer, writeCount);

	/*
	 * If zlib has not yet finished writing all of its data to the
	 * output buffer, then we are done for now.
	 */
	if (status != Z_STREAM_END)
		return upSetDidReadWrite(dev);

	/*
	 * The compressed data is fully written.
	 * Calculate the trailer bytes and set up the phase to write
	 * them to the output device.  Store the CRC and input byte
	 * count values into the trailer buffer.
	 */
	writeAddress = data->trailer;

	*writeAddress++ = data->crc;
	*writeAddress++ = data->crc >> 8;
	*writeAddress++ = data->crc >> 16;
	*writeAddress++ = data->crc >> 24;

	*writeAddress++ = stream->total_in;
	*writeAddress++ = stream->total_in >> 8;
	*writeAddress++ = stream->total_in >> 16;
	*writeAddress++ = stream->total_in >> 24;

	/*
	 * Set up to write those values to the output device.
	 */
	data->phase = PHASE_TRAILER;
	data->dataAddress = data->trailer;
	data->dataLength = TRAILER_SIZE;

	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * We did some successful I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when end of file has been read from the input buffer.
 * Note: We don't actually set end of file on the output buffer yet since
 * we still have more data to write.
 */
static UP_BOOL
transferEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Acknowledge that we have read the end of file from
	 * the input buffer.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * Set our state so that we will write the last bit of the
	 * compressed data to the output buffer.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	data->phase = PHASE_REMAINDER;
	data->dataAddress = NULL;
	data->dataLength = 0;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Here when we are writing some data to the output buffer.
 * This is used for writing the final bit of compressed data
 * or when writing the file header or trailer information.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		cc;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * If we are in the PHASE_REMAINDER phase then the data we need
	 * to write is the final bit of compressed data remaining in
	 * the zlib buffer after the input had reached end of file.
	 * This case is handled specially by transferData.
	 */
	if (data->phase == PHASE_REMAINDER)
		return transferData(dev);

	/*
	 * We are writing our own data that has been stored by us.
	 * Write as much of it as we can to the output buffer.
	 */
	cc = upWriteData(dev, data->dataAddress, data->dataLength);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Adjust our buffer values by the amount written.
	 */
	data->dataAddress += cc;
	data->dataLength -= cc;

	/*
	 * If there is any data left to be written then just return
	 * saying we have done some I/O.
	 */
	if (data->dataLength > 0)
		return upSetDidReadWrite(dev);

	/*
	 * The special data bas been totally written.
	 * Switch on the phase of the data to see what to do now.
	 */
	switch (data->phase)
	{
		case PHASE_HEADER:
			/*
			 * This was the initial header.
			 * Set our state to transfer actual data now.
			 */
			upSetState(dev, UP_STATE_TRANSFER_DATA);

			break;

		case PHASE_TRAILER:
			/*
			 * This was the trailer data.
			 * Write an end of file to the output buffer.
			 */
			upBufferWriteEof(writeBuffer);

			/*
			 * Set our state to read another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Remember that we did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Tell the zlib library to reset itself in preparation for
	 * another compression.
	 */
	if (deflateReset(&data->stream) != Z_OK)
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	/*
	 * Reset the other information about the compression.
	 */
	data->phase = PHASE_HEADER;
	data->dataAddress = NULL;
	data->dataLength = 0;
	data->crc = crc32(0L, Z_NULL, 0);

	upSetState(dev, UP_STATE_TRANSFER_NAME);

	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;
	int	status;
	
	data = DEVTODATA(dev);

	/*
	 * Close the zlib stream.
	 * Ignore errors caused by discarding data.
	 */
	status = deflateEnd(&data->stream);

	if ((status != Z_OK) && (status != Z_DATA_ERROR))
	{
		upDeviceError(dev, UP_ERR_ZLIBFAILED);

		return UP_FALSE;
	}

	return UP_TRUE;
}

#endif
