/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The join device.
 * This device joins two user mode pipelines together so that one feeds
 * data into another.  The connection can later be removed or changed
 * as the application requires.
 */

#include "upPrivate.h"


#define DEVTOUP(dev)  ((UP *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	int	doread(UP_DEV *, char *, int);
static	int	dowrite(UP_DEV *, const char *, int);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_JOIN,
	"Join",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Add a second pipeline in as a component of the current pipeline.
 * It is allowable to add a NULL pipeline, in which case this device
 * will act like the null device.  The pipeline which is added can be
 * changed later.
 */
UP_ID
upAddJoin(UP *	up, UP * upjoin)
{
	UP_DEV	*	dev;

	/*
	 * If there is a joining pipeline then make sure that its mode
	 * matches the one it is connecting to.
	 */
	if (upjoin && (upjoin->mode != up->mode))
	{
		upError(up, UP_ERR_MODEMISMATCH);

		return UP_NULL_ID;
	}

	dev = upAllocDevice(up, &funcs, 0);

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	dev->private = (char *) upjoin;

	return upLinkDevice(up, dev);
}


/*
 * Change the joined pipeline to the specified one (which could be null).
 * The old one is returned unchanged if a non-null pointer is specified.
 * If the old pipeline is not returned, then it is closed.
 * Returns 0 on success, or -1 on error.
 */
int
upChangeJoin(UP * up, UP_ID id, UP * upjoin, UP ** ret_old_upjoin)
{
	UP_DEV	*	dev;
	UP *		old_upjoin;

	if (ret_old_upjoin)
		*ret_old_upjoin = NULL;

	dev = upFindDevice(up, id, UP_TYPE_JOIN);

	if (dev == UP_NULL_DEV)
		return -1;

	old_upjoin = DEVTOUP(dev);

	/*
	 * If there is a joining pipeline then make sure that its mode
	 * matches the one it is connecting to.
	 */
	if (upjoin && (upjoin->mode != up->mode))
	{
		upError(up, UP_ERR_MODEMISMATCH);

		return UP_NULL_ID;
	}

	/*
	 * If the old pipeline is to be returned, then do that.
	 */
	if (ret_old_upjoin)
		*ret_old_upjoin = old_upjoin;

	/*
	 * If there is an old pipeline and it is not being returned,
	 * then close it.
	 */
	if (old_upjoin && (ret_old_upjoin == NULL))
	{
		dev->private = NULL;

		if (upClose(old_upjoin) < 0)
			return -1;
	}

	/*
	 * Set the new pipeline into the device for future use.
	 */
	dev->private = (char *) upjoin;

	return 0;
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	return UP_TRUE;
}


static int
doread(UP_DEV * dev, char * buf, int len)
{
	UP *	up;

	up = DEVTOUP(dev);

	if (up == NULL)
		return 0;

	return upRead(up, buf, len);
}


static int
dowrite(UP_DEV * dev, const char * buf, int len)
{
	UP *	up;

	up = DEVTOUP(dev);

	if (up == NULL)
		return len;

	return upWrite(up, buf, len);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	UP *	up;

	up = DEVTOUP(dev);
	dev->private = NULL;

	if (up == NULL)
		return UP_TRUE;

	return upClose(up);
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	return UP_TRUE;
}
