/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The list file device.
 * This lists the names of files being sent through the pipeline.
 * The list is sent to the specified file name.
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>

#include "upPrivate.h"


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL transferFileName(UP_DEV * dev);

static	void	writeListing(
	UP_DEV * dev, const char * fileName, const UP_STAT * status);


static UP_FUNC	funcs =
{
	UP_TYPE_LIST,
	"List",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef struct
{
	UP_BOOL		listGivenSize;
	UP_BOOL		listTrueSize;
	UP_BOOL		listMode;
	off_t		trueSize;
	FILE *		fp;
} DATA;

#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


UP_ID
upAddList(UP * up, const char * listFileName, UP_FLAG listFlags)
{
	UP_DEV *	dev;
	DATA *		data;
	FILE *		fp;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Open the listing file.
	 */
	fp = fopen(listFileName, "w");

	if (fp == NULL)
	{
		upSystemError(up, "fopen");

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);

	data->listMode = ((listFlags & UP_FLAG_LIST_MODE) != 0);
	data->listGivenSize = ((listFlags & UP_FLAG_LIST_GIVEN_SIZE) != 0);
	data->listTrueSize = ((listFlags & UP_FLAG_LIST_TRUE_SIZE) != 0);
	data->fp = fp;

	/*
	 * Set our initial state so as to transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;
	int		count;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Transfer as much data as we can.
			 * Our state remains the same.
			 */
			count = upTransferData(dev, 0);

			if (count < 0)
				return UP_FALSE;

			/*
			 * Update the true file size.
			 */
			data->trueSize += count;

			break;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred from
			 * the input to the output buffer.
			 */
			if (!transferFileName(dev))
				return UP_FALSE;

			/*
			 * Set our state to transfer data.
			 */
			upSetState(dev, UP_STATE_TRANSFER_DATA);

			break;

		case UP_STATE_TRANSFER_EOF:
			/*
			 * There is an end of file to be transferred
			 * from the input to the output buffer.
			 * If we were listing the true file size
			 * then do so now.
			 */
			if (data->listTrueSize)
			{
				writeListing(dev,
					upBufferGetFileName(readBuffer),
					upBufferGetReadFileStatus(readBuffer));
			}

			/*
			 * Acknowledge the end of file and pass it on.
			 */
			upBufferReadEof(readBuffer);
			upBufferWriteEof(writeBuffer);

			/*
			 * Set our state to transfer another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * There is an end of buffer to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say that we are finished.
			 */
			upSetFinished(dev);

			break;

		case UP_STATE_TRANSFER_FILE_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer the file name information from the input to output
 * buffer, and write its name and information to the listing file.
 * Returns UP_TRUE if successful.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *			data;
	UP_BUF *		readBuffer;
	const char *		fileName;
	const UP_STAT *		status;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the file name and status.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	status = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Write the listing to the log file now if we don't
	 * need the true file size.
	 */
	if (!data->listTrueSize)
		writeListing(dev, fileName, status);

	/*
	 * Initialize the true file size.
	 */
	data->trueSize = 0;

	/*
	 * Tell the output buffer the file name and status.
	 */
	upBufferWriteFileName(dev->writeBuffer, fileName, status);

	/*
	 * We did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write the file listing to the log file.
 * What is written is the available information for the file
 * and what was indicated by the initial flags.
 */
static void
writeListing(UP_DEV * dev, const char * fileName, const UP_STAT * status)
{
	DATA *		data;
	FILE *		fp;
	int		mode;
	char		modeString[12];
	char		givenSizeString[32];

	data = DEVTODATA(dev);
	fp = data->fp;

	/*
	 * If none of the flags are set then just give the file name.
	 */
	if (!data->listGivenSize && !data->listTrueSize && !data->listMode)
	{
		fprintf(fp, "%s\n", fileName);

		return;
	}

	/*
	 * Initialize the strings.
	 */
	givenSizeString[0] = '\0';
	strcpy(modeString, "          ");

	/*
	 * Generate the file type if we can.
	 */
	mode = status->mode;

	if (status->flags & UP_STAT_TYPE)
		modeString[0] = upGetFileType(mode);

	/*
	 * Generate the file permissions if we can.
	 */
	if (status->flags & UP_STAT_MODE)
	{
		modeString[1] = ((mode & 0400) ? 'r' : '-');
		modeString[2] = ((mode & 0200) ? 'w' : '-');
		modeString[3] = ((mode & 0100) ? 'x' : '-');
		modeString[4] = ((mode & 0040) ? 'r' : '-');
		modeString[5] = ((mode & 0020) ? 'w' : '-');
		modeString[6] = ((mode & 0010) ? 'x' : '-');
		modeString[7] = ((mode & 0004) ? 'r' : '-');
		modeString[8] = ((mode & 0002) ? 'w' : '-');
		modeString[9] = ((mode & 0001) ? 'x' : '-');
	}

	/*
	 * Generate the file size if we can.
	 */
	if (status->flags & UP_STAT_SIZE)
		sprintf(givenSizeString, "%ld", status->size);

	/*
	 * Write the parts according to the user's flags.
	 */
	if (data->listMode)
		fprintf(fp, "%s ", modeString);

	if (data->listGivenSize)
		fprintf(fp, "%10s ", givenSizeString);

	if (data->listTrueSize)
		fprintf(fp, "%10ld ", data->trueSize);

	fprintf(fp, "%s\n", fileName);
}


/*
 * Close the listing file.
 */
static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *			data;

	data = DEVTODATA(dev);

	if ((data->fp != NULL) && (fclose(data->fp) == EOF))
	{
		upSystemError(dev->owner, "fclose");

		return UP_FALSE;
	}

	return UP_TRUE;
}
