/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The log file device.
 * This logs traffic through it to a specified log file.
 * File names, file data, or all items can be selected for logging.
 */

#include <stdio.h>
#include "upPrivate.h"


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doFlush(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL transferData(UP_DEV * dev);


static UP_FUNC	funcs =
{
	UP_TYPE_LOG,
	"LogFile",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	doFlush,
	doClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef struct
{
	FILE *		fp;
	UP_BOOL		logFileName;
	UP_BOOL		logFileData;
	UP_BOOL		logItems;
} DATA;

#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


UP_ID
upAddLogFile(UP * up, const char * name, UP_FLAG logFlags)
{
	UP_DEV *	dev;
	FILE *		fp;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Try to open the log file for writing.
	 */
	fp = fopen(name, "w");

	if (fp == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Initialise our data.
	 * This is the file handle and the logging flags.
	 */
	data = DEVTODATA(dev);

	data->fp = fp;
	data->logFileName = ((logFlags & UP_FLAG_LOG_FILE_NAME) != 0);
	data->logFileData = ((logFlags & UP_FLAG_LOG_FILE_DATA) != 0);
	data->logItems = ((logFlags & UP_FLAG_LOG_ITEMS) != 0);

	/*
	 * Set our initial state which is transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Transfer some data.
			 * Our state remains the same.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred from
			 * the input to the output buffer.
			 */
			upBufferReadFileName(readBuffer);

			upBufferWriteFileName(writeBuffer,
				upBufferGetFileName(readBuffer),
				upBufferGetReadFileStatus(readBuffer));

			if (data->logFileName || data->logItems)
			{
				fprintf(data->fp, "[File: \"%s\"]\n",
					upBufferGetFileName(readBuffer));
			}

			/*
			 * Change our state to transfer data.
			 */
			upSetState(dev, UP_STATE_TRANSFER_DATA);

			break;

		case UP_STATE_TRANSFER_EOF:
			/*
			 * There is an end of file to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEof(readBuffer);
			upBufferWriteEof(writeBuffer);

			if (data->logItems)
				fprintf(data->fp, "[End of file]\n");

			/*
			 * Set our state to transfer another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * There is an end of buffer to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			if (data->logItems)
				fprintf(data->fp, "[End of buffer]\n");

			/*
			 * Say that we are finished.
			 */
			upSetFinished(dev);

			break;

		case UP_STATE_TRANSFER_FILE_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer file data between the input and output buffers.
 * Returns true if successful.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	readAddress;
	char *		writeAddress;
	int		readLength;
	int		writeLength;
	int		cc;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the amount of data that can be easily read and written.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Take the minimum of the two lengths.
	 */
	cc = readLength;

	if (cc > writeLength)
		cc = writeLength;

	/*
	 * If there is no data to move, then something is wrong.
	 */
	if (cc <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If we are logging items then log this item.
	 */
	if (data->logItems)
		fprintf(data->fp, "[Data: count=%d]\n", cc);

	/*
	 * If we are logging the file data then write the data
	 * into the log file.
	 */
	if (data->logFileData)
	{
		fwrite(readAddress, 1, cc, data->fp);
	}

	/*
	 * Copy the data from the read buffer to the write buffer.
	 */
	memcpy(writeAddress, readAddress, cc);

	/*
	 * Increment the read and write buffer pointers past the data.
	 */
	upBufferReadAdjust(readBuffer, cc);
	upBufferWriteAdjust(writeBuffer, cc);

	/*
	 * Remember that we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doFlush(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if (fflush(data->fp) == EOF)
	{
		upDeviceSystemError(dev, "fflush");

		return UP_FALSE;
	}

	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Close the log file.
	 */
	if (fclose(data->fp))
	{
		upDeviceSystemError(dev, "fclose");

		return UP_FALSE;
	}

	return UP_TRUE;
}
