/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The make file name list device.
 * This reads all of the input data, and extracts a list of file names
 * from the data, either one per line, or separated by nulls.  The output
 * file names have no data or status.
 */

#include <stdio.h>
#include "upPrivate.h"


/*
 * The size of the initial filename size buffer and the slop size
 * when a new file buffer is needed.
 */
#define	INITIAL_SIZE	1024
#define	SLOP_SIZE	256


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL readData(UP_DEV * dev);
static	UP_BOOL	readEndOfFile(UP_DEV * dev);
static	UP_BOOL	writeFileName(UP_DEV * dev);


static UP_FUNC	funcs =
{
	UP_TYPE_MAKE_FILE_LIST,
	"MakeFileList",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef struct
{
	int		separator;	/* separator char (null or newline)*/
	char *		fileName;	/* current file name */
	int		length;		/* current length of file name */
	int		allocSize;	/* allocated size of file name */
	UP_STAT		emptyStatBuf;	/* an empty status structure */
	char		buffer[INITIAL_SIZE];	/* initial file name buffer */
} DATA;


#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


UP_ID
upAddMakeFileList(UP * up, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialize our data.
	 */
	data = DEVTODATA(dev);

	data->separator = 
		((flags & UP_FLAG_MAKEFILELIST_NULL_SEPARATORS) != 0)
		? '\0': '\n';

	data->fileName = data->buffer;
	data->allocSize = INITIAL_SIZE;
	data->length = 0;

	/*
	 * Set our initial state which is to read a file name.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_READ_NAME:
			/*
			 * A file name can be read.
			 * Read it and ignore it.
			 */
			upBufferReadFileName(readBuffer);

			/*
			 * Change our state to read input data now.
			 */
			upSetState(dev, UP_STATE_READ_DATA);

			break;

		case UP_STATE_READ_DATA:
			/*
			 * Data can be read from the input buffer.
			 */
			return readData(dev);

		case UP_STATE_READ_EOF:
			/*
			 * An end of file can be read.
			 */
			return readEndOfFile(dev);

			break;

		case UP_STATE_READ_END_OF_BUFFER:
			/*
			 * There is an end of buffer.
			 * Transfer it to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say that we are finished.
			 */
			upSetFinished(dev);

			break;

		case UP_STATE_WRITE_NAME:
			/*
			 * We can write a file name now.
			 */
			return writeFileName(dev);

		case UP_STATE_READ_FILE_ERROR:
			/*
			 * There is a read error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read data from the input buffer.
 * This data is a list of file names which can be separated by
 * newlines or nulls.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	const char *	endAddress;
	char *		newFileName;
	int		newAllocSize;
	int		readLength;
	int		appendLength;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the available data in the input buffer.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * There should be some data.
	 */
	if (readLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * See if the separator character is present in the buffer.
	 * If so, then we have a new filename.
	 */
	endAddress = memchr(readAddress, data->separator, readLength);

	/*
	 * Get the amount of data to be appended to the current file name.
	 */
	if (endAddress != NULL)
		appendLength = endAddress - readAddress;
	else
		appendLength = readLength;

	/*
	 * Calculate the filename buffer size we need for the old data
	 * plus the new data plus a trailing null.
	 */
	newAllocSize = data->length + appendLength + 1;

	/*
	 * Reallocate the filename if there is no more room.
	 */
	if (newAllocSize > data->allocSize)
	{
		/*
		 * Add some slop to the size to help prevent reallocs.
		 */
		newAllocSize += SLOP_SIZE;

		/*
		 * If the old file name was our own local buffer then
		 * we need to allocate a new one from scratch and copy
		 * the old filename into it.
		 */
		if (data->fileName == data->buffer)
		{
			newFileName = upDeviceMalloc(dev, newAllocSize);

			if (newFileName == NULL)
				return UP_FALSE;

			memcpy(newFileName, data->fileName, data->length);
		}
		else
		{
			/*
			 * Reallocate the old buffer.
			 */
			newFileName = upDeviceRealloc(dev, data->fileName,
				newAllocSize);

			if (newFileName == NULL)
				return UP_FALSE;
		}

		/*
		 * Store the new filename buffer and its allocated size.
		 */
		data->fileName = newFileName;
		data->allocSize = newAllocSize;
	}

	/*
	 * Append the new data to the file name.
	 */
	memcpy(data->fileName + data->length, readAddress, appendLength);

	/*
	 * Do the accounting.
	 */
	upBufferReadAdjust(readBuffer, appendLength);

	data->length += appendLength;

	/*
	 * If we have seen the separator character then skip over it
	 * and change our state to write the file name.
	 */
	if (endAddress != NULL)
	{
		upBufferReadAdjust(readBuffer, 1);

		upSetState(dev, UP_STATE_WRITE_NAME);
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read the end of file from the input file.
 * Check for a partial file name that is left over.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;

	data = DEVTODATA(dev);

	/*
	 * Read the end of file.
	 */
	upBufferReadEof(dev->readBuffer);

	/*
	 * If there is any filename left then there is an error.
	 */
	if (data->length != 0)
	{
		upDeviceError(dev, UP_ERR_PARTIALLASTLINE);

		return UP_FALSE;
	}

	/*
	 * Set our state to reading another file name.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write a file name to the output buffer.
 * There is only empty status to be written with the name.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;

	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Terminate the file name with a null.
	 * (We know that there has been room allocated for this.)
	 */
	data->fileName[data->length] = '\0';

	/*
	 * Write the file name and an empty status to the output buffer.
	 */
	if (!upBufferWriteFileName(writeBuffer, data->fileName,
		&data->emptyStatBuf))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * Write an immediate end of file since there isn't any data.
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Reset the filename length to be ready to store a new one.
	 */
	data->length = 0;

	/*
	 * Change our state back to reading more input data.
	 */
	upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *		data;

	data = DEVTODATA(dev);

	/*
	 * If the filename was allocated by us, then free it.
	 */
	if (data->fileName != data->buffer)
		free(data->fileName);

	/*
	 * Return success.
	 */
	return UP_TRUE;
}
