/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The external process device.
 * This reads data from the input buffer, sends that data to the STDIN
 * of a specified command, then sends the STDOUT from that process to
 * the output buffer.
 */

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#ifndef __hpux
#include <wait.h>
#endif
#include <poll.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>

#include "upPrivate.h"


/*
 * Definition of signal handlers to help prevent compile warnings.
 * Change as required.
 */
#define	SIGCAST	(void (*)(int))

typedef	void	(*SIGFUNC)(int);


/*
 * Other definitions.
 */
#define	STDIN		0
#define	STDOUT		1
#define	MIN_CLOSE_FD	3
#define	MAX_CLOSE_FD	1000

#define	TERMTIME	10

#define	isblank(ch)	(((ch) == ' ') || ((ch) == '\t'))


/*
 * Returned poll flags that we check for.
 */
#define	READ_POLL	(POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL)
#define	WRITE_POLL	(POLLOUT | POLLERR | POLLHUP | POLLNVAL)


/*
 * In the following structure, the only data which needs freeing is the
 * argBuffer pointer, as the prog and argv pointers point into that buffer.
 */
typedef	struct
{
	char *		argBuffer;	/* buffer area for strings */
	char *		fullPath;	/* full path of program */
	const char *	prog;		/* program to execute */
	const char **	argv;		/* table of argument strings */
	int		argc;		/* number of arguments */
	UP_BOOL		systemPathFlag;	/* only run from system path */
	UP_BOOL		partialFlag;	/* ok if input partially read */
	UP_BOOL		okExit1Flag;	/* ok to have exit status 1 */
	UP_BOOL		okExit2Flag;	/* ok to have exit status 2 */
	UP_BOOL		okExitAnyFlag;	/* ok to have any exit status */
	pid_t		pid;		/* process id of child process */
	int		sendFd;		/* fd to send data to child */
	int		recvFd;		/* fd to receive data from child */
	int *		statusList;	/* exit status list */
	struct pollfd *	readPollEntry;	/* reading polling entry if any */
	struct pollfd * writePollEntry;	/* writing polling entry if any */
	struct pollfd	pollTable[2];	/* polling table */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_PROCESS,
	"Proc",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Other local procedures.
 */
static	UP_ID	commonAddProc(UP *, char *, const char *, int,
		const char **, UP_FLAG, const int *);

static	UP_BOOL	transferFileName(UP_DEV * dev);
static	UP_BOOL	transferEndOfBuffer(UP_DEV * dev);
static	UP_BOOL	transferEndOfFile(UP_DEV * dev);
static	UP_BOOL	transferData(UP_DEV * dev);
static	UP_BOOL	readData(UP_DEV * dev);
static	UP_BOOL	readEndOfFile(UP_DEV * dev);
static	UP_BOOL	writeData(UP_DEV * dev, UP_STATE state);
static	UP_BOOL	startProc(UP_DEV * dev);
static	UP_BOOL	stopProc(UP_DEV * dev);
static	char *	findExecutable(UP *, const char *, const char *);


/*
 * Add the process device with command arguments specified by one string.
 * Spaces or tabs separate the arguments.  No processing of the arguments
 * is performed by this routine.  The command line is not given to a shell,
 * so no wildcarding or quoting is handled.  If this is needed, use the
 * upAddProcessArgs function to execute a shell explicitly.  The statusList
 * argument specifies a list of exit status values (other than zero which
 * is always considered successful) which are not to be considered errors.
 * The list is in the form of pairs of values, representing all status values
 * from the first of a pair to the last of a pair, inclusive.  The value
 * zero as the first value of pair terminates the list.  The statusList
 * can be NULL if zero is the only successful exit status.
 */
UP_ID
upAddProcessString(UP * up, const char * cmd,
	UP_FLAG flags, const int * statusList)
{
	char *		argBuffer;	/* storage for strings and pointers */
	char *		strings;	/* strings */
	const char *	cp;
	const char **	argv;		/* argument table */
	int		argc;
	int		len;
	int		i;

	argBuffer = NULL;

	/*
	 * Count the number of arguments in the command and make sure
	 * that there is something there.
	 */
	while (isblank(*cmd))
		cmd++;

	cp = cmd;
	argc = 0;

	while (*cp)
	{
		argc++;

		while (*cp && !isblank(*cp))
			cp++;

		while (isblank(*cp))
			cp++;
	}

	if (argc == 0)
	{
		upError(up, UP_ERR_NULLCOMMAND);

		return UP_NULL_ID;
	}

	/*
	 * Allocate a buffer large enough to hold the pointer table and a
	 * copy of the input string, and plant pointers to the areas to
	 * hold them.
	 */
	len = strlen(cmd) + 1;

	argBuffer = upMalloc(up, (sizeof(const char *) * (argc + 1)) + len);

	if (argBuffer == NULL)
		return UP_NULL_ID;

	argv = (const char **) argBuffer;
	strings = (char *) (argv + argc + 1);

	/*
	 * Copy the input string for later use.
	 */
	memcpy(strings, cmd, len);

	/*
	 * Fill in the pointer table with pointers to the blank-separated
	 * strings, null-terminating each string as it is found.
	 */
	for (i = 0; i < argc; i++)
	{
		argv[i] = strings;

		while (*strings && !isblank(*strings))
			strings++;

		*strings++ = '\0';

		while (isblank(*strings))
			cp++;
	}

	argv[i] = NULL;

	/*
	 * OK, finish up.
	 */
	return commonAddProc(up, argBuffer, argv[0], argc, argv,
		flags, statusList);
}


/*
 * Add the process device to a pipeline where the arguments are explicitly
 * specified.  This allows the program to be executed to differ from the
 * first argument, and also allows blanks in the arguments.  Arguments are
 * given to the program as specified without the involvement of a shell.
 */
UP_ID
upAddProcessArgs(UP * up, const char * prog,
	int argc, const char ** argv, UP_FLAG flags,
	const int * statusList)
{
	const char **	oldArgv;
	char *		argBuffer;
	char *		strings;
	char *		cp;
	int		i;
	int		len;

	oldArgv = argv;

	/*
	 * Count up the length of all of the strings.
	 * Include the null-terminated bytes at the ends.
	 */
	len = strlen(prog) + 1;

	for (i = 0; i < argc; i++)
		len += strlen(argv[i]) + 1;

	/*
	 * Allocate a buffer large enough to hold the strings and also
	 * the pointers to the strings.
	 */
	argBuffer = upMalloc(up, (sizeof(const char *) * (argc + 1)) + len);

	if (argBuffer == NULL)
		return UP_NULL_ID;

	argv = (const char **) argBuffer;
	strings = (char *) (argv + argc + 1);

	/*
	 * Copy the strings into the argument buffer, and set new
	 * pointers to those strings.
	 */
	cp = strings;

	len = strlen(prog) + 1;

	memcpy(cp, prog, len);

	cp += len;

	for (i = 0; i < argc; i++)
	{
		argv[i] = cp;
		len = strlen(oldArgv[i]) + 1;
		memcpy(cp, oldArgv[i], len);
		cp += len;
	}

	argv[i] = NULL;

	/*
	 * OK, finish up.
	 */
	return commonAddProc(up, argBuffer, prog, argc, argv, flags, statusList);
}


/*
 * Common routine used for both of the two types of process startups.
 * The supplied argBuffer parameter is an malloc'd buffer which we will own,
 * and which contains all of the supplied strings and argument pointers.
 * On an error we will free this buffer.
 */
static UP_ID
commonAddProc(
	UP *		up,
	char *		argBuffer,
	const char *	prog,
	int		argc,
	const char **	argv,
	UP_FLAG		flags,
	const int *	statusList
)
{
	UP_DEV	*	dev;
	DATA *		data;
	const char *	path;
	UP_ID		id;
	UP_BOOL		started;

	started = UP_FALSE;
	id = UP_NULL_ID;

	/*
	 * Allocate a device structure and fill it in.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
	{
		free(argBuffer);

		return UP_NULL_ID;
	}

	/*
	 * Set up our private data.
	 */
	data = DEVTODATA(dev);

	data->argBuffer = argBuffer;
	data->fullPath = NULL;
	data->prog = prog;
	data->argv = argv;
	data->argc = argc;
	data->sendFd = -1;
	data->recvFd = -1;
	data->pid = 0;
	data->readPollEntry = NULL;
	data->writePollEntry = NULL;
	data->statusList = NULL;

	/*
	 * Set the option flags.
	 */
	data->partialFlag = ((flags & UP_FLAG_PROC_PARTIAL_INPUT_OK) != 0);
	data->okExit1Flag = ((flags & UP_FLAG_PROC_EXIT_1_OK) != 0);
	data->okExit2Flag = ((flags & UP_FLAG_PROC_EXIT_2_OK) != 0);
	data->okExitAnyFlag = ((flags & UP_FLAG_PROC_EXIT_ANY_OK) != 0);
	data->systemPathFlag = ((flags & UP_FLAG_PROC_SYSTEM_PATH) != 0);

	/*
	 * Select the path list to be searched for the program.
	 * This is either our own secure path, or the normal one
	 * from the PATH environment variable.
	 */
	if (data->systemPathFlag)
		path = upGetSystemPath();
	else
		path = getenv("PATH");

	/*
	 * Find the full path of the program.
	 */
	data->fullPath = findExecutable(up, path, prog);

 	if (data->fullPath == NULL)
	{
		free(argBuffer);

		return UP_NULL_ID;
	}

	/*
	 * Set our state to begin by transferring a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Point to our polling table.
	 */
	dev->pollTable = data->pollTable;

	/*
	 * Finally link the device into the pipeline.
	 */
	id = upLinkDevice(up, dev);

	if (id == UP_NULL_ID)
	{
		free(argBuffer);
		free(data->fullPath);
	}

	return id;
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *		data;
	UP_STATE	state;
	UP_BOOL		needWritePoll;
	UP_BOOL		needReadPoll;
	struct pollfd *	pollEntry;

	data = DEVTODATA(dev);

	/*
	 * Start off with no polling required.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	data->readPollEntry = NULL;
	data->writePollEntry = NULL;

	needReadPoll = UP_FALSE;
	needWritePoll = UP_FALSE;

	/*
	 * Get our current state.
	 */
	state = upCheckState(dev);

	/*
	 * Switch on what we can do according to the state.
	 */
	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * We can do nothing now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * We want to both read and write to the process.
			 * from the input and output buffers.  However,
			 * we might only need to consume the input buffer
			 * data and so not actually need to write data to
			 * the process.
			 */
			needWritePoll = (data->sendFd >= 0);
			needReadPoll = UP_TRUE;

			break;

		case UP_STATE_WRITE_DATA:
			/*
			 * We want to read data from the process and
			 * write it to the output buffer.
			 */
			needReadPoll = UP_TRUE;

			break;

		default:
			/*
			 * For all other states, we want to run right away
			 * to handle the condition.
			 */
			dev->pollTimeout = 0;

			return UP_TRUE;
	}

	/*
	 * If we want to poll for reading from the process, then set that
	 * up by adding the receiving file descriptor to the poll table.
	 */
	if (needReadPoll)
	{
		/*
		 * Make sure we have a receiving file handle.
		 */
		if (data->recvFd < 0)
		{
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
		}

		/*
		 * Set up the poll entry.
		 */
		pollEntry = &dev->pollTable[dev->pollCount++];

		pollEntry->fd = data->recvFd;
		pollEntry->events = POLLIN;
		pollEntry->revents = 0;

		data->readPollEntry = pollEntry;
	}

	/*
	 * If we want to poll for writing to the process, then set that
	 * up by adding the sending file descriptor to the poll table.
	 */
	if (needWritePoll)
	{
		/*
		 * Make sure we have a sending file handle.
		 */
		if (data->sendFd < 0)
		{
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
		}

		/*
		 * Set up the poll entry.
		 */
		pollEntry = &dev->pollTable[dev->pollCount++];

		pollEntry->fd = data->sendFd;
		pollEntry->events = POLLOUT;
		pollEntry->revents = 0;

		data->writePollEntry = pollEntry;
	}

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	UP_STATE	state;

	/*
	 * Get the current state.
	 */
	state = upCheckState(dev);

	/*
	 * Switch on what can be acted on now.
	 */
	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * Nothing can be done now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * A file name can be transferred.
			 */
			return transferFileName(dev);

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * The end of the buffer has been reached.
			 */
			return transferEndOfBuffer(dev);

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Data can be transferred from the input buffer
			 * to the output buffer.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_EOF:
			/*
			 * An end of file can be read from the input buffer.
			 */
			return transferEndOfFile(dev);

		case UP_STATE_WRITE_DATA:
			/*
			 * Data can be written to the output buffer.
			 * This is trailing data after the end of file
			 * has been read from the input buffer and sent
			 * to the process.
			 */
			return writeData(dev, state);

		case UP_STATE_READ_DATA:
			/*
			 * Data can be read from the input buffer (and
			 * ignored since the process has been killed.)
			 */
			return readData(dev);

		case UP_STATE_READ_EOF:
			/*
			 * An end of file can be read from the input buffer.
			 */
			return readEndOfFile(dev);

		case UP_STATE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}



/*
 * Here to transfer a file name from the input to the output buffer.
 * We also set up the pipeline for use in processing the data for
 * the new file.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const UP_STAT *	statBuf;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Read the current file name and status.
	 * The input filename is ignored.
	 */
	upBufferReadFileName(readBuffer);

	statBuf = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Set the output file name with a default status.
	 */
	upBufferWriteFileName(writeBuffer, "process-output", NULL);

	/*
	 * Try to start up the process.
	 */
	if (!startProc(dev))
		return UP_FALSE;

	/*
	 * The process is started.
	 * Set our state to transfer data.
	 */
	upSetState(dev, UP_STATE_TRANSFER_DATA);

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer the end of the buffer to the output buffer.
 */
static UP_BOOL
transferEndOfBuffer(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Make sure that there is no process still.
	 * There shouldn't be one at this point.
	 */
	if (data->pid != 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		stopProc(dev);

		return UP_FALSE;
	}

	/*
	 * Transfer the end of buffer.
	 */
	upBufferReadEndOfBuffer(readBuffer);
	upBufferWriteEndOfBuffer(writeBuffer);

	/*
	 * Say we are finished.
	 */
	return upSetFinished(dev);
}


/*
 * Transfer data from the input buffer to the output buffer.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	if (!readData(dev))
		return UP_FALSE;

	return writeData(dev, UP_STATE_TRANSFER_DATA);
}


/*
 * Here to transfer an end of file from the input buffer.
 * We close the sending file handle so that the process will see an
 * end of file from its standard input.  When the process eventually
 * exits, it will close its standard output, which will then cause
 * us to finally send an end of file to the output buffer.
 */
static UP_BOOL
transferEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Read the end of file from the input buffer.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * Close the sending end of the pipeline so that the process
	 * will read an end of file from its standard input.
	 */
	(void) close(data->sendFd);
	data->sendFd = -1;

	/*
	 * Remove our polling entry.
	 */
	data->writePollEntry = NULL;

	/*
	 * Set our state so that we are only writing data to the
	 * output buffer.  I.E., we are reading the standard output
	 * from the process until it closes it.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine to read data from the input buffer and attempt to write it
 * to the stdin of the child process.  This should not be called unless
 * we know that there is some data in the input buffer to be sent.
 * Data is only transferred if poll had indicated that the pipe can
 * hold at least one byte of data.  This routine also handes the consuming
 * of extra input data after the child process has exited.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	struct pollfd *	pollEntry;
	const char *	readAddress;
	int		readLength;
	SIGFUNC		oldSignal;
	int		cc;

	/*
	 * Point to our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the amount of data available to be read from the input buffer.
	 * There should be some.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	if (readLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * See if there is still a file descriptor for writing data to
	 * the process.  If not, then we are simply reading and discarding
	 * input data because the process has stopped before all of our
	 * data was sent.
	 */
	if (data->sendFd < 0)
	{
		upBufferReadClear(readBuffer);

		return upSetDidReadWrite(dev);
	}

	/*
	 * The sending file descriptor is still in use.
	 * Check whether our polling information said that there
	 * should be some room to write data to the process.
	 * If not, then return doing nothing.
	 */
	pollEntry = data->writePollEntry;

	if ((pollEntry == NULL) || ((pollEntry->revents & WRITE_POLL) == 0))
		return UP_TRUE;

	/*
	 * We have some data to be written to the stdin of the process.
	 * Set to ignore SIGPIPE errors, saving any old signal handler.
	 */
	oldSignal = signal(SIGPIPE, SIG_IGN);

	/*
	 * Do an iteration here so that we can try the write several times
	 * if necessary.  The first time is for the amount to be written,
	 * the second time is for 8 bytes, and the last time is for 1
	 * byte.
	 */
	while ((cc = write(data->sendFd, readAddress, readLength)) < 0)
	{
		/*
		 * See if the write failed because of a broken pipeline.
		 * If so, and that is allowed, then leave the loop
		 * the cc negative as a flag.
		 */
		if ((errno == EPIPE) && data->partialFlag)
			break;

		/*
		 * See if the write failed because we would block.
		 * If not, then fail.
		 */
		if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
		{
			upDeviceSystemError(dev, "write");

			return UP_FALSE;
		}

		/*
		 * There wasn't room for all of the data.
		 * Reduce the length if required so that maybe it will
		 * fit using a smaller length.  If we tried a single
		 * byte and it still wouldn't fit, then return no
		 * data transferred.
		 */
		if (readLength > 8)
			readLength = 8;
		else if (readLength > 1)
			readLength = 1;
		else
		{
			cc = 0;

			break;
		}
	}

	/*
	 * Restore the signal handler if required.
	 */
	if (oldSignal != SIG_IGN)
		signal(SIGPIPE, SIGCAST oldSignal);

	/*
	 * If the pipeline is broken then we must close the sending
	 * file descriptor.
	 */
	if (cc < 0)
	{
		/*
		 * Close the sending side of the pipeline.
		 */
		(void) close(data->sendFd);
		data->sendFd = -1;

		/*
		 * Remove our polling entry.
		 */
		data->writePollEntry = NULL;

		/*
		 * Consume any input data since it is useless now.
		 */
		upBufferReadClear(readBuffer);

		/*
		 * Return saying we have done I/O.
		 * Leave us in the same state so that we can read the
		 * input end of file.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * Make sure that something got written to the pipeline.
	 * This should have been possible because poll told us that
	 * there was some room.
	 */
	if (cc == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Account for any data processed from the input buffer.
	 */
	upBufferReadAdjust(readBuffer, cc);

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read an end of file from the input buffer after we have consumed
 * any remaining data when the process was killed.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;

	/*
	 * Point to our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(dev->readBuffer);

	/*
	 * Make sure that the process was killed.
	 */
	if (data->pid != 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		stopProc(dev);

		return UP_FALSE;
	}

	/*
	 * Set our state to read another new filename to be sent through
	 * another process.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write data to the output buffer from the stdout of the child process.
 * If end of file is reached from the stdout pipe, the controlled
 * process is killed.  This should only be called when we know that
 * there is room in the output buffer for some data.  Data is only
 * transferred if poll had indicated that the pipe has at least one
 * byte of data to read.  Returns UP_TRUE on success.
 */
static UP_BOOL
writeData(UP_DEV * dev, UP_STATE state)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	struct pollfd *	pollEntry;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	/*
	 * Point to our data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Check whether poll has indicated that something can be
	 * read from the stdout of the process.  If not, then do
	 * nothing.
	 */
	pollEntry = data->readPollEntry;

	if ((pollEntry == NULL) || ((pollEntry->revents & READ_POLL) == 0))

		return UP_TRUE;

	/*
	 * Get the available space in the output buffer to be
	 * stored in the output buffer.  There should be some.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	if (writeLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Try to read some data from the stdout of the process.
	 */
	cc = read(data->recvFd, writeAddress, writeLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "read");

		return UP_FALSE;
	}

	/*
	 * If we read any data from the process then account for it
	 * and return saying we have done I/O.
	 */
	if (cc > 0)
	{
		upBufferWriteAdjust(writeBuffer, cc);

		return upSetDidReadWrite(dev);
	}

	/*
	 * We have read an end of file from the process.
	 * Make sure that the process is killed, and clean up the
	 * file handle and poll information.
	 */
	if (!stopProc(dev))
		return UP_FALSE;

	/*
	 * If we were still able to read data from the input buffer
	 * then set our state to do that.  (That data will be ignored
	 * since the process is killed.)  Otherwise we have already
	 * read all of the input file, and we set our state to transfer
	 * a new filename.
	 */
	if (state == UP_STATE_TRANSFER_DATA)
		upSetState(dev, UP_STATE_READ_DATA);
	else
		upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Tell the output buffer that it has an end of file.
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Close the pipe device.
 */
static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;
	UP_BOOL	status;

	data = DEVTODATA(dev);

	/*
	 * Kill off the process if there is one.
	 */
	status = stopProc(dev);

	/*
	 * Reset the poll table back so that it won't be wrongly freed.
	 */
	dev->pollTable = &dev->pollEntry;

	/*
	 * Free the argument buffer and full path name.
	 */
	free(data->argBuffer);
	free(data->fullPath);

	return status;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Kill off the process if there is one.
	 */
	if (!stopProc(dev))
		return UP_FALSE;

	/*
	 * Then restart the process.
	 */
	if (!startProc(dev))
		return UP_FALSE;

	return UP_TRUE;
}


/*
 * Start the process.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
startProc(UP_DEV * dev)
{
	DATA *	data;
	int	sendFds[2];
	int	recvFds[2];
	pid_t	pid;
	int	fd;

	data = DEVTODATA(dev);

	/*
	 * Initialize some data associated with the process.
	 */
	data->pid = 0;
	data->sendFd = -1;
	data->recvFd = -1;

	/*
	 * Create the sending and receiving pipelines.
	 */
	if (pipe(sendFds) < 0)
	{
		upDeviceSystemError(dev, "pipe");

		return UP_FALSE;
	}

	if (pipe(recvFds) < 0)
	{
		upDeviceSystemError(dev, "pipe");

		(void) close(sendFds[0]);
		(void) close(sendFds[1]);

		return UP_FALSE;
	}

	/*
	 * Fork ourself.
	 * If we are the parent, then remember things and
	 * close the unused sides of the pipelines.
	 */
	fflush(NULL);

	pid = fork();

	if (pid < 0)
	{
		upDeviceSystemError(dev, "fork");

		(void) close(sendFds[0]);
		(void) close(sendFds[1]);
		(void) close(recvFds[0]);
		(void) close(recvFds[1]);

		return UP_FALSE;
	}

	/*
	 * If we are the parent then finish up.
	 */
	if (pid)
	{
		data->pid = pid;
		data->sendFd = sendFds[1];
		data->recvFd = recvFds[0];

		(void) close(sendFds[0]);
		(void) close(recvFds[1]);

		/*
		 * Set the pipe writing file descriptor to be non-blocking.
		 */
		if (fcntl(data->sendFd, F_SETFL, O_NONBLOCK) < 0)
		{
			(void) close(data->sendFd);
			(void) close(data->recvFd);

			return UP_FALSE;
		}

		return UP_TRUE;
	}

	/*
	 * We are the child.  Errors from this point on cause us to exit.
	 * Close the unused side of the pipelines, then set up our STDIN
	 * and STDOUT to the remaining ends of the pipes.
	 */
	(void) close(sendFds[1]);
	(void) close(recvFds[0]);
	(void) close(STDIN);
	(void) close(STDOUT);

	if (dup(sendFds[0]) < 0)
	{
		perror("dup for 0");

		_exit(99);
	}

	(void) close(sendFds[0]);

	if (dup(recvFds[1]) < 0)
	{
		perror("dup for 1");

		_exit(99);
	}

	(void) close(recvFds[1]);

	/*
	 * Close all other file descriptors.
	 */
	for (fd = MIN_CLOSE_FD; fd <= MAX_CLOSE_FD; fd++)
		(void) close(fd);

	/*
	 * Now try to execute the program.
	 */
	execvp(data->fullPath, (char **) data->argv);

	perror("exec");

	_exit(99);
}


/*
 * Dummy signal handler for SIGALRM.
 */
static void
CatchAlarm(int a)
{
}


/*
 * Try to stop the process and check its exit status.
 * Failure includes a nonzero exit status from the process.
 * It is ok to try to stop the process more than once.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
stopProc(UP_DEV * dev)
{
	DATA *	data;
	int	status;
	int	exitStatus;
	pid_t	thePid;
	pid_t	pid;
	int	oldAlarm;
	int	signalNumber;
	UP_BOOL	didTerm;
	UP_BOOL	didKill;
	SIGFUNC	oldSignal;

	data = DEVTODATA(dev);

	/*
	 * If there is no process, then there is nothing to do.
	 * Clear the process id from the structure so we don't do this again.
	 */
	thePid = data->pid;

	if (thePid == 0)
		return UP_TRUE;

	data->pid = 0;

	/*
	 * Close the file descriptors if necessary and clean up their
	 * information.
	 */
	if (data->sendFd >= 0)
		(void) close(data->sendFd);

	if (data->recvFd >= 0)
		(void) close(data->recvFd);

	data->sendFd = -1;
	data->recvFd = -1;

	data->readPollEntry = NULL;
	data->writePollEntry = NULL;

	/*
	 * Initialize flags in preparation for killing the process.
	 */
	didTerm = UP_FALSE;
	didKill = UP_FALSE;

	/*
	 * First try to collect the process without killing it first,
	 * since we usually get here after it has already exited.
	 */
	pid = waitpid(thePid, &status, WNOHANG);

	/*
	 * If the process did not already exit, then we will have to kill it.
	 */
	if (pid == 0)
	{
		/*
		 * First send a SIGTERM to let the process clean up.
		 */
		if (kill(thePid, SIGTERM) < 0)
		{
			upDeviceSystemError(dev, "kill");

			return UP_FALSE;
		}

		didTerm = UP_TRUE;

		/*
		 * Wait for a few seconds if necessary for it to exit.
		 */
		oldAlarm = alarm(0);
		oldSignal = signal(SIGALRM, SIGCAST CatchAlarm);
		alarm(TERMTIME);

		pid = waitpid(thePid, &status, 0);

		signal(SIGALRM, SIGCAST oldSignal);
		alarm(oldAlarm);

		/*
		 * Kill it unconditionally if it has not yet exited.
		 */
		if ((pid < 0) && (errno == EINTR))
		{
			if (kill(thePid, SIGKILL) < 0)
			{
				upDeviceSystemError(dev, "kill");

				return UP_FALSE;
			}

			didKill = UP_TRUE;

			/*
			 * Wait for the new status.
			 */
			pid = waitpid(thePid, &status, 0);
		}
	}

	if (pid < 0)
	{
		upDeviceSystemError(dev, "waitpid");

		return UP_FALSE;
	}

	/*
	 * Check the exit status.
	 * If it exited properly, then check if the status is ok.
	 */
	if (WIFEXITED(status))
	{
		exitStatus = WEXITSTATUS(status);

		/*
		 * Check for some ok exit status values.
		 */
		if ((exitStatus == 0) || data->okExitAnyFlag)
			return UP_TRUE;

		if ((exitStatus == 1) && data->okExit1Flag)
			return UP_TRUE;

		if ((exitStatus == 2) && data->okExit2Flag)
			return UP_TRUE;

		/*
		 * The exit status is not ok.
		 */
		upDeviceError(dev, UP_ERR_EXITERROR);

		return UP_FALSE;
	}

	/*
	 * If it didn't exit properly, then it should have had a signal.
	 * If not, then something is wrong with waitpid.
	 */
	if (!WIFSIGNALED(status))
	{
		upDeviceError(dev, UP_ERR_EXITUNKNOWN);

		return UP_FALSE;
	}

	/*
	 * The process did terminate due to a signal.
	 * If the signal was due to us, then we don't care about it.
	 */
	signalNumber = WTERMSIG(status);

	if (didTerm && (signalNumber == SIGTERM))
		return UP_TRUE;

	if (didKill && (signalNumber == SIGKILL))
		return UP_TRUE;

	/*
	 * It was something else so complain about it.
	 */
	upDeviceError(dev, UP_ERR_EXITSIGNAL);

	return UP_FALSE;
}


/*
 * Find the executable path for the specified program name by
 * possibly walking down the specified colon-separated paths.
 * Returns the found path, or NULL if the executable is not found.
 * The returned path is an allocated string which must be freed by
 * the caller.
 */
static char *
findExecutable(UP * up, const char * pathList, const char * program)
{
	const char *	endPath;
	char *		buffer;
	int		length;

	/*
	 * Make sure the path is usable.
	 */
	if ((pathList == NULL) || (*pathList == '\0'))
	{
		upError(up, UP_ERR_BADPATH);

		return NULL;
	}

	/*
	 * If the program is an absolute path or contains a path
	 * separator then we only look at it as is.
	 */
	if ((*program == '/') || (strchr(program, '/') != NULL))
	{
		/*
		 * See if we can access the path.
		 */
		if (access(program, X_OK) < 0)
		{
			upSystemError(up, "access");

			return NULL;
		}

		/*
		 * Return a copy of the program name.
		 */
		return strdup(program);
	}

	/*
	 * Calculate the maximum path length we need.
	 * This is the sum of the two arguments and two bytes
	 * for the path separator and the terminaing null.
	 */
	length = strlen(pathList) + strlen(program) + 2;

	/*
	 * Allocate the temporary buffer.
	 */
	buffer = malloc(length);

	if (buffer == NULL)
		return NULL;

	/*
	 * Walk down the path list trying each path.
	 */
	while (*pathList)
	{
		/*
		 * Find the end of this path entry.
		 */
		endPath = strchr(pathList, ':');

		if (endPath == NULL)
			endPath = pathList + strlen(pathList);

		length = endPath - pathList;

		/*
		 * Copy the path into the buffer and append a slash
		 * if it is non-empty.
		 */
		memcpy(buffer, pathList, length);

		if (length > 0)
			buffer[length++] = '/';

		/*
		 * Append the program name.
		 */
		strcpy(buffer + length, program);

		/*
		 * If the program is executable then return the path.
		 */
		if (access(buffer, X_OK) == 0)
			return buffer;

		/*
		 * Advance to the next path entry.
		 */
		pathList = endPath;

		if (*pathList)
			pathList++;
	}

	/*
	 * We didn't find any executable.
	 * Free the buffer and return failure.
	 */
	free(buffer);

	upSystemError(up, "access");

	return NULL;
}
