/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The progress device.
 * This provides a callback function giving information about the
 * progress of the pipeline at the devices' location.  This contains
 * items like the current filename, current file status, file count,
 * total and current byte count.
 */

#include <stdio.h>
#include "upPrivate.h"


static	UP_BOOL	doReadWrite(UP_DEV *);

static	UP_BOOL transferFileName(UP_DEV * dev);
static	UP_BOOL transferData(UP_DEV * dev);
static	UP_BOOL transferEndOfFile(UP_DEV * dev);
static	UP_BOOL transferEndOfBuffer(UP_DEV * dev);
static	UP_BOOL transferFileError(UP_DEV * dev);
static	void	updateProgress(UP_DEV *, UP_PROGRESS_ACTION);


static UP_FUNC	funcs =
{
	UP_TYPE_PROGRESS,
	"Progress",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	upDefaultClose,
	upDefaultReset
};


/*
 * Local data for this device.
 */
typedef struct
{
	UP_PROGRESS *		progress;
	UP_PROGRESS_FUNC	func;
	void *			userData;
	UP_PROGRESS		ourProgress;
} DATA;

#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


UP_ID
upAddProgress(UP * up, UP_PROGRESS * progress,
	UP_PROGRESS_FUNC func, void * userData)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * If the caller doesn't want his own progress then use ours.
	 */
	if (progress == NULL)
		progress = &data->ourProgress;

	/*
	 * Set up the static parts of the progress structure.
	 */
	memset(progress, 0, sizeof(UP_PROGRESS));

	progress->up = dev->owner;
	progress->id = dev->id;
	progress->userData = userData;
	progress->lastAction = UP_PROGRESS_ACTION_NONE;

	/*
	 * Save the user's callback function and data.
	 */
	data->progress = progress;
	data->func = func;
	data->userData = userData;

	/*
	 * Set our initial state which is transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_STATE	state;

	data = DEVTODATA(dev);

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Transfer some data.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred from
			 * the input to the output buffer.
			 */
			return transferFileName(dev);

		case UP_STATE_TRANSFER_EOF:
			/*
			 * There is an end of file to be transferred
			 * from the input to the output buffer.
			 */
			return transferEndOfFile(dev);

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * There is an end of buffer to be transferred
			 * from the input to the output buffer.
			 */
			return transferEndOfBuffer(dev);

		case UP_STATE_TRANSFER_FILE_ERROR:
			/*
			 * There has been a file error.
			 */
			return transferFileError(dev);

		case UP_STATE_FATAL_ERROR:
			/*
			 * A buffer has a fatal error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer a file name from the input buffer to the output buffer.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_PROGRESS *	progress;
	const char *	fileName;
	const UP_STAT *	fileStatus;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;

	data = DEVTODATA(dev);
	progress = data->progress;
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Read the file name and status from the input buffer.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	fileStatus = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Tell the output buffer what the file name and status is.
	 */
	upBufferWriteFileName(writeBuffer, fileName, fileStatus);

	/*
	 * Put the file name and status into the progress, and clear
	 * the file offset.
	 */
	progress->fileName = fileName;
	progress->fileStatus = fileStatus;
	progress->fileOffset = 0;

	/*
	 * Update the progress.
	 */
	updateProgress(dev, UP_PROGRESS_ACTION_FILE_NAME);

	/*
	 * Change our state to transfer data.
	 */
	upSetState(dev, UP_STATE_TRANSFER_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer an end of buffer from the input buffer to the output buffer.
 */
static UP_BOOL
transferEndOfBuffer(UP_DEV * dev)
{
	DATA *		data;
	UP_PROGRESS *	progress;

	data = DEVTODATA(dev);
	progress = data->progress;

	/*
	 * Acknowledge the end of buffer and set it on the output buffer.
	 */
	upBufferReadEndOfBuffer(dev->readBuffer);
	upBufferWriteEndOfBuffer(dev->writeBuffer);

	/*
	 * Update the progress to say we are finished.
	 */
	updateProgress(dev, UP_PROGRESS_ACTION_FINISHED);

	/*
	 * Say that we are finished.
	 */
	return upSetFinished(dev);
}


/*
 * Transfer file data between the input and output buffers.
 * Returns true if successful.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *		data;
	UP_PROGRESS *	progress;
	int		cc;
	unsigned long	oldLowByteCount;

	data = DEVTODATA(dev);
	progress = data->progress;

	/*
	 * Transfer data and get the byte count.
	 */
	cc = upTransferData(dev, 0);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Increment the byte count for this file.
	 */
	progress->fileOffset += cc;

	/*
	 * Increment the low and high total byte counts.
	 * This works since we know that the amount transferred
	 * is relatively small, so that on an overflow from the
	 * low count value it will become smaller.
	 */
	oldLowByteCount = progress->lowByteCount;

	progress->lowByteCount += cc;

	if (progress->lowByteCount < oldLowByteCount)
		progress->highByteCount++;

	/*
	 * Update the progress.
	 */
	updateProgress(dev, UP_PROGRESS_ACTION_FILE_DATA);

	/*
	 * Remember that we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer an end of file from the input to the output buffer.
 */
static UP_BOOL
transferEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_PROGRESS *	progress;

	data = DEVTODATA(dev);
	progress = data->progress;

	/*
	 * Read the end of file and set it on the output buffer.
	 */
	upBufferReadEof(dev->readBuffer);
	upBufferWriteEof(dev->writeBuffer);

	/*
	 * Update the progress.
	 */
	progress->fileCount++;

	updateProgress(dev, UP_PROGRESS_ACTION_END_OF_FILE);

	/*
	 * Set our state to transfer another file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer a file error from the input to the output buffer.
 */
static UP_BOOL
transferFileError(UP_DEV * dev)
{
	DATA *		data;
	UP_PROGRESS *	progress;

	data = DEVTODATA(dev);
	progress = data->progress;

	/*
	 * Read the file error and set it on the output buffer.
	 */
	upBufferReadFileError(dev->readBuffer);
	upBufferWriteFileError(dev->writeBuffer);

	/*
	 * Update the progress.
	 */
	progress->fileCount++;
	progress->fileErrors++;

	updateProgress(dev, UP_PROGRESS_ACTION_FILE_ERROR);

	/*
	 * Set our state to transfer another file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Set the progress last action value, increment the change counter,
 * and perform a callback if required.  Other data must have been placed
 * in the progress structure prior to this call.
 */
static void
updateProgress(UP_DEV * dev, UP_PROGRESS_ACTION lastAction)
{
	DATA *		data;
	UP_PROGRESS *	progress;

	data = DEVTODATA(dev);
	progress = data->progress;

	/*
	 * Set the last action.
	 */
	progress->lastAction = lastAction;

	/*
	 * Increment the change counter.
	 */
	if (++(progress->lowChangeCount) == 0)
		progress->highChangeCount++;

	/*
	 * If there is a callback function then call it.
	 */
	if (data->func != 0)
		data->func(progress);
}
