/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The remove data pipe device.
 * This transmits all of the file names and status through the device
 * but removes all of the file's data.
 */

#include <stdio.h>
#include "upPrivate.h"


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL transferFileName(UP_DEV * dev);


static UP_FUNC	funcs =
{
	UP_TYPE_REMOVE_DATA,
	"RemoveData",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	upDefaultClose,
	upDefaultReset
};


UP_ID
upAddRemoveData(UP * up)
{
	UP_DEV *	dev;

	dev = upAllocDevice(up, &funcs, 0);

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Set our initial state which is transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_READ_DATA:
			/*
			 * Some input data can be read.
			 * We just ignore all of it!
			 */
			upBufferReadClear(readBuffer);

			break;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred from
			 * the input to the output buffer.
			 */
			return transferFileName(dev);

			break;

		case UP_STATE_READ_EOF:
			/*
			 * An end of file has been read.
			 * Acknowledge it.
			 */
			upBufferReadEof(readBuffer);

			/*
			 * Set our state to transfer another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * There is an end of buffer to be transferred
			 * from the input to the output buffer.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Say that we are finished.
			 */
			upSetFinished(dev);

			break;

		case UP_STATE_READ_FILE_ERROR:
			/*
			 * There is a read error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * There is a file name to be transferred from the input to the
 * output buffer.  We send it straight through except that we clear
 * the data is present flag in the status.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	fileName;
	UP_STAT		statBuf;

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Read the file name and status from the input buffer.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	statBuf = *upBufferGetReadFileStatus(readBuffer);

	/*
	 * Clear the data is present flag since we are discarding it.
	 */
	statBuf.flags &= ~UP_STAT_DATA;

	/*
	 * Write the file name and new status to the output buffer.
	 */
	if (!upBufferWriteFileName(writeBuffer, fileName, &statBuf))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * Write an immediate end of file for the output file since we
	 * are giving it no data!
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Change our state to read the input file data if any
	 * (which we will ignore).
	 */
	upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}
