/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The rescan device.
 * This buffers data through it making it available for reading again.
 * This is useful for "peeking" at data.
 */

#include "upPrivate.h"


/*
 * This defines a circular rescan buffer.
 * The buffer is allocated larger than the size defined in the structure.
 */
typedef	struct
{
	int	size;		/* size of rescan buffer */
	int	maxPos;		/* maximum position which stores data */
	int	begPos;		/* beginning position of data in buffer */
	int	endPos;		/* end position of data in buffer */
	int	curPos;		/* current position for reading data out */
	long	offset;		/* input offset of beginning of buffer */
	char	buf[1];		/* buffer for data (runs past end of struct) */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	int	doread(UP_DEV *, char *, int);
static	UP_BOOL	doReset(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_RESCAN,
	"Rescan",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	upDefaultClose,
	doReset
};


UP_ID
upAddRescan(UP * up, int size)
{
	UP_DEV *	dev;
	DATA *		data;

	/*
	 * Verify that the size is reasonable.
	 */
	if (size <= 0)
	{
		upError(up, UP_ERR_BADSIZE);

		return UP_NULL_ID;
	}

	/*
	 * Allocate a device, allocating a data area which is large
	 * enough to hold the specified amount of data.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA) + size);

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialise our private data.
	 */
	data = DEVTODATA(dev);

	data->size = size;
	data->begPos = 0;
	data->endPos = 0;
	data->curPos = 0;
	data->maxPos = 0;
	data->offset = 0;

	/*
	 * Link the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

#if 0
	cc = data->endPos - data->curPos;
#endif
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	return UP_TRUE;
}


static int
doread(UP_DEV * dev, char * buf, int len)
{
	DATA *	data;
	int	cc;
	int	append;		/* amount of data to append */
	int	discard;	/* amount of data to discard */
	int	left;

	data = DEVTODATA(dev);

	/*
	 * Get the amount of rescanned data available.
	 * If there is some, then return that before getting new data.
	 */
	cc = data->endPos - data->curPos;

	if (cc > 0)
	{
		if (cc > len)
			cc = len;

		memcpy(buf, data->buf + data->curPos, cc);

		data->curPos += cc;

		return cc;
	}

	/*
	 * We need more data, so read it into the user's buffer.
	 */
	cc = (*dev->next->func->read)(dev->next, buf, len);

	if (cc <= 0)
		return cc;

	/*
	 * Calculate how much of the new data we can append to our buffer.
	 * This is the amount of new data read limited by the size of the
	 * rescan buffer.  If it is limited, adjust the user's buffer
	 * address to point to the beginning of the remaining data.
	 */
	append = cc;

	if (append > data->size)
	{
		buf += (append - data->size);
		append = data->size;
	}

	/*
	 * Discard as much data as is necessary from our buffer in order
	 * to append the the new data to it.
	 */
#if 0
	left = ??;
#endif
	discard = (data->endPos + cc) - data->size;
	data->offset += discard;

	if (discard >= data->size)
	{
		data->endPos = 0;
		data->curPos = 0;
		discard = 0;
	}

	/*
	 * If there is any data left in our buffer after discarding some,
	 * then move the remaining data back up to the front of the buffer.
	 */
	if ((discard > 0) && (left > 0))
	{
		memmove(data->buf, data->buf + discard, left);
		data->endPos -= discard;
		data->curPos -= discard;
	}

	memcpy(data->buf + data->endPos, buf, append);

	data->endPos += append;
	data->curPos += append;

	return cc;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

#if 0
	data->readpos = 0;
	data->count = 0;
	data->totalcount = 0;
#endif

	return UP_TRUE;
}
