/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The switch device.
 * This allows the application to redirect the input device to any one
 * of the connected pipelines.  The switching of the output can be done
 * on any end of file boundary.  A callback function is used to give the
 * application the opportunity to select the new pipeline for output.
 */

#include "upPrivate.h"

/*
 * Typedef for function callback to switch pipelines.
 * The function must return the index of the pipeline to be used,
 * or the special values -1 for sending the data straight through,
 * -2 for ignoring the file's data, or -3 for an error.
 */
typedef	int	(*UP_SWITCH_CALLBACK)
	(UP *, const char * fileName, const UP_STAT * fileStatus, void * data);


/*
 * The maximum number of pipelines which can be attached.
 */
#define	MAX_PIPES	20


/*
 * The data private to this device.
 */
typedef	struct
{
	UP_SWITCH_CALLBACK	func;			/* callback function */
	void *			data;			/* callback data */
	int			pipeCount;		/* connected pipes */
	UP_BOOL			ignore;			/* ignoring file */
	UP *			currentPipe;		/* current pipe */
	UP *			pipeTable[MAX_PIPES];	/* pipeline table */
	BOOL			owned[MAX_PIPES];	/* pipeline owned */
} DATA;


#define DEVTODATA(dev)  ((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	int	doread(UP_DEV *, char *, int);
static	int	dowrite(UP_DEV *, const char *, int);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_SWITCH,
	"Switch",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Add the switch device as the final component of the pipeline.
 * Other pipelines can then be attached to this device.
 */
UP_ID
upAddSwitch(UP * up, UP_SWITCH_CALLBACK callbackFunc, void * callbackData)
{
	UP_DEV	*	dev;
	DATA *		data;

	/*
	 * Make sure there is a valid function.
	 */
	if (func == 0)
	{
		upError(up, UP_ERR_NOCALLBACK);

		return FALSE;
	}

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialise our data.
	 */
	data = DEVTODATA(dev);

	data->currentPipe = UP_NULL_PIPE;
	data->func = func;
	data->data = callbackData;

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


/*
 * Attach another user mode pipeline to this device.
 * The owned flag indicates whether the added pipeline is to be "owned"
 * by our pipeline, and thus also freed when our pipeline is freed.
 * An id is returned which identifies the pipeline that was added.
 */
int
upSwitchAttachPipe(UP * up, UP * newUp, BOOL owned)
{
	UP_DEV *	dev;
	DATA *		data;

	/*
	 * Make sure that the pipeline has been set up.
	 */
	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return -1;
	}

	/*
	 * Get the last device of the pipeline.
	 */
	dev = up->last;

	/*
	 * Make sure that the last device is a switch device.
	 */
	if (dev->func->type != UP_TYPE_SWITCH)
	{
		upError(up, UP_ERR_NOSWITCH);

		return -1;
	}

	/*
	 * Get the data for this device.
	 */
	data = DEVTODATA(dev);

	/*
	 * Check for too many pipes being attached.
	 */
	if (data->pipeCount >= MAX_PIPES)
	{
		upError(up, UP_ERR_TOOMANYPIPES);

		return -1;
	}

	/*
	 * Add the pipeline to our data.
	 */
	data->pipeTable[data->pipeCount] = newUp;
	data->owned[data->pipeCount] = owned;

	data->pipeCount++;

	/*
	 * Return the id which is the current number of pipes.
	 */
	return data->pipeCount;
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *		data;
	struct pollfd *	pollTable;
	int		pollCount;
	int		pollTimeOut;

	data = DEVTODATA(dev);

	/*
	 * Assume that no polling is required and that we can wait forever.
	 */
	dev->pollTable = NULL;
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * If we have a current output pipeline then as ask it what
	 * polling requirements it has.
	 */
	if (data->currentPipe != NULL)
	{
		if (!upPollList(data->currentPipe, &pollTable,
			&pollCount, &timeOut))
		{
			upDeviceError(dev, UP_ERR_SWITCHUP);

			return UP_FALSE;
		}

		/*
		 * Copy its polling requirements into our own.
		 */
		dev->pollTable = pollTable;
		dev->pollCount = pollCount;
		dev->pollTimeout = pollTimeOut;
	}

	/*
	 * If the pipeline wants to run now then that is good enough for us.
	 */
	if (dev->pollTimeout == 0)
		return UP_TRUE;

	/*
	 * Check to see if any input is ready for us.
	 */
	 TODO:
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	up;
	BOOL	success;
	int	index;

	/*
	 * Resetting each of the pipelines in the switch, remembering
	 * whether they failed.
	 */
	success = UP_TRUE;

	for (index = 0; index < data->pipeCount; index++)
	{
		if (!upReset(data->pipeTable[index]))
			status = UP_FALSE;
	}

	/*
	 * Generate an error in our pipeline if they failed.
	 */
	if (!success)
	{
		upDeviceError(dev, UP_ERR_PIPEERROR);

		return UP_FALSE;
	}

	return UP_TRUE;
}
