/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The tar format pipe device.
 * This writes all of its input files into one output tar file.
 */

#include "upPrivate.h"


/*
 * Tar file constants.
 */
#define TAR_BLOCK_SIZE	512
#define TAR_NAME_SIZE	100
#define	TAR_PREFIX_SIZE	155

#define	TAR_MAGIC	"ustar"
#define	TAR_VERSION	"00"

#define	TAR_TYPE_REGULAR	'0'
#define	TAR_TYPE_HARD_LINK	'1'
#define	TAR_TYPE_SOFT_LINK	'2'
#define	TAR_TYPE_CHAR_DEVICE	'3'
#define	TAR_TYPE_BLOCK_DEVICE	'4'
#define	TAR_TYPE_DIRECTORY	'5'


/*
 * The POSIX (and basic GNU) tar header format.
 * This structure is always embedded in a TAR_BLOCK_SIZE sized block
 * with zero padding.
 */
typedef struct
{
	char	name[TAR_NAME_SIZE];
	char	mode[8];
	char	uid[8];
	char	gid[8];
	char	size[12];
	char	mtime[12];
	char	checkSum[8];
	char	typeFlag;
	char	linkName[TAR_NAME_SIZE];
	char	magic[6];
	char	version[2];
	char	uname[32];
	char	gname[32];
	char	devMajor[8];
	char	devMinor[8];
	char	prefix[TAR_PREFIX_SIZE];
} TarHeader;


/*
 * One block of a tar file.
 * This consists of either a header, or a block of data.
 * The end of a tar file is indicated by a block of all zeroes.
 */
typedef	union
{
	TarHeader	header;
	char		data[TAR_BLOCK_SIZE];
} BLOCK;



/*
 * Phases of writing special data.
 * We have to write the file header block, padding the file with zeroes
 * if it is too short, padding bytes after the file data to align it
 * on a block boundary, or the zero end of file block.
 */
typedef enum
{
	PHASE_HEADER,
	PHASE_FILLING,
	PHASE_PADDING,
	PHASE_EOF
} PHASE;


typedef	struct
{
	char *		outputFileName;	/* output file name */
	PHASE		phase;		/* phase of writing special data. */
	off_t		left;		/* data left to write in file */
	off_t		pos;		/* current position in tar file */
	const char  *	dataAddress;	/* current address of data to write */
	int		dataLength;	/* length left of data to write */
	BLOCK		dataBlock;	/* header block to be written */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_TAR,
	"Tar",
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Local procedures.
 */
static	UP_BOOL	readFileName(UP_DEV *);
static	UP_BOOL	readEndOfFile(UP_DEV *);
static	UP_BOOL	transferData(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	void	skipToHeader(UP_DEV * dev);
static	UP_BOOL	putOctal(char * cp, int len, long value);
static	UP_BOOL	formatHeader(TarHeader *, const char *, const UP_STAT *);
static	const char *	findSuffix(const char *);


UP_ID
upAddTar(UP * up, const char * name)
{
	UP_DEV	*	dev;
	DATA	*	data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	data = DEVTODATA(dev);

	/*
	 * Default the output name if required.
	 */
	if (name == NULL)
		name = "output.tar";

	/*
	 * Initialize our private data.
	 */
	data = DEVTODATA(dev);
	data->dataAddress = NULL;
	data->dataLength = 0;
	data->phase = PHASE_HEADER;
	data->left = 0;
	data->pos = 0;

	/*
	 * Copy the output file name.
	 */
	data->outputFileName = upDeviceCopyString(dev, name);

	if (data->outputFileName == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Set our initial state to the writing of a file name.
	 */
	upSetState(dev, UP_STATE_WRITE_NAME);

	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the state which we can act on.
	 */
	state = upCheckState(dev);

	/*
	 * Switch on the state to execute it.
	 */
	switch (state)
	{
	case UP_STATE_NONE:
		/*
		 * We can do nothing right now.
		 */
		return UP_TRUE;

	case UP_STATE_READ_NAME:
		/*
		 * We have a new filename available.
		 */
		return readFileName(dev);

	case UP_STATE_READ_END_OF_BUFFER:
		/*
		 * There are no more files to put into the archive.
		 * Acknowledge the end of buffer.
		 */
		upBufferReadEndOfBuffer(readBuffer);

		/*
		 * Set our state to write out the end of archive block.
		 * This is just a block full of zeroes.
		 */
		data->phase = PHASE_EOF;
		data->dataAddress = NULL;
		data->dataLength = TAR_BLOCK_SIZE;

		upSetState(dev, UP_STATE_WRITE_DATA);

		break;

	case UP_STATE_TRANSFER_DATA:
		/*
		 * We can transfer data from the input to output buffer.
		 */
		return transferData(dev);

	case UP_STATE_TRANSFER_EOF:
		/*
		 * We have an end of file during the transfer.
		 * This means that the input file is too short.
		 */
		upDeviceError(dev, UP_ERR_SHORTFILE);

		return UP_FALSE;

	case UP_STATE_READ_DATA:
		/*
		 * We have some extra data from the input file.
		 * This means that the input file is too long.
		 */
		upDeviceError(dev, UP_ERR_LONGFILE);

		return UP_FALSE;

	case UP_STATE_READ_EOF:
		/*
		 * We have an end of file from the input file.
		 */
		return readEndOfFile(dev);

	case UP_STATE_WRITE_NAME:
		/*
		 * We can write our file name to the output buffer.
		 */
		upBufferWriteFileName(writeBuffer, data->outputFileName, NULL);

		/*
		 * Now set our state to read an input file name.
		 */
		upSetState(dev, UP_STATE_READ_NAME);

		break;

	case UP_STATE_WRITE_DATA:
		/*
		 * We can write some special data to the output buffer.
		 */
		return writeData(dev);

	default:
		/*
		 * We have an unknown state.
		 */
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read the file name, format the tar header for that file name,
 * and set up for that header to be written.
 */
static UP_BOOL
readFileName(UP_DEV * dev)
{
	DATA *			data;
	UP_BUF *		readBuffer;
	const char *		fileName;
	const UP_STAT *		statBuf;
	TarHeader *		header;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the input file's name and status.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	statBuf = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Make sure there is a file name.
	 */
	if (fileName == NULL)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If the file name is too long then do something about it.
	 */
	if (findSuffix(fileName) == NULL)
	{
		fileName = "filename-too-long";
	}

	/*
	 * Point at the header block in our data.
	 */
	header = &data->dataBlock.header;

	/*
	 * Format the header from the file information.
	 */
	if (!formatHeader(header, fileName, statBuf))
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Set up to write the header block out.
	 * Also save the amount to write for the file.
	 */
	data->phase = PHASE_HEADER;
	data->dataAddress = (char *) header;
	data->dataLength = TAR_BLOCK_SIZE;

	if (S_ISDIR(statBuf->mode))
		data->left = 0;
	else
		data->left = statBuf->size;

	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Transfer data from the input buffer to the output buffer.
 * We must be careful to make sure that the amount of data written
 * will match the amount indicated in the file's header.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *		data;
	int		cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Try to transfer some data.
	 */
	cc = upTransferData(dev, UP_LIMIT(data->left));

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Do the accounting.
	 */
	data->pos += cc;
	data->left -= cc;

	/*
	 * If there is no more data to transfer for the file then
	 * set our state to read and discard any additional data.
	 * We expect the read to get an immediate end of file.
	 * (If more data exists, then the file size was incorrect.)
	 */
	if (data->left == 0)
		upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Return saying we did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read an end of file from the input buffer.
 * We set up what is needed to read the next input file.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Read the end of file from the input buffer.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * Set up what is needed for another file.
	 */
	skipToHeader(dev);

	/*
	 * We have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine to write special data to the output buffer.
 * This is either the specified header data, or zeroed data of
 * the specified size.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * If there is data to be written, then write that.
	 * Otherwise write zeroed data.
	 */
	if (data->dataAddress != NULL)
		cc = upWriteData(dev, data->dataAddress, data->dataLength);
	else
		cc = upWriteZeroedData(dev, data->dataLength);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Do accounting.
	 */
	if (data->dataAddress != NULL)
		data->dataAddress += cc;

	data->pos += cc;
	data->dataLength -= cc;

	/*
	 * If there is still more to be written then just return
	 * success for now.
	 */
	if (data->dataLength > 0)
		return upSetDidReadWrite(dev);

	/*
	 * We have finished writing the special data.
	 * Switch on the writing phase to see what we do next.
	 */
	switch (data->phase)
	{
	case PHASE_HEADER:
		/*
		 * If there is any data to be transferred then set our
		 * state to transfer it, otherwise this file is done
		 * and we need to set the state to read any possible
		 * extra data.
		 */
		if (data->left != 0)
			upSetState(dev, UP_STATE_TRANSFER_DATA);
		else
			upSetState(dev, UP_STATE_READ_DATA);

		break;

	case PHASE_FILLING:
	case PHASE_PADDING:
		/*
		 * Decide what we have to do now in order to get to
		 * the next file header position.  If we have to do
		 * nothing, then we set up to read another file name.
		 */
		skipToHeader(dev);

		break;

	case PHASE_EOF:
		/*
		 * We have finished writing the end of tar archive block.
		 * Write the end of file and the end of buffer.
		 */
		upBufferWriteEof(writeBuffer);
		upBufferWriteEndOfBuffer(writeBuffer);

		/*
		 * Mark ourself as finished.
		 */
		return upSetFinished(dev);

	default:
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Return saying we did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Set up to write the next file header by possibly zero filling the rest
 * of the current file and writing possible padding bytes to get to the
 * next block boundary.  If no filling or padding is required, then we
 * set up the state to read the next input file name.
 */
static void
skipToHeader(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Decide what we should do based on our current state.
	 */
	if (data->left > 0)
	{
		/*
		 * There is some data left in the current file.
		 * Set up to zero fill the rest of that data.
		 */
		data->phase = PHASE_FILLING;
		data->dataAddress = NULL;
		data->dataLength = data->left;
		data->left = 0;

		upSetState(dev, UP_STATE_WRITE_DATA);
	}
	else if ((data->pos % TAR_BLOCK_SIZE) != 0)
	{
		/*
		 * The file offset is not at a multiple of a block boundary.
		 * We need to zero pad the rest of the block.
		 */
		data->phase = PHASE_PADDING;
		data->dataAddress = NULL;

		data->dataLength =
			TAR_BLOCK_SIZE - (data->pos % TAR_BLOCK_SIZE);

		upSetState(dev, UP_STATE_WRITE_DATA);
	}
	else
	{
		/*
		 * We are positioned at the correct offset for a file header.
		 * Set up to read another file.
		 */
		data->phase = PHASE_HEADER;
		data->dataAddress = NULL;
		data->dataLength = 0;

		upSetState(dev, UP_STATE_READ_NAME);
	}

	/*
	 * Say we have done I/O.
	 */
	upSetDidReadWrite(dev);
}


/*
 * Format the specified tar header given the specified file name and
 * status information.  Returns UP_TRUE if successful.
 */
static UP_BOOL
formatHeader(TarHeader * header,
	const char * fileName,
	const UP_STAT * statBuf)
{
	const unsigned char *	cp;
	const char *		prefix;
	const char *		suffix;
	int			len;
	long			checkSum;
	size_t			size;
	UP_BOOL			isDir;

	/*
	 * Zero the header block in preparation for filling it in.
	 */
	memset((char *) header, 0, sizeof(*header));

	/*
	 * Split the filename into a prefix and suffix.
	 */
	suffix = findSuffix(fileName);

	if (suffix == NULL)
		return UP_FALSE;

	if (suffix == fileName)
		prefix = NULL;
	else
		prefix = fileName;

	/*
	 * Store the prefix and suffix parts of the file name.
	 */
	strcpy(header->name, suffix);

	if (prefix)
		memcpy(header->prefix, prefix, suffix - prefix - 1);

	/*
	 * See if this is a directory.
	 */
	isDir = S_ISDIR(statBuf->mode);

	/*
	 * Get the size assuming that it is valid.
	 */
	size = statBuf->size;

	/*
	 * If this is a directory then append a slash and make the size zero.
	 */
	if (isDir)
	{
		header->name[strlen(fileName)] = '/';
		size = 0;
	}

	strncpy(header->magic, TAR_MAGIC, sizeof(header->magic));
	strncpy(header->version, TAR_VERSION, sizeof(header->version));

	putOctal(header->mode, sizeof(header->mode), statBuf->mode & 07777);
	putOctal(header->uid, sizeof(header->uid), statBuf->uid);
	putOctal(header->gid, sizeof(header->gid), statBuf->gid);
	putOctal(header->size, sizeof(header->size), size);
	putOctal(header->mtime, sizeof(header->mtime), statBuf->mtime);

	/*
	 * Set the type field appropriately.
	 */
	if (isDir)
		header->typeFlag = TAR_TYPE_DIRECTORY;
	else
		header->typeFlag = TAR_TYPE_REGULAR;

	/*
	 * Calculate and store the checksum.
	 * This is the sum of all of the bytes of the header,
	 * with the checksum field itself treated as blanks.
	 */
	memset(header->checkSum, ' ', sizeof(header->checkSum));

	cp = (const unsigned char *) header;
	len = sizeof(TarHeader);
	checkSum = 0;

	while (len-- > 0)
		checkSum += *cp++;

	putOctal(header->checkSum, sizeof(header->checkSum), checkSum);

	return UP_TRUE;
}


/*
 * Put an octal string into the specified buffer.
 * The number is zero and space padded and possibly null padded.
 * Returns TRUE if successful.
 */
static UP_BOOL
putOctal(char * cp, int len, long value)
{
	int	tempLength;
	char *	tempString;
	char	tempBuffer[32];

	/*
	 * Create a string of the specified length with an initial space,
	 * leading zeroes and the octal number, and a trailing null.
	 */
	tempString = tempBuffer;

	sprintf(tempString, " %0*lo", len - 2, value);

	tempLength = strlen(tempString) + 1;

	/*
	 * If the string is too large, suppress the leading space.
	 */
	if (tempLength > len)
	{
		tempLength--;
		tempString++;
	}

	/*
	 * If the string is still too large, suppress the trailing null.
	 */
	if (tempLength > len)
		tempLength--;

	/*
	 * If the string is still too large, fail.
	 */
	if (tempLength > len)
		return UP_FALSE;

	/*
	 * Copy the string to the field.
	 */
	memcpy(cp, tempString, len);

	return UP_TRUE;
}


/*
 * Check the file name and see if it needs to be split into a prefix
 * and suffix part, returning a pointer to the suffix which remains when
 * the prefix is removed.  If no prefix is required, the original file
 * name is returned.  Returns NULL if the name is too large to be split.
 */
static const char *
findSuffix(const char * fileName)
{
	const char *	endFileName;
	int		length;

	length = strlen(fileName);

	/*
	 * If the filename is small enough then it fits in the suffix.
	 */
	if (length <= TAR_NAME_SIZE)
		return fileName;

	/*
	 * If it is larger than the allowed prefix and suffix then fail.
	 */
	if (length > TAR_NAME_SIZE + TAR_PREFIX_SIZE)
		return NULL;

	/*
	 * Set up to look for a slash which can separate the two parts.
	 */
	endFileName = fileName + length;

	fileName++;

	while (UP_TRUE)
	{
		fileName = strchr(fileName, '/');

		if (fileName == NULL)
			break;

		fileName++;

		/*
		 * If the remaining suffix is small enough then
		 * this will work.
		 */
		if (endFileName - fileName <= TAR_NAME_SIZE)
			return fileName;
	}

	return NULL;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the output file name.
	 */
	free(data->outputFileName);

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	data->dataAddress = NULL;
	data->dataLength = 0;
	data->phase = PHASE_HEADER;
	data->left = 0;
	data->pos = 0;

	memset(&data->dataBlock, 0, sizeof(data->dataBlock));

	/*
	 * Set our initial state to the writing of a file name.
	 */
	upSetState(dev, UP_STATE_WRITE_NAME);

	return UP_TRUE;
}
