/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The unarchive format multi-file device.
 * This reads an archive file to extract its files one by one.
 */

#include <ar.h>

#include "upPrivate.h"


/*
 * Special phases when reading data from the archive file.
 */
typedef enum
{
	PHASE_MAGIC,
	PHASE_HEADER,
	PHASE_IGNORE
} PHASE;


/*
 * Local data for this device.
 */
typedef	struct
{
	PHASE		phase;		/* current input phase */
	char *		dataAddress;	/* address of data left to read */
	off_t		dataLength;	/* length of data left to read */
	off_t		offset;		/* current file offset */
	off_t		left;		/* amount of data left in file */
	struct	ar_hdr	header;		/* current file header */
	char		magic[SARMAG];	/* file magic number */
	char		name[sizeof(struct ar_hdr)];	/* current file name */
	UP_STAT		stat;		/* status of current file */
	UP_BOOL		valid;		/* file information is valid */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_ARCHIVE,
	"Unarchive",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	upDefaultClose,
	doReset
};


/*
 * Other local procedures.
 */
static	UP_BOOL readData(UP_DEV * dev);
static	UP_BOOL readEndOfFile(UP_DEV * dev);
static	UP_BOOL transferData(UP_DEV * dev);
static	UP_BOOL writeFileName(UP_DEV * dev);
static	UP_BOOL	parseHeader(UP_DEV * dev);
static	UP_BOOL	parseNumber(UP_DEV * dev, const char *, int, int, long *);
static	void	skipToHeader(UP_DEV * dev);


UP_ID
upAddUnArchive(UP * up)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Set up to read the magic line at the start of the file.
	 */
	data = DEVTODATA(dev);

	data->phase = PHASE_MAGIC;
	data->dataAddress = data->magic;
	data->dataLength = SARMAG;
	data->offset = 0;
	data->left = 0;

	/*
	 * Set our initial state as reading the archive file name.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_READ_NAME:
			/*
			 * There is a file name to be read.
			 * Read it and ignore it.
			 */
			upBufferReadFileName(readBuffer);

			/*
			 * Set our state to read the magic number in
			 * the header now.
			 */
			upSetState(dev, UP_STATE_READ_DATA);
			data->phase = PHASE_MAGIC;

			break;

		case UP_STATE_READ_DATA:
			/*
			 * We have some special data to be read.
			 * Handle it according to the phase we are in.
			 */
			return readData(dev);

		case UP_STATE_READ_EOF:
			/*
			 * We have an end of file from the file.
			 * Handle it according to the phase we are in.
			 */
			return readEndOfFile(dev);

		case UP_STATE_READ_END_OF_BUFFER:
			/*
			 * We have read the end of the buffer.
			 * Transfer that to the output device.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Set ourself as being finished.
			 */
			return upSetFinished(dev);

		case UP_STATE_TRANSFER_FILE_ERROR:
		case UP_STATE_READ_FILE_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_TRANSFER_DATA:
			/*
			 * Transfer data from the input buffer to the
			 * output buffer.
			 */
			return transferData(dev);

		case UP_STATE_TRANSFER_EOF:
			/*
			 * We got an end of file during the transfer.
			 * Acknowledge the end of file.
			 */
			upBufferReadEof(readBuffer);

			/*
			 * If there was any data left to transfer
			 * then generate an error.
			 */
			if (data->left != 0)
			{
				upDeviceError(dev, UP_ERR_SHORTFILE);

				return UP_FALSE;
			}

			/*
			 * Write an end of file to the output buffer.
			 */
			upBufferWriteEof(writeBuffer);

			/*
			 * Set our state to read another file name now.
			 */
			upSetState(dev, UP_STATE_READ_NAME);

			break;

		case UP_STATE_WRITE_NAME:
			/*
			 * Set the output file name and status for the
			 * output buffer.
			 */
			return writeFileName(dev);

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when there is data which can be transferred
 * between the input and output files.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *	data;
	int	cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * If there is still some more data to transfer in the
	 * current file of the archive, then do that.
	 */
	if (data->left != 0)
	{
		cc = upTransferData(dev, UP_LIMIT(data->left));

		if (cc < 0)
			return UP_FALSE;

		data->offset += cc;
		data->left -= cc;

		return upSetDidReadWrite(dev);
	}

	/*
	 * The current file is finished.
	 * Tell the output buffer that it has an end of file.
	 */
	upBufferWriteEof(dev->writeBuffer);

	/*
	 * The current file of the archive is done.
	 * Set up to read the next file header.
	 */
	skipToHeader(dev);

	return upSetDidReadWrite(dev);
}


/*
 * Routine called when data can be read which is not file data.
 * This parses the magic line at the front of the file and individual
 * file headers.  No output is produced.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;
	int		readCount;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the address and length of the available data.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * Keep examining the available data until there is no more to
	 * process or until we decide to quit looking for special data.
	 */
	while (readLength > 0)
	{
		/*
		 * Calculate the amount of data which we can examine.
		 * This is the minimum of the amount desired and the
		 * amount available.
		 */
		readCount = readLength;

		if (readCount > data->dataLength)
			readCount = data->dataLength;

		/*
		 * If the incoming data should be stored then do that
		 * and advance the pointer address past it.  Otherwise
		 * the incoming data will just be ignored.
		 */
		if (data->dataAddress != NULL)
		{
			memcpy(data->dataAddress, readAddress, readCount);

			data->dataAddress += readCount;
		}

		/*
		 * Do accounting for the reading that we did.
		 */
		upSetDidReadWrite(dev);

		upBufferReadAdjust(readBuffer, readCount);

		readAddress += readCount;
		readLength -= readCount;
		data->dataLength -= readCount;
		data->offset += readCount;

		/*
		 * If there is still some more data to be read for this
		 * phase then we can do nothing more now.
		 */
		if (data->dataLength > 0)
			break;

		/*
		 * All the data has been collected for this phase so that
		 * it is complete.  Process the data that was collected
		 * and then advance to the next phase.
		 */
		switch (data->phase)
		{
		case PHASE_MAGIC:
			/*
			 * We have just read in the first magic header line.
			 * Verify that this is really an archive file.
			 */
			if (memcmp(ARMAG, data->magic, SARMAG))
			{
				upDeviceError(dev, UP_ERR_NOTARCHIVE);

				return UP_FALSE;
			}

			/*
			 * Set up to read a file header now.
			 */
			data->phase = PHASE_HEADER;
			data->dataAddress = (char *) &data->header;
			data->dataLength = sizeof(data->header);

			break;

		case PHASE_HEADER:
			/*
			 * A new file header has been read in.
			 * Examine the header and parse it for the file
			 * name, file size, and other status.
			 */
			if (!parseHeader(dev))
				return UP_FALSE;

			/*
			 * The file is to be retrieved from the archive.
			 * Set up to tell the output buffer about it.
			 */
			data->dataAddress = NULL;
			data->dataLength = 0;

			upSetState(dev, UP_STATE_WRITE_NAME);

			/*
			 * Set our read length to zero to terminate
			 * the loop.
			 */
			readLength = 0;

			break;

		case PHASE_IGNORE:
			/*
			 * We have been ignoring the input because we
			 * needed to consume the padding byte after a file
			 * or need to skip to the next file.
			 */
			skipToHeader(dev);

			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
		}
	}

	/*
	 * Say we have done some I/O and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when we have an end of file while reading special data.
 * We make sure that the end of file happened at the correct place.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Make sure that the end of file occurred when we were
	 * starting to read a file header.
	 */
	if ((data->phase != PHASE_HEADER) ||
		(data->dataLength != sizeof(data->header)))
	{
		upDeviceError(dev, UP_ERR_SHORTFILE);

		return UP_FALSE;
	}

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * Set up to read another archive file.
	 */
	data->phase = PHASE_MAGIC;
	data->dataAddress = data->magic;
	data->dataLength = SARMAG;

	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * Say we have done I/O successfully.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to write the output file name to the output buffer.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * We can write the file name now.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer, data->name, &data->stat))
		return UP_FALSE;

	/*
	 * Set up to transfer the file's data.
	 */
	data->dataAddress = NULL;
	data->dataLength = 0;
	data->offset = 0;
	data->left = data->stat.size;

	upSetState(dev, UP_STATE_TRANSFER_DATA);

	return upSetDidReadWrite(dev);
}


/*
 * Set up to read the next file header by possibly skipping the rest
 * of the current file and a possible padding byte.  If the header is
 * the next thing to read, then we set up the phase to read the header.
 */
static void
skipToHeader(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Set our state to read the file data.
	 */
	upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * If there is any data left in the current file, then set up to
	 * ignore the rest of that data.
	 */
	if (data->left > 0)
	{
		data->phase = PHASE_IGNORE;
		data->dataAddress = NULL;
		data->dataLength = data->left;
		data->left = 0;

		return;
	}

	/*
	 * If the file offset is odd, then we need to ignore one padding
	 * byte to make the file offset even for the next header.
	 */
	if ((data->offset & 0x1) != 0)
	{
		data->phase = PHASE_IGNORE;
		data->dataAddress = NULL;
		data->dataLength = 1;

		return;
	}

	/*
	 * We are positioned at the correct offset for a file header.
	 * Set up to read it.
	 */
	data->phase = PHASE_HEADER;
	data->dataAddress = (char *) &data->header;
	data->dataLength = sizeof(data->header);
}


/*
 * Parse the file header that has just been read, and store the information
 * into our private data.  Returns UP_TRUE if the header looks reasonable.
 */
static UP_BOOL
parseHeader(UP_DEV * dev)
{
	DATA *			data;
	const struct ar_hdr *	head;
	char *			cp;
	long			value;

	data = DEVTODATA(dev);

	data->valid = UP_FALSE;
	head = &data->header;

	/*
	 * Copy the file name from the header and remove trailing
	 * blanks and the trailing slash.
	 */
	memcpy(data->name, head->ar_name, sizeof(head->ar_name));
	data->name[sizeof(data->name) - 1] = '\0';

	cp = data->name + strlen(data->name);

	while ((cp != data->name) && (cp[-1] == ' '))
		cp--;

	if ((cp != data->name) && (cp[-1] == '/'))
		cp--;

	*cp = '\0';

	/*
	 * Set the flags for the information which is available.
	 */
	memset((char *) &data->stat, 0, sizeof(data->stat));

	data->stat.flags = (UP_STAT_TYPE | UP_STAT_MODE | UP_STAT_UID |
		UP_STAT_GID | UP_STAT_SIZE | UP_STAT_MTIME | UP_STAT_DATA);

	data->stat.type = '-';

	/*
	 * Parse the date.
	 */
	if (!parseNumber(dev, head->ar_date, sizeof(head->ar_date), 10, &value))
		return UP_FALSE;

	data->stat.mtime = value;

	/*
	 * Parse the user id.
	 */
	if (!parseNumber(dev, head->ar_uid, sizeof(head->ar_uid), 8, &value))
		return UP_FALSE;

	data->stat.uid = value;

	/*
	 * Parse the group id.
	 */
	if (!parseNumber(dev, head->ar_gid, sizeof(head->ar_gid), 8, &value))
		return UP_FALSE;

	data->stat.gid = value;

	/*
	 * Parse the file mode.
	 */
	if (!parseNumber(dev, head->ar_mode, sizeof(head->ar_mode), 8, &value))
		return UP_FALSE;

	data->stat.mode = (value & 0777) | S_IFREG;

	/*
	 * Parse the file size.
	 */
	if (!parseNumber(dev, head->ar_size, sizeof(head->ar_size), 10, &value))
		return UP_FALSE;

	data->stat.size = value;

	/*
	 * Remember the size of the file, and that we have valid information.
	 */
	data->valid = UP_TRUE;

	return UP_TRUE;
}


/*
 * Parse an octal or decimal string of the specified length.
 * The string may contain leading or trailing spaces, and might end on a null.
 * Returns UP_TRUE on succes.
 */
static UP_BOOL
parseNumber(UP_DEV * dev, const char * cp, int len, int base, long * retvalue)
{
	long	value;
	UP_BOOL	successFlag;

	if ((base < 2) || (base > 10))
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	successFlag = UP_TRUE;
	value = 0;

	while ((len > 0) && (*cp == ' '))
	{
		cp++;
		len--;
	}
#if 0
	if ((len == 0) || (*cp < '0') || (*cp >= '0' + base))
		successFlag = UP_FALSE;
#endif
	while ((len > 0) && (*cp >= '0') && (*cp < '0' + base))
	{
		value = value * base + *cp++ - '0';
		len--;
	}

	while ((len > 0) && (*cp == ' '))
	{
		cp++;
		len--;
	}

	if ((len > 0) && (*cp != '\0'))
		successFlag = UP_FALSE;

	*retvalue = value;

	if (!successFlag)
		upDeviceError(dev, UP_ERR_BADHEADER);

	return successFlag;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Clear all of our data.
	 */
	memset(data, 0, sizeof(data));

	/*
	 * Set up to read the magic line again.
	 */
	data->phase = PHASE_MAGIC;
	data->dataAddress = data->magic;
	data->dataLength = SARMAG;
	data->offset = 0;
	data->left = 0;

	upSetState(dev, UP_STATE_READ_NAME);

	return UP_TRUE;
}
