/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The untar format device.
 */
#include <sys/types.h>
#include <sys/stat.h>

#include "upPrivate.h"


/*
 * Macro to check for an octal digit.
 */
#define	isoctal(ch)	(((ch) >= '0') && ((ch) <= '7'))


/*
 * Tar file constants.
 */
#define TAR_BLOCK_SIZE	512
#define TAR_NAME_SIZE	100
#define	TAR_PREFIX_SIZE	155

#define	TAR_MAGIC	"ustar"
#define	TAR_VERSION	"00"

#define	TAR_TYPE_REGULAR	'0'
#define	TAR_TYPE_HARD_LINK	'1'
#define	TAR_TYPE_SOFT_LINK	'2'
#define	TAR_TYPE_CHAR_DEVICE	'3'
#define	TAR_TYPE_BLOCK_DEVICE	'4'
#define	TAR_TYPE_DIRECTORY	'5'


/*
 * The POSIX (and basic GNU) tar header format.
 * This structure is always embedded in a TAR_BLOCK_SIZE sized block
 * with zero padding.
 */
typedef struct
{
	char	name[TAR_NAME_SIZE];
	char	mode[8];
	char	uid[8];
	char	gid[8];
	char	size[12];
	char	mtime[12];
	char	checkSum[8];
	char	typeFlag;
	char	linkName[TAR_NAME_SIZE];
	char	magic[6];
	char	version[2];
	char	uname[32];
	char	gname[32];
	char	devMajor[8];
	char	devMinor[8];
	char	prefix[TAR_PREFIX_SIZE];
} TarHeader;


/*
 * One block of a tar file.
 * This consists of either a header, or a block of data.
 * The end of a tar file is indicated by a block of all zeroes.
 */
typedef	union
{
	TarHeader	header;
	char		data[TAR_BLOCK_SIZE];
} BLOCK;


/*
 * Phases of reading special data.
 * We have to read file header blocks, possibly ignore file data,
 * and ignore trailing bytes after the file data or after the end
 * of tar block.
 */
typedef enum
{
	PHASE_HEADER,
	PHASE_IGNORE,
	PHASE_PADDING,
	PHASE_TRAILER
} PHASE;


/*
 * Our private data.
 */
typedef	struct
{
	PHASE	phase;		/* current reading phase */
	char *	targetName;	/* target name to find */
	char *	dataAddress;	/* address of data left to read */
	int	dataLength;	/* length of data left to read */
	BLOCK	block;		/* current block containing possible header */
	char	name[TAR_NAME_SIZE + TAR_PREFIX_SIZE + 2];	/* filename */
	UP_STAT	stat;		/* status of current file */
	UP_BOOL	valid;		/* status is valid */
	UP_BOOL	found;		/* file has been found */
	UP_BOOL	eofBlock;	/* have read end of file block */
	long	left;		/* data left to read in current file */
	off_t	offset;		/* current offset in tar file */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_TAR,
	"Untar",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Other local procedures.
 */
static	UP_BOOL	readData(UP_DEV * dev);
static	UP_BOOL	writeFileName(UP_DEV * dev);
static	UP_BOOL	transferData(UP_DEV * dev);
static	UP_BOOL parseHeader(UP_DEV * dev);
static	void	skipToHeader(UP_DEV * dev);
static	long	getOctal(const char * cp, int len);


UP_ID
upAddUnTar(UP * up, const char * fileName)
{
	UP_DEV	*	dev;
	DATA	*	data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Start by reading a header block.
	 */
	data = DEVTODATA(dev);

	data->phase = PHASE_HEADER;
	data->dataAddress = data->block.data;
	data->dataLength = TAR_BLOCK_SIZE;

	/*
	 * We want to start by reading a file name.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * If a file name has been given then copy it.
	 */
	if (fileName != NULL)
	{
		data->targetName = upDeviceCopyString(dev, fileName);

		if (data->targetName == NULL)
		{
			upFreeDevice(dev);

			return UP_NULL_ID;
		}
	}

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the state which we can act on.
	 */
	state = upCheckState(dev);

	/*
	 * Switch on the state to execute it.
	 */
	switch (state)
	{
	case UP_STATE_NONE:
		/*
		 * We can do nothing right now.
		 */
		return UP_TRUE;

	case UP_STATE_READ_NAME:
		/*
		 * There is a new file name to be read.
		 * Read it and ignore it.
		 */
		upBufferReadFileName(readBuffer);

		/*
		 * Set our state to read the tar data now.
		 */
		upSetState(dev, UP_STATE_READ_DATA);

		break;

	case UP_STATE_READ_END_OF_BUFFER:
		/*
		 * The end of the buffer has been reached.
		 * Read it and pass it to the output buffer.
		 */
		upBufferReadEndOfBuffer(readBuffer);
		upBufferWriteEndOfBuffer(writeBuffer);

		/*
		 * Return saying we are finished.
		 */
		return upSetFinished(dev);

	case UP_STATE_WRITE_NAME:
		/*
		 * We can write an output file name now.
		 */
		return writeFileName(dev);

	case UP_STATE_READ_DATA:
		/*
		 * Special data can be read from the input buffer.
		 */
		return readData(dev);

	case UP_STATE_READ_EOF:
		/*
		 * An end of file can be read from the input buffer.
		 * This is only expected after we have read the end of
		 * file block.
		 */
		if (data->phase != PHASE_TRAILER)
		{
			upDeviceError(dev, UP_ERR_SHORTFILE);

			return UP_FALSE;
		}

		/*
		 * Read the end of file from the input buffer.
		 */
		upBufferReadEof(readBuffer);

		/*
		 * Set our state to look for another file name.
		 */
		upSetState(dev, UP_STATE_READ_NAME);

		break;

	case UP_STATE_READ_FILE_ERROR:
	case UP_STATE_TRANSFER_FILE_ERROR:
		/*
		 * A read error has occurred.
		 */
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;

	case UP_STATE_TRANSFER_DATA:
		/*
		 * Data can be transferred from the input buffer to
		 * the output buffer.
		 */
		return transferData(dev);

	case UP_STATE_TRANSFER_EOF:
		/*
		 * An end of file can be read from the input
		 * buffer while file data was being read.
		 * This is not expected.
		 */
		upDeviceError(dev, UP_ERR_SHORTFILE);

		return UP_FALSE;

	case UP_STATE_FATAL_ERROR:
		/*
		 * There was a fatal error of some kind.
		 */
		return UP_FALSE;

	default:
		/*
		 * We have an unknown state.
		 */
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when reading data which is not file data.
 * This is either a file header block, or padding after a file has
 * been read.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	readAddress;
	int		readLength;
	int		readCount;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the address and length of the available data.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * Keep examining the available data until there is no more to
	 * process or until we decide to quit looking for special data.
	 */
	while (readLength > 0)
	{
		/*
		 * Calculate the amount of data which we can examine.
		 * This is the minimum of the amount desired and the
		 * amount available.
		 */
		readCount = readLength;

		if (readCount > data->dataLength)
			readCount = data->dataLength;

		/*
		 * If the incoming data should be stored then do that
		 * and advance the pointer address past it.  Otherwise
		 * the incoming data will just be ignored.
		 */
		if (data->dataAddress != NULL)
		{
			memcpy(data->dataAddress, readAddress, readCount);

			data->dataAddress += readCount;
		}

		/*
		 * Do accounting for the reading that we did.
		 */
		upBufferReadAdjust(readBuffer, readCount);
		upSetDidReadWrite(dev);

		readAddress += readCount;
		readLength -= readCount;
		data->dataLength -= readCount;
		data->offset += readCount;

		/*
		 * If there is still some more data to be read for this
		 * phase then we can do nothing more now.
		 */
		if (data->dataLength > 0)
			break;

		/*
		 * All the data has been collected for this phase so that
		 * it is complete.  Process the data that was collected
		 * and then advance to the next phase.
		 */
		switch (data->phase)
		{
		case PHASE_HEADER:
			/*
			 * A new file header has been read in.
			 * Examine the header and parse it for the file
			 * name, file size, and other status.
			 */
			if (!parseHeader(dev))
				return UP_FALSE;

			/*
			 * If the header block indicates an end of file,
			 * then handle that and return.
			 */
			if (data->eofBlock)
			{
				/*
				 * See if we were supposed to find a
				 * particular file name.
				 */
				if ((data->targetName != NULL) && !data->found)
				{
					upDeviceError(dev, UP_ERR_FILENOTFOUND);

					return UP_FALSE;
				}

				/*
				 * Set our state to consume any additional
				 * data in the tar file.  Just set a large
				 * data count for now.
				 */
				data->phase = PHASE_TRAILER;
				data->dataAddress = NULL;
				data->dataLength = 1024*1024;

				break;
			}

			/*
			 * If a particular file name has been given to be
			 * read, and this file does does not match that one,
			 * then set up to ignore all of the file's data.
			 */
			if ((data->targetName != NULL) &&
				(strcmp(data->targetName, data->name) != 0))
			{
				data->phase = PHASE_IGNORE;
				data->dataAddress = NULL;
				data->dataLength = data->stat.size;
				data->left = 0;

				break;
			}

			/*
			 * We want to transfer this file.
			 * Change our state to write the output file name.
			 */
			upSetState(dev, UP_STATE_WRITE_NAME);

			/*
			 * Return saying we have done I/O.
			 */
			return upSetDidReadWrite(dev);

		case PHASE_IGNORE:
		case PHASE_PADDING:
			/*
			 * We have been ignoring the input data, either
			 * because a file was not wanted, or because we
			 * need to consume the padding bytes after a file.
			 * Set up to read the next header.
			 */
			skipToHeader(dev);

			break;

		case PHASE_TRAILER:
			/*
			 * We are consuming trailing data after the eof
			 * block.  Just continue consuming it until we
			 * reach the end of file.
			 */
			data->dataLength = 1024*1024;

			break;

		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
		}
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when there is data which can be transferred
 * between the input and output files.
 */
static UP_BOOL
transferData(UP_DEV * dev)
{
	DATA *	data;
	int	cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Transfer as much of the data as we can.
	 */
	cc = upTransferData(dev, UP_LIMIT(data->left));

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Do the accounting for the amount transferred.
	 */
	data->offset += cc;
	data->left -= cc;

	/*
	 * If there is still some left to transfer then just
	 * return saying we have done I/O.
	 */
	if (data->left > 0)
		return upSetDidReadWrite(dev);

	/*
	 * The current file is finished.
	 * Tell the output buffer that it has an end of file.
	 */
	upBufferWriteEof(dev->writeBuffer);

	/*
	 * The current file of the archive is done.
	 * Set up to skip to the next file header.
	 */
	skipToHeader(dev);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write the file name and status for the file being untarred,
 * and then set our state to begin the file data transfer.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;

	/*
	 * Point to our private data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Write the filename and status to the output buffer.
	 */
	if (!upBufferWriteFileName(writeBuffer, data->name, &data->stat))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If this is not a regular file or there is no data to be written
	 * for the file then output an end of file and set our state to
	 * read another file name.
	 */
	if (!S_ISREG(data->stat.mode) || (data->stat.size == 0))
	{
		upBufferWriteEof(writeBuffer);

		skipToHeader(dev);

		return upSetDidReadWrite(dev);
	}

	/*
	 * There is some data to be written.
	 * Set up for the file transfer.
	 */
	data->dataAddress = NULL;
	data->dataLength = 0;
	data->left = data->stat.size;
	data->found = UP_TRUE;

	/*
	 * Set our state to transfer the file data.
	 */
	upSetState(dev, UP_STATE_TRANSFER_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Set up to read the next file header by possibly skipping the rest
 * of the current file and possible padding bytes.  If the header is
 * the next thing to read, then we set up the phase to read the header.
 */
static void
skipToHeader(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Decide what we should do based on our current state.
	 */
	if (data->left > 0)
	{
		/*
		 * There is some data left in the current file.
		 * Set up to ignore the rest of that data.
		 */
		data->phase = PHASE_IGNORE;
		data->dataAddress = NULL;
		data->dataLength = data->left;
		data->left = 0;
	}
	else if ((data->offset % TAR_BLOCK_SIZE) != 0)
	{
		/*
		 * The file offset is not at a multiple of a block boundary.
		 * We need to ignore the rest of the block.
		 */
		data->phase = PHASE_PADDING;
		data->dataAddress = NULL;

		data->dataLength =
			TAR_BLOCK_SIZE - (data->offset % TAR_BLOCK_SIZE);
	}
	else
	{
		/*
		 * We are positioned at the correct offset for a file header.
		 * Set up to read it.
		 */
		data->phase = PHASE_HEADER;
		data->dataAddress = data->block.data;
		data->dataLength = TAR_BLOCK_SIZE;
	}

	/*
	 * Set our state to be reading of data.
	 */
	upSetState(dev, UP_STATE_READ_DATA);

	/*
	 * Say we have done I/O.
	 */
	upSetDidReadWrite(dev);
}


/*
 * Parse the file header that has just been read, and store the information
 * into our private data.  Returns UP_TRUE if the header looks reasonable.
 */
static UP_BOOL
parseHeader(UP_DEV * dev)
{
	DATA *			data;
	const TarHeader *	head;
	UP_STAT *		stat;
	const char *		name;
	char *			cp;
	int			checkSum;
	int			cc;
	UP_BOOL			isDir;
	UP_BOOL			isHardLink;
	UP_BOOL			isSoftLink;

	data = DEVTODATA(dev);

	data->valid = UP_FALSE;
	head = &data->block.header;
	stat = &data->stat;
	isDir = UP_FALSE;

	/*
	 * If the block is completely empty, then this is the end of the
	 * archive file.  If the name is null, then just skip this header.
	 */
	name = head->name;

	if (*name == '\0')
	{
		for (cc = TAR_BLOCK_SIZE; cc > 0; cc--)
		{
			if (*name++)
				return UP_TRUE;
		}

		data->eofBlock = UP_TRUE;

		return UP_TRUE;
	}

	/*
	 * Copy the file name and remove any trailing slash for directories.
	 * The file name may have a prefix part.
	 */
	cp = data->name;

	*cp = '\0';

	if (head->prefix[0] != '\0')
	{
		memcpy(cp, head->prefix, TAR_PREFIX_SIZE);
		cp[TAR_PREFIX_SIZE] = '\0';
		cp += strlen(cp);
		*cp++ = '/';
		*cp = '\0';
	}

	memcpy(cp, name, TAR_NAME_SIZE);

	cp[TAR_NAME_SIZE] = '\0';

	cp = strrchr(data->name, '/');

	if ((cp != NULL) && (cp[1] == '\0'))
	{
		isDir = UP_TRUE;
		*cp = '\0';
	}

	if (head->typeFlag == TAR_TYPE_DIRECTORY)
		isDir = UP_TRUE;

	isHardLink = ((head->typeFlag == 1) ||
		(head->typeFlag == TAR_TYPE_HARD_LINK));

	isSoftLink = ((head->typeFlag == 2) ||
		(head->typeFlag == TAR_TYPE_SOFT_LINK));

	/*
	 * Set the flags for the information which is available.
	 */
	memset((char *) stat, 0, sizeof(UP_STAT));

	stat->flags = (UP_STAT_TYPE | UP_STAT_MODE | UP_STAT_UID |
		UP_STAT_GID | UP_STAT_SIZE | UP_STAT_MTIME);

	stat->type = '-';
	stat->isLink = isHardLink;

	/*
	 * Parse the fields in the header.
	 */
	stat->mode = getOctal(head->mode, sizeof(head->mode));
	stat->uid = getOctal(head->uid, sizeof(head->uid));
	stat->gid = getOctal(head->gid, sizeof(head->gid));
	stat->size = getOctal(head->size, sizeof(head->size));
	stat->mtime = getOctal(head->mtime, sizeof(head->mtime));
	checkSum = getOctal(head->checkSum, sizeof(head->checkSum));

	if ((stat->mode < 0) || (stat->uid < 0) || (stat->gid < 0) ||
		(stat->size == -1))
	{
		upDeviceError(dev, UP_ERR_BADHEADER);

		return UP_FALSE;
	}

	/*
	 * Construct the file mode and type as best we can.
	 */
	if (isDir)
	{
		stat->flags &= ~UP_STAT_DATA;
		stat->mode |= S_IFDIR;
		stat->type = 'd';
		stat->size = 0;
	}
#ifdef	S_IFLNK
	else if (isSoftLink)
	{
		stat->mode |= S_IFLNK;
		stat->mode = 'l';
	}
#endif
	else if ((stat->mode & S_IFMT) == 0)
		stat->mode |= S_IFREG;

	data->valid = UP_TRUE;

	return UP_TRUE;
}


/*
 * Read an octal value in a field of the specified width, with optional
 * spaces on both sides of the number and with an optional null character
 * at the end.  Returns -1 on an illegal format.
 */
static long
getOctal(const char * cp, int len)
{
	long	val;

	while ((len > 0) && (*cp == ' '))
	{
		cp++;
		len--;
	}

	if ((len == 0) || !isoctal(*cp))
		return -1;

	val = 0;

	while ((len > 0) && isoctal(*cp))
	{
		val = val * 8 + *cp++ - '0';
		len--;
	}

	while ((len > 0) && (*cp == ' '))
	{
		cp++;
		len--;
	}

	if ((len > 0) && *cp)
		return -1;

	return val;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;
	char *	oldTarget;

	data = DEVTODATA(dev);

	/*
	 * Clear all the fields except for the target name.
	 */
	oldTarget = data->targetName;
	memset(data, 0, sizeof(data));
	data->targetName = oldTarget;

	/*
	 * Set up to read a file header again.
	 */
	data->phase = PHASE_HEADER;
	data->dataAddress = data->block.data;
	data->dataLength = TAR_BLOCK_SIZE;

	upSetState(dev, UP_STATE_READ_NAME);

	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the file name we may have allocated.
	 */
	free(data->targetName);

	return UP_TRUE;
}
