/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The uudecode device.
 */

#include "upPrivate.h"


/*
 * Maximum number of output characters from one line.
 */
#define	OUTSIZE	64


/*
 * Macro to decode a character.
 */
#define	DEC(ch)	(((ch) - ' ') & 077)


/*
 * The beginning of the uuencoded data and its length.
 * There is more on the line after this which needs to be parsed.
 */
#define	BEGIN_STRING	"begin "
#define	BEGIN_LENGTH	6


/*
 * The end of the uuencoded data and its length.
 * This is on a line all by itself.
 */
#define	END_STRING	"end\n"
#define	END_LENGTH	4


#define	isblank(ch)	((ch) == ' ')
#define	isoctal(ch)	(((ch) >= '0') && ((ch) <= '7'))


/*
 * The phases that we can be in while reading data.
 */
typedef enum
{
	PHASE_HEADER,		/* looking for header line */
	PHASE_DATA,		/* reading data */
	PHASE_SKIP		/* skipping data */
} PHASE;


/*
 * The types of lines that are important to us.
 */
typedef enum
{
	TYPE_BEGIN,		/* a begin line */
	TYPE_DATA,		/* a normal data line */
	TYPE_EMPTY,		/* an empty line */
	TYPE_END		/* an end line */
} TYPE;


/*
 * Local data for this device.
 */
typedef	struct
{
	PHASE		phase;		/* current phase */
	UP_BOOL		found;		/* file name was found */
	int		offset;		/* offset of data to output */
	int		count;		/* count of bytes left to output */
	char *		targetName;	/* target file name */
	char *		name;		/* name of current file */
	int		mode;		/* mode of current file */
	char		buf[OUTSIZE];	/* buffer for output data */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))

static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_UUDECODE,
	"Uudecode",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	upDefaultBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Other local procedures.
 */
static	UP_BOOL	readLine(UP_DEV *);
static	UP_BOOL	readEndOfFile(UP_DEV *);
static	UP_BOOL	writeFileName(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	UP_BOOL	ParseHeader(UP_DEV *, const char *, int);
static	TYPE	ClassifyLine(const char *, int);
static	void	FillBuffer(UP_DEV *, const char *, int);


/*
 * Add the uudecode device to the pipeline.
 * This restores one or more uuencoded files from the input stream.
 */
UP_ID
upAddUUDecode(UP * up, const char * fileName)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	data = DEVTODATA(dev);

	/*
	 * Initialize the fields.
	 */
	data->phase = PHASE_HEADER;
	data->name = NULL;
	data->targetName = NULL;
	data->offset = 0;
	data->count = 0;
	data->mode = 0;

	/*
	 * Set our initial state to read a file name.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	/*
	 * If there is a target file name then save it.
	 * If there is no target file name then all uuencoded files
	 * will be found.
	 */
	if (fileName)
	{
		data->targetName = upDeviceCopyString(dev, fileName);

		if (data->targetName == NULL)
		{
			upFreeDevice(dev);

			return UP_NULL_ID;
		}
	}

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	int		state;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_READ_NAME:
			/*
			 * There is a file name to be read.
			 * Read it and ignore it.
			 */
			upBufferReadFileName(readBuffer);

			/*
			 * Set our state to search for the next header line.
			 */
			upSetState(dev, UP_STATE_READ_LINE);
			data->phase = PHASE_HEADER;

			break;

		case UP_STATE_READ_LINE:
			/*
			 * We have a line of data to be read.
			 * Handle it according to the phase we are in
			 * and the type of line that it is.
			 */
			return readLine(dev);

		case UP_STATE_READ_EOF:
			/*
			 * We have an end of file from the file.
			 * Handle it according to the phase we are in.
			 */
			return readEndOfFile(dev);

		case UP_STATE_READ_END_OF_BUFFER:
			/*
			 * We have read the end of the buffer.
			 * Transfer that to the output device.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Set ourself as being finished.
			 */
			return upSetFinished(dev);

		case UP_STATE_READ_FILE_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_WRITE_DATA:
			/*
			 * Write decoded data to the output buffer.
			 */
			return writeData(dev);

		case UP_STATE_WRITE_NAME:
			/*
			 * Set the output file name and status for the
			 * output buffer.
			 */
			return writeFileName(dev);

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to read a line of data from the input file.
 * The data is handled according to our current phase.
 */
static UP_BOOL
readLine(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;
	TYPE		type;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the address of the available line to be read.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLineLength(readBuffer);

	/*
	 * Make sure that there is a line there.
	 */
	if (readLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Check for a line too long problem.
	 */
	if (readLength < 0)
	{
		upDeviceError(dev, UP_ERR_LINETOOLONG);

		return UP_FALSE;
	}

	/*
	 * Classify the line according to its type.
	 */
	type = ClassifyLine(readAddress, readLength);

	/*
	 * Now process it depending on the current reading phase.
	 */
	switch (data->phase)
	{
		case PHASE_HEADER:
			/*
			 * We are searching for the next header line.
			 * If this is not a begin line then ignore it.
			 */
			if (type != TYPE_BEGIN)
				break;

			/*
			 * This is a header line, so parse it.
			 */
			if (!ParseHeader(dev, readAddress, readLength))
				return UP_FALSE;

			/*
			 * If we have a target file name specified and
			 * it matches this one, then we want to start
			 * transferring data.  Otherwise we want to skip
			 * lines until we see the end of this file.
			 */
			if ((data->targetName != NULL) &&
				(strcmp(data->targetName, data->name) != 0))
			{
				data->phase = PHASE_SKIP;

				break;
			}

			/*
			 * We want to transfer this file.
			 * Set up to write the file name to the output buffer.
			 */
			upSetState(dev, UP_STATE_WRITE_NAME);

			break;

		case PHASE_DATA:
			/*
			 * We have been reading some uuencoded data lines.
			 * Check for the end line of the uuencoded data.
			 * If we found it then set end of file for the
			 * write buffer and set up to look for another
			 * header line.
			 */
			if (type == TYPE_END)
			{
				upBufferWriteEof(dev->writeBuffer);

				data->phase = PHASE_HEADER;

				break;
			}

			/*
			 * Check for a duplicate header line.
			 * This means we haven't seen a end line.
			 */
			if (type == TYPE_BEGIN)
			{
				upDeviceError(dev, UP_ERR_DUPLICATEBEGINLINE);

				return UP_FALSE;
			}

			/*
			 * Fill our internal buffer with the data that is
			 * available from this line.
			 */
			FillBuffer(dev, readAddress, readLength);

			break;

		case PHASE_SKIP:
			/*
			 * Check for the end line of the uuencoded data.
			 * If we found it then set the phase to begin
			 * looking for another header line.
			 */
			if (type == TYPE_END)
				data->phase = PHASE_HEADER;

			/*
			 * Check for a duplicate header line.
			 */
			if (type == TYPE_BEGIN)
			{
				upDeviceError(dev, UP_ERR_DUPLICATEBEGINLINE);

				return UP_FALSE;
			}

			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Account for the line that was just read since we
	 * always completely consume it.
	 */
	upBufferReadAdjust(readBuffer, readLength);

	return upSetDidReadWrite(dev);
}


/*
 * Read an end of file from the input buffer.
 * This should occur only between file data since uuencoded data
 * is terminated by an end line.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * We have an end of file.
	 * If we were reading or skipping data then
	 * generate an error.
	 */
	if ((data->phase == PHASE_DATA) ||
		(data->phase == PHASE_SKIP))
	{
		upDeviceError(dev, UP_ERR_MISSINGENDLINE);

		return UP_FALSE;
	}

	/*
	 * Make sure that we found at least one file.
	 */
#if 0
	if (!data->found)
	{
		upDeviceError(dev, UP_ERR_FILENOTFOUND);

		return UP_FALSE;
	}
#endif

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * Set our state for reading the next input file.
	 */
	upSetState(dev, UP_STATE_READ_NAME);

	data->phase = PHASE_HEADER;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called after we have found the beginning of a uuencoded
 * file to write the file name to the output buffer and then set
 * up read the data.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	UP_STAT		statBuf;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Build a status structure.
	 */
	memset(&statBuf, 0, sizeof(statBuf));

	statBuf.flags = UP_STAT_TYPE | UP_STAT_MODE | UP_STAT_DATA;
	statBuf.type = '-';
	statBuf.mode = (data->mode & 0777) | S_IFREG;

	/*
	 * Tell the output buffer what its file name and status is.
	 */
	upBufferWriteFileName(writeBuffer, data->name, &statBuf);

	/*
	 * Set our state to read the file data.
	 */
	upSetState(dev, UP_STATE_READ_LINE);

	data->phase = PHASE_DATA;

	/*
	 * Say we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to write data to the output buffer.
 * This data is the decoded data from one input line.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		cc;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Write as much of our data as we can.
	 */
	cc = upWriteData(dev, &data->buf[data->offset], data->count);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Adjust our buffer values.
	 */
	data->offset += cc;
	data->count -= cc;

	/*
	 * If we have no more data left to write from the buffer then
	 * set up to read another uuencoded line.
	 */
	if (data->count <= 0)
	{
		upSetState(dev, UP_STATE_READ_LINE);

		data->phase = PHASE_DATA;
	}

	/*
	 * Return saying we did some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Fill the data buffer with uudecoded data obtained from the specified
 * line having the specified length, and make it available.  The data
 * length must be within the range 0 to OUTSIZE.
 */
static void
FillBuffer(UP_DEV * dev, const char * buf, int len)
{
	DATA *		data;
	const char *	incp;
	char *		outcp;
	int		cc;

	/*
	 * Set up to parse the line and store the output.
	 */
	data = DEVTODATA(dev);

	incp = buf;
	outcp = data->buf;

	/*
	 * Get the number of bytes to decode on the line.
	 */
	cc = DEC(*incp++);

	if (cc == 0)
	{
		data->offset = 0;
		data->count = 0;

		return;
	}

	while (cc >= 3)
	{
		*outcp++ = (DEC(incp[0]) << 2) | (DEC(incp[1]) >> 4);
		*outcp++ = (DEC(incp[1]) << 4) | (DEC(incp[2]) >> 2);
		*outcp++ = (DEC(incp[2]) << 6) | (DEC(incp[3]));

		incp += 4;
		cc -= 3;
	}

	if (cc >= 1)
		*outcp++ = (DEC(incp[0]) << 2) | (DEC(incp[1]) >> 4);

	if (cc >= 2)
		*outcp++ = (DEC(incp[1]) << 4) | (DEC(incp[2]) >> 2);

	/*
	 * Set up to read the stored data.
	 */
	data->offset = 0;
	data->count = outcp - data->buf;

	/*
	 * Set our state to write the data to the output buffer.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the strings we might have allocated.
	 */
	free(data->name);
	free(data->targetName);

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	free(data->name);

	data->phase = PHASE_HEADER;
	data->name = NULL;
	data->offset = 0;
	data->count = 0;

	upSetState(dev, UP_STATE_READ_NAME);

	return UP_TRUE;
}


/*
 * Classify the specified line having the specified length.
 * The line includes the terminating newline character if any.
 * Returns the type of line found.
 */
static TYPE
ClassifyLine(const char * buf, int len)
{
	/*
	 * Check for an empty line.
	 */
	if ((len == 0) || ((len == 1) && (*buf == '\n')))
		return TYPE_EMPTY;

	/*
	 * Check for a begin line.
	 */
	if ((len > BEGIN_LENGTH) &&
		(memcmp(buf, BEGIN_STRING, BEGIN_LENGTH) == 0))
	{
		return TYPE_BEGIN;
	}

	/*
	 * Check for an end line.
	 */
	if ((len == END_LENGTH) &&
		(memcmp(buf, END_STRING, END_LENGTH) == 0))
	{
		return TYPE_END;
	}

	/*
	 * The line is a normal data line.
	 */
	return TYPE_DATA;
}


/*
 * Parse the specified heaer line to store a file name and mode in the
 * stat structure.  Returns UP_TRUE if the line was parsed successfully.
 * This should only be called for lines which are known to be header lines.
 */
static UP_BOOL
ParseHeader(UP_DEV * dev, const char * buf, int len)
{
	DATA *	data;
	int	mode;

	data = DEVTODATA(dev);

	/*
	 * If the line doesn't start like the beginning of a header line,
	 * then return an error.
	 */
	if ((len < BEGIN_LENGTH) ||
		(memcmp(buf, BEGIN_STRING, BEGIN_LENGTH) != 0))
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * This looks like a header line.
	 * Remove any terminating newline from consideration.
	 */
	if (buf[len - 1] == '\n')
		len--;

	/*
	 * Skip over the beginning word and any spaces.
	 */
	buf += BEGIN_LENGTH;
	len -= BEGIN_LENGTH;

	while ((len > 0) && isblank(*buf))
	{
		buf++;
		len--;
	}

	/*
	 * Parse an octal mode if present.
	 */
	mode = 0;

	while ((len > 0) && isoctal(*buf))
	{
		mode = mode * 8 + *buf++ - '0';
		len--;
	}

	data->mode = mode;

	/*
	 * Skip over any blanks, then get the file name and save it.
	 */
	while ((len > 0) && isblank(*buf))
	{
		buf++;
		len--;
	}

	if (data->name)
		free(data->name);

	data->name = upDeviceMalloc(dev, len + 1);

	if (data->name == NULL)
		return UP_FALSE;

	memcpy(data->name, buf, len);
	data->name[len] = '\0';

	return UP_TRUE;
}
