/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The uuencode device.
 * This uuencodes each input file as an individual output file with
 * the .uue extension added.
 */

#include "upPrivate.h"


/*
 * Maximum number of input characters for one line.
 */
#define	INSIZE	45


/*
 * Maximum number of output characters for one line.
 */
#define	OUTSIZE	64


/*
 * Macro to encode a character.
 */
#define	ENC(c)	((c) ? ((c) & 077) + ' ': '`')


/*
 * Phases that the uuencoding can be in.
 */
typedef	enum
{
	PHASE_BEGIN,		/* writing begin line */
	PHASE_DATA,		/* writing data lines */
	PHASE_END		/* writing end line */
} PHASE;


/*
 * Our private data.
 */
typedef	struct
{
	PHASE	phase;		/* phase of uuencoding */
	int	offset;		/* offset of data to output */
	int	count;		/* count of bytes left to output */
	char *	header;		/* header line to be written */
	char *	data;		/* data to be written (usually buf) */
	char	buf[OUTSIZE];	/* buffer for output data */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_UUENCODE,
	"Uuencode",
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Translation table for file name extensions when uuencoding.
 * The input and output extensions form pairs, and the table is terminated
 * by a NULL string.
 */
static const char * const translateTable[] =
{
	"",	".uue",
	NULL,	NULL
};


/*
 * Local procedures.
 */
static	UP_BOOL	transferFileName(UP_DEV *);
static	UP_BOOL	readData(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	void	FillBuffer(UP_DEV *, const char *, int);


/*
 * Add the uuencode device to the pipeline.
 */
UP_ID
upAddUUEncode(UP * up)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	data = DEVTODATA(dev);

	/*
	 * Initialize the fields.
	 */
	data->phase = PHASE_BEGIN;
	data->offset = 0;
	data->count = 0;
	data->data = data->buf;
	data->header = NULL;

	/*
	 * Set our state to transfer a file name.
	 */
	upSetState(dev, UP_STATE_TRANSFER_NAME);

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *		data;
	UP_STATE	state;

	data = DEVTODATA(dev);

	/*
	 * We don't do any polling, and assume an infinite wait at first.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Get the current state that we can act on, if any.
	 */
	state = upCheckState(dev);

	/*
	 * If the state is reading of data then see if we have enough
	 * bytes available to form a complete output block.  If not,
	 * then we don't want to do anything now.
	 */
	if ((state == UP_STATE_READ_DATA) &&
		(!upBufferReadDataCheck(dev->readBuffer, INSIZE)))
	{
		state = UP_STATE_NONE;
	}

	/*
	 * If we can perform some action then we want to run now.
	 */
	if (state != UP_STATE_NONE)
		dev->pollTimeout = 0;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	UP_STATE	state;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the state that we can act on.
	 */
	state = upCheckState(dev);

	switch (state)
	{
		case UP_STATE_NONE:
			/*
			 * There is nothing we can do right now.
			 */
			return UP_TRUE;

		case UP_STATE_TRANSFER_NAME:
			/*
			 * There is a file name to be transferred.
			 */
			return transferFileName(dev);

		case UP_STATE_TRANSFER_END_OF_BUFFER:
			/*
			 * We have read the end of the buffer.
			 * Acknowledge it and pass it on.
			 */
			upBufferReadEndOfBuffer(readBuffer);
			upBufferWriteEndOfBuffer(writeBuffer);

			/*
			 * Set ourself as being finished.
			 */
			return upSetFinished(dev);

		case UP_STATE_READ_DATA:
			/*
			 * We have a buffer of data to be read.
			 */
			return readData(dev);

		case UP_STATE_READ_EOF:
			/*
			 * We have an end of file from the file.
			 */
			upBufferReadEof(readBuffer);

			/*
			 * Set up to write the ending line.
			 */
			sprintf(data->buf, "%c\nend\n", ENC('\0'));
			data->offset = 0;
			data->data = data->buf;
			data->count = strlen(data->data);
			data->phase = PHASE_END;

			upSetState(dev, UP_STATE_WRITE_DATA);

			break;

		case UP_STATE_READ_FILE_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_STATE_WRITE_DATA:
			/*
			 * Write some data to the output buffer.
			 */
			return writeData(dev);

		case UP_STATE_FATAL_ERROR:
			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * We were successful transfering something.
	 * Say we have done that and return success.
	 */
	return upSetDidReadWrite(dev);
}



/*
 * Routine called when a file name is available for transferring.
 * We then have to set up to write the header line for the uuencoded data.
 */
static UP_BOOL
transferFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	fileName;
	const char *	outputFileName;
	const UP_STAT *	status;
	char *		header;
	int		headerLength;
	int		mode;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Read the file name from the input buffer.
	 */
	upBufferReadFileName(readBuffer);

	/*
	 * Get the file name and status.
	 */
	fileName = upBufferGetFileName(readBuffer);
	status = upBufferGetReadFileStatus(readBuffer);

	/*
	 * Translate the file name.
	 */
	outputFileName = upTranslateFileName(dev, fileName, translateTable);

	if (outputFileName == NULL)
		return UP_FALSE;

	/*
	 * Write the output file name and a default status.
	 */
	upBufferWriteFileName(writeBuffer, outputFileName, NULL);

	/*
	 * If the file has permissions, then get them.
	 * Otherwise use a default permission.
	 */
	if (status->flags & UP_STAT_MODE)
		mode = status->mode & 0777;
	else
		mode = 0666;

	/*
	 * Calculate the header line length (with a bit of slop).
	 */
	headerLength = strlen(fileName) + 16;

	/*
	 * Release any previous header if it had been allocated.
	 */
	if (data->header != data->buf)
		free(data->header);

	data->header = NULL;

	/*
	 * Default the header to point into the data buffer.
	 * But if the filename is too long then we must allocate it.
	 */
	header = data->buf;

	if (headerLength > OUTSIZE)
	{
		/*
		 * Allocate a string large enough for the header line.
		 */
		header = upMalloc(dev->owner, headerLength);

		if (header == NULL)
			return UP_FALSE;
	}

	/*
	 * Format the header line.
	 */
	sprintf(header, "begin %03o %s\n", mode, fileName);

	/*
	 * Store the header into our data structure.
	 */
	data->header = header;

	/*
	 * Set up the data to be written, and that we are in the
	 * writing of the begin line phase.
	 */
	data->offset = 0;
	data->data = header;
	data->count = strlen(header);
	data->phase = PHASE_BEGIN;

	/*
	 * Set our state to write it to the output buffer.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when some data has been read from the input buffer.
 * We only process it when we get a full buffer of data so that
 * the uuencoded line is of maximum length.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * See if a buffer full of data is available.
	 * If it isn't available then we need to wait
	 * for more data to arrive.  In this way we only
	 * will uuencode full lines of data until the
	 * end of the data.
	 */
	if (!upBufferReadDataCheck(readBuffer, INSIZE))
		return UP_TRUE;

	/*
	 * There is a full buffer (or the data at the end of file).
	 * Get the address and length of the data to be read.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * If there is no data available then there is some problem.
	 */
	if (readLength <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Limit the amount of data to the size that we
	 * can handle for one uuencoded line.
	 */
	if (readLength > INSIZE)
		readLength = INSIZE;

	/*
	 * Fill our internal buffer with the data that is available.
	 */
	FillBuffer(dev, readAddress, readLength);

	/*
	 * Account for the data read.
	 */
	upBufferReadAdjust(readBuffer, readLength);

	/*
	 * Set our state to write the generated data to the output buffer.
	 */
	upSetState(dev, UP_STATE_WRITE_DATA);

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to write stored data to the output buffer.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		cc;

	/*
	 * Get values we use.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Write as much of the stored data as we can.
	 */
	cc = upWriteData(dev, &data->data[data->offset], data->count);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Adjust our buffer values by the amount written.
	 */
	data->offset += cc;
	data->count -= cc;

	/*
	 * If there is still more to be written then just stay in the
	 * same state and return saying we have done I/O.
	 */
	if (data->count > 0)
		return upSetDidReadWrite(dev);

	/*
	 * We have finished writing some data.
	 * Decide what to do based on what phase of writing we were doing.
	 */
	switch (data->phase)
	{
		case PHASE_BEGIN:
		case PHASE_DATA:
			/*
			 * Set up to read some more file data.
			 */
			upSetState(dev, UP_STATE_READ_DATA);

			break;

		case PHASE_END:
			/*
			 * Write the end of file to the output buffer.
			 */
			upBufferWriteEof(writeBuffer);

			/*
			 * Set up to transfer another file name.
			 */
			upSetState(dev, UP_STATE_TRANSFER_NAME);

			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Fill the data buffer with the uuencoded line representing the
 * specified data of the specified length, and make it available.
 * The data length must be within the range 0 to INSIZE.
 */
static void
FillBuffer(UP_DEV * dev, const char * buf, int len)
{
	DATA *		data;
  	const char *	incp;
  	char *		outcp;
	int		ch;

	data = DEVTODATA(dev);

	incp = buf;
	outcp = data->buf;

	/*
	 * Encode the number of characters that are being encoded.
	 */
	*outcp++ = ENC(len);

	/*
	 * Now encode the data so as to get printable output.
	 * Every three arbitrary input characters get turned into
	 * four printing characters.  Loop while we have all three
	 * characters.
	 */
	while (len > 2)
	{
		ch = (incp[0] >> 2);
		*outcp++ = ENC(ch);
		ch = (((incp[0] << 4) & 060) | ((incp[1] >> 4) & 017));
		*outcp++ = ENC(ch);
		ch = (((incp[1] << 2) & 074) | ((incp[2] >> 6) & 003));
		*outcp++ = ENC(ch);
		ch = (incp[2] & 077);
		*outcp++ = ENC(ch);
		incp += 3;
		len -= 3;
	}

	/*
	 * Handle non-multiple of three cases by supplying zeroes as
	 * the extra one or two characters.  These extra characters
	 * will be ignored on uudecoding.
	 */
	if (len == 2)
	{
		ch = (incp[0] >> 2);
		*outcp++ = ENC(ch);
		ch = (((incp[0] << 4) & 060) | ((incp[1] >> 4) & 017));
		*outcp++ = ENC(ch);
		ch = ((incp[1] << 2) & 074);
		*outcp++ = ENC(ch);
		*outcp++ = ENC('\0');
	}

	if (len == 1)
	{
		ch = (incp[0] >> 2);
		*outcp++ = ENC(ch);
		ch = ((incp[0] << 4) & 060);
		*outcp++ = ENC(ch);
		*outcp++ = ENC('\0');
		*outcp++ = ENC('\0');
	}

	/*
	 * End the buffer with a newline.
	 */
	*outcp++ = '\n';

	/*
	 * Set up to write the stored data.
	 */
	data->offset = 0;
	data->data = data->buf;
	data->count = outcp - data->buf;
	data->phase = PHASE_DATA;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	data->phase = PHASE_BEGIN;
	data->offset = 0;
	data->count = 0;
	data->data = data->buf;

	upSetState(dev, UP_STATE_TRANSFER_NAME);

	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Free the header line if required.
	 */
	if (data->header != data->buf)
		free(data->header);

	data->header = NULL;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}
