/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The directory reading device.
 * This returns the files within a directory.
 * It can do recursion to a specified depth.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include "upPrivate.h"


/*
 * Reallocation size for file names.
 */
#define	NAME_CHUNK_SIZE		128
#define	LEVEL_REALLOC_SIZE	20


/*
 * Structure to hold information about one directory level.
 */
typedef struct
{
	DIR *	dir;		/* opened directory */
	int	pathLength;	/* length of the directory path */
} DIR_LEVEL;


/*
 * Local data for this device.
 */
typedef	struct
{
	char *		path;		/* current path (including original) */
	int		pathAlloc;	/* Allocated length of path */
	DIR_LEVEL *	levelTable;	/* table of levels */
	int		levelCount;	/* number of levels in table */
	int		levelAlloc;	/* allocated size of table */
	int		fd;		/* file descriptor */
	int		maxDepth;	/* max depth of recursion or zero */
	dev_t		fsDev;		/* file system device */
	UP_BOOL		isRegular;	/* file is a regular file */
	UP_BOOL		ignoreErrors;	/* ignore errors */
	UP_BOOL		namesOnly;	/* only return the file names */
	UP_BOOL		noCrossSystems;	/* don't cross file systems */
	UP_BOOL		noDirectories;	/* don't want any directories */
	UP_BOOL		noRegularFiles;	/* don't want regular files */
	UP_BOOL		noSpecialFiles;	/* don't want any special files */
	UP_STAT		stat;		/* status of current file */
	DIR_LEVEL	easyLevels[LEVEL_REALLOC_SIZE];	/* easy levels */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	writeFileName(UP_DEV * dev);
static	UP_BOOL	writeData(UP_DEV * dev);
static	UP_BOOL	enterDirectory(UP_DEV *);
static	void	leaveDirectory(UP_DEV *);


static	UP_FUNC	funcs =
{
	UP_TYPE_READ_DIRECTORY,
	"ReadDir",
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Add a directory reading device which returns all of the files
 * within a directory.  The maxDepth argument indicates whether we
 * do recursive reading.  If zero, we will recurse arbitrarily
 * deeply.  If positive, we will read that many directory levels
 * (where level 1 means the initially specified directory).
 */
UP_ID
upAddReadDirectory(UP * up, const char * path, int maxDepth, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;
	DIR_LEVEL *	dirLevel;
	int		pathLength;
	struct stat	statBuf;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	pathLength = strlen(path);

	/*
	 * Fill in our private data.
	 */
	data = DEVTODATA(dev);

	data->fd = -1;
	data->maxDepth = maxDepth;

	/*
	 * Set our flags.
	 */
	data->namesOnly = ((flags & UP_FLAG_READDIR_NAMES_ONLY) != 0);
	data->ignoreErrors = ((flags & UP_FLAG_READDIR_IGNORE_ERRORS) != 0);

	data->noDirectories = ((flags & UP_FLAG_READDIR_NO_DIRECTORIES) != 0);

	data->noRegularFiles =
		((flags & UP_FLAG_READDIR_NO_REGULAR_FILES) != 0);

	data->noSpecialFiles =
		((flags & UP_FLAG_READDIR_NO_SPECIAL_FILES) != 0);

	data->noCrossSystems =
		((flags & UP_FLAG_READDIR_NO_CROSS_FILE_SYSTEMS) != 0);

	/*
	 * Check the specified path's status.
	 */
	if (stat(path, &statBuf) < 0)
	{
		upSystemError(up, "stat");

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Make sure that the argument is a directory.
	 */
	if (!S_ISDIR(statBuf.st_mode))
	{
		upError(up, UP_ERR_NOTDIR);

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Remember the file system device.
	 */
	data->fsDev = statBuf.st_dev;

	/*
	 * Copy the initial path into allocated memory.
	 */
	data->pathAlloc = pathLength + NAME_CHUNK_SIZE;
	data->path = upMalloc(up, data->pathAlloc);

	if (data->path == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	memcpy(data->path, path, pathLength + 1);

	/*
	 * Set up the directory levels flag.
	 */
	data->levelTable = data->easyLevels;
	data->levelCount = 0;
	data->levelAlloc = LEVEL_REALLOC_SIZE;

	/*
	 * Open the directory for reading as the first level.
	 * This first directory level must succeed.
	 */
	dirLevel = data->levelTable;

	dirLevel->pathLength = pathLength;
	dirLevel->dir = opendir(path);

	if (dirLevel->dir == NULL)
	{
		upSystemError(up, "opendir");

		free(data->path);
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	data->levelCount++;

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *		data;

	data = DEVTODATA(dev);

	/*
	 * Assume that we have nothing to poll for and that we have an
	 * indefinite timeout.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the type of item that can be written now.
	 * At this point we are only interested in whether we need to wait.
	 * Full error handling will be done in the doReadWrite routine.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Break out of the switch to handle file data.
			 */
			break;

		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * These indicate that there is not enough room right
			 * now for new storage, so we have to wait for the
			 * buffer reader to consume some data.
			 */
			return UP_TRUE;

		default:
			/*
			 * For all other conditions run soon so that we can
			 * check them out.
			 */
			dev->pollTimeout = 0;

			return UP_TRUE;
	}

	/*
	 * Here when file data is wanted by the buffer.
	 * Make sure that we have a file opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If the file descriptor is a regular file then no polling is
	 * required and there is a timeout value of zero since I/O can
	 * be done immediately.
	 */
	if (data->isRegular)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for reading of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLIN;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	/*
	 * Switch on the operation that can be performed on the buffer.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * We can write the next file name now.
			 */
			return writeFileName(dev);

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * We can write file data now.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * The buffer has no room for a file name or data.
			 */
			break;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Try to write some more file data to the buffer.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	/*
	 * Point at our private data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Make sure that a file is opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If we are polling and there is no data ready, then do nothing.
	 */
	if ((dev->pollCount > 0) &&
		((dev->pollEntry.revents & (POLLIN | POLLHUP)) == 0))
	{
		return UP_TRUE;
	}

	/*
	 * Reserve as much space as we can for file data.
	 */
	writeLength = upBufferRequestSpace(writeBuffer, UP_BUFFER_MAX_SIZE);

	if (writeLength < 0)
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If there is no room then do nothing more.
	 */
	if (writeLength == 0)
		return UP_TRUE;

	/*
	 * Get the address and length of the requested space.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Try to read some data.
	 */
	cc = read(data->fd, writeAddress, writeLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "read");

		return UP_FALSE;
	}

	/*
	 * We have done some I/O.
	 */
	upSetDidReadWrite(dev);

	/*
	 * If we read some data then inform the buffer about it.
	 */
	if (cc > 0)
	{
		upBufferWriteAdjust(writeBuffer, cc);

		return UP_TRUE;
	}

	/*
	 * End of file has occurred.
	 * Tell the buffer about the end of file.
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Close the file.
	 */
	if (close(data->fd) < 0)
	{
		upDeviceSystemError(dev, "close");

		data->fd = -1;

		return UP_FALSE;
	}

	data->fd = -1;

	/*
	 * Reset the polling information.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


/*
 * Write the next file name and status to the buffer.  If there are no more
 * files then this will write an end of buffer and mark ourself as finished.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *			data;
	DIR_LEVEL *		dirLevel;
	char *			newPath;
	char *			cp;
	const struct dirent *	dirEntry;
	const char *		entryName;
	int			entryLength;
	int			pathLength;
	UP_BOOL			isDir;
	struct	stat		statBuf;

	data = DEVTODATA(dev);

	/*
	 * Make sure no file is currently opened.
	 */
	if (data->fd >= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Say we have done some I/O.
	 */
	upSetDidReadWrite(dev);

	/*
	 * Read the next directory entry, returning EOF if there are no more.
	 * But for convenience skip the normal "." and ".." entries.
	 */
	dirLevel = &data->levelTable[data->levelCount - 1];

	do
	{
		/*
		 * Read the next directory entry.
		 */
		dirEntry = readdir(dirLevel->dir);

		/*
		 * If there are no more files to be read in this directory
		 * then close this directory down and see if we are done.
		 */
		if (dirEntry == NULL)
		{
			leaveDirectory(dev);

			/*
			 * If there are more directory levels above us
			 * then return for now and we'll continue with the
			 * newest one of them.
			 */
			if (data->levelCount > 0)
				return UP_TRUE;

			/*
			 * This was the last directory level.
			 * Mark the buffer and ourself as being finished.
			 */
			upBufferWriteEndOfBuffer(dev->writeBuffer);

			return upSetFinished(dev);
		}

		/*
		 * Point at the entry's file name.
		 */
		entryName = dirEntry->d_name;

	}
	while ((strcmp(entryName, ".") == 0) || (strcmp(entryName, "..") == 0));

	/*
	 * Calculate the full path length, reallocating it if necessary.
	 * The full path consists of the original path, the entry name, a
	 * separating slash, and a trailing null.
	 */
	entryLength = strlen(entryName);
	pathLength = dirLevel->pathLength + entryLength + 2;

	if (pathLength > data->pathAlloc)
	{
		newPath = upDeviceRealloc(dev, data->path,
			pathLength + NAME_CHUNK_SIZE);

		if (newPath == NULL)
			return UP_FALSE;

		data->path = newPath;
		data->pathAlloc = pathLength + NAME_CHUNK_SIZE;
	}

	/*
	 * Add in a separating slash if required and then the entry name.
	 */
	cp = &data->path[dirLevel->pathLength];

	if ((dirLevel->pathLength > 0) && (cp[-1] != '/'))
		*cp++ = '/';

	memcpy(cp, entryName, entryLength + 1);

	/*
	 * Get the status of the file so we can check its type.
	 */
	if (lstat(data->path, &statBuf) < 0)
	{
		/*
		 * If the error was not file not found and we aren't
		 * ignoring errors then fail.  We always allow ENOENT
		 * since that can arise from races during directory scans.
		 */
		if ((errno != ENOENT) && !data->ignoreErrors)
		{
			upDeviceFileSystemError(dev, "lstat", data->path);

			return UP_FALSE;
		}

		upWarning(dev->owner, "Cannot stat \"%s\": %s", data->path,
			strerror(errno));

		/*
		 * Return for now.
		 * On the next call we will examine the next file.
		 */
		return UP_TRUE;
	}

	/*
	 * Remember whether this is a directory.
	 */
	isDir = S_ISDIR(statBuf.st_mode);

	/*
	 * If this is a directory and we want to recurse into it then
	 * set up to do that.
	 */
	if (isDir &&
		((data->maxDepth == 0) || (data->levelCount < data->maxDepth)))
	{
		/*
		 * See if we are crossing devices and that is allowed.
		 * If not, then don't enter the directory.
		 */
		if ((statBuf.st_dev == data->fsDev) || !data->noCrossSystems)
		{
			if (!enterDirectory(dev))
				return UP_FALSE;
		}
	}

	/*
	 * If we don't want directories and this is a directory then
	 * skip it.
	 */
	if (isDir && data->noDirectories)
		return UP_TRUE;

	/*
	 * If we don't want regular files and this is a regular file
	 * then skip it.
	 */
	if (data->noRegularFiles && S_ISREG(statBuf.st_mode))
		return UP_TRUE;

	/*
	 * If we don't want special files and this is not a
	 * regular file or directory then skip it.
	 */
	if (data->noSpecialFiles && !isDir && !S_ISREG(statBuf.st_mode))
		return UP_TRUE;

	/*
	 * We want this file name.
	 * If we are only interested in names or if this is a directory
	 * then write the name and status without reading the file.
	 */
	if (isDir || data->namesOnly)
	{
		/*
		 * Copy the stat information into our format.
		 */
		upCopyStatus(&data->stat, &statBuf);

		/*
		 * Clear the data exists flag in the status.
		 */
		data->stat.flags &= ~UP_STAT_DATA;

		/*
		 * Tell the buffer what the new file name and status is.
		 */
		if (!upBufferWriteFileName(dev->writeBuffer,
			data->path, &data->stat))
		{
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;
		}

		/*
		 * Write an end of file since we are providing no data.
		 */
		upBufferWriteEof(dev->writeBuffer);

		return UP_TRUE;
	}

	/*
	 * We want the file data.
	 * Open the file for reading without blocking.
	 */
	data->fd = open(data->path, O_RDONLY | O_NONBLOCK);

	if (data->fd < 0)
	{
		/*
		 * If the error was not file not found and we aren't
		 * ignoring open errors then fail.  We always allow ENOENT
		 * since that can arise from races during directory scans.
		 */
		if ((errno != ENOENT) && !data->ignoreErrors)
		{
			upDeviceFileSystemError(dev, "open", data->path);

			return UP_FALSE;
		}

		upWarning(dev->owner, "Cannot open \"%s\": %s", data->path,
			strerror(errno));

		return UP_TRUE;
	}

	/*
	 * Get the new status of the file in case it had changed.
	 */
	if (fstat(data->fd, &statBuf) < 0)
	{
		upDeviceFileSystemError(dev, "fstat", data->path);

		(void) close(data->fd);

		data->fd = -1;

		return UP_FALSE;
	}

	/*
	 * Remember whether the file is a regular file.
	 * We need this fact to know whether the file handle
	 * requires polling when we read it.
	 */
	data->isRegular = S_ISREG(statBuf.st_mode);

	/*
	 * Save the stat information so that it can be returned later.
	 */
	upCopyStatus(&data->stat, &statBuf);

	/*
	 * Tell the output buffer what the new file name and status is.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer, data->path, &data->stat))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If the file is a directory then write an end of file to the buffer
	 * and close the file right now so that we don't try to read it.
	 * (Reading a directory might generate read error anyway.)
	 */
	if (S_ISDIR(statBuf.st_mode))
	{
		upBufferWriteEof(dev->writeBuffer);

		(void) close(data->fd);

		data->fd = -1;
	}

	return UP_TRUE;
}


/*
 * Enter another directory level whose path is stored.
 * This is only called when we know we are at a directory and our
 * maxDepth value is not yet exceeded.
 */
static UP_BOOL
enterDirectory(UP_DEV * dev)
{
	DATA *		data;
	DIR_LEVEL *	newLevelTable;
	DIR_LEVEL *	dirLevel;
	DIR *		dir;
	int		newLevelAlloc;
	int		size;

	data = DEVTODATA(dev);

	/*
	 * Reallocate the level table if necessary.
	 */
	if (data->levelCount >= data->levelAlloc)
	{
		newLevelAlloc = data->levelAlloc + LEVEL_REALLOC_SIZE;
		size = newLevelAlloc * sizeof(DIR_LEVEL);

		/*
		 * Allocate a new table and copy the old one,
		 * or else realloc the old table.
		 */
		if (data->levelTable == data->easyLevels)
		{
			newLevelTable = (DIR_LEVEL *) upDeviceMalloc(dev, size);

			if (newLevelTable == NULL)
				return UP_FALSE;

			memcpy(newLevelTable, data->easyLevels,
				data->levelAlloc * sizeof(DIR_LEVEL));
		}
		else
		{
			newLevelTable = (DIR_LEVEL *) upDeviceRealloc(dev,
				(void *) data->levelTable, size);

			if (newLevelTable == NULL)
				return UP_FALSE;
		}

		data->levelTable = newLevelTable;
		data->levelAlloc = newLevelAlloc;
	}

	/*
	 * Try to open this directory.
	 */
	dir = opendir(data->path);

	if (dir == NULL)
	{
		/*
		 * If the error was not file not found or not a directory
		 * and we aren't ignoring open errors then fail.  We always
		 * allow ENOENT or ENOTDIR since they can arise from races
		 * during directory scans.
		 */
		if ((errno != ENOENT) && (errno != ENOTDIR) &&
			!data->ignoreErrors)
		{
			upDeviceFileSystemError(dev, "opendir", data->path);

			return UP_FALSE;
		}

		/*
		 * Print a warning and return success.
		 */
		upWarning(dev->owner, "Cannot read directory \"%s\": %s",
			data->path, strerror(errno));

		return UP_TRUE;
	}

	/*
	 * Point at the next directory level and fill it in.
	 */
	dirLevel = &data->levelTable[data->levelCount];

	dirLevel->pathLength = strlen(data->path);
	dirLevel->dir = dir;

	data->levelCount++;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


/*
 * Leave the current directory level.
 */
static void
leaveDirectory(UP_DEV * dev)
{
	DATA *		data;
	DIR_LEVEL *	dirLevel;

	data = DEVTODATA(dev);

	if (data->levelCount > 0)
	{
		dirLevel = &data->levelTable[data->levelCount - 1];

		closedir(dirLevel->dir);
		dirLevel->dir = NULL;

		data->levelCount--;
	}
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;
	int	index;

	data = DEVTODATA(dev);

	free(data->path);

	for (index = 0; index < data->levelCount; index++)
		closedir(data->levelTable[index].dir);

	if ((data->fd >= 0) && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	UP_DEVICE_INTERNAL_ERROR(dev);

	return UP_FALSE;
}
