/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The file descriptor reading device.
 * All of the data from the specified file handle is read as one file.
 * When end of file is reached, we possibly will close the file handle.
 * This is an endpoint device.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "upPrivate.h"


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


/*
 * Special file descriptors.
 */
#define	STDIN	0
#define	STDOUT	1
#define	STDERR	2


/*
 * Local data for this device.
 */
typedef	struct
{
	int		fd;		/* file handle reading from */
	UP_BOOL		isRegular;	/* file is a regular file */
	UP_BOOL		canClose;	/* we can close the file handle */
	UP_STAT		stat;		/* status of file */
} DATA;


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_READ_FD,
	"ReadFd",
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


static	UP_BOOL	writeFileName(UP_DEV * dev);
static	UP_BOOL	writeData(UP_DEV * dev);


/*
 * Add the specified file handle for reading from.
 * The canClose flag indicates whether we are to "own" the file handle.
 * If so, then we are responsible for closing it.
 */
UP_ID
upAddReadFd(UP * up, int fd, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;
	struct stat	statbuf;

	/*
	 * Make sure the file descriptor is reasonable.
	 */
	if (fd < 0)
	{
		upError(up, UP_ERR_BADFILEDESCRIPTOR);

		return UP_NULL_ID;
	}

	/*
	 * Validate the file descriptor.
	 */
	if (fstat(fd, &statbuf) < 0)
	{
		upSystemError(up, "fstat");

		return UP_NULL_ID;
	}

	/*
	 * Allocate the device.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Point at the data structure and fill it in.
	 */
	data = DEVTODATA(dev);

	data->fd = fd;
	data->canClose = ((flags & UP_FLAG_READFD_CAN_CLOSE) != 0);
	data->isRegular = S_ISREG(statbuf.st_mode);

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Assume that we have nothing to poll for and that we have an
	 * indefinite timeout.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the type of item that can be written now.
	 * At this point we are only interested in whether we need to wait.
	 * Full error handling will be done in the doReadWrite routine.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Break out of the switch to check file data.
			 */
			break;

		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * These indicate that there is not enough room right
			 * now for new storage, so we have to wait for the
			 * buffer reader to consume some data.
			 */
			return UP_TRUE;

		default:
			/*
			 * For all other conditions run soon so that we can
			 * check them out.
			 */
			dev->pollTimeout = 0;

			return UP_TRUE;
	}

	/*
	 * Here when file data is wanted by the buffer.
	 * Make sure that we have a file opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If the file descriptor is a regular file then no polling is
	 * required and there is a timeout value of zero since I/O can
	 * be done immediately.
	 */
	if (data->isRegular)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for reading of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLIN;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	/*
	 * Switch on the operation that can be performed on the buffer.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * We can write our file name now.
			 */
			return writeFileName(dev);

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * We can write some file data now.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * There is no room for a file name or data now.
			 */
			return UP_TRUE;

		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Try to write some more file data to the buffer.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	/*
	 * Point at our private data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Make sure that a file is opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If we are polling and there is no data ready, then do nothing.
	 */
	if ((dev->pollCount > 0) &&
		((dev->pollEntry.revents &  (POLLIN | POLLHUP)) == 0))
	{
		return UP_TRUE;
	}

	/*
	 * Reserve as much space as we can for file data.
	 */
	writeLength = upBufferRequestSpace(writeBuffer, UP_BUFFER_MAX_SIZE);

	if (writeLength < 0)
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If there is no room then do nothing more.
	 */
	if (writeLength == 0)
		return UP_TRUE;

	/*
	 * Get the address and length of the requested space.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Try to read some data.
	 */
	cc = read(data->fd, writeAddress, writeLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "read");

		return UP_FALSE;
	}

	/*
	 * If we read some data then inform the buffer about it
	 * and say that we have done some I/O.
	 */
	if (cc > 0)
	{
		upBufferWriteAdjust(writeBuffer, cc);

		return upSetDidReadWrite(dev);
	}

	/*
	 * End of file has occurred.
	 * Tell the buffer about the end of file and about the
	 * end of the buffer since we only write one file.
	 */
	upBufferWriteEof(writeBuffer);
	upBufferWriteEndOfBuffer(writeBuffer);

	/*
	 * Mark ourself as having done I/O and being finished.
	 */
	upSetFinished(dev);

	/*
	 * Close the file handle if we are allowed to.
	 */
	if (data->canClose && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		data->fd = -1;

		return UP_FALSE;
	}

	data->fd = -1;

	return UP_TRUE;
}


/*
 * Write the file name and status to the buffer.
 * The file name is a generated one since a file descriptor doesn't
 * have a real file name.  This routine is only called once since we
 * only write one file from the file descriptor.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	char		fileName[20];
	struct	stat	statBuf;

	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the status of the file descriptor.
	 */
	if (fstat(data->fd, &statBuf) < 0)
	{
		upDeviceSystemError(dev, "fstat");

		if (data->canClose)
			(void) close(data->fd);

		data->fd = -1;

		return UP_FALSE;
	}

	/*
	 * Remember whether the file is a regular file.
	 * We need this fact to know whether the file handle
	 * requires polling when we read it.
	 */
	data->isRegular = S_ISREG(statBuf.st_mode);

	/*
	 * Save the stat information in our own format.
	 */
	upCopyStatus(&data->stat, &statBuf);

	/*
	 * Generate a pseudo-name for the file.
	 */
	if (data->fd == STDIN)
		strcpy(fileName, "stdin");
	else if (data->fd == STDOUT)
		strcpy(fileName, "stdout");
	else if (data->fd == STDERR)
		strcpy(fileName, "stderr");
	else
		sprintf(fileName, "fd%d", data->fd);

	/*
	 * Tell the output buffer what the file name and status is.
	 */
	if (!upBufferWriteFileName(writeBuffer, fileName, &data->stat))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If the file is a directory then write an end of file to the buffer
	 * and close the file right now so that we don't try to read it.
	 * (Reading a directory might generate read error anyway.)
	 */
	if (S_ISDIR(statBuf.st_mode))
	{
		upBufferWriteEof(writeBuffer);
		upBufferWriteEndOfBuffer(writeBuffer);

		if (data->canClose)
			(void) close(data->fd);

		data->fd = -1;
	}

	/*
	 * Say that we have done some I/O and return success.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	/*
	 * Close the file handle if we are allowed to.
	 */
	data = DEVTODATA(dev);

	if (data->canClose && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	return UP_TRUE;
}
