/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The file reading device.
 * This returns the specified files.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include "upPrivate.h"


/*
 * Local data for this device.
 */
typedef	struct
{
	const char *	fileName;	/* current file name */
	char **		fileTable;	/* file table */
	int		fileCount;	/* count of files in the table */
	int		fileIndex;	/* index of current file */
	int		fd;		/* file descriptor */
	UP_BOOL		ignoreErrors;	/* ignore open errors */
	UP_BOOL		isRegular;	/* file is a regular file */
	UP_STAT		stat;		/* status of current file */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	writeFileName(UP_DEV * dev);
static	UP_BOOL	writeData(UP_DEV * dev);
static	void	freeTable(UP_DEV * dev);


static	UP_FUNC	funcs =
{
	UP_TYPE_READ_FILES,
	"ReadFiles",
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Read the specified file name.
 * This is a convenience method to upAddReadFileList.
 */
UP_ID
upAddReadFile(UP * up, const char * fileName, UP_FLAG flags)
{
	const char *	fileNameTable[1];

	fileNameTable[0] = fileName;

	/*
	 * Read one file name from the table.
	 */
	return upAddReadFileList(up, fileNameTable, 1, flags);
}


/*
 * Read the specified list of file names.
 * The file name table and the file names themselves are copied
 * into our own memory and so don't need to be preserved.
 */
UP_ID
upAddReadFileList(UP * up,
	const char ** fileNames,
	int fileCount,
	UP_FLAG flags)
{

	UP_DEV *	dev;
	DATA *		data;
	int		index;
	const char *	fileName;
	char *		newFileName;

	/*
	 * Make sure the arguments seem reasonable.
	 */
	if (fileNames == NULL)
	{
		upError(up, UP_ERR_NULLPOINTER);

		return UP_NULL_ID;
	}

	if (fileCount <= 0)
	{
		upError(up, UP_ERR_BADCOUNT);

		return UP_NULL_ID;
	}

	/*
	 * Allocate the device.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Fill in our private data.
	 */
	data = DEVTODATA(dev);

	data->fileName = NULL;
	data->fileCount = 0;
	data->fileIndex = 0;
	data->fd = -1;
	data->isRegular = UP_FALSE;

	data->ignoreErrors = ((flags & (UP_FLAG_READFILES_IGNORE_ERRORS)) != 0);

	/*
	 * Allocate our own table for holding the file names.
	 */
	data->fileTable = (char **)
		upMalloc(dev->owner, sizeof(const char *) * fileCount);

	if (data->fileTable == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Copy each filename into our own table by allocating it.
	 */
	for (index = 0; index < fileCount; index++)
	{
		fileName = fileNames[index];

		newFileName = upDeviceCopyString(dev, fileName);

		if (newFileName == NULL)
		{
			freeTable(dev);

			upFreeDevice(dev);

			return UP_NULL_ID;
		}

		data->fileTable[index] = newFileName;
		data->fileCount++;
	}

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *		data;

	data = DEVTODATA(dev);

	/*
	 * Assume that we have nothing to poll for and that we have an
	 * indefinite timeout.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the type of item that can be written now.
	 * At this point we are only interested in whether we need to wait.
	 * Full error handling will be done in the doReadWrite routine.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Break out of the switch to handle file data.
			 */
			break;

		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * These indicate that there is not enough room right
			 * now for new storage, so we have to wait for the
			 * buffer reader to consume some data.
			 */
			return UP_TRUE;

		default:
			/*
			 * For all other conditions run soon so that we can
			 * check them out.
			 */
			dev->pollTimeout = 0;

			return UP_TRUE;
	}

	/*
	 * Here when file data is wanted by the buffer.
	 * Make sure that we have a file opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If the file descriptor is a regular file then no polling is
	 * required and there is a timeout value of zero since I/O can
	 * be done immediately.
	 */
	if (data->isRegular)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for reading of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLIN;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	/*
	 * Switch on the operation that can be performed on the buffer.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * We can write a file name now.
			 */
			return writeFileName(dev);

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * We can write file data now.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * These states indicate that nothing useful
			 * can be written now.
			 */
			return UP_TRUE;

		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Say that we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Try to write some more file data to the buffer.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	/*
	 * Point at our private data.
	 */
	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Make sure that a file is opened.
	 */
	if (data->fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If we are polling and there is no data ready, then do nothing.
	 */
	if ((dev->pollCount > 0) &&
		((dev->pollEntry.revents & (POLLIN | POLLHUP)) == 0))
	{
		return UP_TRUE;
	}

	/*
	 * Reserve as much space as we can for file data.
	 */
	writeLength = upBufferRequestSpace(writeBuffer, UP_BUFFER_MAX_SIZE);

	if (writeLength < 0)
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * There should have been room for data.
	 */
	if (writeLength == 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Get the address and length of the requested space.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Try to read some data.
	 */
	cc = read(data->fd, writeAddress, writeLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "read");

		return UP_FALSE;
	}

	/*
	 * If we read some data then inform the buffer about it.
	 */
	if (cc > 0)
	{
		upBufferWriteAdjust(writeBuffer, cc);

		/*
		 * Return saying we have done some I/O.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * End of file has occurred.
	 * Tell the buffer about the end of file.
	 */
	upBufferWriteEof(writeBuffer);

	/*
	 * Close the file.
	 */
	if (close(data->fd) < 0)
	{
		upDeviceSystemError(dev, "close");

		data->fd = -1;

		return UP_FALSE;
	}

	data->fd = -1;

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write the next file name and status to the buffer.  If there are no more
 * files then this will write an end of buffer and mark ourself as finished.
 * Returns UP_TRUE on success.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	struct	stat	statBuf;

	data = DEVTODATA(dev);

	/*
	 * Make sure no file is currently opened.
	 */
	if (data->fd >= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If there are no more files to process then mark the output
	 * buffer and ourself as finished.
	 */
	if (data->fileIndex >= data->fileCount)
	{
		upBufferWriteEndOfBuffer(dev->writeBuffer);

		return upSetFinished(dev);
	}

	/*
	 * We have another file to be read.
	 * Get its name from the file table and advance the file index.
	 */
	data->fileName = data->fileTable[data->fileIndex++];

	/*
	 * Open the file for reading without blocking.
	 */
	data->fd = open(data->fileName, O_RDONLY | O_NONBLOCK);

	if (data->fd < 0)
	{
		/*
		 * If we aren't ignoring errors then fail.
		 */
		if (!data->ignoreErrors)
		{
			upDeviceFileSystemError(dev, "open", data->fileName);

			return UP_FALSE;
		}

		upWarning(dev->owner, "Cannot open \"%s\": %s",
			data->fileName, strerror(errno));

		/*
		 * Return saying we have done I/O.
		 * The next call will examine the next file.
		 */
		return upSetDidReadWrite(dev);
	}

	/*
	 * Get the status of the file that we have opened.
	 */
	if (fstat(data->fd, &statBuf) < 0)
	{
		upDeviceFileSystemError(dev, "fstat", data->fileName);

		(void) close(data->fd);

		data->fd = -1;

		return UP_FALSE;
	}

	/*
	 * Save the stat information so that it can be returned later.
	 */
	upCopyStatus(&data->stat, &statBuf);

	/*
	 * Remember whether the file is a regular file.
	 * We need this fact to know whether the file handle
	 * requires polling when we read it.
	 */
	data->isRegular = S_ISREG(statBuf.st_mode);

	/*
	 * Tell the output buffer what the new file name and status is.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer,
		data->fileName, &data->stat))
	{
		upDeviceError(dev, UP_ERR_BUFFER);

		return UP_FALSE;
	}

	/*
	 * If the file is a directory then write an end of file to the buffer
	 * and close the file right now so that we don't try to read it.
	 * (Reading a directory might generate read error anyway.)
	 * Also set the status size to zero to match this.
	 */
	if (S_ISDIR(statBuf.st_mode))
	{
		upBufferWriteEof(dev->writeBuffer);

		(void) close(data->fd);

		data->fd = -1;
		data->stat.size = 0;
	}

	/*
	 * Return success saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Free the table of file names that we may have allocated.
 */
static void
freeTable(UP_DEV * dev)
{
	DATA *	data;
	int	index;

	data = DEVTODATA(dev);

	for (index = 0; index < data->fileCount; index++)
		free(data->fileTable[index]);

	free(data->fileTable);

	data->fileTable = NULL;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if ((data->fd >= 0) && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	freeTable(dev);

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if (data->fd >= 0)
	{
		(void) close(data->fd);

		data->fd = -1;
	}

	freeTable(dev);

	return UP_TRUE;
}
