/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The string reading endpoint device.
 * This returns data from the specified buffer.
 */

#include <stdio.h>
#include "upPrivate.h"


/*
 * Private data for the string reading device.
 */
typedef	struct
{
	char *	buffer;		/* buffer for data */
	int	length;		/* length of buffer */
	int	offset;		/* current offset in buffer */
	UP_BOOL	canFree;	/* we can free the buffer */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);

static	UP_BOOL	writeFileName(UP_DEV * dev);
static	UP_BOOL	writeData(UP_DEV * dev);


static UP_FUNC	funcs =
{
	UP_TYPE_READ_STRING,
	"ReadString",
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Add the specified buffer of data to be read from.
 * The buffer and its data must be preserved during the use of the pipeline.
 * Optionally, the buffer can be owned by us and freed when we are closed.
 */
UP_ID
upAddReadString(UP * up, char * buffer, int length, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Initialize our private data.
	 */
	data = DEVTODATA(dev);

	data->buffer = buffer;
	data->length = length;
	data->offset = 0;
	data->canFree = ((flags & UP_FLAG_READSTRING_CAN_FREE) != 0);

	/*
	 * Link in our device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	/*
	 * Assumne at first that we need to wait.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the type of item that can be written now.
	 * At this point we are only interested in whether we need to wait.
	 * Full error handling will be done in the doReadWrite routine.
	 */
	switch (upBufferGetWriteItemType(dev->writeBuffer))
	{
		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * These indicate that there is not enough room right
			 * now for new storage, so we have to wait for the
			 * buffer reader to consume some data.
			 */
			break;

		default:
			/*
			 * For all other conditions run soon so that we can
			 * check them out.
			 */
			dev->pollTimeout = 0;

			break;
	}

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	UP_ITEM_TYPE	itemType;

	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Switch on the type of data which can be written.
	 */
	itemType = upBufferGetWriteItemType(writeBuffer);

	switch (itemType)
	{
		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * We can write our file name now.
			 */
			return writeFileName(dev);

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * We can write file data now.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_FATAL_ERROR:
			/*
			 * We had a fatal error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	return upSetDidReadWrite(dev);
}


/*
 * Here when we can write our file name and status.
 * This is only done once.
 */
static UP_BOOL
writeFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_STAT		statBuf;

	data = DEVTODATA(dev);

	/*
	 * Fill in our status which is a normal file whose size is known.
	 */
	memset(&statBuf, 0, sizeof(statBuf));

	statBuf.flags =
		UP_STAT_DATA | UP_STAT_TYPE | UP_STAT_MODE | UP_STAT_SIZE;

	statBuf.type = '-';
	statBuf.mode = S_IFREG | 0666;
	statBuf.size = data->length;

	/*
	 * Write the file name and status to the output buffer.
	 */
	if (!upBufferWriteFileName(dev->writeBuffer, "string-data", &statBuf))
		return UP_FALSE;

	/*
	 * Return that we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to write our data to the output buffer.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	writeBuffer;
	int		left;
	int		cc;

	data = DEVTODATA(dev);
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the amount of data left that we have to write.
	 */
	left = data->length - data->offset;

	/*
	 * Check whether the string is completely written.
	 */
	if (left <= 0)
	{
		/*
		 * Write the end of file and end of buffer.
		 */
		upBufferWriteEof(writeBuffer);
		upBufferWriteEndOfBuffer(writeBuffer);

		/*
		 * Return saying we are finished.
		 */
		return upSetFinished(dev);
	}

	/*
	 * We still have some data to be written.
	 * Write as much as we can.
	 */
	cc = upWriteData(dev, data->buffer + data->offset, left);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Adjust the string offset by the amount written.
	 */
	data->offset += cc;

	/*
	 * Remember that we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Close the device.
 * All we do here is possibly free our supplied string.
 */
UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * If we were told that we can free the buffer, then do it.
	 */
	if (data->canFree)
		free(data->buffer);

	data->buffer = NULL;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}
