/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The writing callback device.
 * This is an endpoint device which calls the client's supplied routines
 * in order to deliver data to the client.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "upPrivate.h"


/*
 * The structure which contains the routines we call in order to
 * perform I/O.  Null function address are not called.
 */
typedef	struct
{
	UP_BOOL (*initialize)(void *);
	UP_BOOL	(*writeFileName)(void *, const char *, const UP_STAT *);
	pollfd *(*getPollTable(void *);
	int	(*getPollCount(void *);
	int	(*getPollTimeout(void *);
	int	(*writeData)(void *, const char *, int);
	UP_BOOL	(*fileError)(void *);
	UP_BOOL	(*endOfFile)(void *);
	UP_BOOL	(*endOfBuffer)(void *);
	UP_BOOL	(*close)(void *);
	UP_BOOL	(*reset)(void *);
} UP_WRITE_CALLBACK;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


/*
 * Local data for this device.
 */
typedef	struct
{
	UP_WRITE_CALLBACK	func;		/* callback function table */
	void *			userData;	/* user data */
} DATA;


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_WRITE_CALLBACK,
	"WriteCallback",
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


UP_ID
upAddWriteCallback(UP * up, UP_WRITE_CALLBACK * functionTable, void * userData)
{
	UP_DEV *	dev;
	DATA *		data;
	struct stat	statbuf;

	/*
	 * Allocate the device.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Point at the data structure and fill it in.
	 */
	data = DEVTODATA(dev);

	data->func = functionTable;
	data->userData = userData;

	/*
	 * Call the user's initialization routine if there is one.
	 */
	if ((func->initialize != 0) && !func->initialize(userData))
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *				data;
	const UP_DATA_CALLBACK *	func;
	void *				userData;

	data = DEVTODATA(dev);
	func = &data->func;
	userData = data->userData;

	/*
	 * Assume at first that there is nothing to poll for,
	 * and that we can wait indefinitely.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the data that can be read from the input buffer.
	 */
	switch (upBufferGetReadItemType(dev->readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			dev->pollTimeout = 0;

			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_DATA:
			break;

		case UP_ITEM_TYPE_FILE_ERROR:
		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;
		
		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
	}

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *			data;
	UP_BUF *		readBuffer;
	UP_DATA_CALLBACK *	func;
	void *			userData;
	UP_BOOL			status;
	int			cc;

	/*
	 * Get our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;
	func = &data->func;
	userData = data->userData;
	status = UP_TRUE;

	/*
	 * Switch on the state of the input buffer.
	 */
	switch (upBufferGetReadItemType(readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			/*
			 * Nothing can be read right now.
			 */
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * Acknowledge the file name that was set.
			 */
			upBufferReadFileName(readBuffer);

			/*
			 * Give the file name and status to the callback
			 * function if there is one.
			 */
			if (func->writeFileName == 0)
				break;

			status = func->writeFileName(userData,
				upBufferGetFileName(readBuffer),
				upBufferGetReadFileStatus(readBuffer));

			break;

		case UP_ITEM_TYPE_END_OF_FILE:
			/*
			 * Acknowledge the end of file.
			 */
			upBufferReadEof(readBuffer);

			/*
			 * Tell the callback about the end of file if
			 * it is interested.
			 */
			if (func->endOfFile != 0)
				status = func->endOfFile(userData);

			break;

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Write some data to the output callback if it
			 * wants it.
			 */
			if (func->writeData == 0)
				break;

			cc = func->writeData(userData,
				upBufferGetReadAddress(readBuffer),
				upBufferGetReadLength(readBuffer));

			status = (cc >= 0);

			/*
			 * If any data was written account for it.
			 */
			if (cc >= 0)
				upBufferReadAdjust(readBuffer, cc);

			break;

		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * The input buffer is finally done.
			 * Acknowledge the end of buffer and mark
			 * ourself as being finished.
			 */
			upBufferReadEndOfBuffer(readBuffer);

			/*
			 * Tell the callback that the buffer is done if
			 * it wants to know.
			 */
			if (func->endOfBuffer)
				status = func->endOfBuffer(userData))

			if (!status)
				break;

			return upSetFinished(dev);

		case UP_ITEM_TYPE_FATAL_ERROR:
		case UP_ITEM_TYPE_FILE_ERROR:
			/*
			 * There is a fatal error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
	}

	/*
	 * Remember that we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	if ((data->func.close != 0) && !data->func.close(data->userData))
		return UP_FALSE;

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	if ((data->func.reset != 0) && !data->func.reset(data->userData))
		return UP_FALSE;

	return UP_TRUE;
}
