/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The directory writing device.
 * This writes files into a directory.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <utime.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include "upPrivate.h"


/*
 * Local data for this device.
 */
typedef	struct
{
	char *		dirName;		/* directory path */
	char *		fullPath;		/* current full path name */
	int		fullPathAlloc;		/* allocated full path size */
	int		dirLength;		/* length of directory path */
	int		fd;			/* file descriptor */
	UP_BOOL		isRegular;		/* file is a regular file */
	UP_BOOL		ignoreFlag;		/* ignore all file data */
	UP_BOOL		relativeFlag;		/* paths are all relative */
	UP_BOOL		noCreateFlag;		/* don't create directories */
	UP_BOOL		noOverwriteFlag;	/* don't overwrite files */
	UP_BOOL		okOpenErrorFlag;	/* open errors are ok */
	UP_BOOL		setModeFlag;		/* set file modes */
	UP_BOOL		setOwnerFlag;		/* set file owners */
	UP_BOOL		setTimesFlag;		/* set file times */
} DATA;


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	doReset(UP_DEV *);
static	UP_BOOL	readFileName(UP_DEV * dev);
static	UP_BOOL	readData(UP_DEV * dev);
static	UP_BOOL	readEndOfFile(UP_DEV * dev);
static	UP_BOOL makeDirs(UP_DEV * dev, const char * path);


static	UP_FUNC	funcs =
{
	UP_TYPE_WRITE_DIRECTORY,
	"WriteDir",
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	doReset
};


/*
 * Add a directory writing device which writes files into a directory.
 * A NULL directory means the current directory.
 */
UP_ID
upAddWriteDirectory(UP * up, const char * dirName, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;
	struct stat	statBuf;

	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * If the directory name is NULL then use the current directory.
	 */
	if ((dirName == NULL) || (*dirName == '\0'))
		dirName = ".";

	/*
	 * Make sure that the directory name is accessible and that
	 * it is a directory.
	 */
	if (stat(dirName, &statBuf) < 0)
	{
		upSystemError(up, "stat");

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Make sure that the path is really a directory.
	 */
	if (!S_ISDIR(statBuf.st_mode))
	{
		upError(up, UP_ERR_NOTDIR);

		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Fill in some private data.
	 */
	data = DEVTODATA(dev);

	data->fd = -1;
	data->dirLength = strlen(dirName);
	data->ignoreFlag = UP_FALSE;

	/*
	 * Save the directory name for later use.
	 */
	data->dirName = upDeviceCopyString(dev, dirName);

	if (data->dirName == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Set our flag values.
	 */
	data->relativeFlag = ((flags & UP_FLAG_WRITEDIR_RELATIVE) != 0);

	data->noCreateFlag =
		((flags & UP_FLAG_WRITEDIR_NO_CREATE_DIRECTORIES) != 0);

	data->noOverwriteFlag = ((flags & UP_FLAG_WRITEDIR_NO_OVERWRITE) != 0);

	data->okOpenErrorFlag =
		((flags & UP_FLAG_WRITEDIR_ALLOW_OPEN_ERRORS) != 0);

	data->setModeFlag = ((flags & UP_FLAG_WRITEDIR_SET_MODE) != 0);
	data->setOwnerFlag = ((flags & UP_FLAG_WRITEDIR_SET_OWNER) != 0);
	data->setTimesFlag = ((flags & UP_FLAG_WRITEDIR_SET_TIMES) != 0);

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Assume that we have nothing to poll for and that we have
	 * an indefinite timeout.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * If there is no data available from the input buffer then no
	 * polling or timeout is required.
	 */
	if (!upBufferIsReadAvailable(dev->readBuffer))
		return UP_TRUE;

	/*
	 * See if no file is opened or if the file is to be ignore or if
	 * this file descriptor is a regular file.  If so, then no polling
	 * is required and there is a timeout value of zero since I/O can
	 * be done immediately.
	 */
	if ((data->fd < 0) || data->isRegular || data->ignoreFlag)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for writing of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLOUT;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;

	/*
	 * Point at our private data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Switch on the current state of the input buffer.
	 */
	switch (upBufferGetReadItemType(readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
			return readFileName(dev);

		case UP_ITEM_TYPE_FILE_DATA:
			return readData(dev);

		case UP_ITEM_TYPE_FILE_ERROR:
		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		case UP_ITEM_TYPE_END_OF_FILE:
			return readEndOfFile(dev);

		case UP_ITEM_TYPE_END_OF_BUFFER:
			return upSetFinished(dev);

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read a file name and open it for writing.
 */
static UP_BOOL
readFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	fileName;
	const UP_STAT *	fileStatus;
	UP_BOOL		prependFlag;
	char *		cp;
	int		openFlags;
	int		length;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the file name and file status from the input buffer.
	 */
	upBufferReadFileName(readBuffer);

	fileName = upBufferGetFileName(readBuffer);
	fileStatus = upBufferGetReadFileStatus(readBuffer);

	/*
	 * If they are not valid then fail.
	 */
	if ((fileName == NULL) || (fileStatus == NULL))
	{
		upDeviceError(dev, UP_ERR_INTERNAL);

		return UP_FALSE;
	}

	/*
	 * Determine whether or not we want to prepend the directory path.
	 * This is dependant on whether the file name is already
	 * absolute, and on whether we want to make all paths relative.
	 */
	prependFlag = ((*fileName != '/') || data->relativeFlag);

	/*
	 * Calculate the length of the full path name we need.
	 * This depends on whether we are prepending the path or not.
	 */
	if (prependFlag)
		length = data->dirLength + strlen(fileName) + 2;
	else
		length = strlen(fileName) + 1;

	/*
	 * Make sure the allocated path name is large enough.
	 * If not, then reallocate it.
	 */
	if (length > data->fullPathAlloc)
	{
		cp = upRealloc(dev->owner, data->fullPath, length);

		if (cp == NULL)
			return UP_FALSE;

		data->fullPath = cp;
		data->fullPathAlloc = length;
	}

	/*
	 * Store the full path name depending on whether we are
	 * prepending the directory name or not.
	 */
	cp = data->fullPath;

	if (prependFlag)
	{
		memcpy(cp, data->dirName, data->dirLength);

		cp += data->dirLength;

		if ((data->dirLength > 0) && (cp[-1] != '/'))
			*cp++ = '/';
	}

	strcpy(cp, fileName);

	/*
	 * From now on the full file name is used.
	 */
	fileName = data->fullPath;

	/*
	 * Create any needed directories if we are allowed to.
	 */
	if (!data->noCreateFlag && !makeDirs(dev, fileName))
		return UP_FALSE;

	/*
	 * If the file to be created is a directory then do that.
	 * Ignore already existing errors.
	 */
	if (S_ISDIR(fileStatus->mode))
	{
		if (mkdir(fileName, fileStatus->mode & 0777) < 0)
		{
			if (errno != EEXIST)
			{
				upDeviceFileSystemError(dev, "mkdir", fileName);

				return UP_FALSE;
			}
		}

		data->ignoreFlag = UP_TRUE;
		data->isRegular = UP_FALSE;

		return UP_TRUE;
	}

	/*
	 * If the file is a regular file, then try to create it.
	 */
	if (S_ISREG(fileStatus->mode))
	{
		/*
		 * Set flags for normal files.
		 */
		data->ignoreFlag = UP_FALSE;
		data->isRegular = UP_TRUE;

		/*
		 * Get the normal open flags.
		 */
		openFlags = O_WRONLY | O_CREAT | O_NONBLOCK | O_TRUNC;

		/*
		 * If we don't want to overfile files than add that flag.
		 */
		if (data->noOverwriteFlag)
			openFlags |= O_EXCL;

		/*
		 * Try to create the file.
		 */
		data->fd = open(fileName, openFlags, fileStatus->mode & 0777);

		/*
		 * If we are successful then return saying we did I/O.
		 */
		if (data->fd >= 0)
			return upSetDidReadWrite(dev);

		/*
		 * Set the ignore data flag now since we can't write it.
		 */
		data->ignoreFlag = UP_TRUE;

		/*
		 * We failed to open the file.
		 * If we are not allowing overwriting that this is the
		 * error, then warn about it and proceed.  Its data
		 * will be ignored.
		 */
		if (data->noOverwriteFlag && (errno == EEXIST))
		{
			upWarning(dev->owner,"Not overwriting file \"%s\"",
				fileName);

			return upSetDidReadWrite(dev);
		}

		/*
		 * If we are ignore open failures then just warn about
		 * the error and proceed.  Its data will be ignored.
		 */
		if (data->okOpenErrorFlag)
		{
			upWarning(dev->owner,
				"Cannot create file \"%s\", ignored",
				fileName);

			return upSetDidReadWrite(dev);
		}

		/*
		 * The open error is fatal.
		 */
		upDeviceFileSystemError(dev, "open", fileName);

		return UP_FALSE;
	}

	/*
	 * This is a strange file type such as a device file.
	 * Ignore it for now.
	 */
	upWarning(dev->owner, "Ignoring non-regular file \"%s\"", fileName);

	data->ignoreFlag = UP_TRUE;
	data->isRegular = UP_FALSE;

	return upSetDidReadWrite(dev);
}


/*
 * Read data from the input buffer and try to write it to the output file.
 */
static UP_BOOL
readData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;
	int		cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the address and length of the data that is available.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * If there is no data right now then something is wrong.
	 */
	if (readLength <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If we are ignoring the file data then just consume all
	 * of the input data.
	 */
	if (data->ignoreFlag)
	{
		upBufferReadClear(readBuffer);

		return upSetDidReadWrite(dev);
	}

	/*
	 * If we are polling and there is no data ready, then do nothing.
	 */
	if ((dev->pollCount > 0) && ((dev->pollEntry.revents & POLLIN) == 0))
		return UP_TRUE;

	/*
	 * Try to write some data.
	 */
	cc = write(data->fd, readAddress, readLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "write");

		return UP_FALSE;
	}

	/*
	 * We have written some data.
	 * Tell the input buffer about the data that we have consumed.
	 */
	upBufferReadAdjust(readBuffer, cc);

	/*
	 * Return saying we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read an end of file from the input buffer and close the current
 * file, and possibly set its attributes.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *		data;
	const UP_STAT *	fileStatus;
	UP_BUF *	readBuffer;
	struct	utimbuf	timeBuf;
	uid_t		uid;
	gid_t		gid;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(readBuffer);

	/*
	 * If we were ignoring the input file then do nothing.
	 */
	if (data->ignoreFlag)
		return upSetDidReadWrite(dev);

	/*
	 * Get the file name and status of what we were reading.
	 */
	fileStatus = upBufferGetReadFileStatus(readBuffer);

	/*
	 * If we are to set the file times (and they are available)
	 * then do so.
	 */
	if (data->setTimesFlag &&
		((fileStatus->flags & (UP_STAT_ATIME | UP_STAT_MTIME)) != 0))
	{
		/*
		 * Store the appropriate access time.
		 */
		if (fileStatus->flags & UP_STAT_ATIME)
			timeBuf.actime = fileStatus->atime;
		else
			timeBuf.actime = time(0);

		/*
		 * Store the appropriate modification time.
		 */
		if (fileStatus->flags & UP_STAT_MTIME)
			timeBuf.modtime = fileStatus->mtime;
		else
			timeBuf.modtime = time(0);

		/*
		 * Set the file time.
		 * If this fails, give a warning.
		 */
		if (utime(data->fullPath, &timeBuf) < 0)
		{
			upWarning(dev->owner, "Cannot set times for \"%s\"",
				data->fullPath);
		}
	}

	/*
	 * If we are to change the file mode (and it is available)
	 * then do so.
	 */
	if (data->setModeFlag && ((fileStatus->flags & UP_STAT_MODE) != 0))
	{
		if (fchmod(data->fd, fileStatus->mode & 0777) < 0)
		{
			upWarning(dev->owner, "Cannot set mode for \"%s\"",
				data->fullPath);
		}
	}

	/*
	 * If we are to change the file owner ids (and they are available)
	 * then do so.
	 */
	if (data->setOwnerFlag &&
		((fileStatus->flags & (UP_STAT_UID | UP_STAT_GID)) != 0))
	{
		/*
		 * Store the appropriate user and group ids,
		 * with -1 for no change if they are not known.
		 */
		uid = -1;
		gid = -1;

		if (fileStatus->flags & UP_STAT_UID)
			uid = fileStatus->uid;

		if (fileStatus->flags & UP_STAT_GID)
			gid = fileStatus->gid;

		/*
		 * Set the file owner and group.
		 * If this fails, give a warning.
		 */
		if (fchown(data->fd, uid, gid) < 0)
		{
			upWarning(dev->owner, "Cannot set owner id for \"%s\"",
				data->fullPath);
		}
	}

	/*
	 * Close the open file if there was one opened.
	 */
	if ((data->fd >= 0) && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		data->fd = -1;

		return UP_FALSE;
	}

	data->fd = -1;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Make sure that all the directories exist down to the last
 * path component of the specified path (which is ignored).
 * Returns UP_TRUE on success.
 */
static UP_BOOL
makeDirs(UP_DEV * dev, const char * path)
{
	const char *	endPath;
	int		length;
	UP_BOOL		moreFlag;
	struct stat	statBuf;
	char *		cp;
	static char *	buffer;
	static int	bufferAlloc;

	/*
	 * Find the last slash in the path.
	 * If there isn't one then there is no directory to check.
	 */
	endPath = strrchr(path, '/');

	if (endPath == NULL)
		return UP_TRUE;

	/*
	 * Get the length of the path up to the last component.
	 * If this is just the root directory then we assume it exists.
	 */
	length = endPath - path;

	if (length == 0)
		return UP_TRUE;

	/*
	 * Grow our static buffer if necessary to hold the path.
	 */
	if (length > bufferAlloc)
	{
		cp = upRealloc(dev->owner, buffer, length);

		if (cp == NULL)
			return UP_FALSE;

		buffer = cp;
		bufferAlloc = length;
	}

	/*
	 * Copy the directory path into the buffer and null terminate it.
	 * (Here the last slash is replaced by a null.)
	 */
	memcpy(buffer, path, length);

	buffer[length] = '\0';

	/*
	 * If the full path can be accessed then just verify that it
	 * is a directory.
	 */
	if (stat(buffer, &statBuf) == 0)
	{
		if (!S_ISDIR(statBuf.st_mode))
		{
			upDeviceError(dev, UP_ERR_NOTDIR);

			return UP_FALSE;
		}

		/*
		 * The full directory path exists.
		 */
		return UP_TRUE;
	}

	/*
	 * The full path cannot be accessed.
	 * Make sure that the error is that the path is non-existent.
	 */
	if (errno != ENOENT)
	{
		upDeviceSystemError(dev, "stat");

		return UP_FALSE;
	}

	/*
	 * Somewhere along the path a directory does not exist.
	 * So we have to walk them one by one and create them
	 * as needed.
	 */
	cp = buffer;
	moreFlag = UP_TRUE;

	while (moreFlag)
	{
		/*
		 * Find the next path separator or the end of the path
		 * and remember whether another component is coming.
		 */
		cp = strchr(cp, '/');

		if (cp == NULL)
		{
			cp = buffer + length;
			moreFlag = UP_FALSE;
		}

		/*
		 * Null terminate the partial path.
		 */
		*cp = '\0';

		/*
		 * Check out whether this much of the path is accessible.
		 */
		if (stat(buffer, &statBuf) < 0)
		{
			/*
			 * The stat failed.  The error must be that the
			 * directory does not exist.
			 */
			if (errno != ENOENT)
			{
				upDeviceSystemError(dev, "stat");

				return UP_FALSE;
			}

			/*
			 * The directory does not exist.
			 * Try to create it.
			 */
			if (mkdir(buffer, 0777) < 0)
			{
				upDeviceSystemError(dev, "mkdir");

				return UP_FALSE;
			}
		}
		else
		{
			/*
			 * The path exists.
			 * Make sure that this path is a directory.
			 */
			if (!S_ISDIR(statBuf.st_mode))
			{
				upDeviceError(dev, UP_ERR_NOTDIR);

				return UP_FALSE;
			}
		}

		/*
		 * Restore the separator character and skip over it
		 * in preparation for checking the next component.
		 */
		*cp++ = '/';
	}

	/*
	 * All of the directory paths exist now.
	 */
	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	free(data->dirName);
	free(data->fullPath);

	if ((data->fd >= 0) && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	return UP_TRUE;
}


static UP_BOOL
doReset(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	if (data->fd >= 0)
	{
		(void) close(data->fd);

		data->fd = -1;
	}

	data->ignoreFlag = UP_FALSE;

	return UP_TRUE;
}
