/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The file descriptor writing device.
 * This is an endpoint device.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "upPrivate.h"


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


/*
 * Local data for this device.
 */
typedef	struct
{
	int		fd;		/* file handle to write to */
	UP_BOOL		isRegular;	/* file handle is for regular file */
	UP_BOOL		canClose;	/* we need to close the file handle */
} DATA;


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_WRITE_FD,
	"WriteFd",
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


UP_ID
upAddWriteFd(UP * up, int fd, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;
	struct stat	statbuf;

	/*
	 * Get information about the file descriptor.
	 */
	if (fstat(fd, &statbuf) < 0)
	{
		upSystemError(up, "fstat");

		return UP_NULL_ID;
	}

	/*
	 * Allocate the device.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Point at the data structure and fill it in.
	 * We remember the file handle, whether or not we are to own it,
	 * and whether or not the file is a regular type file.
	 */
	data = DEVTODATA(dev);

	data->fd = fd;
	data->canClose = ((flags & UP_FLAG_WRITEFD_CAN_CLOSE) != 0);
	data->isRegular = S_ISREG(statbuf.st_mode);

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Assume at first that there is nothing to poll for,
	 * and that we can wait indefinitely.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the data that can be read from the input buffer.
	 */
	switch (upBufferGetReadItemType(dev->readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			dev->pollTimeout = 0;

			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_DATA:
			break;

		case UP_ITEM_TYPE_FILE_ERROR:
		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;
		
		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
	}

	/*
	 * See if this file descriptor is a regular file.
	 * If so, then there is a timeout value of zero since I/O
	 * can be done immediately.
	 */
	if (data->isRegular)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for writing of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLOUT;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	UP_BUF *	readBuffer;

	readBuffer = dev->readBuffer;

	/*
	 * Switch on the state of the input buffer.
	 */
	switch (upBufferGetReadItemType(readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			/*
			 * Nothing can be read right now.
			 */
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * Acknowledge the file name that was set
			 * but otherwise ignore it.
			 */
			upBufferReadFileName(readBuffer);

			break;

		case UP_ITEM_TYPE_END_OF_FILE:
			/*
			 * Acknowledge the end of file that was set
			 * but otherwise ignore it.
			 */
			upBufferReadEof(readBuffer);

			break;

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Write some data to the output file.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * The input buffer is finally done.
			 * Acknowledge the end of buffer and mark
			 * ourself as being finished.
			 */
			upBufferReadEndOfBuffer(readBuffer);

			return upSetFinished(dev);

		case UP_ITEM_TYPE_FATAL_ERROR:
		case UP_ITEM_TYPE_FILE_ERROR:
			/*
			 * There is a fatal error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
	}

	/*
	 * Remember that we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Write some data to the file from the input buffer.
 * Returns UP_TRUE if successful.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;
	int		cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	readBuffer = dev->readBuffer;

	/*
	 * Get the address and length of the data in the input buffer.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * Write some data from the input buffer to the file descriptor.
	 */
	cc = write(data->fd, readAddress, readLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "write");

		return UP_FALSE;
	}

	/*
	 * Skip over the data that was successfully written.
	 */
	upBufferReadAdjust(readBuffer, cc);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	/*
	 * Clear the polling table so that it isn't wrongly freed
	 * by the main level code.
	 */
	dev->pollTable = NULL;
	dev->pollCount = 0;

	/*
	 * Close the file handle if we can.
	 */
	data = DEVTODATA(dev);

	if (data->canClose && (close(data->fd) < 0))
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	return UP_TRUE;
}
