/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The file writing device.
 * This is an endpoint device.
 * Each input file is written to one or more output files which are
 * specified either explicitly, or else by using a pattern.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "upPrivate.h"


#define	DEVTODATA(dev)	((DATA *) ((dev)->private))


/*
 * Local data for this device.
 */
typedef	struct
{
	int		fd;		/* file handle to write to */
	char *		fileName;	/* file name or pattern */
	long		sequence;	/* sequence number for pattern */
	UP_BOOL		isRegular;	/* file handle is for regular file */
	UP_BOOL		isPattern;	/* the file name is a pattern */
} DATA;


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	readFileName(UP_DEV *);
static	UP_BOOL	readEndOfFile(UP_DEV *);
static	UP_BOOL	readEndOfBuffer(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	UP_BOOL	openFile(UP_DEV *, const char *);
static	UP_BOOL	closeFile(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_WRITE_FILE,
	"WriteFile",
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Write an output file either to the specified file name,
 * or to multiple files generated by a filename pattern.
 */
UP_ID
upAddWriteFile(UP * up, const char * fileName, UP_FLAG flags)
{
	UP_DEV *	dev;
	DATA *		data;

	/*
	 * Allocate the device.
	 */
	dev = upAllocDevice(up, &funcs, sizeof(DATA));

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Point at the data structure and fill it in.
	 */
	data = DEVTODATA(dev);

	data->fd = -1;
	data->isPattern = ((flags & UP_FLAG_WRITEFILE_PATTERN) != 0);
	data->isRegular = UP_TRUE;

	/*
	 * Copy the file name into our own memory.
	 */
	data->fileName = upCopyString(up, fileName);

	if (data->fileName == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * If the file name is not a pattern then open it now.
	 */
	if (!data->isPattern)
	{
		if (!openFile(dev, fileName))
		{
			free(data->fileName);

			upFreeDevice(dev);

			return UP_NULL_ID;
		}
	}

	/*
	 * OK, link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	DATA *	data;

	data = DEVTODATA(dev);

	/*
	 * Assume at first that there is nothing to poll for,
	 * and that we can wait indefinitely.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * Switch on the data that can be read from the input buffer.
	 */
	switch (upBufferGetReadItemType(dev->readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * For these cases set the timeout to zero so
			 * that we run quickly to handle them.
			 */
			dev->pollTimeout = 0;

			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Break out to handle this case.
			 */
			break;

		case UP_ITEM_TYPE_FILE_ERROR:
		case UP_ITEM_TYPE_FATAL_ERROR:
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;
		
		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
	}

	/*
	 * See if this file descriptor is a regular file.
	 * If so, then there is a timeout value of zero since I/O
	 * can be done immediately.
	 */
	if (data->isRegular)
	{
		dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * Polling for the file descriptor is required and we can wait
	 * indefinitely for it.  Point at our poll entry for writing of
	 * the file descriptor.
	 */
	dev->pollEntry.fd = data->fd;
	dev->pollEntry.events = POLLOUT;
	dev->pollEntry.revents = 0;

	dev->pollCount = 1;
	dev->pollTimeout = -1;

	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	UP_BUF *	readBuffer;

	readBuffer = dev->readBuffer;

	/*
	 * Switch on the state of the input buffer.
	 */
	switch (upBufferGetReadItemType(readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			/*
			 * Nothing can be read right now.
			 */
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * Another file name has arrived.
			 */
			return readFileName(dev);

		case UP_ITEM_TYPE_END_OF_FILE:
			/*
			 * Handle the end of the input file.
			 */
			return readEndOfFile(dev);

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * Write some data to the output file.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * We have reached the end of the buffer.
			 */
			return readEndOfBuffer(dev);

		case UP_ITEM_TYPE_FATAL_ERROR:
		case UP_ITEM_TYPE_FILE_ERROR:
			/*
			 * There is a fatal error.
			 */
			upDeviceError(dev, UP_ERR_BUFFER);

			return UP_FALSE;

		default:
			upDeviceError(dev, UP_ERR_INTERNAL);

			return UP_FALSE;
	}

	/*
	 * Remember that we have done some I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when another file name has arrived.
 * We either ignore it, or else we close the current file and
 * create another one from the pattern string.
 */
static UP_BOOL
readFileName(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	char *		fileName;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Read the file name from the input buffer.
	 */
	upBufferReadFileName(readBuffer);

	/*
	 * If our file name is not a pattern,
	 * then we are concatenating all of the input files
	 * together.  In this case, just say that we have
	 * done I/O and return.
	 */
	if (!data->isPattern)
		return upSetDidReadWrite(dev);

	/*
	 * We need to generate a new file name.
	 * Allocate the next one using the next sequence number.
	 */
	fileName = upGenerateFileName(dev, data->fileName, data->sequence++);

	/*
	 * If we couldn't allocate the file name then fail.
	 */
	if (fileName == NULL)
		return UP_FALSE;

	/*
	 * Open the file for writing.
	 */
	if (!openFile(dev, fileName))
	{
		free(fileName);

		return UP_FALSE;
	}

	/*
	 * Free the generated file name.
	 */
	free(fileName);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called when the end of the buffer has been reached.
 * We close any output file that we had been writing.
 */
static UP_BOOL
readEndOfBuffer(UP_DEV * dev)
{
	DATA *		data;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Acknowledge the end of buffer.
	 */
	upBufferReadEndOfBuffer(dev->readBuffer);

	/*
	 * If we had not been using file patterns then we need
	 * to close the output file now.
	 */
	if (!data->isPattern && !closeFile(dev))
		return UP_FALSE;

	/*
	 * Mark ourself as being finished.
	 */
	return upSetFinished(dev);
}


/*
 * Write some data to the file from the input buffer.
 * Returns UP_TRUE if successful.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	DATA *		data;
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;
	int		cc;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Get the address and length of the data in the input buffer.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * Write some data from the input buffer to the file descriptor.
	 */
	cc = write(data->fd, readAddress, readLength);

	if (cc < 0)
	{
		upDeviceSystemError(dev, "write");

		return UP_FALSE;
	}

	/*
	 * Skip over the data that was successfully written.
	 */
	upBufferReadAdjust(readBuffer, cc);

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Read the end of file from the input buffer.
 * We either ignore it, or close the current output file.
 * Which we do depends on whether the filename is a pattern.
 */
static UP_BOOL
readEndOfFile(UP_DEV * dev)
{
	DATA *	data;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Acknowledge the end of file.
	 */
	upBufferReadEof(dev->readBuffer);

	/*
	 * If we were using filename patterns, then we want to
	 * close the current output file.
	 */
	if (data->isPattern && !closeFile(dev))
		return UP_FALSE;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Open the specified file name for writing to and determine
 * whether it is a regular file.
 */
static UP_BOOL
openFile(UP_DEV * dev, const char * fileName)
{
	DATA *		data;
	struct stat	statBuf;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Try to create the file.
	 */
	data->fd = open(fileName, O_CREAT | O_TRUNC | O_WRONLY, 0666);

	if (data->fd < 0)
	{
		upDeviceFileSystemError(dev, "open", fileName);

		return UP_FALSE;
	}

	/*
	 * Get information about the file descriptor.
	 */
	if (fstat(data->fd, &statBuf) < 0)
	{
		upDeviceFileSystemError(dev, "fstat", fileName);

		return UP_FALSE;
	}

	/*
	 * Remember whether the file is a regular file.
	 */
	data->isRegular = S_ISREG(statBuf.st_mode);

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


/*
 * Close the opened file.  This should only be called when we know that
 * a file is opened.  Returns UP_TRUE on success.
 */
static UP_BOOL
closeFile(UP_DEV * dev)
{
	DATA *	data;
	int	fd;

	/*
	 * Point at our data.
	 */
	data = DEVTODATA(dev);

	/*
	 * Get the file handle and clear it in our structure.
	 */
	fd = data->fd;

	data->fd = -1;

	/*
	 * Make sure something was opened.
	 */
	if (fd < 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * Try to close the file.
	 */
	if (close(fd) < 0)
	{
		upDeviceSystemError(dev, "close");

		return UP_FALSE;
	}

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doClose(UP_DEV * dev)
{
	DATA *	data;

	/*
	 * Clear the polling table so that it isn't wrongly freed
	 * by the main level code.
	 */
	dev->pollTable = NULL;
	dev->pollCount = 0;

	/*
	 * Close the file handle if needed.
	 */
	data = DEVTODATA(dev);

	if ((data->fd >= 0) && !closeFile(dev))
		return UP_FALSE;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}
