/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * The string writing endpoint device.
 * This stores data into a memory buffer which is reallocated as necessary.
 */

#include <stdio.h>
#include "upPrivate.h"


/*
 * The minimum and initial sizes of the string buffer.
 */
#define	MINIMUM_SIZE	256
#define	INITIAL_SIZE	8192


#define	DEVTODATA(dev)	((UP_WRITE_STRING *) ((dev)->private))


static	UP_BOOL	doBuildPollTable(UP_DEV *);
static	UP_BOOL	doReadWrite(UP_DEV *);
static	UP_BOOL	doClose(UP_DEV *);
static	UP_BOOL	writeData(UP_DEV *);
static	UP_BOOL	growBuffer(UP_DEV *);


static UP_FUNC	funcs =
{
	UP_TYPE_WRITE_STRING,
	"WriteString",
	UP_FALSE,
	UP_TRUE,
	UP_FALSE,
	UP_FALSE,
	UP_FALSE,
	UP_TRUE,
	doBuildPollTable,
	doReadWrite,
	upDefaultFlush,
	doClose,
	upDefaultReset
};


/*
 * Add the specified string buffer to be written to.
 * The supplied UP_WRITE_STRING structure is initialized and updated
 * by this device as it runs.  It must be supplied by the caller and
 * remain valid until the pipeline is finished.
 */
UP_ID
upAddWriteString(UP * up, UP_WRITE_STRING * data, int maxBufferSize)
{
	UP_DEV *	dev;
	int		initialBufferSize;

	/*
	 * Validate the arguments.
	 */
	if (data == NULL)
	{
		upError(up, UP_ERR_NULLPOINTER);

		return UP_NULL_ID;
	}

	if (maxBufferSize < MINIMUM_SIZE)
	{
		upError(up, UP_ERR_BADSIZE);

		return UP_NULL_ID;
	}

	/*
	 * Allocate a new device.
	 */
	dev = upAllocDevice(up, &funcs, 0);

	if (dev == UP_NULL_DEV)
		return UP_NULL_ID;

	/*
	 * Remember the caller's structure as our private data.
	 * Note: We must make sure we NULL it out when we are closed
	 * since we did not allocate it.
	 */
	dev->private = data;

	/*
	 * Calculate the initial buffer size.
	 * This is the minimum of our initial size and the specified
	 * maximum size.
	 */
	initialBufferSize = maxBufferSize;

	if (initialBufferSize > INITIAL_SIZE)
		initialBufferSize = INITIAL_SIZE;

	/*
	 * Allocate the initial buffer now.
	 */
	data->buffer = upMalloc(up, initialBufferSize);

	if (data->buffer == NULL)
	{
		upFreeDevice(dev);

		return UP_NULL_ID;
	}

	/*
	 * Fill in the rest of the information.
	 */
	data->bufferSize = initialBufferSize;
	data->maxBufferSize = maxBufferSize;
	data->byteCount = 0;
	data->hasFinished = UP_FALSE;
	data->hasError = UP_FALSE;

	/*
	 * Link in the device.
	 */
	return upLinkDevice(up, dev);
}


static UP_BOOL
doBuildPollTable(UP_DEV * dev)
{
	/*
	 * Set that we have no poll entries and that we can wait.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * If the input buffer has any items to be read then we want
	 * to run now.
	 */
	if (upBufferIsReadAvailable(dev->readBuffer))
		dev->pollTimeout = 0;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


static UP_BOOL
doReadWrite(UP_DEV * dev)
{
	UP_WRITE_STRING *	data;
	UP_BUF *		readBuffer;

	data = DEVTODATA(dev);
	readBuffer = dev->readBuffer;

	/*
	 * Switch on the item which can be read.
	 */
	switch (upBufferGetReadItemType(readBuffer))
	{
		case UP_ITEM_TYPE_NONE:
			/*
			 * Nothing can be done right now.
			 */
			return UP_TRUE;

		case UP_ITEM_TYPE_FILE_NAME:
			/*
			 * There is a file name available.
			 * Just read it and ignore it.
			 */
			upBufferReadFileName(readBuffer);

			break;

		case UP_ITEM_TYPE_FILE_DATA:
			/*
			 * There is data to be stored.
			 */
			return writeData(dev);

		case UP_ITEM_TYPE_END_OF_FILE:
			/*
			 * There is an end of file.
			 * Read it and ignore it.
			 */
			upBufferReadEof(readBuffer);

			break;

		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * There is an end of buffer.
			 * Read the end of buffer, mark the string as
			 * done and ourself as finished.
			 */
			upBufferReadEndOfBuffer(readBuffer);

			data->hasFinished = UP_TRUE;

			return upSetFinished(dev);

		case UP_ITEM_TYPE_FATAL_ERROR:
		case UP_ITEM_TYPE_FILE_ERROR:
			/*
			 * There is a file or fatal error.
			 */
			data->hasError = UP_TRUE;

			return UP_FALSE;

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			data->hasError = UP_TRUE;

			return UP_FALSE;
	}

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to store data into the string.
 */
static UP_BOOL
writeData(UP_DEV * dev)
{
	UP_WRITE_STRING *	data;
	int			cc;

	data = DEVTODATA(dev);

	/*
	 * Get the amount of space left in the string buffer.
	 */
	cc = data->bufferSize - data->byteCount;

	/*
	 * If there is no more room then we must try to allocate more.
	 */
	if (cc <= 0)
	{
		if (!growBuffer(dev))
			return UP_FALSE;

		/*
		 * Recalculate the amount of space left in the string.
		 */
		cc = data->bufferSize - data->byteCount;
	}

	/*
	 * Copy some data into the string buffer.
	 */
	cc = upReadData(dev, data->buffer + data->byteCount, cc);

	if (cc < 0)
		return UP_FALSE;

	/*
	 * Increment the number of bytes stored in the string.
	 */
	data->byteCount += cc;

	/*
	 * Return saying we have done I/O.
	 */
	return upSetDidReadWrite(dev);
}


/*
 * Routine called to grow the string buffer if possible.
 */
static UP_BOOL
growBuffer(UP_DEV * dev)
{
	UP_WRITE_STRING *	data;
	char *			newBuffer;
	int			newBufferSize;

	data = DEVTODATA(dev);

	/*
	 * If we have reached the maximum buffer size then fail.
	 */
	if (data->bufferSize >= data->maxBufferSize)
	{
		upDeviceError(dev, UP_ERR_TOOMUCHDATA);

		data->hasError = UP_TRUE;

		return UP_FALSE;
	}

	/*
	 * We can grow the buffer.
	 * First calculate a new size which is 1.5 times as large
	 * as the old one.
	 */
	newBufferSize = (int) ((((long) data->bufferSize) * 3) / 2);

	/*
	 * Restrict the new buffer size to the maximum buffer size.
	 */
	if (newBufferSize > data->maxBufferSize)
		newBufferSize = data->maxBufferSize;

	/*
	 * Reallocate the buffer to the new size.
	 */
	newBuffer = upRealloc(dev->owner, data->buffer, newBufferSize);

	if (newBuffer == NULL)
	{
		data->hasError = UP_TRUE;

		return UP_FALSE;
	}

	/*
	 * Ok, save the new buffer.
	 */
	data->buffer = newBuffer;
	data->bufferSize = newBufferSize;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}


/*
 * Close the device.
 * We do NOT free the allocated memory buffer since that is stored
 * in the caller's structure, and he wants to use it after the
 * pipeline is finished.
 */
static UP_BOOL
doClose(UP_DEV * dev)
{
	/*
	 * Clear out the private pointer since we did not allocate it.
	 */
	dev->private = NULL;

	/*
	 * Return success.
	 */
	return UP_TRUE;
}
