/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline utility.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * Program to run a configurable user mode pipeline.
 * This allows a pipeline to be created according to the options given
 * on the command line.  The devices in the pipeline are provided by
 * the "up" library.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>

#include "up.h"


/*
 * Exit status values.
 */
#define	EXIT_NORMAL	0	/* exited normally */
#define	EXIT_WARNING	1	/* successful run with warnings */
#define	EXIT_PARAM	2	/* parameter error during setup */
#define	EXIT_FATAL	3	/* fatal error */


/*
 * Special file descriptors.
 */
#define	STDIN	0
#define	STDOUT	1


/*
 * The number of arguments for a device to mean there is no limit.
 */
#define	MANYARGS	-1


/*
 * The default device separator string.
 */
#define	SEPARATOR	"="


/*
 * The default memory buffer size for PipeFindSize before it uses
 * temporary files.
 */
#define	FINDSIZE_SIZE	(8 * 1024 * 1024)


/*
 * Function called to add a particular device.
 */
typedef UP_BOOL	(*FUNC)(UP * up, int argc, const char ** argv);


/*
 * Structure to describe one device that can be used.
 */
typedef	struct
{
	const char *	name;		/* device name */
	const FUNC	func;		/* function to call to open it */
	int		minArgs;	/* minimum number of arguments */
	int		maxArgs;	/* maximum number of arguments */
	UP_BOOL		readEndpoint;	/* this is a reading endpoint */
	UP_BOOL		writeEndpoint;	/* this is a writing endpoint */
	const char *	argString;	/* possible arguments */
} DEV;


/*
 * Functions which are in the device table.
 */
static	UP_BOOL	addGrep(UP * up, int argc, const char ** argv);
static	UP_BOOL	addLog(UP * up, int argc, const char ** argv);
static	UP_BOOL	addMakeFileList(UP * up, int argc, const char ** argv);
static	UP_BOOL	addReadNull(UP * up, int argc, const char ** argv);
static	UP_BOOL	addCat(UP *, int argc, const char ** argv);
static	UP_BOOL	addFindSize(UP *, int argc, const char ** argv);
static	UP_BOOL	addRemoveData(UP *, int argc, const char ** argv);
static	UP_BOOL	addFilter(UP *, int argc, const char ** argv);
static	UP_BOOL	addArchive(UP * up, int argc, const char ** argv);
static	UP_BOOL	addUnArchive(UP * up, int argc, const char ** argv);
static	UP_BOOL	addReadFiles(UP * up, int argc, const char ** argv);
static	UP_BOOL	addWriteFile(UP * up, int argc, const char ** argv);
static	UP_BOOL	addFindFileData(UP * up, int argc, const char ** argv);
static	UP_BOOL	addList(UP * up, int argc, const char ** argv);
static	UP_BOOL	addUudecode(UP * up, int argc, const char ** argv);
static	UP_BOOL	addUuencode(UP * up, int argc, const char ** argv);
static	UP_BOOL	addGnuZip(UP * up, int argc, const char ** argv);
static	UP_BOOL	addGnuUnzip(UP * up, int argc, const char ** argv);
static	UP_BOOL	addWriteNull(UP * up, int argc, const char ** argv);
static	UP_BOOL	addReadDirectory(UP * up, int argc, const char ** argv);
static	UP_BOOL	addUnTar(UP * up, int argc, const char ** argv);
static	UP_BOOL	addTar(UP * up, int argc, const char ** argv);
static	UP_BOOL	addProcess(UP * up, int argc, const char ** argv);
static	UP_BOOL	addCompress(UP * up, int argc, const char ** argv);
static	UP_BOOL	addUnCompress(UP * up, int argc, const char ** argv);
static	UP_BOOL	addWriteDirectory(UP * up, int argc, const char ** argv);
static	UP_BOOL	addProgress(UP * up, int argc, const char ** argv);
static	UP_BOOL	addReadString(UP * up, int argc, const char ** argv);



const	DEV	devTable[] =
{
	{
		"cat",		addCat,		0,	1,
		UP_FALSE,	UP_FALSE,	"[fileName]"
	},
	{
		"findsize",	addFindSize,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"filter",	addFilter,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"readfiles",	addReadFiles,	1,	MANYARGS,
		UP_TRUE,	UP_FALSE,	"[-i] <fileName> ..."
	},
	{
		"writefile",	addWriteFile,	1,	2,
		UP_FALSE,	UP_TRUE,	"[-p] <fileName>"
	},
	{
		"uudecode",	addUudecode,	0,	1,
		UP_FALSE,	UP_FALSE,	"[fileName]"
	},
	{
		"uuencode",	addUuencode,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"gzip",		addGnuZip,	0,	1,
		UP_FALSE,	UP_FALSE,	"[level]"
	},
	{
		"gunzip",	addGnuUnzip,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"compress",	addCompress,	0,	1,
		UP_FALSE,	UP_FALSE,	"[bits]"
	},
	{
		"uncompress",	addUnCompress,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"tar",		addTar,		0,	1,
		UP_FALSE,	UP_FALSE,	"[fileName]"
	},
	{
		"untar",	addUnTar,	0,	1,
		UP_FALSE,	UP_FALSE,	"[fileName]"
	},
	{
		"archive",	addArchive,	0,	1,
		UP_FALSE,	UP_FALSE,	"[fileName]"
	},
	{
		"unarchive",	addUnArchive,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"proc",		addProcess,	1,	MANYARGS,
		UP_FALSE,	UP_FALSE,
		"[-p] [-s] [-1] [-2] [-a] <program> [argument] ..."
	},
	{
		"makefilelist",	addMakeFileList, 0,	1,
		UP_FALSE,	UP_FALSE,	"[-0]"
	},
	{
		"findfiledata",	addFindFileData, 0,	MANYARGS,
		UP_FALSE,	UP_FALSE,	"[-i] [-r]"
	},
	{
		"removedata",	addRemoveData,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		"readdir",	addReadDirectory, 1,	MANYARGS,
		UP_TRUE,	UP_FALSE,
		"[-n] [-d] [-s] [-r] [-i] [-x] [-l <levels>] <dirName>"
	},
	{
		"readstring",	addReadString,	1,	MANYARGS,
		UP_TRUE,	UP_FALSE,	"[-n] [-0] <string> ..."
	},
	{
		"writedir",	addWriteDirectory, 1,	MANYARGS,
		UP_FALSE,	UP_TRUE,
		"[-o] [-c] [-r] [-m] [-t] [-u] [-d] <dirName>"
	},
	{
		"readnull",	addReadNull,	0,	0,
		UP_TRUE,	UP_FALSE,	NULL
	},
	{
		"writenull",	addWriteNull,	0,	0,
		UP_FALSE,	UP_TRUE,	NULL
	},
	{
		"grep",		addGrep,	1,	MANYARGS,
		UP_FALSE,	UP_FALSE,	"[-fnisvwl] <pattern>"
	},
	{
		"log",		addLog,		1,	1,
		UP_FALSE,	UP_FALSE,	"<fileName>"
	},
	{
		"list",		addList,	1,	MANYARGS,
		UP_FALSE,	UP_FALSE,	"[-lsm] <fileName>"
	},
	{
		"progress",	addProgress,	0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	},
	{
		NULL,		0,		0,	0,
		UP_FALSE,	UP_FALSE,	NULL
	}
};


/*
 * The device which was just added to the pipeline.
 */
static	const	DEV *	latestDevice;


/*
 * The separator string between devices.
 */
static	const char *	separatorString = SEPARATOR;


/*
 * Other private procedures.
 */
static int addDevice(
	UP *		up,
	const char *	name,
	int		argc,
	const char **	argv
);

static	void	showProgress(const UP_PROGRESS * progress);
static	int	parseInt(const char * str);
static	void	usage(void);
static	void	printHelp(void);
static	void	fatal(const char * format, ...);
static	void	badParameter(const char * format, ...);


int
main(argc, argv)
	int		argc;
	const char **	argv;
{
	UP *		up;
	const char *	name;
	int		count;
	UP_DEBUG_FLAG	debugFlags;
	long		warningCount;
	long		maximumWarnings;

	/*
	 * Check for sufficient arguments.
	 * If none are given then print the usage.
	 */
	if (argc < 2)
	{
		usage();

		return EXIT_PARAM;
	}

	argc--;
	argv++;

	/*
	 * Allocate the pipeline.
	 */
	up = upAllocate();

	if (up == UP_NULL)
		fatal("upAllocate Failed");

	/*
	 * Initialize our debugging flags.
	 */
	debugFlags = UP_DEBUG_FLAG_NONE;

	/*
	 * Look for our program options.
	 */
	while ((argc > 0) && (**argv == '-'))
	{
		name = *argv++;
		argc--;

		/*
		 * Look for the help option.
		 */
		if (strcmp(name, "--help") == 0)
		{
			printHelp();

			return EXIT_NORMAL;
		}

		/*
		 * Look for the debugging flags.
		 */
		if (strcmp(name, "--debugState") == 0)
		{
			debugFlags |= UP_DEBUG_FLAG_STATE;

			continue;
		}

		if (strcmp(name, "--debugPoll") == 0)
		{
			debugFlags |= UP_DEBUG_FLAG_POLL;

			continue;
		}

		if (strcmp(name, "--debugLoop") == 0)
		{
			debugFlags |= UP_DEBUG_FLAG_LOOP;

			continue;
		}

		if (strcmp(name, "--debugAll") == 0)
		{
			debugFlags |= UP_DEBUG_FLAG_ALL;

			continue;
		}

		if (strcmp(name, "--separator") == 0)
		{
			if (argc <= 0)
				badParameter("Missing --separator argument");

			separatorString = *argv++;
			argc--;

			continue;
		}

		/*
		 * Check for the warnings limit option.
		 * A value of 0 disables all warning printing.
		 */
		if (strcmp(name, "--warnings") == 0)
		{
			if (argc <= 0)
				badParameter("Missing --warnings argument");

			upSetMaximumWarnings(up, atol(*argv++));
			argc--;

			continue;
		}

		badParameter("Unknown option %s", name);
	}

	/*
	 * Set the debugging flags.
	 */
	upSetDebugFlags(up, debugFlags);

	/*
	 * Keep adding devices until there are no more arguments left.
	 * Each device consumes one or more of the arguments.
	 * The separatorString argument separates devices.
	 */
	while (argc-- > 0)
	{
		name = *argv++;

		count = addDevice(up, name, argc, argv);

		argc -= count;
		argv += count;
	}

	/*
	 * If the latest device added is not an write endpoint device
	 * then add the stdout device as the default writing endpoint.
	 */
	if (latestDevice && !latestDevice->writeEndpoint)
	{
		if (upAddWriteFd(up, STDOUT, UP_FLAG_NONE) == UP_NULL_ID)
		{
			upPrintError(up, "stdout");

			exit(EXIT_PARAM);
		}
	}

	/*
	 * Tell the pipeline that setup is complete.
	 */
	if (!upSetupDone(up))
	{
		upPrintError(up, "Pipeline setup failed");

		return EXIT_FATAL;
	}

	/*
	 * Now run the pipeline until it is done!
	 * Use an infinite timeout value so we block until
	 * some data is ready.
	 */
	while (upReadWrite(up, -1))
	{
		/*
		 * If the pipeline isn't finished then continue.
		 */
		if (!upIsFinished(up))
			continue;

		/*
		 * The pipeline is complete.
		 * Close it and check for errors.
		 */
		if (!upClose(up))
		{
			upPrintError(up, "Error closing pipeline");

			return EXIT_FATAL;
		}

		/*
		 * Get the warning values and print the total if required.
		 */
		warningCount = upGetWarningCount(up);
		maximumWarnings = upGetMaximumWarnings(up);

		if ((warningCount > 0) && (maximumWarnings != 0))
		{
			fprintf(stderr,	"UP warning: Total warnings: %ld\n",
				warningCount);
		}

		/*
		 * Free the pipeline and return either normal exit status
		 * or the status for having had warnings.
		 */
		upFree(up);

		if (warningCount > 0)
			return EXIT_WARNING;

		return EXIT_NORMAL;
	}

	/*
	 * The pipeline has had a fatal error.
	 */
	upPrintError(up, "Pipeline failed");

	upClose(up);

	return EXIT_FATAL;
}


/*
 * Routine to add a device to the pipeline.
 * This collects the arguments for the device and calls the device
 * specific function.  It returns the number of command line
 * arguments consumed by the device routine.
 */
static int
addDevice(UP * up, const char * name, int argc, const char ** argv)
{
	const DEV *	dev;
	int		argCount;

	/*
	 * Search for the device name in the table.
	 */
	for (dev = devTable; dev->name; dev++)
	{
		if (strcmp(dev->name, name) == 0)
			break;
	}

	/*
	 * If it isn't known, then complain.
	 */
	if (dev->name == NULL)
		badParameter("Unknown device type \"%s\"", name);

	/*
	 * Calculate the number of arguments that are present.
	 * Arguments end when the special separatorString argument is seen.
	 */
	argCount = 0;

	while ((argCount < argc) &&
		(strcmp(argv[argCount], separatorString) != 0))
	{
		argCount++;
	}

	/*
	 * Verify the number of arguments that the function wants.
	 */
	if (argCount < dev->minArgs)
		badParameter("Too few arguments for device \"%s\"", name);

	if ((argCount > dev->maxArgs) && (dev->maxArgs != MANYARGS))
		badParameter("Too many arguments for device \"%s\"", name);

	/*
	 * If this is the first device and it is not a reading endpoint
	 * device then first add the stdin device as the default reading
	 * endpoint.
	 */
	if ((latestDevice == NULL) && !dev->readEndpoint)
	{
		if (upAddReadFd(up, STDIN, UP_FLAG_NONE) == UP_NULL_ID)
		{
			upPrintError(up, "stdin");

			exit(EXIT_PARAM);
		}
	}

	/*
	 * Call the function to process the arguments.
	 * If it fails, then print an error message.
	 */
	if (!dev->func(up, argCount, argv))
	{
		upPrintError(up, name);

		exit(EXIT_PARAM);
	}

	/*
	 * Save this device as being the latest one defined.
	 */
	latestDevice = dev;

	/*
	 * Return the number of arguments consumed.
	 * Account for the separator argument if present.
	 */
	if (argCount < argc)
		argCount++;

	return argCount;
}


static UP_BOOL
addCat(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	fileName;

	fileName = NULL;

	if (argc > 0)
		fileName = *argv;

	id = upAddCat(up, fileName, UP_FLAG_NONE);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addRemoveData(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddRemoveData(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addFindSize(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddFindSize(up, FINDSIZE_SIZE,
		UP_FLAG_FINDSIZE_TRUNCATE | UP_FLAG_FINDSIZE_PAD);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addFilter(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	UP_FILTER_ITEM	item[2];

	item[0].type = UP_FILTER_TYPE_HAS_PATH_SUFFIX;
	item[0].trueIndex = 1;
	item[0].falseIndex = UP_FILTER_INDEX_REJECT;
	item[0].string = ".o";

	item[1].type = UP_FILTER_TYPE_MATCHES_FILE;
	item[1].trueIndex = UP_FILTER_INDEX_REJECT;
	item[1].falseIndex = UP_FILTER_INDEX_ACCEPT;
	item[1].string = ".*i.*";

	id = upAddFilter(up, item, 2);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addLog(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddLogFile(up, *argv, UP_FLAG_LOG_ITEMS);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addList(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	UP_FLAG		flags;
	const char *	str;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'l':
				flags |= (UP_FLAG_LIST_MODE |
					UP_FLAG_LIST_GIVEN_SIZE);
				break;

			case 's':
				flags |= UP_FLAG_LIST_TRUE_SIZE;
				break;
			case 'm':
				flags |= UP_FLAG_LIST_MODE;
				break;
			default:
				badParameter("Unknown list option");
		}
	}

	if (argc < 1)
		badParameter("Missing list file name");

	if (argc > 1)
		badParameter("Only one list file name allowed");

	id = upAddList(up, *argv, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addReadString(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	UP_BOOL		terminate;
	int		separator;
	char *		buffer;
	char *		cp;
	int		len;
	int		i;
	int		bufferLength;

	terminate = UP_TRUE;
	separator = '\n';

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'n':
				terminate = UP_FALSE;
				break;
			case '0':
				separator = '\0';
				break;
			default:
				badParameter("Unknown readString flag");
		}
	}

	/*
	 * Allocate a string large enough to hold all of the arguments
	 * plus separating characters.
	 */
	bufferLength = argc + 1;

	for (i = 0; i < argc; i++)
		bufferLength += strlen(argv[i]);

	buffer = malloc(bufferLength);

	if (buffer == NULL)
		fatal("Memory allocation failed");

	/*
	 * Copy all of the strings separated by the separator character
	 * into the buffer.
	 */
	cp = buffer;

	for (i = 0; i < argc; i++)
	{
		if (i != 0)
			*cp++ = separator;

		len = strlen(argv[i]);

		memcpy(cp, argv[i], len);

		cp += len;
	}

	/*
	 * Add the trailing separator if required.
	 */
	if (terminate)
		*cp++ = separator;

	bufferLength = cp - buffer;

	/*
	 * Add the device saying that it can free the string.
	 */
	id = upAddReadString(up, buffer, bufferLength,
		UP_FLAG_READSTRING_CAN_FREE);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addReadNull(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddReadNull(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addWriteNull(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddWriteNull(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addGrep(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	UP_FLAG		flags;
	const char *	str;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
		case 'i':
			flags |= UP_FLAG_GREP_IGNORE_CASE;
			break;
		case 'n':
			flags |= UP_FLAG_GREP_SHOW_LINE_NUMBERS;
			break;
		case 'f':
			flags |= UP_FLAG_GREP_SHOW_FILE_NAMES;
			break;
		case 'l':
			flags |= UP_FLAG_GREP_MATCH_WHOLE_LINE;
			break;
		case 'w':
			flags |= UP_FLAG_GREP_MATCH_WHOLE_WORD;
			break;
		case 's':
			flags |= UP_FLAG_GREP_STRING_MATCH;
			break;
		case 'v':
			flags |= UP_FLAG_GREP_SHOW_NON_MATCHES;
			break;
		}
	}

	if (argc < 1)
		badParameter("Pattern required");

	if (argc > 1)
		badParameter("Only one pattern allowed");

	id = upAddGrep(up, *argv, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addMakeFileList(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	UP_FLAG		flags;
	const char *	str;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case '0':
				flags |= UP_FLAG_MAKEFILELIST_NULL_SEPARATORS;
				break;
			default:
				badParameter("Unknown option for makefilelist");
		}
	}

	if (argc > 0)
		badParameter("No arguments allowed for filelist");

	id = upAddMakeFileList(up, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addUuencode(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddUUEncode(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addGnuZip(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	int		level;
	int		argCount;

	argCount = 0;

	if (argc == 0)
		level = 9;
	else
	{
		str = *argv;

		if ((*str++ != '-') || (*str < '0') ||
			(*str > '9') || (str[1] != '\0'))
		{
			badParameter("Invalid gzip level");
		}

		level = *str - '0';
		argCount++;
	}

	id = upAddGnuZip(up, level);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addGnuUnzip(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddGnuUnzip(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addProcess(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	UP_FLAG		flags;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'p':
				flags |= UP_FLAG_PROC_PARTIAL_INPUT_OK;
				break;
			case 'a':
				flags |= UP_FLAG_PROC_EXIT_ANY_OK;
				break;
			case '1':
				flags |= UP_FLAG_PROC_EXIT_1_OK;
				break;
			case '2':
				flags |= UP_FLAG_PROC_EXIT_2_OK;
				break;
			case 's':
				flags |= UP_FLAG_PROC_SYSTEM_PATH;
				break;
			default:
				badParameter("Unknown process flag");
		}
	}

	if (argc < 1)
		badParameter("Missing program name");

	id = upAddProcessArgs(up, *argv, argc, argv, flags, NULL);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addProgress(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddProgress(up, NULL, showProgress, NULL);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addCompress(UP * up, int argc, const char ** argv)
{
	UP_ID	id;
	int	bits;

	bits = 0;

	if (argc > 0)
		bits = parseInt(*argv);

	id = upAddCompress(up, bits);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addUnCompress(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddUnCompress(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addArchive(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	name;

	name = NULL;

	if (argc > 0)
		name = *argv;

	id = upAddArchive(up, name);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addUnArchive(UP * up, int argc, const char ** argv)
{
	UP_ID	id;

	id = upAddUnArchive(up);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addUudecode(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	name;

	name = NULL;

	if (argc > 0)
		name = *argv;

	id = upAddUUDecode(up, name);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addTar(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	name;

	name = NULL;

	if (argc > 0)
		name = *argv;

	id = upAddTar(up, name);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addUnTar(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	name;

	name = NULL;

	if (argc > 0)
		name = *argv;

	id = upAddUnTar(up, name);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addReadDirectory(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	int		depth;
	UP_FLAG		flags;

	flags = UP_FLAG_NONE;
	depth = 0;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'n':
				flags |= UP_FLAG_READDIR_NAMES_ONLY;
				break;
			case 's':
				flags |= UP_FLAG_READDIR_NO_SPECIAL_FILES;
				break;
			case 'd':
				flags |= UP_FLAG_READDIR_NO_DIRECTORIES;
				break;
			case 'f':
				flags |= UP_FLAG_READDIR_NO_REGULAR_FILES;
				break;
			case 'x':
				flags |= UP_FLAG_READDIR_NO_CROSS_FILE_SYSTEMS;
				break;
			case 'i':
				flags |= UP_FLAG_READDIR_IGNORE_ERRORS;
				break;
			case 'l':
				if ((argc < 1) || (**argv == '-'))
					badParameter("Missing level");

				depth = parseInt(*argv++);
				argc--;
				break;

			default:
				badParameter("Unknown readdirectory flag");
		}
	}

	if (argc < 1)
		badParameter("Missing directory name");

	if (argc > 1)
		badParameter("Only one directory name allowed");

	id = upAddReadDirectory(up, *argv, depth, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addReadFiles(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	UP_FLAG		flags;
	const char *	str;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'i':
				flags |= UP_FLAG_READFILES_IGNORE_ERRORS;
				break;
			default:
				badParameter("Unknown readfiles option");
		}
	}

	if (argc < 1)
		badParameter("No files for ReadFiles");

	id = upAddReadFileList(up, argv, argc, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addWriteFile(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	UP_FLAG		flags;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'p':
				flags |= UP_FLAG_WRITEFILE_PATTERN;
				break;
			default:
				badParameter("Unknown writefile flag");
		}
	}

	if (argc < 1)
		badParameter("Missing writefile name");

	if (argc > 1)
		badParameter("Too many writefile arguments");

	id = upAddWriteFile(up, *argv, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addWriteDirectory(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	UP_FLAG		flags;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'o':
				flags |= UP_FLAG_WRITEDIR_NO_OVERWRITE;
				break;
			case 'c':
				flags |= UP_FLAG_WRITEDIR_ALLOW_OPEN_ERRORS;
				break;
			case 'r':
				flags |= UP_FLAG_WRITEDIR_RELATIVE;
				break;
			case 'd':
				flags |= UP_FLAG_WRITEDIR_NO_CREATE_DIRECTORIES;
				break;
			case 'm':
				flags |= UP_FLAG_WRITEDIR_SET_MODE;
				break;
			case 'u':
				flags |= UP_FLAG_WRITEDIR_SET_OWNER;
				break;
			case 't':
				flags |= UP_FLAG_WRITEDIR_SET_TIMES;
				break;
			default:
				badParameter("Unknown writedirectory flag");
		}
	}

	if (argc < 1)
		badParameter("Missing output directory");

	if (argc > 1)
		badParameter("Too many output directories");

	id = upAddWriteDirectory(up, *argv, flags);

	return (id != UP_NULL_ID);
}


static UP_BOOL
addFindFileData(UP * up, int argc, const char ** argv)
{
	UP_ID		id;
	const char *	str;
	UP_FLAG		flags;

	flags = UP_FLAG_NONE;

	while ((argc > 0) && (**argv == '-'))
	{
		str = *argv++;
		argc--;

		while (*++str) switch (*str)
		{
			case 'i':
				flags |= UP_FLAG_FINDFILEDATA_IGNORE_ERRORS;
				break;
			case 'r':
				flags |= UP_FLAG_FINDFILEDATA_ALWAYS_READ;
				break;
			default:
				badParameter("Unknown findfiledata option");
		}
	}

	if (argc > 0)
		badParameter("Too many arguments for addfiledata");

	id = upAddFindFileData(up, flags);

	return (id != UP_NULL_ID);
}


/*
 * Callback function for the PipeProgress device to display progress.
 */
static void
showProgress(const UP_PROGRESS * progress)
{
	/*
	 * Switch on the reason for the callback.
	 */
	switch (progress->lastAction)
	{
	case UP_PROGRESS_ACTION_FILE_NAME:
		printf("File \"%s\" <", progress->fileName);
		break;

	case UP_PROGRESS_ACTION_END_OF_FILE:
		printf("%ld>\n", progress->fileOffset);
		break;

	case UP_PROGRESS_ACTION_FILE_ERROR:
		printf("%ld error>\n", progress->fileOffset);
		break;

	default:
		return;
	}

	fflush(stdout);
}


/*
 * Convert a string into an integer.
 * Doen't return on an error.
 */
static int
parseInt(const char * str)
{
	int	result;
	UP_BOOL	isNeg;

	result = 0;
	isNeg = UP_FALSE;

	if (*str == '-')
	{
		str++;
		isNeg = UP_TRUE;
	}

	if (*str == '\0')
		badParameter("Bad number");

	while (((*str >= '0') && (*str <= '9')) && (result >= 0))
		result = result * 10 + (*str++ - '0');

	if (*str || (result < 0))
		badParameter("Bad number");

	if (isNeg)
		result = -result;

	return result;
}


/*
 * Print out a fatal error and exit.
 * The error message can be formatted.
 */
static void
fatal(const char * format, ...)
{
	va_list	ap;
	char	buf[8192];

	va_start(ap, format);
	vsprintf(buf, format, ap);
	va_end(ap);

	fprintf(stderr, "%s\n", buf);

	exit(EXIT_FATAL);
}


/*
 * Print out a bad parameter error and exit.
 * The error message can be formatted.
 */
static void
badParameter(const char * format, ...)
{
	va_list	ap;
	char	buf[8192];

	va_start(ap, format);
	vsprintf(buf, format, ap);
	va_end(ap);

	fprintf(stderr, "%s\n", buf);

	exit(EXIT_PARAM);
}


/*
 * Print usage information to stderr.
 */
static void
usage(void)
{
	int	major;
	int	minor;

	upGetVersion(&major, &minor);

	fprintf(stderr, "\n");
	fprintf(stderr, "up - User mode pipeline utility (version %d.%d)\n",
		major, minor);

	fprintf(stderr, "\n");
	fprintf(stderr,
"Usage: up [options] <device> [argument] ... [= <device> [argument] ...] ...\n");

	fprintf(stderr, "\n");
	fprintf(stderr, "Use the --help option for more information\n");
	fprintf(stderr, "\n");
}


/*
 * The help lines for the program.
 * The list of lines must be terminated by a NULL string.
 */
static	const char * const	helpLines[] =
{
"",
"Usage: up [options] <device> [argument] ... [= <device> [argument] ...] ...",
"",
"This program constructs a \"user-mode\" pipeline which can process data",
"similarly to a normal UNIX pipeline, but which can handle multiple files.",
"Most of the data which is processed stays within this single process.",
"",
"Each stage of the pipeline is called a \"device\".  Data flows from the",
"left-most device through to the right-most device.  The left-most device",
"must read data from some input source, and the right-most device must write",
"data to some output destination.  If the devices specified for the pipeline",
"do not meet this requirement, then devices which read from stdin or which",
"write to stdout are automatically inserted into the pipeline.",
"",
"Devices are specified by name and are followed by possible arguments.",
"The devices are separated by a special argument value appearing by itself.",
"By default, this argument is a single equals sign.  This can be changed",
"if necessary so that the equals sign can be used as a device argument.",
"",
"The following example shows how to read a tar file, compress each file",
"in the tar file individually, and write out a new tar file containing the",
"compressed files:",
"",
"  up readfiles input.tar = untar = gzip = tar = writefile output.tar",
"",
"Options global to the program are specified before the first device name.",
"These can be any of the following:",
"",
"  --separator str   Set the device separator string value",
"  --warnings num    Set the maximum number of warning messages printed",
"  --debugState      Print debugging messages about device state changes",
"  --debugPoll       Print debugging messages about the use of poll",
"  --debugLoop       Print debugging messages about each I/O loop call",
"  --debugAll        Print all debugging messages",
"  --help            Print this help information and exit",
"",
NULL
};


/*
 * Print out additional help for using the program.
 * This output goes to stdout.
 */
static void
printHelp(void)
{
	const char * const *	text;
	const DEV *		dev;

	/*
	 * Print the help lines.
	 */
	for (text = helpLines; *text; text++)
		printf("%s\n", *text);

	/*
	 * Print the reading enpoint devices.
	 */
	printf("Available reading devices are:\n");
	printf("\n");

	for (dev = devTable; dev->name; dev++)
	{
		if (!dev->readEndpoint)
			continue;

		if (dev->argString)
			printf("  %s %s\n", dev->name, dev->argString);
		else
			printf("  %s\n", dev->name);
	}

	/*
	 * Print the writing endpoint devices.
	 */
	printf("\n");
	printf("Available writing devices are:\n");
	printf("\n");

	for (dev = devTable; dev->name; dev++)
	{
		if (!dev->writeEndpoint)
			continue;

		if (dev->argString)
			printf("  %s %s\n", dev->name, dev->argString);
		else
			printf("  %s\n", dev->name);
	}

	/*
	 * Print the pipe-like devices.
	 */
	printf("\n");
	printf("Available filtering devices are:\n");
	printf("\n");

	for (dev = devTable; dev->name; dev++)
	{
		if (dev->readEndpoint || dev->writeEndpoint)
			continue;

		if (dev->argString)
			printf("  %s %s\n", dev->name, dev->argString);
		else
			printf("  %s\n", dev->name);
	}

	printf("\n");
}
