/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * Public definitions for user mode pipelines.
 * The code supporting these definitions is in upPublic.c.
 */

#ifndef	UP_H
#define	UP_H

#include <sys/types.h>
#include <poll.h>


/*
 * Some basic typedefs.
 * Some of these are only defined privately within upPrivate.h.
 */
typedef	struct	UP	UP;
typedef	struct	UP_DEV	UP_DEV;
typedef	struct	UP_BUF	UP_BUF;
typedef	struct	UP_POLL	UP_POLL;
typedef	unsigned int	UP_STATFLAGS;
typedef	int		UP_BOOL;
typedef	int		UP_TYPE;
typedef	int		UP_ID;
typedef	int		UP_FLAG;
typedef	int		UP_DEBUG_FLAG;
typedef	int		UP_FILTER_TYPE;


/*
 * Booleans.
 */
#define	UP_TRUE		((UP_BOOL) 1)
#define	UP_FALSE	((UP_BOOL) 0)


/*
 * The structure which holds data about the user mode pipeline.
 * Note: All of the data within this structure is considered private.
 * Clients should only reference the data by using the access methods.
 */
struct	UP
{
	UP_DEV	*	first;		/* first device in pipeline */
	UP_DEV	*	last;		/* last device in pipeline */
	struct pollfd *	pollTable;	/* systen poll table */
	int		pollCount;	/* entries used in poll table */
	int		pollAllocCount;	/* allocated size of poll table */
	int		pollTimeout;	/* polling timeout value */
	int		deviceCount;	/* count of devices in pipeline */
	int		errorCode;	/* last error code */
	int		systemErrno;	/* saved system errno value */
	UP_ID		errorDeviceId;	/* device id with error */
	const char *	errorDeviceName;/* name of device with error */
	const char *	systemName;	/* system call name causing error */
	char *		systemFileName;	/* file name causing error */
	const char *	errorFileName;	/* file name of internal error */
	int		errorLineNumber;/* line number of internal error */
	UP_BOOL		setupDone;	/* setup is done */
	UP_BOOL		pollBlab;	/* describe poll calls */
	UP_BOOL		loopBlab;	/* describe main I/O loop */
	UP_BOOL		stateBlab;	/* describe state changes */
	long		maxWarningCount;/* max number of warnings printed */
	long		warningCount;	/* counter of warnings */
	char		padding[64];	/* room for future changes */
};


/*
 * Structure returned when reading files from a device.
 * This returns the information available from the device about the file.
 * Information which is not known by the device is set to a zero value.
 * In addition, the flags element indicates which information is available.
 */
typedef	struct
{
	UP_STATFLAGS	flags;		/* which fields are valid */
	int		type;		/* char for file type (like 'ls') */
	UP_BOOL		isLink;		/* is a hard link to previous file */
	mode_t		mode;		/* file mode */
	dev_t		rdev;		/* real device number */
	dev_t		dev;		/* device number */
	ino_t		inode;		/* inode number */
	nlink_t		nlink;		/* number of links */
	uid_t		uid;		/* owner id */
	gid_t		gid;		/* group id */
	off_t		size;		/* file size */
	time_t		atime;		/* last accessed time */
	time_t		mtime;		/* last modified time */
	time_t		ctime;		/* last inode change time */
} UP_STAT;


/*
 * Bits in the flags field of the UP_STAT structure which indicate
 * which types of information is available.  These are OR'd together.
 */
#define	UP_STAT_DATA	((UP_STATFLAGS) 0x0001)
#define	UP_STAT_TYPE	((UP_STATFLAGS) 0x0002)
#define	UP_STAT_MODE	((UP_STATFLAGS) 0x0004)
#define	UP_STAT_DEV	((UP_STATFLAGS) 0x0008)
#define	UP_STAT_INODE	((UP_STATFLAGS) 0x0010)
#define	UP_STAT_NLINK	((UP_STATFLAGS) 0x0020)
#define	UP_STAT_UID	((UP_STATFLAGS) 0x0040)
#define	UP_STAT_GID	((UP_STATFLAGS) 0x0080)
#define	UP_STAT_RDEV	((UP_STATFLAGS) 0x0100)
#define	UP_STAT_SIZE	((UP_STATFLAGS) 0x0200)
#define	UP_STAT_ATIME	((UP_STATFLAGS) 0x0400)
#define	UP_STAT_MTIME	((UP_STATFLAGS) 0x0800)
#define	UP_STAT_CTIME	((UP_STATFLAGS) 0x1000)


/*
 * All the stat flags.
 */
#define	UP_STAT_ALLFLAGS \
	(UP_STAT_DATA | UP_STAT_TYPE | UP_STAT_MODE | UP_STAT_DEV | \
	UP_STAT_INODE | UP_STAT_NLINK | UP_STAT_UID | UP_STAT_GID | \
	UP_STAT_RDEV | UP_STAT_SIZE | UP_STAT_ATIME | UP_STAT_MTIME | \
	UP_STAT_CTIME)


/*
 * The system path list that we allow programs to be run from when secure.
 */
#ifndef	UP_SYSTEM_PATH
#define	UP_SYSTEM_PATH	"/bin:/usr/bin:/sbin:/usr/sbin:/usr/local/bin"
#endif


/*
 * Types of devices.
 */
#define	UP_TYPE_NONE		((UP_TYPE) 0)
#define	UP_TYPE_READ_NULL	((UP_TYPE) 1)
#define	UP_TYPE_WRITE_NULL	((UP_TYPE) 2)
#define	UP_TYPE_READ_ENDPOINT	((UP_TYPE) 3)
#define	UP_TYPE_WRITE_ENDPOINT	((UP_TYPE) 4)
#define	UP_TYPE_READ_JOIN	((UP_TYPE) 5)
#define	UP_TYPE_WRITE_JOIN	((UP_TYPE) 6)
#define	UP_TYPE_READ_FD		((UP_TYPE) 7)
#define	UP_TYPE_WRITE_FD	((UP_TYPE) 8)
#define	UP_TYPE_READ_FILES	((UP_TYPE) 9)
#define	UP_TYPE_WRITE_FILE	((UP_TYPE) 10)
#define	UP_TYPE_READ_STRING	((UP_TYPE) 11)
#define	UP_TYPE_WRITE_STRING	((UP_TYPE) 12)
#define	UP_TYPE_PROCESS		((UP_TYPE) 13)
#define	UP_TYPE_ARCHIVE		((UP_TYPE) 14)
#define	UP_TYPE_UNARCHIVE	((UP_TYPE) 15)
#define	UP_TYPE_TAR		((UP_TYPE) 16)
#define	UP_TYPE_UNTAR		((UP_TYPE) 17)
#define	UP_TYPE_LOG		((UP_TYPE) 18)
#define	UP_TYPE_REMOVE_DATA	((UP_TYPE) 19)
#define	UP_TYPE_GREP		((UP_TYPE) 20)
#define	UP_TYPE_READ_DIRECTORY	((UP_TYPE) 21)
#define	UP_TYPE_UUDECODE	((UP_TYPE) 22)
#define	UP_TYPE_UUENCODE	((UP_TYPE) 23)
#define	UP_TYPE_GZIP		((UP_TYPE) 24)
#define	UP_TYPE_GUNZIP		((UP_TYPE) 25)
#define	UP_TYPE_CAT		((UP_TYPE) 26)
#define	UP_TYPE_LIST		((UP_TYPE) 27)
#define	UP_TYPE_WRITE_DIRECTORY	((UP_TYPE) 28)
#define	UP_TYPE_FIND_SIZE	((UP_TYPE) 29)
#define	UP_TYPE_FILTER		((UP_TYPE) 30)
#define	UP_TYPE_PROGRESS	((UP_TYPE) 31)
#define	UP_TYPE_FIND_FILE_DATA	((UP_TYPE) 32)
#define	UP_TYPE_MAKE_FILE_LIST	((UP_TYPE) 33)

#define	UP_TYPE_USER		((UP_TYPE) 1000)


/*
 * Some null values.
 */
#define	UP_NULL		((UP *) 0)
#define	UP_NULL_ID	((UP_ID) 0)


/*
 * Various flags for some devices.
 * Each device type which uses flags has its own independent set of
 * definitions, which contains its device name.  The flags are OR'd
 * together as required.
 */

/*
 * No flags set at all value for all devices.
 */
#define	UP_FLAG_NONE	((UP_FLAG) 0)


/*
 * ReadDirectory device flags.
 */
#define	UP_FLAG_READDIR_NAMES_ONLY		((UP_FLAG) 0x01)
#define	UP_FLAG_READDIR_NO_REGULAR_FILES	((UP_FLAG) 0x02)
#define	UP_FLAG_READDIR_NO_SPECIAL_FILES	((UP_FLAG) 0x04)
#define	UP_FLAG_READDIR_NO_DIRECTORIES		((UP_FLAG) 0x08)
#define	UP_FLAG_READDIR_NO_CROSS_FILE_SYSTEMS	((UP_FLAG) 0x10)
#define	UP_FLAG_READDIR_IGNORE_ERRORS		((UP_FLAG) 0x20)


/*
 * ReadFiles device flags.
 */
#define	UP_FLAG_READFILES_IGNORE_ERRORS		((UP_FLAG) 0x01)


/*
 * PipeFindFileData device flags.
 */
#define	UP_FLAG_FINDFILEDATA_IGNORE_ERRORS	((UP_FLAG) 0x01)
#define	UP_FLAG_FINDFILEDATA_ALWAYS_READ	((UP_FLAG) 0x02)


/*
 * WriteDirectory device flags.
 */
#define	UP_FLAG_WRITEDIR_RELATIVE		((UP_FLAG) 0x01)
#define	UP_FLAG_WRITEDIR_NO_CREATE_DIRECTORIES	((UP_FLAG) 0x02)
#define	UP_FLAG_WRITEDIR_NO_OVERWRITE		((UP_FLAG) 0x04)
#define	UP_FLAG_WRITEDIR_ALLOW_OPEN_ERRORS	((UP_FLAG) 0x08)
#define	UP_FLAG_WRITEDIR_SET_MODE		((UP_FLAG) 0x10)
#define	UP_FLAG_WRITEDIR_SET_OWNER		((UP_FLAG) 0x20)
#define	UP_FLAG_WRITEDIR_SET_TIMES		((UP_FLAG) 0x40)


/*
 * PipeLogFile device flags.
 */
#define	UP_FLAG_LOG_FILE_NAME	((UP_FLAG) 0x01)
#define	UP_FLAG_LOG_FILE_DATA	((UP_FLAG) 0x02)
#define	UP_FLAG_LOG_ITEMS	((UP_FLAG) 0x04)


/*
 * PipeList device flags.
 */
#define	UP_FLAG_LIST_MODE	((UP_FLAG) 0x01)
#define	UP_FLAG_LIST_GIVEN_SIZE	((UP_FLAG) 0x02)
#define	UP_FLAG_LIST_TRUE_SIZE	((UP_FLAG) 0x04)


/*
 * PipeFindSize device flags.
 */
#define	UP_FLAG_FINDSIZE_TRUNCATE	((UP_FLAG) 0x01)
#define	UP_FLAG_FINDSIZE_PAD		((UP_FLAG) 0x02)
#define	UP_FLAG_FINDSIZE_BUFFER		((UP_FLAG) 0x04)


/*
 * PipeProcess device flags.
 */
#define	UP_FLAG_PROC_PARTIAL_INPUT_OK	((UP_FLAG) 0x01)
#define	UP_FLAG_PROC_EXIT_1_OK		((UP_FLAG) 0x02)
#define	UP_FLAG_PROC_EXIT_2_OK		((UP_FLAG) 0x04)
#define	UP_FLAG_PROC_EXIT_ANY_OK	((UP_FLAG) 0x08)
#define	UP_FLAG_PROC_SYSTEM_PATH	((UP_FLAG) 0x10)


/*
 * PipeGrep device flags.
 */
#define	UP_FLAG_GREP_SHOW_FILE_NAMES	((UP_FLAG) 0x01)
#define	UP_FLAG_GREP_SHOW_LINE_NUMBERS	((UP_FLAG) 0x02)
#define	UP_FLAG_GREP_SHOW_NON_MATCHES	((UP_FLAG) 0x04)
#define	UP_FLAG_GREP_IGNORE_CASE	((UP_FLAG) 0x08)
#define	UP_FLAG_GREP_STRING_MATCH	((UP_FLAG) 0x10)
#define	UP_FLAG_GREP_MATCH_WHOLE_WORD	((UP_FLAG) 0x20)
#define	UP_FLAG_GREP_MATCH_WHOLE_LINE	((UP_FLAG) 0x40)


/*
 * ReadFd device flags.
 */
#define	UP_FLAG_READFD_CAN_CLOSE	((UP_FLAG) 0x01)


/*
 * WriteFd device flags.
 */
#define	UP_FLAG_WRITEFD_CAN_CLOSE	((UP_FLAG) 0x01)


/*
 * WriteFile device flags.
 */
#define	UP_FLAG_WRITEFILE_PATTERN	((UP_FLAG) 0x01)


/*
 * ReadString device flags.
 */
#define	UP_FLAG_READSTRING_CAN_FREE	((UP_FLAG) 0x01)


/*
 * MakeFileList device flags.
 */
#define	UP_FLAG_MAKEFILELIST_NULL_SEPARATORS	((UP_FLAG) 0x01)


/*
 * A data item for the PipeFilter device which is a test for a file.
 * An array of these determines whether a file should be passed
 * through the device.  Each test generates a TRUE or FALSE condition.
 * These select the index in the array of the next test.  Both indexes
 * must refer to an item further down in the array, except for special
 * cases of 0 and -1 (see UP_FILTER_INDEX descriptions below.)
 * Each item has a type, which can be a string, an integer, or a file
 * offset. (Only the appropriate value is referenced.)  String values
 * and the whole filter item array are copied when the PipeFilter device
 * is opened and so do not have to be preserved by the caller.
 * Some filter items are modifiers which change the file name or
 * status of the file.  The file data is never modified.
 */
typedef	struct
{
	UP_FILTER_TYPE	type;		/* type of test to make */
	UP_FLAG		flags;		/* modifier flags */
	int		trueIndex;	/* index of next test if true */
	int		falseIndex;	/* index of next test if false */
	char *		string;		/* string value for test */
	long		integer;	/* integer value for test */
	off_t		offset;		/* file size value */
	time_t		time;		/* time value */
} UP_FILTER_ITEM;


/*
 * Flags which can be set in a filter item to modify the test.
 * The ignore case flag is only used for string and pattern values.
 * The comparison flags are only used for string or integer values,
 * not for patterns or booleans.  If none of LT, EQ, or GT are set
 * then EQ is assumed.
 */
#define	UP_FILTER_ITEM_FLAG_IGNORE_CASE		((UP_FLAG) 0x01)
#define	UP_FILTER_ITEM_FLAG_LT			((UP_FLAG) 0x02)
#define	UP_FILTER_ITEM_FLAG_EQ			((UP_FLAG) 0x04)
#define	UP_FILTER_ITEM_FLAG_GT			((UP_FLAG) 0x08)


/*
 * Special index values for termination of the filtering checks.
 * An index value of UP_FILTER_INDEX_ACCEPT stops examining of the
 * UP_FILTER_ITEM table, and transfers the file to the next device.
 * An index value of UP_FILTER_INDEX_REJECT stops examining of the
 * UP_FILTER_ITEM table, and consumes the file without passing it to
 * the next device.  The other filter value is only used internally.
 * All of thest filter values must be non-positive since positive
 * index values refer to other items in the filter table.
 */
#define	UP_FILTER_INDEX_ACCEPT	0
#define	UP_FILTER_INDEX_REJECT	-1
#define	UP_FILTER_INDEX_ERROR	-2


/*
 * Types of filtering tests.
 * These do not modify anything.
 * PATH is the complete incoming name to be tested.
 * ENTRY is one component of the path (i.e., separated by slashes).
 * DIR is all entries of the path except the last (empty if none).
 * FILE is the last entry of the path.
 * BASE is the part of the FILE prior to the last dot (whole FILE if none).
 * EXTENSION is the part of the FILE after the last dot (empty if none).
 */
#define	UP_FILTER_TYPE_NONE			0	/* always succeeds */
#define	UP_FILTER_TYPE_FILE_NUMBER		1	/* file number */

#define	UP_FILTER_TYPE_HAS_PATH			2	/* given path */
#define	UP_FILTER_TYPE_HAS_ENTRY		3	/* given entry */
#define	UP_FILTER_TYPE_HAS_DIR			4	/* given dir */
#define	UP_FILTER_TYPE_HAS_FILE			5	/* given file */
#define	UP_FILTER_TYPE_HAS_BASE			6	/* given base */
#define	UP_FILTER_TYPE_HAS_EXTENSION		7	/* given extension */

#define	UP_FILTER_TYPE_HAS_PATH_PREFIX		8	/* path prefix */
#define	UP_FILTER_TYPE_HAS_ENTRY_PREFIX		9	/* entry prefix */
#define	UP_FILTER_TYPE_HAS_DIR_PREFIX		10	/* dir prefix */
#define	UP_FILTER_TYPE_HAS_FILE_PREFIX		11	/* file prefix */
#define	UP_FILTER_TYPE_HAS_BASE_PREFIX		12	/* base prefix */
#define	UP_FILTER_TYPE_HAS_EXTENSION_PREFIX	13	/* extension prefix */

#define	UP_FILTER_TYPE_HAS_PATH_SUFFIX		14	/* path suffix */
#define	UP_FILTER_TYPE_HAS_ENTRY_SUFFIX		15	/* entry suffix */
#define	UP_FILTER_TYPE_HAS_DIR_SUFFIX		16	/* dir suffix */
#define	UP_FILTER_TYPE_HAS_FILE_SUFFIX		17	/* file suffix */
#define	UP_FILTER_TYPE_HAS_BASE_SUFFIX		18	/* base suffix */
#define	UP_FILTER_TYPE_HAS_EXTENSION_SUFFIX	19	/* extension suffix */

#define	UP_FILTER_TYPE_MATCHES_PATH		20	/* matches path */
#define	UP_FILTER_TYPE_MATCHES_ENTRY		21	/* matches entry */
#define	UP_FILTER_TYPE_MATCHES_DIR		22	/* matches dir */
#define	UP_FILTER_TYPE_MATCHES_FILE		23	/* matches file */
#define	UP_FILTER_TYPE_MATCHES_BASE		24	/* matches base */
#define	UP_FILTER_TYPE_MATCHES_EXTENSION	25	/* matches extension */

#define	UP_FILTER_TYPE_IS_ABSOLUTE		26	/* name is absolute */
#define	UP_FILTER_TYPE_IS_ENTRY			27	/* name is entry */
#define	UP_FILTER_TYPE_ENTRY_COUNT		28	/* number of entries */

#define	UP_FILTER_TYPE_IS_REGULAR		29	/* file is regular */
#define	UP_FILTER_TYPE_IS_DIRECTORY		30	/* file is directory */
#define	UP_FILTER_TYPE_IS_SPECIAL		31	/* file is some special file */
#define	UP_FILTER_TYPE_FILE_SIZE		32	/* file size */
#define	UP_FILTER_TYPE_FILE_UID			33	/* uid */
#define	UP_FILTER_TYPE_FILE_GID			34	/* gid */
#define	UP_FILTER_TYPE_FILE_ATIME		35	/* access time */
#define	UP_FILTER_TYPE_FILE_MTIME		36	/* modify time */
#define	UP_FILTER_TYPE_FILE_CTIME		37	/* change time */
#define	UP_FILTER_TYPE_FILE_DEV			38	/* device */
#define	UP_FILTER_TYPE_FILE_INODE		39	/* inode */
#define	UP_FILTER_TYPE_FILE_NLINKS		40	/* links */
#define	UP_FILTER_TYPE_HAS_ANY_MODE_BITS	41	/* has any mode bits */
#define	UP_FILTER_TYPE_HAS_ALL_MODE_BITS	42	/* has all mode bits */
#define	UP_FILTER_TYPE_HAS_DATA			43	/* data is provided */


/*
 * Types of filtering modifications.
 * These modify the file name or status of the file.
 */
#define	UP_FILTER_TYPE_SET_UID			1000	/* set uid */
#define	UP_FILTER_TYPE_SET_GID			1001	/* set gid */
#define	UP_FILTER_TYPE_SET_MODE			1002	/* set mode bits */
#define	UP_FILTER_TYPE_ADD_MODE_BITS		1003	/* add mode bits */
#define	UP_FILTER_TYPE_CLEAR_MODE_BITS		1004	/* clear mode bits */
#define	UP_FILTER_TYPE_PREPEND_PATH		1005	/* add path in front */
#define	UP_FILTER_TYPE_APPEND_PATH		1006	/* add path in back */
#define	UP_FILTER_TYPE_PRUNE_PATH_FRONT		1007	/* prune path in front */
#define	UP_FILTER_TYPE_PRUNE_PATH_BACK		1008	/* prune path in back */
#define	UP_FILTER_TYPE_REDUCE_DEPTH_FRONT	1009	/* reduce path depth */
#define	UP_FILTER_TYPE_REDUCE_DEPTH_BACK	1010	/* reduce path depth */
#define	UP_FILTER_TYPE_REMOVE_PATH		1011	/* remove path */
#define	UP_FILTER_TYPE_REPLACE_DIR		1012	/* replace directory */
#define	UP_FILTER_TYPE_REPLACE_FILE		1013	/* replace file */
#define	UP_FILTER_TYPE_REPLACE_EXTENSION	1014	/* replace extension */
#define	UP_FILTER_TYPE_REPLACE_BASE		1015	/* replace base */
#define	UP_FILTER_TYPE_ADD_EXTENSION		1016	/* add extension */
#define	UP_FILTER_TYPE_REMOVE_EXTENSION		1017	/* remove extension */


/*
 * Progress actions used in the UP_PROGRESS structure for the
 * PipeProgress device.  This indicates why the progress information
 * was updated.
 */
typedef enum
{
	UP_PROGRESS_ACTION_NONE,	/* nothing has happened yet */
	UP_PROGRESS_ACTION_FILE_NAME,	/* file name transferred */
	UP_PROGRESS_ACTION_FILE_DATA,	/* file data transferred */
	UP_PROGRESS_ACTION_END_OF_FILE,	/* end of file */
	UP_PROGRESS_ACTION_FILE_ERROR,	/* file had error (implies EOF) */
	UP_PROGRESS_ACTION_FINISHED	/* device is finished */
} UP_PROGRESS_ACTION;


/*
 * Structure containing current progress from the PipeProgress device.
 * This is either a user-specified structure which we update, or else
 * our own internal structure.  Changes to this information can result
 * in a call-back if requested.  The lastAction value indicates why
 * the progress was updated.  The file-specific data remains valid
 * until the next file name is seen.  The low and high count values
 * are the binary top and bottom halfs of the actual value.  For
 * example, if a long was 32 bits then the actual byte count would
 * be calculated as (highByteCount * 2^32) + lowByteCount.
 */
typedef	struct
{
	UP *		up;		/* owning pipeline */
	UP_ID		id;		/* PipeProgress device id */
	UP_PROGRESS_ACTION lastAction;	/* reason progress is updated */
	unsigned long	lowChangeCount;	/* count of changes to info */
	unsigned long	highChangeCount;
	const char *	fileName;	/* current file name */
	const UP_STAT *	fileStatus;	/* current file status */
	off_t		fileOffset;	/* byte count for current file */
	long		fileCount;	/* total files transferred */
	long		fileErrors;	/* files which had errors */
	unsigned long	highByteCount;	/* total byte count for all files */
	unsigned long	lowByteCount;
	void *		userData;	/* user-specified data */
} UP_PROGRESS;


/*
 * The user's callback function which provides progress updates.
 */
typedef	void	(*UP_PROGRESS_FUNC)(const UP_PROGRESS *);


/*
 * Structure used by the WriteString device which holds information
 * about the string being written.  This structure is provided by the
 * caller when the device is added, and must remain allocated while the
 * pipeline is running.  The data in the structure is initialized and
 * updated by the device.  The buffer data is allocated by the device but
 * is never freed by the device.  The caller is responsible for freeing
 * the buffer after the pipeline is finished.  The byteCount value
 * indicates the amount of data that is stored in the buffer.  The
 * hasFinished flag is set when the last bit of data is written.  The
 * hasError flag is set if there was an error (such as the maximum buffer
 * size being exceeded).  The structure is only valid if upAddWriteString
 * returns UP_TRUE.
 */
typedef struct
{
	char *	buffer;		/* allocated buffer containing data */
	int	bufferSize;	/* allocated buffer size */
	int	maxBufferSize;	/* maximum allowed buffer size */
	int	byteCount;	/* count of bytes stored in buffer */
	UP_BOOL	hasFinished;	/* buffer is finished with no error */
	UP_BOOL	hasError;	/* there was an error writing the buffer */
} UP_WRITE_STRING;


/*
 * Error codes and error variable.
 * These must be allocated sequentually from 0 since the error strings
 * are stored in a static table indexed by these values.
 */
#define	UP_ERR_NONE		0	/* no error */
#define	UP_ERR_SYSTEM		1	/* system call failed */
#define	UP_ERR_NOMEM		2	/* cannot allocate memory */
#define	UP_ERR_INTERNAL		3	/* internal consistency check failed */
#define	UP_ERR_NOTSETUP		4	/* pipeline has not yet been set up */
#define	UP_ERR_ALREADYSETUP	5	/* pipeline has already been set up */
#define	UP_ERR_NODEVICES	6	/* no devices are in pipeline */
#define	UP_ERR_NOTREADER	7	/* reading end of pipe is not reader */
#define	UP_ERR_NOTWRITER	8	/* writing end of pipe is not writer */
#define	UP_ERR_BADENDPOINT	9	/* endpoint is wrong for operation */
#define	UP_ERR_INTERNALREADER	10	/* pipeline has an internal reader */
#define	UP_ERR_INTERNALWRITER	11	/* pipeline has an internal writer */
#define	UP_ERR_BADCOUNT		12	/* negative read or write count */
#define	UP_ERR_BADID		13	/* bad or unknown device id */
#define	UP_ERR_NOTARCHIVE	14	/* data is not an archive file */
#define	UP_ERR_BADHEADER	15	/* bad header */
#define	UP_ERR_SHORTFILE	16	/* premature end of file reading file */
#define	UP_ERR_BADSIZE		17	/* bad buffer size */
#define	UP_ERR_EXITERROR	18	/* process exited with non-zero status */
#define	UP_ERR_EXITSIGNAL	19	/* process terminated by a signal */
#define	UP_ERR_EXITUNKNOWN	20	/* process exited for unknown reason */
#define	UP_ERR_NULLCOMMAND	21	/* process command line is null */
#define	UP_ERR_BADGZIPLEVEL	22	/* illegal gzip compression level */
#define	UP_ERR_BADCOMPRESSBITS	23	/* illegal bits value for compress */
#define	UP_ERR_BUFFERFULL	24	/* string buffer is full */
#define	UP_ERR_UNKNOWNDEVID	25	/* unknown device id specified */
#define	UP_ERR_WRONGDEVTYPE	26	/* wrong device type for operation */
#define	UP_ERR_MISSINGENDLINE	27	/* missing end line for uudecode file */
#define	UP_ERR_ZLIBFAILED	28	/* zlib library call failed */
#define	UP_ERR_NOTGZIP		29	/* data is not a gzip file */
#define	UP_ERR_GZIPRESERVED	30	/* gzip file reserved header flags set */
#define	UP_ERR_FILENOTFOUND	31	/* file name was not found */
#define	UP_ERR_MULTIMISMATCH	32	/* multi output connected to single input */
#define	UP_ERR_DUPLICATEBEGINLINE 33	/* duplicate begin line for uudecode file */
#define	UP_ERR_BUFFER		34	/* buffer internal error */
#define	UP_ERR_LONGFILE		35	/* file data is longer than expected */
#define	UP_ERR_NULLPOINTER	36	/* null pointer given */
#define	UP_ERR_BADFILEDESCRIPTOR 37	/* bad file descriptor */
#define	UP_ERR_BADPATH		38	/* invalid PATH value */
#define	UP_ERR_NOTDIR		39	/* path is not a directory */
#define	UP_ERR_TOOMUCHDATA	40	/* too much data was supplied */
#define	UP_ERR_PARTIALLASTLINE	41	/* last line is missing terminator */
#define	UP_ERR_BADINDEX		42	/* bad index value */
#define	UP_ERR_BADITEM		43	/* bad item type */
#define	UP_ERR_NODATA		44	/* file has no data */
#define	UP_ERR_LINETOOLONG	45	/* line is too long */
#define	UP_ERR_BADPATTERN	46	/* bad pattern string */

#define	UP_ERR_ERRORCOUNT	47	/* not an error - count of errors */


/*
 * Debugging flags.
 * These are set using upSetDebugFlags and cause internal information
 * about the running of the pipeline to be printed to stderr.
 * These are OR'd together.
 */
#define	UP_DEBUG_FLAG_NONE	((UP_DEBUG_FLAG) 0x00)
#define	UP_DEBUG_FLAG_POLL	((UP_DEBUG_FLAG) 0x01)
#define	UP_DEBUG_FLAG_LOOP	((UP_DEBUG_FLAG) 0x02)
#define	UP_DEBUG_FLAG_STATE	((UP_DEBUG_FLAG) 0x04)
#define	UP_DEBUG_FLAG_ALL	((UP_DEBUG_FLAG) -1)


/*
 * Non-device specific public procedures.
 */
extern	UP *		upAllocate(void);
extern	UP_BOOL		upSetupDone(UP *);
extern	UP_BOOL		upClose(UP *);
extern	void		upFree(UP *);
extern	int		upGetDeviceCount(const UP * up);
extern	UP_BOOL		upPollList(UP *, struct pollfd **, int *, int *);
extern	int		upPoll(UP *, int timeout);
extern	int		upReadWrite(UP *, int timeout);
extern	int		upRead(UP *, char *, int);
extern	int		upWrite(UP *, const char *, int);
extern	UP_BOOL		upFlush(UP *);
extern	UP_BOOL		upCanRead(UP *);
extern	UP_BOOL		upCanWrite(UP *);
extern	UP_BOOL		upIsFinished(const UP *);
extern	UP_BOOL		upSetSystemPath(const char *);
extern	const char *	upGetSystemPath(void);
extern	UP_BOOL		upSetTemporaryDirectoryPath(const char *);
extern	const char *	upGetTemporaryDirectoryPath(void);


/*
 * Device selection routines.
 * These are endpoint devices which must be at the appropriate reading
 * or writing ends of the pipeline.
 */
extern	UP_ID	upAddReadEndpoint(UP *);
extern	UP_ID	upAddWriteEndpoint(UP *);
extern	UP_ID	upAddReadJoin(UP *, UP *);
extern	UP_ID	upAddWriteJoin(UP *, UP *);
extern	UP_ID	upAddReadNull(UP *);
extern	UP_ID	upAddWriteNull(UP *);
extern	UP_ID	upAddReadFd(UP *, int, UP_FLAG);
extern	UP_ID	upAddWriteFd(UP *, int, UP_FLAG);
extern	UP_ID	upAddReadDirectory(UP *, const char *, int, UP_FLAG);
extern	UP_ID	upAddWriteDirectory(UP *, const char *, UP_FLAG);
extern	UP_ID	upAddReadString(UP *, char *, int, UP_FLAG);
extern	UP_ID	upAddWriteString(UP *, UP_WRITE_STRING *, int);
extern	UP_ID	upAddReadFile(UP *, const char *, UP_FLAG);
extern	UP_ID	upAddReadFileList(UP *, const char **, int, UP_FLAG);
extern	UP_ID	upAddWriteFile(UP *, const char *, UP_FLAG);


/*
 * Device selection routines.
 * These are pass-through or "filter" devices which must not be at
 * the reading or writing ends of the pipeline.
 */
extern	UP_ID	upAddProcessString(UP *, const char *, UP_FLAG, const int *);

extern	UP_ID	upAddProcessArgs(UP *, const char *, int, const char **,
		UP_FLAG, const int *);

extern	UP_ID	upAddProgress(UP *, UP_PROGRESS *, UP_PROGRESS_FUNC, void *);
extern	UP_ID	upAddLogFile(UP *, const char *, UP_FLAG);
extern	UP_ID	upAddCat(UP *, const char *, UP_FLAG);
extern	UP_ID	upAddRescan(UP *, int);
extern	UP_ID	upAddGnuZip(UP *, int);
extern	UP_ID	upAddGnuUnzip(UP *);
extern	UP_ID	upAddCompress(UP *, int);
extern	UP_ID	upAddUnCompress(UP *);
extern	UP_ID	upAddGrep(UP *, const char *, UP_FLAG);
extern	UP_ID	upAddUUDecode(UP *, const char *);
extern	UP_ID	upAddUUEncode(UP *);
extern	UP_ID	upAddArchive(UP *, const char *);
extern	UP_ID	upAddUnArchive(UP *);
extern	UP_ID	upAddTar(UP *, const char *);
extern	UP_ID	upAddUnTar(UP *, const char *);
extern	UP_ID	upAddList(UP *, const char *, UP_FLAG);
extern	UP_ID	upAddFindSize(UP *, int, UP_FLAG);
extern	UP_ID	upAddRemoveData(UP *);
extern	UP_ID	upAddFilter(UP *, const UP_FILTER_ITEM *, int);
extern	UP_ID	upAddFindFileData(UP *, UP_FLAG);
extern	UP_ID	upAddMakeFileList(UP *, UP_FLAG);


/*
 * Other device routines.
 */
extern	int	upChangeJoin(UP *, UP_ID, UP *, UP **);


/*
 * Public error handling and related routines.
 */
extern	int		upGetErrorCode(const UP * up);
extern	const char *	upGetErrorString(const UP *);
extern	void		upPrintError(const UP *, const char *, ...);
extern	void		upClearError(UP * up);
extern	void		upGetVersion(int * major, int * minor);
extern	void		upSetMaximumWarnings(UP *, long);
extern	long		upGetMaximumWarnings(const UP *);
extern	long		upGetWarningCount(const UP *);
extern	void		upSetDebugFlags(UP *, UP_DEBUG_FLAG);

#endif
