/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * Buffer handling routines.
 * These provide a buffer which can be written to and read from,
 * and which provides an end of file capability with multiple files.
 *
 * Tag values and lengths for file names, stat information, file data,
 * and EOF indicators are stored in a fixed size circular table of items.
 * Data associated with items are stored in an allocated data buffer.
 * Indexes into the item table and the data buffer are maintained for
 * both the writer and reader.  The writer must block if there is no more
 * room for an item or data.  The reader must block if there is no item
 * or data to be read.
 *
 * The file data item is special in that it can be updated multiple times
 * by the writer and reader.  This allows an arbitrarily large amount of
 * data to be passed through the buffer using just one item.
 *
 * The data in the buffer is always contiguous.  When the end of the buffer
 * is reached for writing, existing data in the buffer is shifted up to the
 * top of the buffer to create the maximum amount of room for new data.
 */

#include <stdio.h>
#include <memory.h>

#include "upBuffer.h"
#include "upPrivate.h"


/*
 * Produce the next item offset in a buffer given the specified offset.
 * The items are stored in a circular list which wraps around.
 */
#define	NEXT_ITEM_OFFSET(offset) \
	(((offset) >= UP_MAX_ITEMS + UP_SPARE_ITEMS - 1) ? 0 : (offset) + 1)


/*
 * Produce the previous item offset in a buffer given the specified offset.
 * The items are stored in a circular list which wraps around.
 */
#define	PREVIOUS_ITEM_OFFSET(offset) \
	(((offset) > 0) ? ((offset) - 1) : UP_MAX_ITEMS + UP_SPARE_ITEMS - 1)



/*
 * Private routines.
 */
static	UP_BOOL		upBufferCheckWriteItem(const UP_BUF * bp,
				UP_ITEM_TYPE type);

static	void		upBufferError(UP_BUF * bp, const char * errorText);

static	UP_BOOL		upBufferWriteItem(UP_BUF * bp, UP_ITEM_TYPE type,
				const void * data, int length);

static	UP_BOOL		upBufferSaveFileName(UP_BUF * bp,
				const char * fileName);

static	void		upBufferPack(UP_BUF * bp);

static	void		upBufferReadConsumeItem(UP_BUF * bp,
				UP_ITEM_TYPE type);

static	void		upBufferConsumeEmptyTextItems(UP_BUF * bp);

static	UP_ITEM *	upBufferGetReadItem(UP_BUF * bp, UP_ITEM_TYPE type);

static	UP_ITEM *	upBufferGetLastFileDataItem(UP_BUF * bp);


/*
 * ----------------
 * GENERIC ROUTINES
 * ----------------
 */


/*
 * Allocate a new buffer which can contain the specified amount of data.
 * The link count for the buffer is set to 1.
 * Returns a pointer to the allocated buffer, or NULL on failure.
 */
UP_BUF *
upBufferAlloc(int size)
{
	UP_BUF *	bp;

	/*
	 * Validate the size of the buffer.
	 * It must be at least as large as a file name can be since
	 * we store file names into the buffer.
	 */
	if ((size < UP_BUFFER_MIN_SIZE) || (size > UP_BUFFER_MAX_SIZE))
		return NULL;

	/*
	 * Allocate the structure.
	 */
	bp = (UP_BUF *) malloc(sizeof(UP_BUF));

	if (bp == NULL)
		return NULL;

	/*
	 * Clear out all of its fields.
	 */
	memset((char *) bp, 0, sizeof(UP_BUF));

	/*
	 * Allocate the data buffer.
	 */
	bp->buffer = malloc(size);

	if (bp->buffer == NULL)
	{
		free((char *) bp);

		return NULL;
	}

	/*
	 * Initialise the other fields.
	 */
	bp->errorText = NULL;
	bp->linkCount = 1;
	bp->bufferSize = size;
	bp->lastWriteItemType = UP_ITEM_TYPE_NONE;
	bp->fileIsValid = UP_FALSE;
	bp->maxFileNameLength = UP_BUFFER_MIN_FILE_SIZE;

	/*
	 * Point the file name at the local file name buffer.
	 * If the file name grows too large later, a special buffer will
	 * be allocated to hold it.
	 */
	bp->fileNameAlloc = sizeof(bp->fileNameBuffer);
	bp->fileName = bp->fileNameBuffer;

	/*
	 * Return the new buffer.
	 */
	return bp;
}


/*
 * Link to the specified buffer.
 * This does nothing but increment the link counter and return the same
 * buffer pointer.  In this way, the buffer won't actually be freed until each
 * pointer has been freed.  This allows a reader and a writer to both have a
 * pointer to the buffer without giving either one exclusive ownership.
 */
UP_BUF *
upBufferLink(UP_BUF * bp)
{
	if (bp)
		bp->linkCount++;

	return bp;
}


/*
 * Free the allocated buffer.  The link counter is decremented, and if it
 * reaches zero, the buffer is freed.  This does nothing if a NULL pointer
 * is given.
 */
void
upBufferFree(UP_BUF * bp)
{
	int	i;

	/*
	 * Check for null or stale pointers.
	 */
	if ((bp == NULL) || (bp->linkCount <= 0))
		return;

	/*
	 * Decrement the link counter, and if it still nonzero then that's
	 * that's all we do for now since someone else is still using it.
	 */
	bp->linkCount--;

	if (bp->linkCount != 0)
		return;

	/*
	 * This is the last use of the buffer.
	 * Free any allocated memory and then ourself.
	 */
	free(bp->buffer);

	for (i = 0; i < UP_MAX_ITEMS + UP_SPARE_ITEMS; i++)
		free(bp->itemTable[i].data);

	if (bp->fileName != bp->fileNameBuffer)
		free(bp->fileName);

	free((char *) bp);
}


/*
 * Totally reset the buffer so that it contains no data.
 * After this the buffer is just as it was after being newly allocated.
 */
void
upBufferReset(UP_BUF * bp)
{
	bp->errorText = NULL;
	bp->readOffset = 0;
	bp->writeOffset = 0;
	bp->readItemOffset = 0;
	bp->writeItemOffset = 0;
	bp->itemCount = 0;
	bp->lastWriteItemType = UP_ITEM_TYPE_NONE;
	bp->maxFileNameLength = UP_BUFFER_MIN_FILE_SIZE;
	bp->fileIsValid = UP_FALSE;
	bp->lineLength = 0;
	bp->byteCount = 0;
	bp->totalFiles = 0;
	bp->totalErrors = 0;
	bp->totalBytes = 0;
	bp->fileName[0] = '\0';

	memset(&bp->fileStatus, 0, sizeof(UP_STAT));
}


/*
 * Set a fatal error for the buffer by remembering the specified error text.
 * The error text must be a constant string that is permanently allocated.
 * A fatal error will prevent any further use of the buffer, and will report
 * the fatal error to the reader and writer at every opportunity.
 * Only the first fatal error for the buffer is saved.
 */
static void
upBufferError(UP_BUF * bp, const char * textString)
{
	if (textString == NULL)
		textString = "Unknown error";

	if (bp->errorText == NULL)
		bp->errorText = textString;

	fprintf(stderr, "Setting buffer error: %s\n", textString);
}


/*
 * Return whether or not there has been a fatal error using the buffer.
 * Except for memory allocation errors, fatal errors are only caused by
 * the client by providing an illegal argument or failing to follow the
 * proper protocol in using the buffer.
 */
UP_BOOL
upBufferHasError(const UP_BUF * bp)
{
	return (bp->errorText != NULL);
}


/*
 * Return a text string describing a possible fatal error in the use of
 * the buffer.  An empty string is returned if there is no error.
 */
const char *
upBufferGetErrorText(const UP_BUF * bp)
{
	if (bp->errorText == NULL)
		return "";

	return bp->errorText;
}


/*
 * Return whether or not the buffer is totally finished.
 * This means that the end of buffer flag has been written and then
 * been consumed by the reader.
 */
UP_BOOL
upBufferIsFinished(const UP_BUF * bp)
{
	return ((bp->lastWriteItemType == UP_ITEM_TYPE_END_OF_BUFFER) &&
		(bp->itemCount == 0));
}


/*
 * -----------------------
 * BUFFER WRITING ROUTINES
 * -----------------------
 */


/*
 * Return the next item type that could be written to the buffer.
 * This routine can be called at any time without causing an error.
 * The following return values are possible:
 *
 * UP_ITEM_TYPE_FILE_NAME
 *	A file name can be written for a new file now, or an end of
 *	buffer can be written now.
 * UP_ITEM_TYPE_END_OF_BUFFER
 *	There is no room to write a new file name now (but maybe
 *	later), or an end of buffer can be written now.
 * UP_ITEM_TYPE_FILE_DATA
 *	There is room to write some file data now, an end of file
 *	now, or a file error now.
 * UP_ITEM_TYPE_END_OF_FILE
 *	There is no room to write new file data now (but maybe later),
 *	or an end of file or file error can be written now.
 * UP_ITEM_TYPE_NONE
 *	The buffer has been marked finished and cannot be used anymore.
 * UP_ITEM_TYPE_FATAL_ERROR
 *	The buffer has had a fatal error and cannot be used anymore.
 */
UP_ITEM_TYPE
upBufferGetWriteItemType(const UP_BUF * bp)
{
	UP_ITEM_TYPE	nextItem;
	int		availableBytes;

	/*
	 * If there had been a fatal error then always return that fact.
	 */
	if (bp->errorText)
		return UP_ITEM_TYPE_FATAL_ERROR;

	/*
	 * Get the number of available bytes in the buffer.
	 */
	availableBytes = bp->bufferSize - (bp->writeOffset - bp->readOffset);


	/*
	 * Get the item type which can now be written based on the item
	 * which had just been written and whether there is room for storing
	 * a desired type.
	 */
	switch (bp->lastWriteItemType)
	{
		case UP_ITEM_TYPE_FILE_NAME:
		case UP_ITEM_TYPE_FILE_STATUS:
		case UP_ITEM_TYPE_FILE_DATA:
			if (availableBytes > 0)
				nextItem = UP_ITEM_TYPE_FILE_DATA;
			else
				nextItem = UP_ITEM_TYPE_END_OF_FILE;
			break;

		case UP_ITEM_TYPE_NONE:
		case UP_ITEM_TYPE_FILE_ERROR:
		case UP_ITEM_TYPE_END_OF_FILE:
			/*
			 * Assume that the current maximum file name length
			 * and a status structure need to be stored.  If
			 * there is that much room left then say we can store
			 * the next file name.
			 */
			if (availableBytes <
				bp->maxFileNameLength + sizeof(UP_STAT))
			{
				nextItem = UP_ITEM_TYPE_END_OF_BUFFER;
				break;
			}

			/*
			 * We could store a file name unless there are no
			 * more items available.
			 */
			if (bp->itemCount < UP_MAX_ITEMS)
				nextItem = UP_ITEM_TYPE_FILE_NAME;
			else
				nextItem = UP_ITEM_TYPE_END_OF_BUFFER;
			break;

		default:
			nextItem = UP_ITEM_TYPE_NONE;
			break;
	}

	return nextItem;
}


/*
 * Request that the desired amount of contiguous space be made available for
 * writing file data into the buffer.  If the requested amount of space is
 * not immediately available, then the data in the buffer will be packed to
 * create the maximum amount of space.  Returns the amount of available
 * space available after possible packing (limited by requestLength).
 * The returned length may be less than requestLength (zero if there is no
 * room at all in the buffer for new data).  Returns -1 on a fatal error
 * which includes calling this routine when writing file data is not valid.
 * The address and length for actual writing can be obtained by calling
 * the upBufferGetWriteAddress and upBufferGetWriteLength functions.
 */
int
upBufferRequestSpace(UP_BUF * bp, int requestLength)
{
	UP_ITEM *	item;
	int		availableLength;

	/*
	 * Make sure the request length is valid.
	 */
	if (requestLength < 0)
	{
		upBufferError(bp, "Bad length for upBufferRequestSpace");

		return -1;
	}

	/*
	 * Make sure that we can write data to the buffer.
	 */
	if (!upBufferCheckWriteItem(bp, UP_ITEM_TYPE_FILE_DATA))
	{
		upBufferError(bp,
			"Called upBufferRequestSpace when not writable");

		return -1;
	}

	/*
	 * If there is no data in the buffer then set the read and write
	 * offsets back to the top to guarantee the maximum amount of room.
	 */
	if (bp->readOffset == bp->writeOffset)
	{
		bp->readOffset = 0;
		bp->writeOffset = 0;
	}

	/*
	 * We need a file data item to reference the allocated space.
	 * See if there is already one there in the item table.
	 */
	if (bp->lastWriteItemType != UP_ITEM_TYPE_FILE_DATA)
	{
		/*
		 * We need to allocate a new file data item to the item table.
		 * If there are no more items available, then return zero to
		 * indicate nothing can be written now.
		 */
		if (bp->itemCount >= UP_MAX_ITEMS)
			return 0;

		/*
		 * Allocate a file data item having no data yet.
		 * The data will be filled in later by the caller.
		 * (If no data is ever stored before the end of file or a
		 * file error item is added, then this item will be deleted.)
		 */
		item = &bp->itemTable[bp->writeItemOffset];

		item->type = UP_ITEM_TYPE_FILE_DATA;
		item->length = 0;
		item->data = NULL;

		bp->writeItemOffset = NEXT_ITEM_OFFSET(bp->writeItemOffset);
		bp->lastWriteItemType = UP_ITEM_TYPE_FILE_DATA;
		bp->itemCount++;
	}

	/*
	 * If the amount of buffer space left after the write offset is
	 * as large as the request then return the requested length.
	 */
	if (requestLength <= bp->bufferSize - bp->writeOffset)
		return requestLength;

	/*
	 * The requested length is larger than the available space after the
	 * write offset.  Try to pack the data in the buffer so maximize the
	 * available space.
	 */
	upBufferPack(bp);

	/*
	 * Now calculate the amount of available space and limit it by the
	 * requested length.
	 */
	availableLength = bp->bufferSize - bp->writeOffset;

	if (availableLength > requestLength)
		availableLength = requestLength;

	/*
	 * Return the amount of available space.
	 */
	return availableLength;
}


/*
 * Return the address of the largest contiguous amount of data that can
 * be written into the buffer after possibly packing the existing data.
 * The buffer is only packed if there is no more room after the write offset.
 * The amount of space actually available for writing is returned by calling
 * upBuffergetWriteLength.  For best results the upBufferRequestSpace function
 * should be used before this call to maximize the amount of data that can be
 * written to the buffer.  A valid pointer is returned even if no space is
 * available.  The returned address is only valid until the next call to any
 * routine which reads, writes, or reserves data in the buffer.
 * This should only be called when writing data is legal.
 */
char *
upBufferGetWriteAddress(UP_BUF * bp)
{
	/*
	 * Request space for one byte of data.
	 * Usually there will be lots more room than that.
	 */
	(void) upBufferRequestSpace(bp, 1);

	/*
	 * Return the address of the available space.
	 */
	return bp->buffer + bp->writeOffset;
}


/*
 * Return the size of the largest contiguous amount of data that can be
 * written into the buffer after possibly packing the existing data.  The
 * buffer is only packed if there is no more room after the write offset.
 * For best results the upBufferRequestSpace function should be used before
 * this call to maximize the amount of data that can be written to the buffer.
 * This should only be called when writing data is legal.
 * Returns zero if no data can be written.
 */
int
upBufferGetWriteLength(UP_BUF * bp)
{
	/*
	 * Request space for one byte of data.
	 * Usually there will be lots more room than that.
	 * If this fails then indicate that there is no room at all.
	 */
	if (upBufferRequestSpace(bp, 1) <= 0)
		return 0;

	/*
	 * Return the amount of available data.
	 */
	return bp->bufferSize - bp->writeOffset;
}


/*
 * Adjust the write address in the buffer to take into account data that
 * had been stored there after calls to the upBufferGetWriteAddress and
 * upBufferGetWriteLength routines.  The specified value is the amount of
 * data actually stored.  This can only be called when writing data is legal.
 */
void
upBufferWriteAdjust(UP_BUF * bp, int size)
{
	UP_ITEM *	item;

	/*
	 * Make sure that we had just written a file data item.
	 */
	if (bp->lastWriteItemType != UP_ITEM_TYPE_FILE_DATA)
	{
		upBufferError(bp,
			"Called upBufferWriteAdjust when not writing data");

		return;
	}

	/*
	 * Make sure that the size adjustment isn't too small or too large.
	 */
	if ((size < 0) || (size > bp->bufferSize - bp->writeOffset))
	{
		upBufferError(bp, "Illegal upBufferWriteAdjust size");

		return;
	}

	/*
	 * Get the last file data item that was last used.
	 */
	item = upBufferGetLastFileDataItem(bp);

	if (item == NULL)
		return;

	/*
	 * Add the adjusted amount to the length of the file data item.
	 */
	item->length += size;

	/*
	 * Adjust the write offset to make more data available for reading
	 * and do some accounting.  Hopefully the caller has actually stored
	 * data into the buffer before calling this routine!
	 */
	bp->writeOffset += size;
	bp->byteCount += size;
	bp->totalBytes += size;
}


/*
 * Set the file name and file status for the current file that we are
 * starting to write to.  The supplied status information can be NULL
 * if there is none to set, in which case a minimal status structure
 * is set which contains a few default values.  The supplied file name
 * and status are copied and so do not need to be preserved.
 * This routine should only be called when it is known that a file
 * name can be stored.  Returns FALSE if the file name could not be
 * saved with a fatal error set.
 */
UP_BOOL
upBufferWriteFileName(
	UP_BUF *	bp,
	const char *	fileName,
	const UP_STAT *	status
)
{
	int	fileNameLength;
	UP_STAT	defaultStatus;

	/*
	 * Make sure we are allowed to store a file name now.
	 */
	if (!upBufferCheckWriteItem(bp, UP_ITEM_TYPE_FILE_NAME))
	{
		upBufferError(bp, "Invalid state for writing file name");

		return UP_FALSE;
	}

	/*
	 * Make sure that there is an available item.
	 */
	if (bp->itemCount >= UP_MAX_ITEMS)
	{
		upBufferError(bp, "No available item for writing file name");

		return UP_FALSE;
	}

	/*
	 * Get the length of the file name with its terminating null.
	 */
	fileNameLength = strlen(fileName) + 1;

	/*
	 * Store the file name into the buffer.
	 */
	if (!upBufferWriteItem(bp, UP_ITEM_TYPE_FILE_NAME, fileName,
		fileNameLength))
	{
		return UP_FALSE;
	}

	/*
	 * If no status structure was supplied then make a default one.
	 * This is for a normal regular file which is readable and
	 * writable by everyone.
	 */
	if (status == NULL)
	{
		memset(&defaultStatus, 0, sizeof(UP_STAT));

		defaultStatus.flags =
			(UP_STAT_TYPE | UP_STAT_MODE | UP_STAT_DATA);

		defaultStatus.type = '-';
		defaultStatus.mode = 0666 | S_IFREG;

		status = &defaultStatus;
	}

	/*
	 * Store the status structure.
	 */
	if (!upBufferWriteItem(bp, UP_ITEM_TYPE_FILE_STATUS,
			status, sizeof(UP_STAT)))
	{
		return UP_FALSE;
	}

	/*
	 * For future efficiency, if this file name length is larger than the
	 * maximum file name length so far, then increase the maximum length
	 * to help avoid allocating memory for further file names.
	 */
	if ((fileNameLength <= bp->bufferSize) &&
		(bp->maxFileNameLength < fileNameLength))
	{
		bp->maxFileNameLength = fileNameLength;
	}

	/*
	 * All done.
	 */
	return UP_TRUE;
}


/*
 * Write an item of the specified type with the associated data into the
 * buffer while validating that it is legal and that an item is available.
 * If the buffer doesn't have room to store the data then it will be
 * allocated specially and stored in the item.  This routine cannot be
 * called with the UP_ITEM_TYPE_FILE DATA item type since that requires
 * special handling.  Returns UP_TRUE if the item was successfully stored,
 * or UP_FALSE if there is some error.
 */
static UP_BOOL
upBufferWriteItem(
	UP_BUF *	bp,
	UP_ITEM_TYPE	type,
	const void *	data,
	int		length
)
{
	UP_ITEM *	item;
	char *		allocData;
	int		minLength;
	int		maxLength;
	int		maxItems;

	/*
	 * Make sure that it is legal to store the new item.
	 */
	if (!upBufferCheckWriteItem(bp, type))
	{
		upBufferError(bp, "Invalid state to write buffer item");

		return UP_FALSE;
	}

	/*
	 * Set some default limits until we know otherwise.
	 */
	minLength = 0;
	maxLength = bp->bufferSize;
	maxItems = UP_MAX_ITEMS;
	allocData = NULL;

	/*
	 * Switch on the type of item to update the limits.
	 */
	switch (type)
	{
		case UP_ITEM_TYPE_END_OF_FILE:
		case UP_ITEM_TYPE_FILE_ERROR:
		case UP_ITEM_TYPE_END_OF_BUFFER:
			/*
			 * These items have no data and can use the
			 * reserved items in the buffer.
			 */
			maxItems = UP_MAX_ITEMS + UP_SPARE_ITEMS;
			maxLength = 0;

			break;

		case UP_ITEM_TYPE_FILE_NAME:
			minLength = 1;

			break;

		case UP_ITEM_TYPE_FILE_STATUS:
			minLength = sizeof(UP_STAT);
			maxLength = sizeof(UP_STAT);
			maxItems = UP_MAX_ITEMS + 1;
			break;

		case UP_ITEM_TYPE_FILE_DATA:
			upBufferError(bp,
				"Called upBufferWriteItem for file data");

			return UP_FALSE;

		default:
			upBufferError(bp,
				"Illegal item type for upBufferWriteItem");

			return UP_FALSE;
	}

	/*
	 * Verify that a free item is available.
	 */
	if (bp->itemCount >= maxItems)
	{
		upBufferError(bp, "No items available");

		return UP_FALSE;
	}

	/*
	 * Check that the data length is within bounds.
	 */
	if ((length < minLength) || (length > maxLength))
	{
		upBufferError(bp, "Bad length in upBufferWriteItem");

		return UP_FALSE;
	}

	/*
	 * Check that the data length and pointer are consistent.
	 */
	if ((length > 0) && (data == NULL))
	{
		upBufferError(bp, "Null address in upBufferWriteItem");

		return UP_FALSE;
	}

	/*
	 * If there is no data in the buffer then reset the read and write
	 * offsets to maximize the available room.
	 */
	if (bp->readOffset == bp->writeOffset)
	{
		bp->readOffset = 0;
		bp->writeOffset = 0;
	}

	/*
	 * If the data will not fit into the buffer then allocate some
	 * memory specially for this item.  Otherwise pack the buffer
	 * if required to make room.
	 */
	if (length > bp->bufferSize - (bp->writeOffset - bp->readOffset))
	{
		allocData = malloc(length);

		if (allocData == NULL)
		{
			upBufferError(bp, "Cannot allocate memory for item");

			return UP_FALSE;
		}
	}
	else if (length > bp->bufferSize - bp->writeOffset)
		upBufferPack(bp);

	/*
	 * Check whether the previous item type was file data which had
	 * no actual data stored in it. If so we must overwrite it.
	 * (Such an item can come about in various ways.)  If we didn't
	 * consume it here, then the reader would never consume it since
	 * it had no data bytes to read, and so he would stall forever.
	 */
	if (bp->lastWriteItemType == UP_ITEM_TYPE_FILE_DATA)
	{
		/*
		 * Get the last file data item.
		 */
		item = upBufferGetLastFileDataItem(bp);

		if (item == NULL)
			return UP_FALSE;

		/*
		 * If the file data item contains no data then remove it.
		 * The item will immediately be overwritten below.
		 */
		if (item->length == 0)
		{
			bp->itemCount--;

			bp->writeItemOffset =
				PREVIOUS_ITEM_OFFSET(bp->writeItemOffset);
		}
	}

	/*
	 * Get the next item to be written and fill it in.
	 */
	item = &bp->itemTable[bp->writeItemOffset];

	item->type = type;
	item->length = length;
	item->data = allocData;

	/*
	 * Account for the new item and increment the writing item index to
	 * the next entry in the table (wrapping around if necessary).
	 */
	bp->writeItemOffset = NEXT_ITEM_OFFSET(bp->writeItemOffset);
	bp->lastWriteItemType = type;
	bp->itemCount++;

	/*
	 * Store the data too if there is any.
	 * This is stored either in the buffer or in our allocated memory.
	 */
	if (length > 0)
	{
		if (allocData)
			memcpy(allocData, data, length);
		else
		{
			memcpy(&bp->buffer[bp->writeOffset], data, length);

			bp->writeOffset += length;
		}
	}

	return UP_TRUE;
}


/*
 * Get the file data item that was last written to the buffer.
 * This should only be called when there is expected to be such an item.
 * Returns NULL if there is no such item with a fatal error generated.
 */
static UP_ITEM *
upBufferGetLastFileDataItem(UP_BUF * bp)
{
	UP_ITEM *	item;

	/*
	 * Make sure the last item written was file data.
	 */
	if (bp->lastWriteItemType != UP_ITEM_TYPE_FILE_DATA)
	{
		upBufferError(bp,
			"No file data in upBufferGetLastFileDataItem");

		return NULL;
	}

	/*
	 * Make sure the item count is non-zero.
	 */
	if (bp->itemCount <= 0)
	{
		upBufferError(bp, "No items when last item is file data");

		return NULL;
	}

	/*
	 * Get the item which was just written.
	 */
	item = &bp->itemTable[PREVIOUS_ITEM_OFFSET(bp->writeItemOffset)];

	/*
	 * Make sure the item is file data.
	 */
	if (item->type != UP_ITEM_TYPE_FILE_DATA)
	{
		upBufferError(bp, "Last file data item isn't file data");

		return NULL;
	}

	/*
	 * OK, return the item.
	 */
	return item;
}



/*
 * Macro to create a single numeric value from a pair of item types.
 */
#define	TWOVAL(from, to) \
	(((from) * (UP_ITEM_TYPE_FIRST_ILLEGAL + 1)) + (to))


/*
 * Check whether a change from the current writing item type to the new
 * specified writing item type is legal.  Returns UP_TRUE if so.  No error
 * is generated by this routine.
 */
static UP_BOOL
upBufferCheckWriteItem(const UP_BUF * bp, UP_ITEM_TYPE type)
{
	/*
	 * Check the range of the type.
	 */
	if ((type < 0) || (type >= UP_ITEM_TYPE_FIRST_ILLEGAL))
		return UP_FALSE;

	/*
	 * The type itself is legal.
	 * Check whether the change from the current item type to this new
	 * one is legal.  This depends on old and new pair of item types.
	 */
	switch (TWOVAL(bp->lastWriteItemType, type))
	{
	case TWOVAL(UP_ITEM_TYPE_NONE, UP_ITEM_TYPE_FILE_NAME):
	case TWOVAL(UP_ITEM_TYPE_NONE, UP_ITEM_TYPE_END_OF_BUFFER):
	case TWOVAL(UP_ITEM_TYPE_FILE_NAME, UP_ITEM_TYPE_FILE_STATUS):
	case TWOVAL(UP_ITEM_TYPE_FILE_NAME, UP_ITEM_TYPE_FILE_DATA):
	case TWOVAL(UP_ITEM_TYPE_FILE_NAME, UP_ITEM_TYPE_END_OF_FILE):
	case TWOVAL(UP_ITEM_TYPE_FILE_NAME, UP_ITEM_TYPE_FILE_ERROR):
	case TWOVAL(UP_ITEM_TYPE_FILE_STATUS, UP_ITEM_TYPE_FILE_DATA):
	case TWOVAL(UP_ITEM_TYPE_FILE_STATUS, UP_ITEM_TYPE_END_OF_FILE):
	case TWOVAL(UP_ITEM_TYPE_FILE_STATUS, UP_ITEM_TYPE_FILE_ERROR):
	case TWOVAL(UP_ITEM_TYPE_FILE_DATA, UP_ITEM_TYPE_FILE_DATA):
	case TWOVAL(UP_ITEM_TYPE_FILE_DATA, UP_ITEM_TYPE_END_OF_FILE):
	case TWOVAL(UP_ITEM_TYPE_FILE_DATA, UP_ITEM_TYPE_FILE_ERROR):
	case TWOVAL(UP_ITEM_TYPE_END_OF_FILE, UP_ITEM_TYPE_FILE_NAME):
	case TWOVAL(UP_ITEM_TYPE_END_OF_FILE, UP_ITEM_TYPE_END_OF_BUFFER):
	case TWOVAL(UP_ITEM_TYPE_FILE_ERROR, UP_ITEM_TYPE_FILE_NAME):
	case TWOVAL(UP_ITEM_TYPE_FILE_ERROR, UP_ITEM_TYPE_END_OF_BUFFER):
		break;

	default:
		/*
		 * The item type change is not legal.
		 */
		return UP_FALSE;
	}

	/*
	 * The item is legal.
	 */
	return UP_TRUE;
}


/*
 * Write the end of file item for the file currently being written.
 * Calling this routine at the incorrect time sets a fatal error for
 * the buffer.  There is no need to wait for buffer space when calling
 * this routine since the item table has a reserved slot for this
 * item type.
 */
void
upBufferWriteEof(UP_BUF * bp)
{
	(void) upBufferWriteItem(bp, UP_ITEM_TYPE_END_OF_FILE, NULL, 0);
}


/*
 * Write a file error item for the file currently being written.
 * Calling this routine at the incorrect time sets a fatal error for
 * the buffer.  There is no need to wait for buffer space when calling
 * this routine since the item table has a reserved slot for this
 * item type.
 */
void
upBufferWriteFileError(UP_BUF * bp)
{
	(void) upBufferWriteItem(bp, UP_ITEM_TYPE_FILE_ERROR, NULL, 0);
}


/*
 * Write the end of buffer item for the buffer.  Calling this routine at
 * the incorrect time sets a fatal error for the buffer.  There is no need
 * to wait for buffer space when calling this routine since the item
 * table has a reserved slot for this item type.
 */
void
upBufferWriteEndOfBuffer(UP_BUF * bp)
{
	(void) upBufferWriteItem(bp, UP_ITEM_TYPE_END_OF_BUFFER, NULL, 0);
}


/*
 * Write the specified data to the buffer, making it available to be read.
 * Only as much data is written as can fit into the available space in the
 * buffer.  Returns the number of bytes stored, which may be zero if the
 * buffer is full.  Returns -1 on an error.
 */
int
upBufferWrite(UP_BUF * bp, const char * buf, int length)
{
	char *	dataAddress;
	int	dataLength;

	/*
	 * Validate the arguments.
	 */
	if ((buf == NULL) || (length < 0))
	{
		upBufferError(bp,
			"Calling upBufferWrite with bad argument");

		return -1;
	}

	/*
	 * If no data is being written then ignore the call.
	 */
	if (length == 0)
		return 0;

	/*
	 * Request as much contiguous space as we want to write.
	 * Then get the data address and length of the available space.
	 */
	if (upBufferRequestSpace(bp, length) < 0)
		return -1;

	dataAddress = upBufferGetWriteAddress(bp);
	dataLength = upBufferGetWriteLength(bp);

	/*
	 * Limit the amount of data to write to the minimum of the amount
	 * requested and the amount available.
	 */
	if (length > dataLength)
		length = dataLength;

	/*
	 * Store the data into the buffer and account for it.
	 */
	memcpy(dataAddress, buf, length);

	upBufferWriteAdjust(bp, length);

	/*
	 * Return the amount actually written.
	 */
	return length;
}


/*
 * Pack any data left in the buffer so that the writer to the buffer has
 * the full amount of room available for new data.
 */
static void
upBufferPack(UP_BUF * bp)
{
	int	usedLength;

	/*
	 * If the data is already packed then there is nothing to do.
	 */
	if (bp->readOffset == 0)
		return;

	/*
	 * If the buffer has no data then set the read and write offsets
	 * back to the top to guarantee the maximium amount of room.
	 */
	if (bp->readOffset == bp->writeOffset)
	{
		bp->readOffset = 0;
		bp->writeOffset = 0;

		return;
	}

	/*
	 * There is some data in the buffer.
	 * Get the amount of data used in the buffer.
	 */
	usedLength = bp->writeOffset - bp->readOffset;

	/*
	 * Move the data up to the top of the buffer.
	 */
	memmove(bp->buffer, bp->buffer + bp->readOffset, usedLength);

	bp->readOffset = 0;
	bp->writeOffset = usedLength;
}


/*
 * -----------------------
 * BUFFER READING ROUTINES
 * -----------------------
 */


/*
 * Routine which returns UP_TRUE if there is an item which can be
 * read from the buffer.  This is just a convenience function which
 * checks whether upBufferGetReadItemType returns UP_ITEM_TYPE_NONE.
 */
UP_BOOL
upBufferIsReadAvailable(UP_BUF * bp)
{
	return (upBufferGetReadItemType(bp) != UP_ITEM_TYPE_NONE);
}


/*
 * Routine which returns the next item that can be read from the buffer.
 * Returns UP_ITEM_TYPE_NONE if there is nothing available for reading.
 * If the buffer has had a fatal error then UP_ITEM_TYPE_FATAL_ERROR will
 * be returned instead.  If any empty text items are found with a following
 * item then they are consumed.  This is the main routine for the reader
 * to call to determine what needs to be done next.
 */
UP_ITEM_TYPE
upBufferGetReadItemType(UP_BUF * bp)
{
	UP_ITEM *	item;

	if (bp->errorText)
		return UP_ITEM_TYPE_FATAL_ERROR;

	upBufferConsumeEmptyTextItems(bp);

	if (bp->itemCount == 0)
		return UP_ITEM_TYPE_NONE;

	item = &bp->itemTable[bp->readItemOffset];

	if ((item->type == UP_ITEM_TYPE_FILE_DATA) && (item->length == 0))
		return UP_ITEM_TYPE_NONE;

	return item->type;
}


/*
 * Return the address of the current item being read which must be of the
 * specified type (if it is not UP_ITEM_TYPE_NONE).  Returns NULL if there
 * is no item or if its type does not match the one specified.  This does
 * not generate an error.
 */
static UP_ITEM *
upBufferGetReadItem(UP_BUF * bp, UP_ITEM_TYPE type)
{
	UP_ITEM *	item;

	/*
	 * First consume any empty text item.
	 */
	upBufferConsumeEmptyTextItems(bp);

	/*
	 * Then see if the buffer contains no items.
	 */
	if (bp->itemCount <= 0)
		return NULL;

	/*
	 * There is at least one item stored in the buffer.
	 * Get the pointer to the one currently being read.
	 */
	item = &bp->itemTable[bp->readItemOffset];

	/*
	 * If a type is specified then see whether it is different than
	 * the item's type.
	 */
	if ((type != UP_ITEM_TYPE_NONE) && (type != item->type))
		return NULL;

	/*
	 * There is an item being read of the specified type.
	 * Return it.
	 */
	return item;
}


/*
 * Consume any empty text items which are followed by another item.
 * We can't consume the last text item since that is needed by the
 * writer to possibly store more file data.
 */
static void
upBufferConsumeEmptyTextItems(UP_BUF * bp)
{
	UP_ITEM *	item;

	/*
	 * Only loop while there is more than one item.
	 */
	while (bp->itemCount > 1)
	{
		item = &bp->itemTable[bp->readItemOffset];

		/*
		 * If this item is not a text item or it has some
		 * data stored in it, then we must leave it.
		 */
		if ((item->type != UP_ITEM_TYPE_FILE_DATA) ||
			(item->length > 0))
		{
			return;
		}

		/*
		 * Comsume this empty text item.
		 */
		bp->readItemOffset = NEXT_ITEM_OFFSET(bp->readItemOffset);
		bp->itemCount--;
	}
}


/*
 * Return the address of the largest contiguous amount of data that can
 * be read from the buffer.  An address is still returned if no data
 * is available.  Using upBufferGetReadLength can be used to get the
 * size of the buffer.  The returned buffer is only valid until more data
 * is read or written to this buffer.
 */
const char *
upBufferGetReadAddress(UP_BUF * bp)
{
	return bp->buffer + bp->readOffset;
}


/*
 * Return the size of the largest contiguous amount of file data that
 * can be read from the buffer.  Returns zero if no data is available.
 * This amount of data valid until data is read from the buffer.
 * This routine can be called at any time without any error.
 */
int
upBufferGetReadLength(UP_BUF * bp)
{
	UP_ITEM *	item;

	/*
	 * Get the item we are reading file data from.
	 */
	item = upBufferGetReadItem(bp, UP_ITEM_TYPE_FILE_DATA);

	/*
	 * Return zero if we aren't reading file data now.
	 */
	if (item == NULL)
		return 0;

	/*
	 * Return the amount of data contained in the data item.
	 */
	return item->length;
}


/*
 * Return whether or not there is the specified amount of data available
 * to be read from the buffer, or else an end of file or error is pending
 * and so that much data will never be available.  (The actual amount
 * available can be obtained using upBufferGetReadLength.)  Returns UP_TRUE
 * if the specified amount of data is available, or if the size is larger
 * than the amount the buffer can hold.
 */
UP_BOOL
upBufferReadDataCheck(UP_BUF * bp, int length)
{
	int	type;
	int	usedLength;

	/*
	 * Get the type of item that can be read.
	 */
	type = upBufferGetReadItemType(bp);

	/*
	 * If there is no item stored then nothing can be read now.
	 */
	if (type == UP_ITEM_TYPE_NONE)
		return UP_FALSE;

	/*
	 * If the item is not file data then it is readable.
	 */
	if (upBufferGetReadItemType(bp) != UP_ITEM_TYPE_FILE_DATA)
		return UP_TRUE;

	/*
	 * Get the amount of data stored in the data item.
	 */
	usedLength = bp->itemTable[bp->readItemOffset].length;

	/*
	 * If the requested amount of data is available, we are asking
	 * for more than the buffer can hold, or there is another item
	 * following this one, then return that data is readable.
	 */
	if ((length <= usedLength) || (length > bp->bufferSize) ||
		(bp->itemCount > 1))
	{
		return UP_TRUE;
	}

	/*
	 * The required amount of data isn't present.
	 */
	return UP_FALSE;
}


/*
 * Return the length of the first line of data that can be read from
 * the buffer.  This length includes the terminating newline character
 * if any.  Returns zero if a complete line is not available (even if a
 * partial line is present.)  However, if the buffer is completely full so
 * that a line cannot possibly fit, then -1 is returned to indicate that
 * the line is too long.  (Note that this might not be an error if the
 * caller knows how to handle the problem.)
 */
int
upBufferGetReadLineLength(UP_BUF * bp)
{
	UP_ITEM *	item;
	const char *	cpBegin;
	const char *	cpEnd;
	int		left;

	/*
	 * Find information about the current file data item if any.
	 */
	item = upBufferGetReadItem(bp, UP_ITEM_TYPE_FILE_DATA);

	/*
	 * If we are not reading file data then there is none available.
	 */
	if (item == NULL)
		return 0;

	/*
	 * Check whether the current line length is already correct
	 * because the buffer has a newline at the right place.
	 * If so, then return the line length.
	 */
	if ((bp->lineLength > 0) && (bp->lineLength <= item->length) &&
		(bp->buffer[bp->readOffset + bp->lineLength - 1] == '\n'))
	{
		return bp->lineLength;
	}

	/*
	 * We don't know the full line length yet.
	 * Calculate the amount of data which we haven't scanned yet.
	 */
	left = item->length - bp->lineLength;

	/*
	 * If there is any unscanned data, then search it for a newline.
	 */
	if (left > 0)
	{
		cpBegin = bp->buffer + bp->readOffset + bp->lineLength;

		cpEnd = memchr(cpBegin, '\n', left);

		/*
		 * If the new data contained a newline, then calculate
		 * the length of the full line including the newline
		 * and return that.
		 */
		if (cpEnd)
		{
			bp->lineLength += cpEnd - cpBegin + 1;

			return bp->lineLength;
		}

		/*
		 * The new data didn't have a newline.
		 * Increase the line length to remember that we have
		 * now scanned that much.
		 */
		bp->lineLength += left;
	}

	/*
	 * A newline is not present in any of the data scanned so far.
	 * But if another item is stored in the buffer then return what
	 * is left anyway since there is an end of file or file error
	 * waiting to be read.
	 */
	if (bp->itemCount > 1)
		return bp->lineLength;

	/*
	 * If the whole buffer is full of data then return -1 to indicate
	 * that there is no way that the line will ever be complete.
	 * The caller can decide what to do about it.
	 */
	if (item->length >= bp->bufferSize)
		return -1;

	/*
	 * Neither end of file, a complete line, or a full buffer is present.
	 * Return zero to indicate that no complete line is available.
	 */
	return 0;
}


/*
 * Adjust the read address in the buffer to take into account data
 * that has been read from there after a call to upBufferGetReadAddress.
 * The specified value is the amount of data actually consumed.
 */
void
upBufferReadAdjust(UP_BUF * bp, int size)
{
	UP_ITEM *	item;

	/*
	 * Get the pointer to the file data item if it exists.
	 */
	item = upBufferGetReadItem(bp, UP_ITEM_TYPE_FILE_DATA);

	if (item == NULL)
	{
		upBufferError(bp,
			"Called upBufferReadAdjust when no data available");

		return;
	}

	/*
	 * Validate the size of the adjustment.
	 */
	if ((size < 0) || (size > item->length))
	{
		upBufferError(bp, "Illegal upBufferReadAdjust size");

		return;
	}

	/*
	 * Make sure there is no data allocated specially for the item.
	 * This is just an integrity check since file data items never
	 * allocate their own memory.
	 */
	if (item->data)
	{
		upBufferError(bp, "File data item has allocated data");

		return;
	}

	/*
	 * Skip over the requested amount of data.
	 */
	item->length -= size;

	bp->readOffset += size;
	bp->lineLength -= size;

	if (bp->lineLength < 0)
		bp->lineLength = 0;

	/*
	 * If there is any more data left in the item then we can do
	 * nothing more now.
	 */
	if (item->length > 0)
		return;

	/*
	 * All of the data from the item has been consumed.
	 * If there is another item following this one then there is no
	 * more data in the file, and so we can consume this data item.
	 * Otherwise we have to leave the data item there in case the
	 * writer wants to put more data in the item for us to read.
	 */
	if (bp->itemCount > 1)
	{
		bp->readItemOffset = NEXT_ITEM_OFFSET(bp->readItemOffset);
		bp->itemCount--;
	}
}


/*
 * Clear all of the file data present in the buffer.
 * This is the same result as reading the data and ignoring it.
 */
void
upBufferReadClear(UP_BUF * bp)
{
	UP_ITEM *	item;

	/*
	 * Get the current item which must be file data.
	 */
	item = upBufferGetReadItem(bp, UP_ITEM_TYPE_FILE_DATA);

	/*
	 * If there is a file data item then consume its data.
	 */
	if ((item != NULL) && (item->length > 0))
		upBufferReadAdjust(bp, item->length);
}


/*
 * Read data from the buffer into the supplied area of the specified size.
 * Returns the actual number of bytes returned, or zero if no data is
 * currently present in the buffer to be read.  The upBufferGetReadItemType
 * routine can be used to determine whether zero represents the end of file.
 * This routine can be called at any time.
 */
int
upBufferRead(UP_BUF * bp, char * buf, int length)
{
	const char *	dataAddress;
	int		dataLength;

	/*
	 * If we aren't reading file data then return no data.
	 */
	if (upBufferGetReadItemType(bp) != UP_ITEM_TYPE_FILE_DATA)
		return 0;

	/*
	 * If there is nothing to read then do nothing.
	 */
	if (length <= 0)
		return 0;

	/*
	 * Get the address and length of any available data.
	 */
	dataAddress = upBufferGetReadAddress(bp);
	dataLength = upBufferGetReadLength(bp);

	/*
	 * Limit the amount of data returned to the minimum of what we wanted
	 * and what is actually available.
	 */
	if (length > dataLength)
		length = dataLength;

	/*
	 * Copy the data from the buffer into the user's area.
	 */
	memcpy(buf, dataAddress, length);

	/*
	 * Consume the data that we read from the buffer.
	 */
	upBufferReadAdjust(bp, length);

	/*
	 * Return the amount that was read.
	 */
	return length;
}


/*
 * Consume the current item being read from the buffer, and the buffer
 * space that it uses.  The item must match the specified type otherwise
 * a fatal error will be set.
 */
static void
upBufferReadConsumeItem(UP_BUF * bp, UP_ITEM_TYPE type)
{
	UP_ITEM *	item;

	/*
	 * Get the current item being read and verify its type.
	 */
	item = upBufferGetReadItem(bp, type);

	if (item == NULL)
	{
		upBufferError(bp, "Incorrect item type being consumed");

		return;
	}

	/*
	 * Remove any data in the buffer that this item referenced.
	 * Either we free the specially allocated memory for this item,
	 * or we advance the read offset in the buffer to skip over it.
	 */
	if (item->length > 0)
	{
		if (item->data)
		{
			free(item->data);
			item->data = NULL;
		}
		else
		{
			bp->readOffset += item->length;

			if (bp->readOffset == bp->writeOffset)
			{
				bp->readOffset = 0;
				bp->writeOffset = 0;
			}
		}
	}

	/*
	 * Increment the reading item offset to the next item to be read
	 * (wrapping around if necessary), and decrement the item count.
	 */
	bp->readItemOffset = NEXT_ITEM_OFFSET(bp->readItemOffset);
	bp->itemCount--;
}


/*
 * Read an end of file item from the buffer.
 * Calling this routine when an end of file is not present sets a
 * fatal error for the buffer.
 */
void
upBufferReadEof(UP_BUF * bp)
{
	upBufferReadConsumeItem(bp, UP_ITEM_TYPE_END_OF_FILE);
}


/*
 * Read a file error item from the buffer.
 * Calling this routine when a file error is not present sets a
 * fatal error for the buffer.
 */
void
upBufferReadFileError(UP_BUF * bp)
{
	upBufferReadConsumeItem(bp, UP_ITEM_TYPE_FILE_ERROR);

	bp->totalErrors++;
}


/*
 * Read the end of buffer item from the buffer to indicate that the writer knows
 * that the buffer use is complete.  This invalidates the saved file name.
 * Calling this routine when the buffer is not finished sets a fatal error
 * for the buffer.
 */
void
upBufferReadEndOfBuffer(UP_BUF * bp)
{
	upBufferReadConsumeItem(bp, UP_ITEM_TYPE_END_OF_BUFFER);

	bp->fileIsValid = UP_FALSE;
}


/*
 * Read the next file name and possible file status items from the buffer
 * and store them in the buffer for convenient access.  The extracted file
 * name and status can then be obtained using the upBufferGetFileName and
 * upBufferGetFileStatus methods.  Furthermore, the file data can then be
 * read using other calls.  Calling this routine when a file name is not
 * present sets a fatal error for the buffer.
 */
void
upBufferReadFileName(UP_BUF * bp)
{
	UP_ITEM *	item;
	const char *	fileName;
	const UP_STAT *	fileStatus;

	/*
	 * Invalidate the file name until we have a new one.
	 */
	bp->fileIsValid = UP_FALSE;

	/*
	 * Get the file name item to be read (if present).
	 * If not present generate an error.
	 */
	item = upBufferGetReadItem(bp, UP_ITEM_TYPE_FILE_NAME);

	if (item == NULL)
	{
		upBufferError(bp,
			"Called upBufferReadFileName when no file name");

		return;
	}

	/*
	 * Get the file name from this item if it was stored there,
	 * otherwise get it from the read location in the buffer area.
	 */
	fileName = item->data;

	if (fileName == NULL)
		fileName = bp->buffer + bp->readOffset;

	/*
	 * Save the file name for later use.
	 */
	if (!upBufferSaveFileName(bp, fileName))
		return;

	/*
	 * Consume the file name item.
	 */
	upBufferReadConsumeItem(bp, UP_ITEM_TYPE_FILE_NAME);

	/*
	 * Now look for an optional file status item.
	 * If we find it, then copy the file status into our buffer and
	 * consume it.  Otherwise, zero the status buffer.
	 */
	item = upBufferGetReadItem(bp, UP_ITEM_TYPE_FILE_STATUS);

	if (item != NULL)
	{
		/*
		 * Get the file status from the item itself or from the
		 * data buffer.
		 */
		fileStatus = (const UP_STAT *) item->data;

		if (fileStatus == NULL)
		{
			fileStatus = (const UP_STAT *)
				(bp->buffer + bp->readOffset);
		}

		/*
		 * Copy the file status and then consume it.
		 */
		memcpy(&bp->fileStatus, fileStatus, sizeof(UP_STAT));

		upBufferReadConsumeItem(bp, UP_ITEM_TYPE_FILE_STATUS);
	}
	else
		memset(&bp->fileStatus, 0, sizeof(UP_STAT));

	/*
	 * The file name and status are saved away and valid now.
	 * The reader can collect this information anytime until the
	 * next file name is read or until the buffer is done.
	 * Also update some statistics.
	 */
	bp->fileIsValid = UP_TRUE;
	bp->lineLength = 0;
	bp->byteCount = 0;
	bp->totalFiles++;
}


/*
 * Save the specified file name into our special variable so that it can
 * be returned later.  The storage for the file name is reallocated if
 * necessary.  Returns TRUE on success.
 */
static UP_BOOL
upBufferSaveFileName(UP_BUF * bp, const char * fileName)
{
	char *	newFileName;
	int	newFileNameAlloc;
	int	length;

	length = strlen(fileName) + 1;

	/*
	 * Check the length of the file name against our file name string.
	 * If it is larger, then allocate a new one string to hold it.
	 */
	if (length > bp->fileNameAlloc)
	{
		/*
		 * Calculate the length of the new string to allocate.
		 * Add a bit more to allow for slightly longer file names
		 * in future calls.
		 */
		newFileNameAlloc = length + 32;

		/*
		 * Allocate or reallocate the file name string according to
		 * whether the original file name was pointing to our local
		 * buffer or an allocated string.
		 */
		if (bp->fileName == bp->fileNameBuffer)
			newFileName = malloc(newFileNameAlloc);
		else
			newFileName = realloc(bp->fileName, newFileNameAlloc);

		/*
		 * Check for a failure.
		 */
		if (newFileName == NULL)
		{
			upBufferError(bp, "Failed to allocate file name");

			return UP_FALSE;
		}

		/*
		 * Update our file name information.
		 */
		bp->fileName = newFileName;
		bp->fileNameAlloc = newFileNameAlloc;
	}

	/*
	 * Copy the file name into our string.
	 */
	memcpy(bp->fileName, fileName, length);

	return UP_TRUE;
}


/*
 * Return the saved file name set for the buffer (if any).
 * This is the file name whose data is currently being read.
 * The file name is valid until a new file name has been read.
 * Returns NULL if there is no current file name.
 * This routine can be called at any time.
 */
const char *
upBufferGetFileName(const UP_BUF * bp)
{
	if (!bp->fileIsValid)
		return NULL;

	return bp->fileName;
}


/*
 * Return the status of the current file that was being read.
 * If there is no file currently being read then NULL is returned.
 * If no status had been provided, then all zeroes will be returned.
 * The file status is valid until a new file name has been read.
 * This routine can be called at any time.
 */
const UP_STAT *
upBufferGetReadFileStatus(const UP_BUF * bp)
{
	if (!bp->fileIsValid)
		return NULL;

	return &bp->fileStatus;
}


/*
 * Return the number of files that have been transferred through the buffer.
 * This includes the file that is currently being transferred.
 */
long
upBufferGetFileCount(const UP_BUF * bp)
{
	return bp->totalFiles;
}


/*
 * Return the count of files which had errors.
 */
long
upBufferGetErrorCount(const UP_BUF * bp)
{
	return bp->totalErrors;
}


/*
 * Return the number of bytes that were read so far from the current file.
 * After the end of file or a file error has been read, this will be the total
 * size of the file.
 */
long
upBufferGetByteCount(const UP_BUF * bp)
{
	if (!bp->fileIsValid)
		return 0;

	return bp->byteCount;
}
