/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * Private definitions for user mode pipeline circular buffers.
 * Buffers hold data and meta-data about files being sent between devices.
 * Each buffer is shared between a writer and a writer device.
 * The writer inserts data into the buffer, and the reader extracts it.
 * Multiple files can be in the buffer at the same time.
 */

#ifndef	UP_BUF_H
#define	UP_BUF_H

#include <stdio.h>
#include "up.h"


/*
 * The minimum and maximum buffer size we allow.
 * The minimum buffer size must be at least as long as any file name.
 * The minimum file size is the minimum amount of space to reserve in
 * the buffer for file names, and is grown as longer names are seen.
 */
enum
{
	UP_BUFFER_MIN_SIZE = (1024 * 8),
	UP_BUFFER_MAX_SIZE = (1024 * 1024 * 8),
	UP_BUFFER_MIN_FILE_SIZE = 32
};


/*
 * Types of items which are in the buffer.
 * These item types are also returned to the writer and reader of the buffer
 * to indicate what type of data item can be written or read next.
 */
typedef	enum
{
	UP_ITEM_TYPE_NONE,		/* no type (not actually stored) */
	UP_ITEM_TYPE_FILE_NAME,		/* file name */
	UP_ITEM_TYPE_FILE_STATUS,	/* file status */
	UP_ITEM_TYPE_FILE_DATA,		/* file data */
	UP_ITEM_TYPE_FILE_ERROR,	/* error in file data */
	UP_ITEM_TYPE_END_OF_FILE,	/* end of file data */
	UP_ITEM_TYPE_END_OF_BUFFER,	/* end of buffer */
	UP_ITEM_TYPE_FATAL_ERROR,	/* buffer has a fatal error */
	UP_ITEM_TYPE_FIRST_ILLEGAL	/* first illegal type (must be last) */
} UP_ITEM_TYPE;


/*
 * Data about one item stored in the buffer.
 * This gives the type of the item and the length of any associated data.
 * This also contains a pointer to possible allocated data for the item.
 * If length is positive and the data pointer is NULL, then the actual
 * data is stored in the appropriate offset in our buffer.
 */
typedef	struct
{
	UP_ITEM_TYPE	type;
	int		length;
	char *		data;
} UP_ITEM;


/*
 * Maximum number of items normally stored in the buffer and
 * the number of spare items for storing exceptional items.
 * The spare items allow for the storing of file name and status
 * together, for the storing of an end of file or error status,
 * and for storing the end of buffer status.
 */
enum
{
	UP_MAX_ITEMS = 32,
	UP_SPARE_ITEMS = 4
};


/*
 * The structure to hold a buffer of data which can be written to
 * and read from.  This also provides for the phasing of multiple files
 * through the buffer, including remembering of file names, file status,
 * and end of files.  This has already been typedefed in up.h.
 */
struct UP_BUF
{
	char *		buffer;		/* allocated buffer data */
	const char *	errorText;	/* error text (never freed) */
	int		linkCount;	/* number of links to this buffer */
	int		bufferSize;	/* size of allocated data buffer */
	int		readOffset;	/* offset for data to be read */
	int		writeOffset;	/* offset for data to be written */
	int		readItemOffset;	/* offset of item to be read */
	int		writeItemOffset;/* offset of item to be written */
	int		itemCount;	/* count of items in table */
	UP_ITEM		itemTable[UP_MAX_ITEMS + UP_SPARE_ITEMS];

	/*
	 * The following data is saved here for convenience for writers.
	 */
	UP_ITEM_TYPE	lastWriteItemType;	/* last item type written */
	int		maxFileNameLength;	/* longest file name seen */

	/*
	 * The following data is saved here for convenience for readers.
	 * It contains data about the current file being read.
	 */
	UP_BOOL		fileIsValid;	/* file information is valid */
	UP_STAT		fileStatus;	/* current file status */
	char *		fileName;	/* current file name */
	int		fileNameAlloc;	/* allocated size of file name */
	int		lineLength;	/* length of current line so far */
	long		byteCount;	/* bytes read so far in file */
	char		fileNameBuffer[128];	/* local file name storage */

	/*
	 * Accounting values.
	 */
	long		totalFiles;	/* total files read */
	long		totalErrors;	/* total files which had errors */
	long		totalBytes;	/* total data bytes read */
};


/*
 * Generic buffer routines.
 */
extern	UP_BUF *	upBufferAlloc(int size);
extern	UP_BUF *	upBufferLink(UP_BUF * bp);
extern	void		upBufferFree(UP_BUF * bp);
extern	void		upBufferReset(UP_BUF * bp);
extern	const char *	upBufferGetErrorText(const UP_BUF * bp);
extern	UP_BOOL		upBufferHasError(const UP_BUF * bp);
extern	UP_BOOL		upBufferIsFinished(const UP_BUF * bp);


/*
 * Buffer writing routines.
 */
extern	UP_ITEM_TYPE	upBufferGetWriteItemType(const UP_BUF * bp);
extern	int		upBufferRequestSpace(UP_BUF * bp, int requestLength);
extern	char *		upBufferGetWriteAddress(UP_BUF * bp);
extern	int		upBufferGetWriteLength(UP_BUF * bp);
extern	void		upBufferWriteAdjust(UP_BUF * bp, int size);
extern	void		upBufferWriteEof(UP_BUF * bp);
extern	void		upBufferWriteFileError(UP_BUF * bp);
extern	void		upBufferWriteEndOfBuffer(UP_BUF * bp);

extern	int		upBufferWrite(
	UP_BUF *	bp,
	const char *	buf,
	int		length);

extern	UP_BOOL		upBufferWriteFileName(
	UP_BUF *	bp,
	const char *	fileName,
	const UP_STAT *	status
);



/*
 * Buffer reading routines.
 */
extern	UP_BOOL		upBufferIsReadAvailable(UP_BUF * bp);
extern	UP_ITEM_TYPE	upBufferGetReadItemType(UP_BUF * bp);
extern	const char *	upBufferGetReadAddress(UP_BUF * bp);
extern	int		upBufferGetReadLength(UP_BUF * bp);
extern	UP_BOOL		upBufferReadDataCheck(UP_BUF * bp, int length);
extern	int		upBufferGetReadLineLength(UP_BUF * bp);
extern	void		upBufferReadAdjust(UP_BUF * bp, int size);
extern	void		upBufferReadClear(UP_BUF * bp);
extern	int		upBufferRead(UP_BUF * bp, char * buf, int length);
extern	void		upBufferReadEof(UP_BUF * bp);
extern	void		upBufferReadFileError(UP_BUF * bp);
extern	void		upBufferReadEndOfBuffer(UP_BUF * bp);
extern	void		upBufferReadFileName(UP_BUF * bp);
extern	const char *	upBufferGetFileName(const UP_BUF * bp);
extern	const UP_STAT *	upBufferGetReadFileStatus(const UP_BUF * bp);
extern	long		upBufferGetFileCount(const UP_BUF * bp);
extern	long		upBufferGetErrorCount(const UP_BUF * bp);
extern	long		upBufferGetByteCount(const UP_BUF * bp);

#endif
