/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * User pipelines private support routines.
 * These routines should not be called by clients.
 */

#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>

#include "upPrivate.h"


/*
 * A unique id for a device.
 * This is a positive monotonically increasing number.
 */
static	UP_ID	nextid;


/*
 * A table of state names.
 * This table must correspond to the order of the UP_STATE enums
 * in upPrivate.h.
 */
static	const char * const	stateTable[UP_STATE_FIRST_ILLEGAL] =
{
	"none", "readName", "readEndOfBuffer", "writeName",
	"transferName", "transferEndOfBuffer", "readLine",
	"readData", "readFileError", "readEOF", "writeData",
	"transferData", "transferFileError", "transferEOF",
	"fatalError"
};


/*
 * Local procedures.
 */
static	const char *	getStateName(UP_STATE state);
static	UP_STATE	checkState(UP_DEV * dev);
static	UP_STATE	returnCheckStateError(UP_DEV * dev);


/*
 * Allocate a new device structure for a user pipeline.
 * The function table is saved, and the private data storage is allocated
 * if the size is nonzero.  This call is only allowed if the pipeline
 * has not yet been set up.  Returns NULL on an error.
 */
UP_DEV *
upAllocDevice(UP * up, UP_FUNC * func, int size)
{
	UP_DEV *	dev;
	char *		cp;

	/*
	 * A new device cannot be allocated if the pipeline is
	 * already marked as being set up.
	 */
	if (up->setupDone)
	{
		upError(up, UP_ERR_ALREADYSETUP);

		return UP_NULL_DEV;
	}

	/*
	 * Allocate the new device structure.
	 */
	dev = (UP_DEV *) upMalloc(up, sizeof(UP_DEV));

	if (dev == UP_NULL_DEV)
		return UP_NULL_DEV;

	/*
	 * Initialize its fields.
	 */
	dev->func = func;
	dev->owner = up;
	dev->id = 0;
	dev->next = NULL;
	dev->readBuffer = NULL;
	dev->writeBuffer = NULL;
	dev->didReadWrite = UP_FALSE;
	dev->finished = UP_FALSE;
	dev->failed = UP_FALSE;
	dev->private = NULL;
	dev->pollTable = &dev->pollEntry;
	dev->pollCount = 0;
	dev->pollTimeout = -1;
	dev->state = UP_STATE_NONE;

	/*
	 * If the private data size is nonzero, then allocate and
	 * zero that too.
	 */
	if (size > 0)
	{
		cp = upDeviceMalloc(dev, size);

		if (cp == NULL)
		{
			free((char *) dev);

			return UP_NULL_DEV;
		}

		memset(cp, 0, size);

		dev->private = cp;
	}

	return dev;
}


/*
 * Link a newly opened device into a user pipeline.
 * If the device requires that its file sizes must be known, then we
 * link in the FindSize device ahead of this one to calculate sizes.
 * The unique device id is returned, or UP_NULL_ID on a failure.
 */
UP_ID
upLinkDevice(UP * up, UP_DEV * dev)
{
	const UP_DEV *	testDev;
	UP_BOOL		goodFileSize;

	/*
	 * If this device requires file sizes to be known up front
	 * then see if the preceeding devices meet this requirement.
	 */
	if (dev->func->requiresFileSize)
	{
		goodFileSize = UP_FALSE;

		for (testDev = up->first; testDev; testDev = testDev->next)
		{
			if (testDev->func->suppliesFileSize)
				goodFileSize = UP_TRUE;
			else if (!testDev->func->preservesFileSize)
				goodFileSize = UP_FALSE;
		}

		/*
		 * If the file sizes from the previous device is not
		 * known, then we need to add the FindSize device in
		 * front of this one to calculate them for us.
		 */
		if (upAddFindSize(up, 1024*1024, UP_FLAG_NONE) == UP_NULL_ID)
		{
			/*
			 * We failed, release our new device and
			 * return failure.
			 */
			dev->func->close(dev);
			upFreeDevice(dev);

			return UP_NULL_ID;
		}
	}

	/*
	 * It is ok to link in the device now.
	 */
	dev->owner = up;
	dev->id = ++nextid;
	dev->next = NULL;

	if (up->first == NULL)
		up->first = dev;
	else
		up->last->next = dev;

	up->last = dev;

	up->deviceCount++;

	return dev->id;
}


/*
 * Attach a buffer to connect the specified pair of devices.
 * The first device specified is closer to the front of the pipeline,
 * and is thus the "writing" device.  The second device is closer to
 * the end of the pipeline, and is thus the "reading" device.  This
 * routine checks to make sure that the two devices are not external
 * readers or writers endpoints which wouldn't make use of the buffer.
 * Returns UP_TRUE on success.
 */
UP_BOOL
upAttachBuffer(UP * up, UP_DEV * writeDevice, UP_DEV * readDevice)
{
	UP_BUF *	bp;

	/*
	 * Make sure the devices don't have buffers already.
	 */
	if (writeDevice->writeBuffer || readDevice->readBuffer)
	{
		UP_INTERNAL_ERROR(up);

		return UP_FALSE;
	}

	/*
	 * Make sure that the devices aren't external readers or writers
	 * for their directions which are being connected.
	 */
	if (readDevice->func->externalReader)
	{
		UP_INTERNAL_ERROR(up);

		return UP_FALSE;
	}

	if (writeDevice->func->externalWriter)
	{
		UP_INTERNAL_ERROR(up);

		return UP_FALSE;
	}

	/*
	 * Allocate a buffer of a standard size.
	 */
	bp = upBufferAlloc(UP_BUFFER_MIN_SIZE);

	if (bp == NULL)
	{
		upError(up, UP_ERR_NOMEM);

		return UP_FALSE;
	}

	/*
	 * OK, set the reading and writing sides of the devices to
	 * this buffer.  Increment the link counter of the buffer
	 * so that each device has equal ownership of it.
	 */
	writeDevice->writeBuffer = bp;
	readDevice->readBuffer = upBufferLink(bp);

	return UP_TRUE;
}


/*
 * Free a device structure and any private data that it allocated.
 */
void
upFreeDevice(UP_DEV * dev)
{
	/*
	 * If the poll table is not the convenient one, then free it.
	 */
	if (dev->pollTable != &dev->pollEntry)
		free((char *) dev->pollTable);

	/*
	 * Free the reading and writing buffers.
	 */
	upBufferFree(dev->readBuffer);
	upBufferFree(dev->writeBuffer);

	/*
	 * Free any private data for the device.
	 */
	free(dev->private);

	/*
	 * Finally free the whole device structure.
	 */
	free((char *) dev);
}


/*
 * Find the specified device in a user mode pipeline.
 * If a type other than UP_TYPE_NONE is given, then the found
 * device must be of that type.  Returns a pointer to the device,
 * or UP_NULL_DEV on an error.
 */
UP_DEV *
upFindDevice(UP * up, UP_ID id, UP_TYPE type)
{
	UP_DEV *	dev;

	for (dev = up->first; dev != UP_NULL_DEV; dev = dev->next)
	{
		if (dev->id == id)
		{
			if ((type != UP_TYPE_NONE) &&
				(type != dev->func->type))
			{
				upError(up, UP_ERR_WRONGDEVTYPE);

				return UP_NULL_DEV;
			}

			return dev;
		}
	}

	upError(up, UP_ERR_UNKNOWNDEVID);

	return UP_NULL_DEV;
}


/*
 * Routine to perform I/O for one device.  If the device is
 * safely capable of doing multiple I/O, then we repeatedly call it
 * until it can do no more.  Returns UP_TRUE on success.  Whether the
 * had done any I/O or is finished can be checked after this call.
 */
UP_BOOL
upDeviceReadWrite(UP_DEV * dev)
{
	UP *	up;
	UP_BOOL	didReadWrite;

	up = dev->owner;
	didReadWrite = UP_FALSE;

	/*
	 * Keep calling the device's I/O routine until we shouldn't
	 * do it anymore.  If the device does any I/O at all then
	 * remember that.
	 */
	do
	{
		if (up->loopBlab)
		{
			fprintf(stderr, "%s[%d]: readWrite called\n",
				dev->func->name, dev->id);
		}

		/*
		 * Ask the device to try to perform some I/O.
		 */
		dev->didReadWrite = UP_FALSE;

		if (!dev->func->readWrite(dev))
			return UP_FALSE;

		/*
		 * If the device has failed then return now.
		 * The caller can then check whether we failed.
		 */
		if (dev->failed)
			return UP_FALSE;

		/*
		 * If the device did some I/O then remember that
		 * something got done.
		 */
		if (dev->didReadWrite)
		{
			if (up->loopBlab)
			{
				fprintf(stderr, "%s[%d]: did I/O\n",
					dev->func->name, dev->id);
			}

			didReadWrite = UP_TRUE;
		}

		/*
		 * If the device is finished remember that we did
		 * some I/O this call and stop now.
		 */
		if (dev->finished)
		{
			if (up->loopBlab)
			{
				fprintf(stderr,
					"%s[%d]: finished\n",
					dev->func->name, dev->id);
			}

			didReadWrite = UP_TRUE;

			break;
		}
	}
	while (dev->didReadWrite && dev->func->multiReadWriter);

	/*
	 * Store the result of whether the device did any I/O back
	 * into the device for the caller's information.
	 */
	dev->didReadWrite = didReadWrite;

	return UP_TRUE;
}


/*
 * Default routine for devices to build an empty poll table with a
 * timeout value which depends on the availability of data in the
 * input and output buffers.  This works only fpr devices which do
 * not require any actual polling for a device, and which can always
 * produce or consume the required items.  (Most Pipe-type devices
 * meet this requirement, and some external writers, but no external
 * readers.)  Returns UP_TRUE on success.
 */
UP_BOOL
upDefaultBuildPollTable(UP_DEV * dev)
{
	UP_STATE	state;

	/*
	 * Set that there are no file descriptors for polling,
	 * and that there is an infinite timeout value.
	 */
	dev->pollCount = 0;
	dev->pollTimeout = -1;

	/*
	 * If this is an external reader then generate an error since
	 * we don't know how to check it for readyness.
	 */
	if (dev->func->externalReader)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return UP_FALSE;
	}

	/*
	 * If this is an external writer then see if the read buffer
	 * has an item ready.  If so, then it can run now.
	 */
	if (dev->func->externalWriter)
	{
		if (upBufferIsReadAvailable(dev->readBuffer))
			dev->pollTimeout = 0;

		return UP_TRUE;
	}

	/*
	 * This is a Pipe-type device with both read and write buffers.
	 * Get the state that this device can act on, if any.
	 */
	state = upCheckState(dev);

	/*
	 * If the state indicates that something can be done then
	 * clear the timeout value so that we can run soon.
	 */
	if (state != UP_STATE_NONE)
		dev->pollTimeout = 0;

	return UP_TRUE;
}


/*
 * The default flush routine for devices.
 * This does nothing.
 */
UP_BOOL
upDefaultFlush(UP_DEV * dev)
{
	return UP_TRUE;
}


/*
 * The default reset routine for devices.
 * This does nothing.
 */
UP_BOOL
upDefaultReset(UP_DEV * dev)
{
	return UP_TRUE;
}


/*
 * The default close routine for devices.
 * This does nothing.
 */
UP_BOOL
upDefaultClose(UP_DEV * dev)
{
	return UP_TRUE;
}


/*
 * Translate the specified file name by altering its extension according
 * to the specified table of conversions.  The table is in the form of
 * pairs of extensions (including the dots if required), with the first
 * pair replaced by the second pair.  The table is terminated with a NULL
 * string.  The returned file name is valid until the next call.  Returns
 * NULL on an error.
 */
const char *
upTranslateFileName(
	UP_DEV *		dev,
	const char *		fileName,
	const char * const *	conversionTable
)
{
	int		fileNameLength;
	const char *	inputExtension;
	const char *	outputExtension;
	char *		newFileName;
	int		newAlloc;
	int		testLength;
	int		inputLength;
	int		outputLength;
	int		keptLength;

	static char *	savedFileName;
	static int	savedAlloc;

	/*
	 * Get the length of the file name to translate.
	 * Also assume there is no translation needed until we find one.
	 */
	fileNameLength = strlen(fileName);
	keptLength = fileNameLength;
	inputLength = 0;
	outputLength = 0;
	inputExtension = "";
	outputExtension = "";

	/*
	 * Examine pairs of table entries looking for a match.
	 */
	while ((conversionTable[0] != NULL) && (conversionTable[1] != NULL))
	{
		/*
		 * Get the input and output extensions, and the input
		 * extension's length.
		 */
		inputExtension = *conversionTable++;
		outputExtension = *conversionTable++;

		testLength = strlen(inputExtension);

		/*
		 * If the file name is too short then skip this extension.
		 */
		if (fileNameLength < testLength)
			continue;

		/*
		 * If the file name ends in this extension, then set the
		 * input and output lengths and break out of the loop.
		 */
		if (memcmp(fileName + (fileNameLength - testLength),
			inputExtension, testLength) == 0)
		{
			keptLength = fileNameLength - testLength;
			inputLength = testLength;
			outputLength = strlen(outputExtension);

			break;
		}
	}

	/*
	 * Calculate the size we need to store the translated string,
	 * and reallocate the static string if necessary.
	 */
	newAlloc = keptLength + outputLength + 1;

	if (newAlloc > savedAlloc)
	{
		newFileName = upDeviceRealloc(dev, savedFileName, newAlloc);

		if (newFileName == NULL)
			return NULL;

		savedFileName = newFileName;
		savedAlloc = newAlloc;
	}

	/*
	 * Store the old file name with the new extension.
	 */
	memcpy(savedFileName, fileName, keptLength);
	memcpy(savedFileName + keptLength, outputExtension, outputLength);

	savedFileName[keptLength + outputLength] = '\0';

	/*
	 * Return the final file name.
	 */
	return savedFileName;
}


/*
 * Set the fact that we have done I/O for the device.
 * For convenience, this routine always returns UP_TRUE so that this
 * can be called to provide a successful return value at the end of
 * a device routine.
 */
UP_BOOL
upSetDidReadWrite(UP_DEV * dev)
{
	dev->didReadWrite = UP_TRUE;

	return UP_TRUE;
}


/*
 * Set the fact that we are finished doing I/O for the device.
 * This also sets that we have done I/O, and sets our state to the
 * finished value.  For convenience, this routine always returns UP_TRUE
 * to provide a successful return value at end of a device routine.
 */
UP_BOOL
upSetFinished(UP_DEV * dev)
{
	dev->finished = UP_TRUE;
	dev->didReadWrite = UP_TRUE;
	dev->state = UP_STATE_FINISHED;

	return UP_TRUE;
}


/*
 * Set the failed flag for the device which indicates that it has had a
 * fatal error and can no longer perform any I/O.  The fatal flag is also
 * set by calling any of the device error routines.
 */
void
upSetFailed(UP_DEV * dev)
{
	dev->failed = UP_TRUE;
}


/*
 * Return the current state value for the device.
 */
UP_STATE
upGetState(UP_DEV * dev)
{
	return dev->state;
}


/*
 * Set the state value for the device.
 */
void
upSetState(UP_DEV * dev, UP_STATE newState)
{
	/*
	 * Print state changes if required and available.
	 */
	if ((dev->owner != NULL) && dev->owner->stateBlab)
	{
		fprintf(stderr, "%s[%d]: setState %s => %s\n",
			dev->func->name, dev->id, getStateName(dev->state),
			getStateName(newState));
	}

	dev->state = newState;
}


/*
 * Return whether the device allows its current state to be acted on now.
 * This can only be called for devices which are not endpoint devices and
 * which do not need to perform any actual poll call.  The state is one of
 * nothing, reading a file name, writing a file name, transferring a file
 * name, reading file data, writing file data, or transferring file data.
 * The state which can actually be executed is returned, with UP_STATE_NONE
 * indicating that nothing can be done.  Reading, writing, and transferring
 * types of states are never mixed up, so that for example, if a transfer
 * state is supplied as input, a read or write state is never returned.
 * Similarly, file name and file data states are never mixed up.
 * Inconsistent buffer states are detected and generate an error.
 * The incoming and outgoing states are logged to stderr if required.
 */
UP_STATE
upCheckState(UP_DEV * dev)
{
	UP_STATE	state;

	/*
	 * Actually check the state.
	 */
	state = checkState(dev);

	/*
	 * If debugging is turned on or if the state is a fatal error
	 * then log the state.
	 */
	if ((dev->owner->stateBlab) || (state == UP_STATE_FATAL_ERROR))
	{
		fprintf(stderr, "%s[%d]: checkState %s => %s\n",
			dev->func->name, dev->id,
			getStateName(dev->state),
			getStateName(state));
	}

	return state;
}


/*
 * Internal version of upCheckState which does not log anything.
 */
static UP_STATE
checkState(UP_DEV * dev)
{
	UP_STATE	readNameState;
	UP_STATE	readDataState;
	UP_STATE	writeState;
	UP_ITEM_TYPE	readItemType;
	UP_ITEM_TYPE	writeItemType;
	UP_BOOL		canWrite;

	/*
	 * If this is an external reader or writer then fail.
	 */
	if (dev->func->externalReader || dev->func->externalWriter)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return returnCheckStateError(dev);
	}

	/*
	 * If the device has finished then we should not have been called.
	 */
	if (dev->finished)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return returnCheckStateError(dev);
	}

	/*
	 * Initialize the read and write states.
	 */
	readNameState = UP_STATE_NONE;
	readDataState = UP_STATE_NONE;
	writeState = UP_STATE_NONE;
	canWrite = UP_FALSE;

	/*
	 * Switch on the item type of the read buffer and set the state
	 * indicating what kind of data should be readable or not
	 * readable now.
	 */
	readItemType = upBufferGetReadItemType(dev->readBuffer);

	switch (readItemType)
	{
		case UP_ITEM_TYPE_NONE:
			break;

		case UP_ITEM_TYPE_FILE_NAME:
			readNameState = UP_STATE_READ_NAME;
			break;

		case UP_ITEM_TYPE_END_OF_BUFFER:
			readNameState = UP_STATE_READ_END_OF_BUFFER;
			break;

		case UP_ITEM_TYPE_FILE_DATA:
			readDataState = UP_STATE_READ_DATA;
			break;

		case UP_ITEM_TYPE_FILE_ERROR:
			readDataState = UP_STATE_READ_FILE_ERROR;
			break;

		case UP_ITEM_TYPE_END_OF_FILE:
			readDataState = UP_STATE_READ_EOF;
			break;

		case UP_ITEM_TYPE_FATAL_ERROR:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return returnCheckStateError(dev);

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return returnCheckStateError(dev);
	}

	/*
	 * Switch on the state of the write buffer and set the states
	 * indicating what kind of data should be writable or not
	 * writable now.
	 */
	writeItemType = upBufferGetWriteItemType(dev->writeBuffer);

	switch (writeItemType)
	{
		case UP_ITEM_TYPE_FILE_NAME:
			writeState = UP_STATE_WRITE_NAME;
			canWrite = UP_TRUE;
			break;

		case UP_ITEM_TYPE_FILE_DATA:
			writeState = UP_STATE_WRITE_DATA;
			canWrite = UP_TRUE;
			break;

		case UP_ITEM_TYPE_END_OF_BUFFER:
			writeState = UP_STATE_WRITE_NAME;
			break;

		case UP_ITEM_TYPE_END_OF_FILE:
			writeState = UP_STATE_WRITE_DATA;
			break;

		case UP_ITEM_TYPE_FATAL_ERROR:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return returnCheckStateError(dev);

		default:
			UP_DEVICE_INTERNAL_ERROR(dev);

			return returnCheckStateError(dev);
	}

	/*
	 * Now switch on the device's current state and check the states
	 * which indicate the result.
	 */
	switch (dev->state)
	{
		case UP_STATE_NONE:
			return UP_STATE_NONE;

		case UP_STATE_READ_NAME:
			if (readDataState != UP_STATE_NONE)
				break;

			return readNameState;

		case UP_STATE_WRITE_NAME:
			if (writeState == UP_STATE_WRITE_DATA)
				break;

			if (canWrite)
				return writeState;

			return UP_STATE_NONE;

		case UP_STATE_TRANSFER_NAME:
			if (readDataState != UP_STATE_NONE)
				break;

			if (writeState == UP_STATE_WRITE_DATA)
				break;

			if ((readNameState == UP_STATE_READ_NAME) && canWrite)
				return UP_STATE_TRANSFER_NAME;

			if (readNameState == UP_STATE_READ_END_OF_BUFFER)
				return UP_STATE_TRANSFER_END_OF_BUFFER;

			return UP_STATE_NONE;

		case UP_STATE_READ_DATA:
			if (readNameState != UP_STATE_NONE)
				break;

			return readDataState;

		case UP_STATE_READ_LINE:
			if (readNameState != UP_STATE_NONE)
				break;

			if (readDataState != UP_STATE_READ_DATA)
				return readDataState;

			/*
			 * Data is available.
			 * Check whether a line is ready.
			 */
			if (upBufferGetReadLineLength(dev->readBuffer) != 0)
				return UP_STATE_READ_LINE;

			return UP_STATE_NONE;

		case UP_STATE_WRITE_DATA:
			if (writeState == UP_STATE_WRITE_NAME)
				break;

			if (canWrite)
				return writeState;

			return UP_STATE_NONE;

		case UP_STATE_TRANSFER_DATA:
			if (readNameState != UP_STATE_NONE)
				break;

			if (writeState == UP_STATE_WRITE_NAME)
				break;

			if (readDataState == UP_STATE_READ_EOF)
				return UP_STATE_TRANSFER_EOF;

			if (readDataState == UP_STATE_READ_FILE_ERROR)
				return UP_STATE_TRANSFER_FILE_ERROR;

			if (readDataState == UP_STATE_READ_DATA && canWrite)
				return UP_STATE_TRANSFER_DATA;

			return UP_STATE_NONE;

		case UP_STATE_FINISHED:
			/*
			 * We were supposedly finished.
			 */
			UP_DEVICE_INTERNAL_ERROR(dev);

			return returnCheckStateError(dev);

		default:
			/*
			 * The state is unknown.
			 */
			UP_DEVICE_INTERNAL_ERROR(dev);

			return returnCheckStateError(dev);
	}

	/*
	 * The buffer states are inconsistent with the requested state.
	 */
	UP_DEVICE_INTERNAL_ERROR(dev);

	return returnCheckStateError(dev);
}


/*
 * Return a fatal error status for checkState to return.
 * This is a subroutine so that a breakpoint can easily be put on it.
 */
static UP_STATE
returnCheckStateError(UP_DEV * dev)
{
	return UP_STATE_FATAL_ERROR;
}


/*
 * Transfer file data between the input and output buffers.
 * The amount transferred is limited both by the available buffer
 * space and also on the specified limit value, if non-zero.
 * The input and output buffers are adjusted by the amount transferred.
 * This should only be called when it is known that the input
 * buffer has data and that the output buffer has room for data.
 * Returns the actual number of bytes transferred or -1 on an error.
 */
int
upTransferData(UP_DEV * dev, int limitLength)
{
	UP_BUF *	readBuffer;
	UP_BUF *	writeBuffer;
	const char *	readAddress;
	char *		writeAddress;
	int		readLength;
	int		writeLength;
	int		cc;

	readBuffer = dev->readBuffer;
	writeBuffer = dev->writeBuffer;

	/*
	 * Get the amount of data that can be easily read and written.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Take the minimum of the two lengths.
	 */
	cc = readLength;

	if (cc > writeLength)
		cc = writeLength;

	/*
	 * If an additional transfer limit was specified,
	 * then also limit the transfer length to that size.
	 */
	if ((limitLength != 0) && (cc > limitLength))
		cc = limitLength;

	/*
	 * If there is no data to move, then something is wrong.
	 * (This routine should not have been called in this case.)
	 */
	if (cc <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return -1;
	}

	/*
	 * Copy the data from the read buffer to the write buffer.
	 */
	memcpy(writeAddress, readAddress, cc);

	/*
	 * Increment the read and write buffer pointers past the data.
	 */
	upBufferReadAdjust(readBuffer, cc);
	upBufferWriteAdjust(writeBuffer, cc);

	/*
	 * Remember that we have done I/O.
	 */
	dev->didReadWrite = UP_TRUE;

	/*
	 * Return the number of bytes transferred.
	 */
	return cc;
}


/*
 * Read data from the input buffer into the specified address.
 * The amount read is limited both by the available input buffer
 * space and also on the specified data length.  The read location
 * is adjusted by the amount read.  This should only be called when
 * it is known that the input buffer has some data to be read.
 * Returns the actual number of bytes read or -1 on an error.
 */
int
upReadData(UP_DEV * dev, char * dataAddress, int dataLength)
{
	UP_BUF *	readBuffer;
	const char *	readAddress;
	int		readLength;
	int		cc;

	readBuffer = dev->readBuffer;

	/*
	 * Get the amount of data that can be easily read.
	 */
	readAddress = upBufferGetReadAddress(readBuffer);
	readLength = upBufferGetReadLength(readBuffer);

	/*
	 * Take the minimum of the available input buffer data and the
	 * amount to be read.
	 */
	cc = dataLength;

	if (cc > readLength)
		cc = readLength;

	/*
	 * If there is no data to move, then something is wrong.
	 * (This routine should not have been called in this case.)
	 */
	if (cc <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return -1;
	}

	/*
	 * Copy the data from the input buffer to the user's buffer.
	 */
	memcpy(dataAddress, readAddress, cc);

	/*
	 * Increment the read buffer pointer past the data.
	 */
	upBufferReadAdjust(readBuffer, cc);

	/*
	 * Remember that we have done I/O.
	 */
	dev->didReadWrite = UP_TRUE;

	/*
	 * Return the number of bytes transferred.
	 */
	return cc;
}


/*
 * Write data from the specified address to the output buffer.
 * The amount written is limited both by the available output buffer
 * space and also on the specified data length.  The write location
 * is adjusted by the amount written.  This should only be called
 * when it is known that the output buffer has room for data.
 * Returns the actual number of bytes written or -1 on an error.
 */
int
upWriteData(UP_DEV * dev, const char * dataAddress, int dataLength)
{
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	writeBuffer = dev->writeBuffer;

	/*
	 * Get the amount of data that can be easily written.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Take the minimum of the available buffer space and the
	 * amount to be written.
	 */
	cc = dataLength;

	if (cc > writeLength)
		cc = writeLength;

	/*
	 * If there is no data to move, then something is wrong.
	 * (This routine should not have been called in this case.)
	 */
	if (cc <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return -1;
	}

	/*
	 * Copy the data into the write buffer.
	 */
	memcpy(writeAddress, dataAddress, cc);

	/*
	 * Increment the write buffer pointer past the data.
	 */
	upBufferWriteAdjust(writeBuffer, cc);

	/*
	 * Remember that we have done I/O.
	 */
	dev->didReadWrite = UP_TRUE;

	/*
	 * Return the number of bytes transferred.
	 */
	return cc;
}


/*
 * Try to write zeroed data to the output buffer of the specified length.
 * The amount zeroed is limited both by the available output buffer
 * space and also on the specified length.  The write location is
 * adjusted by the amount zeroed.  This should only be called when it
 * is known that the output buffer has room for data.  Returns the
 * actual number of bytes zeroed or -1 on an error.
 */
int
upWriteZeroedData(UP_DEV * dev, int zeroLength)
{
	UP_BUF *	writeBuffer;
	char *		writeAddress;
	int		writeLength;
	int		cc;

	writeBuffer = dev->writeBuffer;

	/*
	 * Get the amount of data that can be easily written.
	 */
	writeAddress = upBufferGetWriteAddress(writeBuffer);
	writeLength = upBufferGetWriteLength(writeBuffer);

	/*
	 * Take the minimum of the available buffer space and the
	 * amount to be written.
	 */
	cc = zeroLength;

	if (cc > writeLength)
		cc = writeLength;

	/*
	 * If there is no data to move, then something is wrong.
	 * (This routine should not have been called in this case.)
	 */
	if (cc <= 0)
	{
		UP_DEVICE_INTERNAL_ERROR(dev);

		return -1;
	}

	/*
	 * Zero the data in the write buffer.
	 */
	memset(writeAddress, 0, cc);

	/*
	 * Increment the write buffer pointer past the data.
	 */
	upBufferWriteAdjust(writeBuffer, cc);

	/*
	 * Remember that we have done I/O.
	 */
	dev->didReadWrite = UP_TRUE;

	/*
	 * Return the number of bytes zeroed.
	 */
	return cc;
}


/*
 * Copy the information from the system stat structure to our own
 * stat structure.  This sets the file type flag and sets the
 * available flags to all, and does some special directory fixes.
 */
void
upCopyStatus(UP_STAT * ourStat, const struct stat * systemStat)
{
	memset((char *) ourStat, 0, sizeof(UP_STAT));

	ourStat->flags = UP_STAT_ALLFLAGS;
	ourStat->type = upGetFileType(systemStat->st_mode);
	ourStat->mode = systemStat->st_mode;
	ourStat->dev = systemStat->st_dev;
	ourStat->inode = systemStat->st_ino;
	ourStat->nlink = systemStat->st_nlink;
	ourStat->uid = systemStat->st_uid;
	ourStat->gid = systemStat->st_gid;
	ourStat->rdev = systemStat->st_rdev;
	ourStat->size = systemStat->st_size;
	ourStat->atime = systemStat->st_atime;
	ourStat->mtime = systemStat->st_mtime;
	ourStat->ctime = systemStat->st_ctime;

	/*
	 * If this is a directory then clear the data present flag.
	 */
	if (S_ISDIR(systemStat->st_mode))
		ourStat->flags &= ~UP_STAT_DATA;
}


/*
 * Get the state name for a specified state.
 * If the state is illegal we return a string which is only valid
 * for two calls (so that a pair of states can be printed).
 */
static const char *
getStateName(UP_STATE state)
{
	char *		str;
	static	int	index;
	static	char	buffer[30 * 2];

	/*
	 * If the state is in range then return the string.
	 */
	if ((state >= 0) && (state < UP_STATE_FIRST_ILLEGAL))
		return stateTable[state];

	/*
	 * We must manufacture the state.
	 * Point to the appropriate place in the buffer.
	 */
	if (++index >= 2)
		index = 0;

	str = &buffer[index * 30];

	sprintf(str, "state%d", state);

	return str;
}


/*
 * Create a new temporary file and return the file descriptor for
 * reading and writing the file.  The file is deleted as soon as it
 * is opened so that no cleanup is required later other than closing
 * file file descriptor.  Returns -1 on an error.
 */
int
upCreateTemporaryFile(UP_DEV * dev)
{
	const char *	tmpDir;
	char *		fullPath;
	char *		changePath;
	int		length;
	int		fd;
	long		sequence;

	/*
	 * Get the directory which contains temporary files.
	 */
	tmpDir = upGetTemporaryDirectoryPath();

	/*
	 * Allocate a buffer for the full path name which is
	 * large enough for the temporary directory path
	 * plus our entry name.
	 */
	length = strlen(tmpDir);

	fullPath = upDeviceMalloc(dev, length + 64);

	if (fullPath == NULL)
		return -1;

	memcpy(fullPath, tmpDir, length);

	changePath = fullPath + length;

	/*
	 * Add the beginning of the entry name to the
	 * directory path, using our process id to help
	 * make it unique.
	 */
	sprintf(changePath, "/up-%ld-", (long) getpid());

	changePath += strlen(changePath);

	/*
	 * Keep looping trying new sequence numbers until we
	 * manage to open a file or fail.
	 */
	sequence = 0;

	while (UP_TRUE)
	{
		sprintf(changePath, "%ld.tmp", sequence++);

		/*
		 * Try to create the file for read/write,
		 * using exclusive open so as to not overwrite
		 * anything, and with a mode of zero so that
		 * noone else can read or write the file.
		 */
		fd = open(fullPath, O_CREAT|O_RDWR|O_EXCL, 0);

		/*
		 * If we opened the file then break out.
		 */
		if (fd >= 0)
			break;	

		/*
		 * If the error was not that the file existed,
		 * then fail.
		 */
		if (errno != EEXIST)
		{
			upDeviceFileSystemError(dev, "open", fullPath);

			free(fullPath);

			return -1;
		}
	}

	/*
	 * We have the file opened.
	 * Delete it so that we don't have any cleanup.
	 */
	if (unlink(fullPath) < 0)
	{
		upDeviceFileSystemError(dev, "unlink", fullPath);

		(void) close(fd);

		free(fullPath);

		return -1;
	}

	free(fullPath);

	/*
	 * Return the file descriptor.
	 */
	return fd;
}


/*
 * Get the file type character from the file mode bits.
 * These flags are the same as the ones returned by the "ls" program.
 * A question mark character is returned if the mode is unknown.
 */
int
upGetFileType(int mode)
{
	if (S_ISREG(mode))
		return '-';

	if (S_ISDIR(mode))
		return 'd';

	if (S_ISBLK(mode))
		return 'b';

	if (S_ISCHR(mode))
		return 'c';

#ifdef S_ISFIFO
	if (S_ISFIFO(mode))
		return 'p';
#endif

#ifdef	S_ISSOCK
	if (S_ISSOCK(mode))
		return 's';
#endif

#ifdef	S_ISLNK
	if (S_ISLNK(mode))
		return 'l';
#endif

	return '?';
}


/*
 * Generate a sequential file name based on the specified pattern.
 * A percent sign in the pattern is replaced with a sequential number.
 * If no percent sign is present, the sequentual number is appended.
 * If the value is negative, our own sequentual number is used,
 * otherwise the indicated sequentual value is used.  The returned
 * file name is allocated and must eventually be freed.
 * Returns NULL on an error with an error set.
 */
char *
upGenerateFileName(UP_DEV * dev, const char * pattern, long value)
{
	static	long	sequentialValue;
	char *		fileName;
	char *		cp;
	char		buffer[32];

	/*
	 * Allocate the next number if required.
	 */
	if (value < 0)
		value = ++sequentialValue;

	/*
	 * Format the number.
	 */
	sprintf(buffer, "%06ld", value);

	/*
	 * Allocate a string large enough to hold the result.
	 */
	fileName = upDeviceMalloc(dev, strlen(pattern) + strlen(buffer) + 4);

	if (fileName == NULL)
		return NULL;

	/*
	 * Look for the pattern character.
	 */
	cp = strchr(pattern, '%');

	/*
	 * Generate the file name according to whether the pattern
	 * character was found.
	 */
	if (cp != NULL)
	{
		memcpy(fileName, pattern, cp - pattern);
		strcpy(fileName + (cp - pattern), buffer);
		strcat(fileName, cp + 1);
	}
	else
	{
		strcpy(fileName, pattern);
		strcat(fileName, "-");
		strcat(fileName, buffer);
	}

	/*
	 * Return the generated file name.
	 */
	return fileName;
}


/*
 * Allocate memory, setting an error if the malloc failed.
 * Returns a pointer to the allocated memory, or NULL on failure.
 * This is the same as upMalloc, except that it accepts a UP_DEV
 * argument so that the error can be localized.
 */
char *
upDeviceMalloc(UP_DEV * dev, int len)
{
	char *	cp;

	cp = malloc(len);

	if (cp == NULL)
	{
		upDeviceError(dev, UP_ERR_NOMEM);

		return NULL;
	}

	return cp;
}


/*
 * Reallocate memory, setting an error if the realloc failed.
 * Returns a pointer to the allocated memory, or NULL on failure.
 * It accepts a UP_DEV argument so that the error can be localized.
 */
char *
upDeviceRealloc(UP_DEV * dev, char * cp, int len)
{
	cp = realloc(cp, len);

	if (cp == NULL)
	{
		upDeviceError(dev, UP_ERR_NOMEM);

		return NULL;
	}

	return cp;
}


/*
 * Copy a null-terminated string into malloc'd memory.
 * It accepts a UP_DEV argument so that the error can be localized.
 * A NULL pointer is accepted and generates an error.
 */
char *
upDeviceCopyString(UP_DEV * dev, const char * str)
{
	char *	cp;
	int	len;

	if (str == NULL)
	{
		upDeviceError(dev, UP_ERR_NULLPOINTER);

		return NULL;
	}

	len = strlen(str) + 1;

	cp = upDeviceMalloc(dev, len);

	if (cp == NULL)
		return NULL;

	memcpy(cp, str, len);

	return cp;
}


/*
 * Allocate memory, setting an error if the malloc failed.
 * Returns a pointer to the allocated memory, or NULL on failure.
 */
char *
upMalloc(UP * up, int len)
{
	char *	cp;

	cp = malloc(len);

	if (cp == NULL)
	{
		upError(up, UP_ERR_NOMEM);

		return NULL;
	}

	return cp;
}


/*
 * Reallocate memory, setting an error if the realloc failed.
 * Returns a pointer to the allocated memory, or NULL on failure.
 */
char *
upRealloc(UP * up, char * cp, int len)
{
	cp = realloc(cp, len);

	if (cp == NULL)
	{
		upError(up, UP_ERR_NOMEM);

		return NULL;
	}

	return cp;
}


/*
 * Copy a null-terminated string into malloc'd memory.
 */
char *
upCopyString(UP * up, const char * str)
{
	char *	cp;
	int	len;

	len = strlen(str) + 1;

	cp = upMalloc(up, len);

	if (cp == NULL)
		return NULL;

	memcpy(cp, str, len);

	return cp;
}


/*
 * Either allocate a new table of elements, or else reallocate an existing
 * table of elements, depending on whether the existing table matches
 * the specified argument.  This is used for devices which have a "quick"
 * non-malloc'd local table of elements for the common cases, but which
 * might need to be able to grow beyond that size.  Returns the newly
 * allocated array or NULL on an error.
 */
char *
upGrowTable(UP_DEV * dev, void * oldTable, void * quickTable,
	int oldElementCount, int newElementCount, size_t elementSize)
{
	void *	newTable;
	size_t	newSize;

	newSize = newElementCount * elementSize;

	/*
	 * If we can reallocate the table then just do that.
	 */
	if ((oldTable == NULL) || (oldTable != quickTable))
		return upDeviceRealloc(dev, oldTable, newSize);

	/*
	 * We must allocate a totally fresh table and copy the data
	 * from the quick table into it.
	 */
	newTable = upDeviceMalloc(dev, newSize);

	if (newTable == NULL)
		return NULL;

	memcpy(newTable, oldTable, oldElementCount * elementSize);

	return newTable;
}


/*
 * Set the specified error code which is not the result of a system call
 * failure or a standard library error, but which is one of ours.
 */
void
upError(UP * up, int errorCode)
{
	up->errorCode = errorCode;
	up->systemErrno = 0;
	up->systemName = "";
	up->errorDeviceId = UP_NULL_ID;
	up->errorDeviceName = "";
}


/*
 * Set the error code to indicate a system call failure, saving the
 * current errno value and the specified system call name.  The system
 * call name must be permanent.
 */
void
upSystemError(UP * up, const char * systemName)
{
	up->errorCode = UP_ERR_SYSTEM;
	up->systemErrno = errno;
	up->systemName = systemName;
	up->errorDeviceId = UP_NULL_ID;
	up->errorDeviceName = "";
}


/*
 * Set an internal error which holds the file name and line number
 * which generated the error.  This is called by the UP_INTERNAL_ERROR
 * macro.
 */
void
upInternalError(UP * up, const char * fileName, int lineNumber)
{
	up->errorCode = UP_ERR_INTERNAL;
	up->systemErrno = 0;
	up->systemName = "";
	up->errorDeviceId = UP_NULL_ID;
	up->errorDeviceName = "";
	up->errorFileName = fileName;
	up->errorLineNumber = lineNumber;
}


/*
 * Set the error code for a device which is not the result of a system call
 * failure or a standard library error, but which is one of ours.
 * The failed flag is set for the device which will stop further I/O.
 */
void
upDeviceError(UP_DEV * dev, int errorCode)
{
	UP *	up;

	up = dev->owner;

	up->errorCode = errorCode;
	up->systemErrno = 0;
	up->systemName = "";
	up->errorDeviceId = dev->id;
	up->errorDeviceName = dev->func->name;

	/*
	 * Remember that we have failed.
	 */
	dev->failed = UP_TRUE;
}


/*
 * Set the error code for a device to indicate a system call failure,
 * saving the current errno value and the specified system call name.
 * The system call name must be permanent.  The failed flag is set for
 * the device which will stop further I/O.
 */
void
upDeviceSystemError(UP_DEV * dev, const char * systemName)
{
	UP *	up;

	up = dev->owner;

	up->errorCode = UP_ERR_SYSTEM;
	up->systemErrno = errno;
	up->systemName = systemName;
	up->errorDeviceId = dev->id;
	up->errorDeviceName = dev->func->name;

	/*
	 * Clear any file name.
	 */
	free(up->systemFileName);
	up->systemFileName = NULL;

	/*
	 * Remember that we have failed.
	 */
	dev->failed = UP_TRUE;
}


/*
 * Set the error code for a device to indicate a system call failure,
 * saving the current errno value, the specified system call name,
 * and a possible file name.  The system call name must be permanent.
 * The file name is copied if possible.  The failed flag is set for
 * the device which will stop further I/O.
 */
void
upDeviceFileSystemError(UP_DEV * dev, const char * systemName,
	const char * fileName)
{
	UP *	up;

	up = dev->owner;

	up->errorCode = UP_ERR_SYSTEM;
	up->systemErrno = errno;
	up->systemName = systemName;
	up->errorDeviceId = dev->id;
	up->errorDeviceName = dev->func->name;

	/*
	 * Free any old file name and duplicate the new one.
	 */
	free(up->systemFileName);
	up->systemFileName = strdup(fileName);

	/*
	 * Remember that we have failed.
	 */
	dev->failed = UP_TRUE;
}


/*
 * Set an internal error for a device which holds the file name and
 * line number which generated the error.  This is called by the
 * UP_DEVICE_INTERNAL_ERROR macro.  The failed flag is set for the
 * device which will stop further I/O.
 */
void
upDeviceInternalError(UP_DEV * dev, const char * fileName, int lineNumber)
{
	UP *	up;

	up = dev->owner;

	up->errorCode = UP_ERR_INTERNAL;
	up->systemErrno = 0;
	up->systemName = "";
	up->errorDeviceId = dev->id;
	up->errorDeviceName = dev->func->name;
	up->errorFileName = fileName;
	up->errorLineNumber = lineNumber;

	/*
	 * Remember that we have failed.
	 */
	dev->failed = UP_TRUE;
}


/*
 * Set a warning message which doesn't stop processing, but which
 * might be seen by the user or affect a program's exit status.
 * Whether warnings are printed is dependent on the flag set by the
 * upPrintWarnings function.  If warnings are printed, they are
 * sent to standard error.  A counter of warnings is incremented
 * which can be checked later to see if there were any warnings.
 */
void
upWarning(UP * up, const char * message, ...)
{
	va_list	ap;
	char	buf[8192];

	/*
	 * Increment the warning count.
	 */
	up->warningCount++;

	/*
	 * If the maximum warning value is to be used and if we have
	 * reached the limit on the number of warnings then return.
	 */
	if ((up->maxWarningCount >= 0) &&
		(up->warningCount > up->maxWarningCount))
	{
		return;
	}

	/*
	 * Format and print the warning.
	 */
	if (message == NULL)
		message = "up";

	va_start(ap, message);
	vsprintf(buf, message, ap);
	va_end(ap);

	fprintf(stderr, "UP warning: %s\n", buf);
}
