/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * Private definitions for user mode pipelines.
 * These definitions should not be referenced by the users of the up library.
 */

#ifndef	UP_PRIV_H
#define	UP_PRIV_H

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "up.h"
#include "upBuffer.h"


/*
 * The null device pointer value.
 */
#define	UP_NULL_DEV	((UP_DEV *) 0)


/*
 * The function table for a device.
 * This provides the methods for operations on the device.
 * It also provides some data values which are constants for the device.
 *
 * ExternalReader indicates whether this device must be at the "left"
 * end of the pipeline since it doesn't read data from any previous
 * device but instead reads it from some external data source.
 *
 * ExternalWriter indicates whether this device must be at the "right"
 * end of the pipeline since it doesn't write data to any next device
 * but instead writes it to some external data sink.
 *
 * RequiresFileSize indicates that this device must know the size of
 * its input files when the file names are given.
 *
 * SuppliesFileSize indicates that this device produces output files
 * whose size is known when its file names are given.
 *
 * PreservesFileSize indicates that this device produces output files
 * whose sizes are the same as its input files.
 *
 * MultiReadWriter means that the readWrite routine for the device can
 * safely be called repeatedly after just one call to its buildPollTable
 * routine, and when this is done it will eventually stop doing I/O without
 * performing an arbitrarily large amount of I/O.
 */
typedef	struct
{
	UP_TYPE		type;
	const char *	name;
	UP_BOOL		externalReader;
	UP_BOOL		externalWriter;
	UP_BOOL		requiresFileSize;
	UP_BOOL		suppliesFileSize;
	UP_BOOL		preservesFileSize;
	UP_BOOL		multiReadWriter;
	UP_BOOL		(*buildPollTable)(UP_DEV *);
	UP_BOOL		(*readWrite)(UP_DEV *);
	UP_BOOL		(*flush)(UP_DEV *);
	UP_BOOL		(*close)(UP_DEV *);
	UP_BOOL		(*reset)(UP_DEV *);
} UP_FUNC;


/*
 * The read and write buffer states required in order to do I/O now.
 * This is used for Pipe devices and other devices which do not
 * require true polling, but which only need to examine the state
 * of the read and write buffers to decide if I/O is possible.
 * Reading of file names also implies reading of end of buffer.
 * Reading of data or lines implies reading of end of file or file errors.
 * Devices should only specify the states which are left justified below,
 * but be prepared to handle any of the states which are returned.
 * The UP_STATE_FIRST_ILLEGAL value is not a true state and must be last.
 * The order and number of these values must correspond to the strings
 * in the stateTable array in upPrivate.c.
 */
typedef	enum
{
	UP_STATE_NONE,
	UP_STATE_READ_NAME,
		UP_STATE_READ_END_OF_BUFFER,
	UP_STATE_WRITE_NAME,
	UP_STATE_TRANSFER_NAME,
	UP_STATE_TRANSFER_END_OF_BUFFER,
	UP_STATE_READ_LINE,
	UP_STATE_READ_DATA,
		UP_STATE_READ_FILE_ERROR,
		UP_STATE_READ_EOF,
	UP_STATE_WRITE_DATA,
	UP_STATE_TRANSFER_DATA,
		UP_STATE_TRANSFER_FILE_ERROR,
		UP_STATE_TRANSFER_EOF,
		UP_STATE_FATAL_ERROR,
		UP_STATE_FINISHED,
		UP_STATE_FIRST_ILLEGAL
} UP_STATE;


/*
 * Structure for one particular device.
 * This has already been typedefed in up.h.
 */
struct	UP_DEV
{
	UP_ID		id;		/* unique id for this device */
	UP *		owner;		/* owning pipeline */
	UP_DEV *	next;		/* next device in chain */
	UP_FUNC *	func;		/* table of function pointers */
	UP_BUF *	readBuffer;	/* buffer to read data from */
	UP_BUF *	writeBuffer;	/* buffer to write data to */
	UP_BOOL		didReadWrite;	/* read/write was done this iteration */
	UP_BOOL		finished;	/* all data transfers are finished */
	UP_BOOL		failed;		/* device has had a fatal error */
	UP_STATE	state;		/* current state of device */
	void *		private;	/* private data for this device */
	int		pollCount;	/* current entries in polling table */
	int		pollTimeout;	/* timeout value for polling */
	struct pollfd *	pollTable;	/* current polling table */
	struct pollfd	pollEntry;	/* convenient poll entry */
};


/*
 * Macro to specify a transfer limit given a long or off_t value.
 * The limit is reduced to a reasonable value which fits in an int.
 */
#define	UP_LIMIT_SIZE	(16 * 1024)

#define	UP_LIMIT(value) \
	(((value) > UP_LIMIT_SIZE) ? UP_LIMIT_SIZE : ((int) (value)))


/*
 * Macros to generate internal errors.
 * These supply the file name and line number of the error.
 */
#define	UP_INTERNAL_ERROR(up) \
	upInternalError((up), __FILE__, __LINE__)

#define	UP_DEVICE_INTERNAL_ERROR(dev) \
	upDeviceInternalError((dev), __FILE__, __LINE__)


/*
 * Internal routines.
 */
extern	UP_DEV *	upAllocDevice(UP *, UP_FUNC *, int);
extern	UP_BOOL		upAttachBuffer(UP * up, UP_DEV *, UP_DEV *);
extern	UP_DEV *	upFindDevice(UP *, UP_ID, UP_TYPE);
extern	UP_ID		upLinkDevice(UP *, UP_DEV *);
extern	void		upFreeDevice(UP_DEV *);
extern	UP_BOOL		upDeviceReadWrite(UP_DEV *);
extern	UP_BOOL		upDefaultBuildPollTable(UP_DEV * dev);
extern	UP_BOOL		upDefaultFlush(UP_DEV * dev);
extern	UP_BOOL		upDefaultReset(UP_DEV * dev);
extern	UP_BOOL		upDefaultClose(UP_DEV * dev);
extern	int		upTransferData(UP_DEV * dev, int limitLength);
extern	int		upReadData(UP_DEV * dev, char * data, int length);

extern	int		upWriteData(UP_DEV * dev, const char * data,
				int length);

extern	int		upWriteZeroedData(UP_DEV * dev, int length);
extern	int		upGetFileType(int mode);
extern	UP_BOOL		upSetDidReadWrite(UP_DEV * dev);
extern	UP_BOOL		upSetFinished(UP_DEV * dev);
extern	void		upSetFailed(UP_DEV * dev);

extern	UP_BOOL		upWriteFileName(UP *, const char * fileName,
			const UP_STAT * status);

extern	UP_BOOL		upReadFileName(UP *, const char ** fileName,
			const UP_STAT ** status);

extern	void		upCopyStatus(UP_STAT * ourStat,
			const struct stat * systemStat);

extern	const char *	upTranslateFileName(UP_DEV * dev,
				const char * fileName,
				const char * const * conversionTable);

extern	char *		upGenerateFileName(UP_DEV *, const char * pattern,
				long value);

extern	UP_STATE	upCheckState(UP_DEV * dev);
extern	UP_STATE	upGetState(UP_DEV * dev);
extern	void		upSetState(UP_DEV * dev, UP_STATE newState);

extern	char *		upDeviceMalloc(UP_DEV * dev, int);
extern	char *		upDeviceRealloc(UP_DEV * dev, char *, int);
extern	char *		upDeviceCopyString(UP_DEV * dev, const char *);
extern	char *		upGrowTable(UP_DEV *, void *, void *, int, int, size_t);
extern	char *		upMalloc(UP * up, int);
extern	char *		upRealloc(UP * up, char *, int);
extern	char *		upCopyString(UP * up, const char *);
extern	int		upCreateTemporaryFile(UP_DEV * dev);
extern	void		upError(UP * up, int errorCode);
extern	void		upSystemError(UP * up, const char * systemName);
extern	void		upInternalError(UP * up, const char * file, int line);
extern	void		upDeviceError(UP_DEV * dev, int errorCode);


extern	void		upDeviceSystemError(UP_DEV * dev,
				const char * systemName);

extern	void		upDeviceFileSystemError(UP_DEV * dev,
				const char * systemName,
				const char * fileName);

extern	void		upDeviceInternalError(UP_DEV * dev, const char *, int);
extern	void		upWarning(UP *, const char *, ...);

#endif
