/*
 * -------------------------------------------------------------------------
 * DBell's user-mode pipeline library.
 * Copyright (C) 2001  David I. Bell (dbell@canb.auug.org.au)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 *
 * User pipelines top level public routines.
 * These routines are callable by the users of the pipeline library.
 * These are prototyped in up.h.
 */

#include <errno.h>
#include <stdarg.h>

#include "upPrivate.h"


/*
 * Version numbers.
 * Major version changes mean code changes for the clients.
 * Minor version changes are just enhancements or bug fixes.
 */
#define	MAJOR_VERSION	0
#define	MINOR_VERSION	9


/*
 * The extra amount of room to allocate for the polling table
 * when it needs to be grown.
 */
#define	POLL_ALLOC_SLOP	20


/*
 * The table of error messages.
 * The number and order of these messages must the definitions in up.h.
 */
static	const char * const	errorTable[UP_ERR_ERRORCOUNT] =
{
	"No error",
	"System call failed",
	"Cannot allocate memory",
	"Internal consistency check failed",
	"Pipeline has not yet been set up",
	"Pipeline has already been set up",
	"No devices are in pipeline",
	"Reading end of pipe is not reader",
	"Writing end of pipe is not writer",
	"Endpoint is wrong for operation",
	"Pipeline has an internal reader",
	"Pipeline has an internal writer",
	"Negative read or write count",
	"Bad or unknown device id",
	"Data is not an archive file",
	"Bad header",
	"Premature end of file reading file",
	"Bad buffer size",
	"Process exited with non-zero status",
	"Process terminated by a signal",
	"Process exited for unknown reason",
	"Process command line is null",
	"Illegal gzip compression level",
	"Illegal bits value for compress",
	"String buffer is full",
	"Unknown device id specified",
	"Wrong device type for operation",
	"Missing end line for uudecode file",
	"Zlib library call failed",
	"Data is not a gzip file",
	"Gzip file reserved header flags set",
	"File name was not found",
	"Multiple output device connected to single input device",
	"Duplicate begin line for uudecode file",
	"Buffer internal error",
	"File size is larger then status indicated",
	"Null pointer argument",
	"Bad file descriptor",
	"Invalid PATH value",
	"Path is not a directory",
	"Too much data",
	"Partial last line",
	"Bad index value",
	"Bad item type",
	"File has no data",
	"Line is too long",
	"Bad pattern string"
};


/*
 * The system path for running external processes securely.
 * This can be used instead of the user's PATH environment variable.
 * A NULL value means that the compiled-in path is being used.
 */
static	char *	systemPath;


/*
 * The directory which contains temporary files.
 * This can be used instead of the user's TMPDIR environment variable.
 * A NULL value means that TMPDIR (or stdio's tempdir) is used.
 */
static	char *	tmpdirPath;


/*
 * Get the version numbers.
 */
void
upGetVersion(int * major, int * minor)
{
	*major = MAJOR_VERSION;
	*minor = MINOR_VERSION;
}


/*
 * Allocate a new user mode pipeline.
 * Returns NULL if memory allocation failed.
 * Note: This routine cannot set an error code because all error codes
 * are stored an the UP structure (which is not yet allocated!)
 */
UP *
upAllocate(void)
{
	UP *	up;

	up = (UP *) malloc(sizeof(UP));

	if (up == UP_NULL)
		return UP_NULL;

	/*
	 * Initialise its fields.
	 */
	up->first = 0;
	up->last = 0;
	up->pollTable = NULL;
	up->pollAllocCount = 0;
	up->pollCount = 0;
	up->deviceCount = 0;
	up->errorCode = 0;
	up->systemErrno = 0;
	up->systemName = "";
	up->systemFileName = NULL;
	up->errorFileName = "";
	up->errorLineNumber = 0;
	up->warningCount = 0;
	up->maxWarningCount = -1;
	up->setupDone = UP_FALSE;
	up->pollBlab = UP_FALSE;
	up->loopBlab = UP_FALSE;

	/*
	 * Return the structure.
	 */
	return up;
}


/*
 * Close the user mode pipeline and free all of its devices.
 * After this call the pipeline can be rebuilt for a new use.
 * Returns UP_TRUE if all of the devices closed successfully.
 */
UP_BOOL
upClose(UP * up)
{
	UP_DEV *	dev;
	UP_DEV *	nextdev;
	UP_BOOL		status;

	status = UP_TRUE;

	/*
	 * Iterate over all of the devices, closing and freeing each one.
	 */
	nextdev = up->first;

	while (nextdev)
	{
		dev = nextdev;
		nextdev = dev->next;

		if (!dev->func->close(dev))
			status = UP_FALSE;

		upFreeDevice(dev);
	}

	/*
	 * Reset the pipeline for reuse.
	 */
	up->first = 0;
	up->last = 0;
	up->pollCount = 0;
	up->deviceCount = 0;
	up->setupDone = UP_FALSE;

	return status;
}


/*
 * Close and then free the user mode pipeline.
 * Errors during the close are ignored.
 * A NULL pointer can be given which will be ignored.
 */
void
upFree(UP * up)
{
	if (up == UP_NULL)
		return;

	(void) upClose(up);

	free(up->systemFileName);
	free(up->pollTable);
	free(up);
}


/*
 * Mark the setup of the pipeline as having been complete.
 * Returns UP_TRUE if the pipeline is successfully ready for use.
 * Nothing is done if the pipeline is already marked set up.
 */
UP_BOOL
upSetupDone(UP * up)
{
	UP_DEV *	writeDevice;
	UP_DEV *	readDevice;

	if (up->setupDone)
		return UP_TRUE;

	/*
	 * If there are no devices, then fail.
	 */
	if (up->deviceCount == 0)
	{
		upError(up, UP_ERR_NODEVICES);

		return UP_FALSE;
	}

	/*
	 * Make sure that the first and last devices in the pipeline are
	 * external reading and writing devices.
	 */
	if (!up->first->func->externalReader)
	{
		upError(up, UP_ERR_NOTREADER);

		return UP_FALSE;
	}

	if (!up->last->func->externalWriter)
	{
		upError(up, UP_ERR_NOTWRITER);

		return UP_FALSE;
	}

	/*
	 * Walk down the pairs of adjacent devices and allocate a
	 * buffer between each pair, and check whether there are any
	 * internal endpoint devices.
	 */
	writeDevice = up->first;
	readDevice = writeDevice->next;

	while (readDevice && writeDevice)
	{
		if (!upAttachBuffer(up, writeDevice, readDevice))
			return UP_FALSE;

		writeDevice = readDevice;
		readDevice = readDevice->next;
	}

	/*
	 * OK, we are ready for use.
	 */
	up->setupDone = UP_TRUE;

	return UP_TRUE;
}


/*
 * Perform one cycle of reading and writing throughout the pipeline.
 * This attempts to push some data from the first to the last device,
 * while waiting up to the specified number of milliseconds (-1 for forever).
 * If the last device in the pipeline has marked itself as finished then
 * this call returns immediately.  Returns UP_TRUE if the attempt was
 * successful, or UP_FALSE if there was an error.
 */
UP_BOOL
upReadWrite(UP * up, int timeout)
{
	UP_DEV *	dev;
	UP_BOOL		moreRequired;
	UP_BOOL		didSome;
	int		pollResult;

	/*
	 * Make sure the pipeline setup is done.
	 */
	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return UP_FALSE;
	}

	if (up->loopBlab)
		fprintf(stderr, "upReadWrite: called\n");

	/*
	 * If the last device has finished I/O, then return immediately.
	 */
	if (up->last->finished)
		return UP_TRUE;

	/*
	 * Perform a poll of the devices to see what I/O is ready,
	 * waiting if required.  This only returns on an error, a
	 * timeout, or when some device can do I/O.
	 */
	pollResult = upPoll(up, timeout);

	/*
	 * If poll failed, return that error.
	 */
	if (pollResult < 0)
		return UP_FALSE;

	/*
	 * If a timeout occurred then return success.
	 */
	if (pollResult == 0)
		return UP_TRUE;

	/*
	 * Poll indicates that some device should now be ready to do
	 * some I/O.  Clear the didReadWrite flag for each device so
	 * that we will know whether the device has moved some data yet.
	 */
	for (dev = up->first; dev; dev = dev->next)
		dev->didReadWrite = UP_FALSE;

	/*
	 * Make multiple passes through the pipeline trying to force
	 * data through it from the writing to the reading end.
	 * Only let each device transfer data once and don't do
	 * any I/O for devices which have finished.  Also make sure
	 * that some device actually did some I/O to detect programming
	 * bugs which would otherwise hang up the pipeline.
	 */
	didSome = UP_FALSE;

	do
	{
		if (up->loopBlab)
			fprintf(stderr, "upReadWrite: loop begin\n");

		moreRequired = UP_FALSE;

		/*
		 * Examine each device from the first one to the last one.
		 */
		for (dev = up->first; dev; dev = dev->next)
		{
			/*
			 * If this device has done any I/O this cycle
			 * or is all finished, then skip it.
			 */
			if (dev->didReadWrite || dev->finished)
				continue;

			/*
			 * Ask the device to try to perform some I/O.
			 */
			if (!upDeviceReadWrite(dev))
				return UP_FALSE;

			/*
			 * If the device did some I/O then remember that
			 * something got done, and that we should loop
			 * again so that a previous device might be able to
			 * fill its possibly emptied buffer with new data.
			 */
			if (dev->didReadWrite || dev->finished)
			{
				didSome = UP_TRUE;
				moreRequired = UP_TRUE;
			}
		}
	}
	while (moreRequired);

	if (up->loopBlab)
		fprintf(stderr, "upReadWrite: loop end\n");

	/*
	 * Make sure that something got done.
	 * This catches programming errors in devices which say that
	 * they are ready for I/O, but when asked, perform none.
	 * A bug such as this is very bad since the program will loop
	 * forever, calling upReadWrite repeatedly without accomplishing
	 * anything.
	 */
	if (!didSome)
	{
		UP_INTERNAL_ERROR(up);

		return UP_FALSE;
	}

	/*
	 * OK, this cycle is complete.
	 */
	return UP_TRUE;
}


/*
 * Wait for some part of the pipeline to have data to be processed,
 * or until the specified timeout in milliseconds has elapsed.
 * Returns 1 if some data is ready for processing, 0 if a timeout
 * occurred, or -1 on an error (which could be EINTR).  This does
 * not transfer any data between the devices.
 */
int
upPoll(UP * up, int timeout)
{
	UP_DEV *	dev;
	struct pollfd *	pollTable;
	int		pollCount;
	int		pollTimeout;
	int		status;
	int		i;

	/*
	 * Ask the devices to construct the polling list.
	 */
	if (!upPollList(up, &pollTable, &pollCount, &pollTimeout))
		return -1;

	/*
	 * If any device has indicated that it can run now, then
	 * return immediately with an indication that I/O is ready.
	 */
	if (pollTimeout == 0)
		return 1;

	/*
	 * If the user timeout value is zero and no devices actually
	 * want to poll right now, then return immediately saying
	 * nothing was done.
	 */
	if ((timeout == 0) && (pollCount == 0))
		return 0;

	/*
	 * If the poll timeout value is less than the requested timeout,
	 * then use that value instead.
	 */
	if ((timeout < 0) || ((pollTimeout >= 0) && (timeout > pollTimeout)))
		timeout = pollTimeout;

	if (up->pollBlab)
	{
		fprintf(stderr, "poll: entries %d, timeout %d)\n",
			pollCount, timeout);
	}

	/*
	 * Actually do the poll.
	 * We block here until a timeout, error, or I/O is ready.
	 */
	status = poll(pollTable, pollCount, timeout);

	if (up->pollBlab)
		fprintf(stderr, "poll: status %d)\n", status);

	if (status < 0)
	{
		upSystemError(up, "poll");

		return -1;
	}

	/*
	 * Copy the poll results back into the results fields in the
	 * device polling tables so that the devices can examine them.
	 */
	for (dev = up->first; dev; dev = dev->next)
	{
		for (i = 0; i < dev->pollCount; i++)
		{
			if (up->pollBlab)
			{
				fprintf(stderr,
					"%s[%d]: poll fd %d rev %x\n",
					dev->func->name, dev->id,
					pollTable->fd, pollTable->revents);
			}

			dev->pollTable[i].revents = pollTable->revents;
			pollTable++;
		}
	}

	/*
	 * OK, return whether there are any file handles ready,
	 * or whether there was a timeout.
	 */
	return (status > 0);
}


/*
 * Return the table of file descriptors within the pipeline which need
 * to be polled in order to transfer data through it, and the maximum
 * time interval to delay for some device..  This table is in the form
 * of a struct pollfd array which is ready for use by poll.
 * This table is valid until the next call to this routine or any
 * other UP routine which reads or writes.  Returns a pointer to the
 * table and the number of entries through the supplied indirect pointers,
 * and the maximum timeout in milliseconds before another check for I/O
 * is required for some device.  A timeout value of zero means that some
 * device wants to do I/O now.  A returned timeout value of -1 means that
 * no device wants to do I/O until poll indicates that one of the specified
 * file descriptors is ready.  Returns UP_TRUE on success.
 */
UP_BOOL
upPollList(
	UP *			up,
	struct pollfd **	retPollTable,
	int *			retPollCount,
	int *			retPollTimeout
)
{
	UP_DEV *	dev;
	struct pollfd *	pollTable;
	int		pollCount;
	int		pollTimeout;
	int		allocCount;
	int		i;

	/*
	 * Clear these in case of errors.
	 */
	*retPollTable = NULL;
	*retPollCount = 0;
	*retPollTimeout = 0;

	up->pollCount = 0;
	up->pollTimeout = 0;

	/*
	 * Make sure the pipeline setup is done.
	 */
	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return UP_FALSE;
	}

	/*
	 * Collect all of the polling information that is required for
	 * the devices and count the number of polling entries required.
	 */
	pollCount = 0;
	pollTimeout = -1;

	for (dev = up->first; dev; dev = dev->next)
	{
		/*
		 * If the device has finished all of its I/O,
		 * then no polling is needed for it.
		 */
		if (dev->finished)
		{
			dev->pollCount = 0;
			dev->pollTimeout = 0;

			continue;
		}

		/*
		 * Ask the device whether it needs any polling done.
		 * The device will store this information into its
		 * pollCount and pollTable data members.
		 */
		if (!dev->func->buildPollTable(dev))
			return UP_FALSE;

		/*
		 * Increment the number of poll entries needed.
		 */
		pollCount += dev->pollCount;

		/*
		 * Check whether the device requires a timeout.
		 * If so, reduce our timeout value to the minimum of its
		 * current value and this one.
		 */
		if ((dev->pollTimeout >= 0) &&
			((pollTimeout < 0) ||
				(dev->pollTimeout < pollTimeout)))
		{
			pollTimeout = dev->pollTimeout;
		}
	}

	/*
	 * If no polling entries are required then return now.
	 */
	if (pollCount == 0)
		return UP_TRUE;

	/*
	 * Make sure our poll structure is large enough to hold the
	 * required number of entries.  If not, then reallocate it
	 * with some spare room for future growth.
	 */
	if (pollCount > up->pollAllocCount)
	{
		allocCount = pollCount + POLL_ALLOC_SLOP;

		pollTable = (struct pollfd *) realloc(
			(void *) up->pollTable,
			allocCount * sizeof(*pollTable));

		if (pollTable == NULL)
		{
			upError(up, UP_ERR_NOMEM);

			return -1;
		}

		up->pollTable = pollTable;
		up->pollAllocCount = allocCount;
	}

	/*
	 * Loop through the devices copying their polling specifications
	 * into our single polling table.  Clear the returned events field
	 * in case of errors.
	 */
	pollTable = up->pollTable;

	for (dev = up->first; dev; dev = dev->next)
	{
		for (i = 0; i < dev->pollCount; i++)
		{
			pollTable->fd = dev->pollTable[i].fd;
			pollTable->events = dev->pollTable[i].events;
			pollTable->revents = 0;

			dev->pollTable[i].revents = 0;

			if (up->pollBlab)
			{
				fprintf(stderr, "%s[%d] poll fd %d ev %x\n",
					dev->func->name, dev->id,
					pollTable->fd, pollTable->events);
			}

			pollTable++;
		}

	}

	/*
	 * Save the number of poll entries and the desired poll timeout.
	 */
	up->pollCount = pollCount;
	up->pollTimeout = pollTimeout;

	/*
	 * Return the results.
	 */
	*retPollTable = up->pollTable;
	*retPollCount = up->pollCount;
	*retPollTimeout = up->pollTimeout;

	return UP_TRUE;
}


/*
 * Read data from the end of the user mode pipeline.
 * This routine can only be called if the last device of the pipeline
 * is the special readEndpoint device.  Returns the amount of data read,
 * zero on end of the current file, or -1 on an error.
 */
int
upRead(UP * up, char * data, int len)
{
	UP_DEV *	dev;
	UP_BUF *	buf;

	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return -1;
	}

	if (len < 0)
	{
		upError(up, UP_ERR_BADCOUNT);

		return -1;
	}

	/*
	 * Make sure the last device is the reader endpoint device.
	 */
	dev = up->last;

	if (dev->func->type != UP_TYPE_READ_ENDPOINT)
	{
		upError(up, UP_ERR_NOTREADER);

		return -1;
	}

	/*
	 * Check for no data to be read or that we are finished.
	 */
	if ((len == 0) || dev->finished)
		return 0;

	/*
	 * Keep polling until there is some data available for reading
	 * from the input buffer or until there is an end of file set
	 * for the buffer.
	 */
#if 0
	buf = dev->readBuffer;

	while (!upBufferReadAvailable(buf))
	{
		/*
		 * Wait for I/O to be ready and then do some.
		 */
		if (!upReadWrite(up, -1))
			return -1;
	}

	/*
	 * Check for an end of file condition.
	 */
	if (upBufferReadHasEof(buf))
	{
		dev->finished = UP_TRUE;

		return 0;
	}
#endif
	/*
	 * OK, read the available data.
	 */
	return upBufferRead(buf, data, len);
}


/*
 * Write some data to a user mode pipeline.
 * This can only be called if the first device of the pipeline is the
 * special writeEndpoint device.  Returns the amount of data written
 * or -1 on an error.
 */
int
upWrite(UP * up, const char * data, int len)
{
	UP_DEV *	dev;
	UP_BUF *	buf;

	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return -1;
	}

	if (len < 0)
	{
		upError(up, UP_ERR_BADCOUNT);

		return -1;
	}

	/*
	 * Make sure the first device is the writer endpoint device.
	 */
	dev = up->first;

	if (dev->func->type != UP_TYPE_WRITE_ENDPOINT)
	{
		upError(up, UP_ERR_NOTWRITER);

		return -1;
	}

	/*
	 * Check for no data to be written.
	 */
	if (len == 0)
		return 0;
#if 0
	/*
	 * Keep polling until there is some data available for writing
	 * to the output buffer.
	 */
	buf = dev->writeBuffer;

	while (!upBufferWriteAvailable(buf))
	{
		/*
		 * Wait for some I/O to be ready and then do some.
		 */
		if (!upReadWrite(up, -1))
			return -1;
	}
#endif
	/*
	 * Write the data.
	 */
	return upBufferWrite(buf, data, len);
}


/*
 * Try to flush data through the pipeline.
 */
UP_BOOL
upFlush(UP * up)
{
	UP_DEV *	dev;
	UP_BOOL		status;

	/*
	 * Make sure we are set up.
	 */
	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return UP_FALSE;
	}

	/*
	 * Loop over all of the devices, flushing them if we can
	 * and remembering whether all of them were successful.
	 */
	status = UP_TRUE;

	for (dev = up->first; dev; dev = dev->next)
	{
		if (dev->finished)
			continue;

		if (!dev->func->flush(dev))
			status = UP_FALSE;
	}

	return status;
}


/*
 * Return whether data can be read from the pipeline using upRead
 * without blocking.  This attempts to do I/O to get some data
 * if required.
 */
UP_BOOL
upCanRead(UP * up)
{
	UP_DEV *	dev;

	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return UP_FALSE;
	}

	/*
	 * Make sure the last device is the reader endpoint device.
	 */
	dev = up->last;

	if (dev->func->type != UP_TYPE_READ_ENDPOINT)
	{
		upError(up, UP_ERR_NOTREADER);

		return UP_FALSE;
	}

	/*
	 * See if any data is already readable from the input buffer.
	 */
	if (upBufferReadDataCheck(dev->readBuffer, 1))
		return UP_TRUE;

	/*
	 * There is no data already present.
	 * Try to transfer data to produce some more but do not block.
	 */
	if (!upReadWrite(up, 0))
		return UP_FALSE;

	/*
	 * OK, now return whether the buffer has data.
	 */
	return upBufferReadDataCheck(dev->readBuffer, 1);
}


/*
 * Return whether data can be written to the pipeline using upWrite
 * without blocking.  This attempts to do I/O to provide some room
 * if required.
 */
UP_BOOL
upCanWrite(UP * up)
{
	UP_DEV *	dev;

	if (!up->setupDone)
	{
		upError(up, UP_ERR_NOTSETUP);

		return UP_FALSE;
	}

	/*
	 * Make sure the first device is the writer endpoint device.
	 */
	dev = up->first;

	if (dev->func->type != UP_TYPE_WRITE_ENDPOINT)
	{
		upError(up, UP_ERR_NOTWRITER);

		return UP_FALSE;
	}

	/*
	 * See if the output buffer already has room for some data.
	 */
	if (upBufferRequestSpace(dev->writeBuffer, 1) > 0)
		return UP_TRUE;

	/*
	 * There is no room in the buffer right now.
	 * Try to transfer data to produce some room but do not block.
	 */
	if (!upReadWrite(up, 0))
		return UP_FALSE;

	/*
	 * OK, now return whether the buffer has room for data.
	 */
	return (upBufferRequestSpace(dev->writeBuffer, 1) > 0);
}


/*
 * Return whether or not the pipeline has completed all its I/O.
 * This is determined by the last device having been marked as finished.
 */
UP_BOOL
upIsFinished(const UP * up)
{
	/*
	 * If we are not set up yet, then return no EOF yet.
	 */
	if (!up->setupDone)
		return UP_FALSE;

	/*
	 * See if the last device has finished its I/O yet.
	 */
	return up->last->finished;
}


/*
 * Set the secure system path list to the specified colon-separated
 * list of paths.  When desired, this can be used instead of the
 * path list in the user's PATH environment variable.  This allows
 * devices which require the execution of external programs to be a
 * bit safer.  The supplied path is copied into internal storage.
 * Specifying NULL sets the compiled-in secure path list.
 * The value set is a global setting for all the pipelines.
 * Returns UP_BOOL if the path was successfully set.
 */
UP_BOOL
upSetSystemPath(const char * path)
{
	char *	newPath;

	newPath = NULL;

	/*
	 * If a path was specified then duplicate it.
	 */
	if (path != NULL)
	{
		newPath = strdup(path);

		if (newPath == NULL)
			return UP_FALSE;
	}

	/*
	 * Free any old path and save the new one.
	 */
	free(systemPath);

	systemPath = newPath;

	return UP_TRUE;
}


/*
 * Return the current secure system path list.  This is the path
 * list last set by upSetSystemPath or else the compiled-in secure
 * path list if one was not set.  The returned string belongs to us
 * and should not be freed by the caller.
 */
const char *
upGetSystemPath()
{
	if (systemPath)
		return systemPath;

	return UP_SYSTEM_PATH;
}


/*
 * Set the temporary file directory to the indicated path.
 * When desired, this can be used instead of the TMPDIR environment
 * variable.  If NULL is supplied, the TMPDIR (or stdio's temp
 * directory) is used.  Returns UP_TRUE on success.
 */
UP_BOOL
upSetTemporaryDirectoryPath(const char * path)
{
	char *	newPath;

	newPath = NULL;

	/*
	 * If a path was specified then duplicate it.
	 */
	if (path != NULL)
	{
		newPath = strdup(path);

		if (newPath == NULL)
			return UP_FALSE;
	}

	/*
	 * Free any old path and save the new one.
	 */
	free(tmpdirPath);

	tmpdirPath = newPath;

	return UP_TRUE;
}


/*
 * Return the temporary file directory to be used.
 * This is the path last set by upSetTemporaryDirectoryPath or else
 * the path defined in TMPDIR or from stdio's definition.
 */
const char *
upGetTemporaryDirectoryPath()
{
	const char *	path;

	if (tmpdirPath)
		return tmpdirPath;

	path = getenv("TMPDIR");

	if (path == NULL)
		path = P_tmpdir;

	return path;
}


/*
 * Clear the last error which occurred on the specified UP structure.
 */
void
upClearError(UP * up)
{
	up->errorCode = UP_ERR_NONE;
	up->systemErrno = 0;
	up->systemName = "";
	up->errorFileName = "";
	up->errorLineNumber = 0;
}


/*
 * Return the error code for the last error on the UP device.
 * This will be nonzero if there was an error stored.
 */
int
upGetErrorCode(const UP * up)
{
	return up->errorCode;
}


/*
 * Return the number of devices which are in the pipeline.
 */
int
upGetDeviceCount(const UP * up)
{
	return up->deviceCount;
}


/*
 * Return an error string for the last error on the specified UP device.
 * The returned string is in a static buffer and is overwritten each call.
 */
const char *
upGetErrorString(const UP * up)
{
	char *		cp;
	static	char	buf[8192];

	/*
	 * Set up to append into the buffer.
	 */
	cp = buf;

	/*
	 * Start off with a description of who is making the error.
	 */
	strcpy(cp, "UP error");
	cp += strlen(cp);

	/*
	 * If this was a device error, then store the device name and id.
	 */
	if (up->errorDeviceId != UP_NULL_ID)
	{
		sprintf(cp, " (device \"%s\", id %d)", up->errorDeviceName,
			up->errorDeviceId);

		cp += strlen(cp);
	}

	/*
	 * Add spacing.
	 */
	strcpy(cp, ": ");
	cp += strlen(cp);

	/*
	 * Store the description of the error code.
	 */
	if ((up->errorCode >= 0) && (up->errorCode < UP_ERR_ERRORCOUNT))
		strcat(cp, errorTable[up->errorCode]);
	else
		sprintf(cp, "Error %d", up->errorCode);

	cp += strlen(cp);

	/*
	 * If the error code was a system call error, then the system call
	 * name which failed, an optional file name, and the description
	 * of the system error code.
	 */
	if (up->errorCode == UP_ERR_SYSTEM)
	{
		sprintf(cp, ": \"%s\" call", up->systemName);
		cp += strlen(cp);

		if (up->systemFileName)
		{
			cp += strlen(cp);
			strcpy(cp, ": \"");
			cp += strlen(cp);
			strcpy(cp, up->systemFileName);
			cp += strlen(cp);
			*cp++ = '"';
		}

		strcpy(cp, ": ");
		cp += strlen(cp);

		strcpy(cp, strerror(up->systemErrno));
		cp += strlen(cp);
	}

	/*
	 * If the error code was an internal error, then store the
	 * file name and line number of the error.
	 */
	if (up->errorCode == UP_ERR_INTERNAL)
	{
		strcpy(cp, ": ");
		cp += strlen(cp);

		sprintf(cp, "file \"%s\", line %d", up->errorFileName,
			up->errorLineNumber);
	}

	/*
	 * Return the address of the error text.
	 */
	return buf;
}


/*
 * Print out the last error on a UP structure to stderr.
 * Errors are remembered in the stderr FILE handle.
 * The specified string is printed before the message, if not NULL.
 * If NULL, a generic message is printed.
 */
void
upPrintError(const UP * up, const char * message, ...)
{
	va_list	ap;
	char	buf[8192];

	if (message == NULL)
		message = "up";

	va_start(ap, message);
	vsprintf(buf, message, ap);
	va_end(ap);

	fprintf(stderr, "%s: %s\n", buf, upGetErrorString(up));
}


/*
 * Return the number of warnings which have occurred.
 */
long
upGetWarningCount(const UP * up)
{
	return up->warningCount;
}


/*
 * Return the limit on how many warnings are printed.
 */
long
upGetMaximumWarnings(const UP * up)
{
	return up->maxWarningCount;
}


/*
 * Specify a limit on how many total warnings are printed.
 * A value of zero disables all new warnings from being printed.
 * A value of -1 indicates no limit.
 */
void
upSetMaximumWarnings(UP * up, long maxWarningCount)
{
	up->maxWarningCount = maxWarningCount;
}


/*
 * Set the debug flags.
 */
void
upSetDebugFlags(UP * up, UP_DEBUG_FLAG flags)
{
	up->pollBlab = ((flags & UP_DEBUG_FLAG_POLL) != 0);
	up->loopBlab = ((flags & UP_DEBUG_FLAG_LOOP) != 0);
	up->stateBlab = ((flags & UP_DEBUG_FLAG_STATE) != 0);
}
